/* 
 Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore
 
 Dr. Preston B. Moore
 Associate Director, Center for Molecular Modeling (CMM)
 University of Pennsylvania, Department of Chemistry, Box 188 
 231 S. 34th St. Philadelphia, PA 19104-6323 USA
 EMAIL: moore@cmm.chem.upenn.edu  
 WWW: http://www.cmm.upenn.edu/~moore
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 */


#include "md.h"

#define RCGBOND 3.71   /* 3.71 have to make this an input variable */

#define HEALINGREGION_SO3
/* #define PRINT_FORCE */

/* #define DEBUG  */
/* #define KCALS  */
/* #define CHECK  */

void ptensorchk(SIMPARMS *simparms,COORDS *coords,INTER *inter,NGBR *ngbr,int iflg);
void ngradv(SIMPARMS *simparms,COORDS *coords,INTER *inter,NGBR *ngbr,int iflg);
void get_p2mt(int,double[],double[],double[],double[],double[9]);

enum funcs {BOND,BEND,XBOND};

/*------------------------------------------------------------------*/
void force(SIMPARMS *simparms,COORDS *coords,INTER *inter,NGBR *ngbr,MYTIMING *timing,int iflg)
{
  int i,n3;
  double *fxp;
#ifdef PARA
  double *rec_buf;
#endif
	
  n3 = simparms->natoms*3;
	
  switch(iflg){
    case 0: /* long range intermolecular */
      myclockon(timing,"FORCE LONG RANGE",3);
      if(simparms->iperd!=0) gethinv9(coords->hmat,coords->hmati);
      fxp = coords->fxs;  for(i=0;i<n3;i++) fxp[i] = 0.;
      fxp = coords->fxl;  for(i=0;i<n3;i++) fxp[i] = 0.;
      force_ter(simparms,coords,inter,ngbr,iflg);
      force_ter_e(simparms,coords,inter,ngbr,iflg);
      if(simparms->iperd == 3){
        fk_ewald(simparms,coords);
        ecorr(simparms,coords);
      }
      rmass(simparms->natoms,coords->fxl,coords->fyl,coords->fzl,coords->amass);
      rmass(simparms->natoms,coords->fxs,coords->fys,coords->fzs,coords->amass);
      myclockoff(timing,"FORCE LONG RANGE",3);
      break;
      
      case 1:  /* short range intermolecular */
      myclockon(timing,"FORCE SHORT RANGE",4);
      if(simparms->iperd!=0)gethinv9(coords->hmat,coords->hmati);
      fxp = coords->fxs;
      for(i=0;i<n3;i++) fxp[i] = 0.;
      force_ter(simparms,coords,inter,ngbr,iflg);
      force_ter_e(simparms,coords,inter,ngbr,iflg);
      rmass(simparms->natoms,coords->fxs,coords->fys,coords->fzs,coords->amass);
      myclockoff(timing,"FORCE SHORT RANGE",4);
      break;
      
      case 3: /* intra molecular */
      myclockon(timing,"FORCE BONDS BENDS",5);
      fxp = coords->fxa;
      for(i=0;i<n3;i++) fxp[i] = 0.;
      fbond(simparms,coords);
      fbend(simparms,coords);
      fbondx(simparms,coords);
      /*
       fbendquart(simparms,coords);
       */
      fcolvar_cntr(simparms,coords,1); /* 1 is to add force to particles */
#ifdef TWO
      ftors(simparms,coords);
      fonfo(simparms,coords); 
      if(simparms->iextern==1) force_extern(simparms,coords,inter->ntable);
#endif    
      
#ifdef PARA
      rec_buf = coords->scr_rec;
      for(i=0;i<n3;i++) rec_buf[i] = 0.;
      MPI_Allreduce(fxp,rec_buf,n3,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
      for(i=0;i<n3;i++) fxp[i] = rec_buf[i];
#endif
      rmass(simparms->natoms,coords->fxa,coords->fya,coords->fza,coords->amass);
      myclockoff(timing,"FORCE BONDS BENDS",5);
      break;
      
      case 2: /* (torsion & 1-4) molecular */
      myclockon(timing,"FORCE TORSION ONE-FOUR",6);
      fxp = coords->fxr;
      for(i=0;i<n3;i++) fxp[i] = 0.;
      
#ifndef TWO
      ftors(simparms,coords);
      fonfo(simparms,coords); 
      if(simparms->iextern==1) force_extern(simparms,coords,inter->ntable);
#ifdef PARA
      rec_buf = coords->scr_rec;
      for(i=0;i<n3;i++) rec_buf[i] = 0.;
      MPI_Allreduce(fxp,rec_buf,n3,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
      for(i=0;i<n3;i++) fxp[i] = rec_buf[i];
#endif
#endif
      rmass(simparms->natoms,coords->fxr,coords->fyr,coords->fzr,coords->amass);
      myclockoff(timing,"FORCE TORSION ONE-FOUR",6);
      break;
      
      default:
      fprintf(stderr,"ERROR: in forces!!!????\n");
      exit(1);
  }
  if(simparms->nfreeze>0) freeze_atom(simparms,coords);
  if(simparms->iensemble==3) {
    fxp = coords->vx;
    for(i=0;i<n3;i++) fxp[i] = 0.;
  }
  
#ifdef PRINT_FORCE
  printf("force iflag = %d\n",iflg);
  for(i=0;i<n3;i++) printf("force %d %g\n",i,fxp[i]);
#endif
}

/*------------------------------------------------------------------*/
void force_hybrid(SIMPARMS *simparms,COORDS *coords,INTER *inter,NGBR *ngbr,SO3 *so3,MYTIMING *timing,int iflg)
{
  int i,ii,n3,*isel,nsel;
  double *fxp,*fyp,*fzp,*fxp2,*fyp2,*fzp2,(*tor)[3];
  double *dr,*vdsnow,*vdsprev,*vbuf;
#ifdef PARA
  double *rec_buf;
#endif
	
  vbuf = simparms->vbuf + (1-coords->lcg)*simparms->ncg; 
  n3 = coords->natoms*3;
  
  switch(iflg){
    case 1:  /* short range intermolecular */
      myclockon(timing,"FORCE SHORT RANGE",4);
      if(simparms->iperd!=0)gethinv9(coords->hmat,coords->hmati);
      
      
      /* SO3 rotation stuff */
//      if(simparms->iso3dyn){
//        if(coords->lcg == 0){
          
          /* zero also the CG part of fxb to store the SO3 coupling contribution */
//          n3 = coords->next->natoms*3;
//          fxp = coords->next->fxb; 
//          for(i=0;i<n3;i++) fxp[i] = 0.;
          
//          n3 = coords->natoms*3;      
//          fxp = coords->fxb; 
//          for(i=0;i<n3;i++) fxp[i] = 0.;
          
//          tor = so3->tor_s;
//          for(i=0;i<simparms->ncg;i++){
//            tor[i][0] = tor[i][1] = tor[i][2] = 0.;
//          }
          
//          nsel = coords->next->ninactive+coords->next->nschizo;
//          isel = coords->next->ischizo;
//          get_relat_aapos(coords->next,simparms->icglist,simparms->napcg,isel,nsel);
//          zero_potlso3();     /* zero total energy (handy when parallelized later) */
//          fnonbond_so3(simparms,coords,inter,so3,ngbr->nprr,ngbr->inpr,ngbr->jnpr,0);
//          get_abs_aapos(coords->next,simparms->icglist,simparms->napcg,isel,nsel);
//        }
//      }
      
      
      fxp = coords->fxs;
      if(coords->lcg){
        for(i=0;i<n3;i++) fxp[i] = 0.;
      }else{
        fyp = coords->fys; fzp = coords->fzs;
        isel = coords->iselect;
        for(i=0;i<coords->nselect;i++){
          ii=isel[i];
          fxp[ii]=fyp[ii]=fzp[ii]=0.;
        }
      }
      for(i=0;i<n3;i++) fxp[i] = 0.;  /* why again??? */
      force_ter(simparms,coords,inter,ngbr,iflg);
      force_ter_e(simparms,coords,inter,ngbr,iflg);
      if(coords->lcg){
        /* add the SO3 coupling contrib to the forces and divide by masses */
//        if(simparms->iso3dyn){
//          for(i=0;i<n3;i++) fxp[i]+=coords->fxb[i];
//        }
        rmass(coords->natoms,coords->fxs,coords->fys,coords->fzs,coords->amass);
      }else{
//        for(i=0;i<n3;i++) fxp[i]+=coords->fxb[i];  /* in de COMcounterforce plan add so3 force to the rest */
        rmassselect(coords->iselect,coords->nselect,coords->fxs,coords->fys,coords->fzs,coords->amass);
      }
      myclockoff(timing,"FORCE SHORT RANGE",4);
      
      break;
      case 0: /* long range intermolecular */
      myclockon(timing,"FORCE LONG RANGE",3);
      dr = simparms->ddr_lrf + (1-coords->lcg)*simparms->ncg;
      vdsnow = simparms->vds_lrf + (1-coords->lcg)*simparms->ncg; 
      vdsprev = simparms->vds_lrfprev + (1-coords->lcg)*simparms->ncg; 
      for(i=0;i<coords->natoms;i++) vdsprev[i] = vdsnow[i];
      for(i=0;i<coords->natoms;i++) vdsnow[i] = 0.;    /* since I only use the HR part, I don' have to zero really */
      
      if(simparms->iperd!=0) gethinv9(coords->hmat,coords->hmati);


      /* SO3 rotation stuff */
//      if(simparms->iso3dyn){
//        if(coords->lcg == 0){
          
          /* zero also the CG part of fxb to store the SO3 coupling contribution */
//          n3 = coords->next->natoms*3;
//          fxp = coords->next->fxb;
//          for(i=0;i<n3;i++) fxp[i] = 0.;
          
//          n3 = coords->natoms*3;
//          fxp = coords->fxb;
//          for(i=0;i<n3;i++) fxp[i] = 0.;
          
//          tor = so3->tor_s;
//          for(i=0;i<simparms->ncg;i++){
//            tor[i][0] = tor[i][1] = tor[i][2] = 0.;
//          }
          
//          nsel = coords->next->ninactive+coords->next->nschizo;
//          isel = coords->next->ischizo;
//          get_relat_aapos(coords->next,simparms->icglist,simparms->napcg,isel,nsel);
//          zero_potlso3();     /* zero total energy (handy when parallelized later) */
//          fnonbond_so3(simparms,coords,inter,so3,ngbr->nprr,ngbr->inpr,ngbr->jnpr,0);
          
//          getMitAA(simparms,coords,so3); 
          
//          get_abs_aapos(coords->next,simparms->icglist,simparms->napcg,isel,nsel);
//        }
//      }
      
      
      fxp = coords->fxl;  fxp2 = coords->fxs;
      if(coords->lcg){
        for(i=0;i<n3;i++) fxp[i] = fxp2[i] = 0.;
      }else{
        fyp = coords->fyl; fzp = coords->fzl;	  fyp2 = coords->fys; fzp2 = coords->fzs;
        isel = coords->iselect;
        for(i=0;i<coords->nselect;i++){
          ii=isel[i];
          fxp[ii]=fyp[ii]=fzp[ii]=fxp2[ii]=fyp2[ii]=fzp2[ii]=0.;
        }
      }
      force_ter(simparms,coords,inter,ngbr,iflg);
      force_ter_e(simparms,coords,inter,ngbr,iflg);
      if(coords->iperd == 3){
        fk_ewald(simparms,coords);
        ecorr(simparms,coords);
      }
      if(coords->lcg){
        /* add the SO3 coupling contrib to the forces and divide by masses */
//        if(simparms->iso3dyn){
//          for(i=0;i<n3;i++) fxp[i]+=coords->fxb[i];
//        }
        rmass(coords->natoms,coords->fxl,coords->fyl,coords->fzl,coords->amass);
        rmass(coords->natoms,coords->fxs,coords->fys,coords->fzs,coords->amass);
      }else{
//        for(i=0;i<n3;i++) fxp[i]+=coords->fxb[i];  /* in de COMcounterforce plan add so3 force to the rest */
        rmassselect(coords->iselect,coords->nselect,coords->fxl,coords->fyl,coords->fzl,coords->amass);
        rmassselect(coords->iselect,coords->nselect,coords->fxs,coords->fys,coords->fzs,coords->amass);
      }
      
      for(i=0;i<coords->nschizo;i++){
        ii = coords->ischizo[i];
        vbuf[ii] += 0.5*(vdsnow[ii]+vdsprev[ii])*dr[ii];
      }
      
      for(i=0;i<coords->natoms;i++) dr[i] = 0.;
#ifdef FIXHYBRIDSWAPS
      inter->nswptot = 0;
#endif
      
      myclockoff(timing,"FORCE LONG RANGE",3);
      break;
      case 3: /* intra molecular */  
      myclockon(timing,"FORCE BONDS BENDS",5);
      dr = simparms->ddr_intra + (1-coords->lcg)*simparms->ncg;
      vdsnow = simparms->vds_intra + (1-coords->lcg)*simparms->ncg; 
      vdsprev = simparms->vds_intraprev + (1-coords->lcg)*simparms->ncg; 
      for(i=0;i<coords->natoms;i++) vdsprev[i] = vdsnow[i];
      for(i=0;i<coords->natoms;i++) vdsnow[i] = 0.;    /* since I only use the HR part, I don' have to zero really */  
      
      /* SO3 rotation stuff */
      if(simparms->iso3dyn){
        if(coords->lcg == 0){
          
          /* zero also the CG part of fxt to store the SO3 coupling contribution */
          n3 = coords->next->natoms*3;
          fxp = coords->next->fxt;
          for(i=0;i<n3;i++) fxp[i] = 0.;
          
          n3 = coords->natoms*3;      
          fxp = coords->fxt; 
          for(i=0;i<n3;i++) fxp[i] = 0.;
          
          tor = so3->tor;
          for(i=0;i<simparms->ncg;i++){
            tor[i][0] = tor[i][1] = tor[i][2] = 0.;
          }
          
          nsel = coords->next->ninactive+coords->next->nschizo;
          isel = coords->next->ischizo;
          get_relat_aapos(coords->next,simparms->icglist,simparms->napcg,isel,nsel);
                    
          fbond_so3(simparms,coords,so3,1); /* healing region part */
          fbond_so3(simparms,coords,so3,0); /* CG region part */
          fbend_so3(simparms,coords,so3,1); /* healing region part */
          fbend_so3(simparms,coords,so3,0); /* CG region part */
          fbondx_so3(simparms,coords,so3,1); /* healing region part */
          fbondx_so3(simparms,coords,so3,0); /* CG region part */
          ftors_so3(simparms,coords,so3,1); /* healing region part */
          ftors_so3(simparms,coords,so3,0); /* CG region part */
          fonfo_so3(simparms,coords,so3,1); /* healing region part */
          fonfo_so3(simparms,coords,so3,0); /* CG region part */
          
          zero_potlso3();     /* zero total energy (handy when parallelized later) */
          fnonbond_so3(simparms,coords,inter,so3,ngbr->nprr,ngbr->inpr,ngbr->jnpr,0);
          
//          tor = so3->tor;
//          for(i=0;i<simparms->ncg;i++){
//            tor[i][0] += so3->tor_s[i][0];
//            tor[i][1] += so3->tor_s[i][1];
//            tor[i][2] += so3->tor_s[i][2];
//          }
          
          getMitAA(simparms,coords,so3); 
          
          get_abs_aapos(coords->next,simparms->icglist,simparms->napcg,isel,nsel);
        }
      }
      
      fxp = coords->fxa; 
      for(i=0;i<n3;i++) fxp[i] = 0.;
      
      fbond_hr(simparms,coords);
      fbend_hr(simparms,coords);
      fbondx_hr(simparms,coords);
			
      /*
       fbendquart(simparms,coords);
       */
      fcolvar_cntr(simparms,coords,1); /* 1 is to add force to particles */
#ifdef TWO
      ftors_hr(simparms,coords);
      fonfo_hr(simparms,coords); 
      if(simparms->iextern==1) force_extern(simparms,coords,inter->ntable);
#endif    
      
#ifdef PARA
      rec_buf = coords->scr_rec;
      for(i=0;i<n3;i++) rec_buf[i] = 0.;
      MPI_Allreduce(fxp,rec_buf,n3,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
      for(i=0;i<n3;i++) fxp[i] = rec_buf[i];
#endif
      if(coords->lcg){
        /* add the SO3 coupling contrib to the forces and divide by masses */
        if(simparms->iso3dyn){
          for(i=0;i<n3;i++) fxp[i]+=coords->fxt[i];
        }
        rmass(coords->natoms,coords->fxa,coords->fya,coords->fza,coords->amass);
      }else{
        for(i=0;i<n3;i++) fxp[i]+=coords->fxt[i];  /* in de COMcounterforce plan add so3 force to the rest */
        rmassselect(coords->iselect,coords->nselect,coords->fxa,coords->fya,coords->fza,coords->amass);
      }
      
      for(i=0;i<coords->nschizo;i++){
        ii = coords->ischizo[i];
        vbuf[ii] += 0.5*(vdsnow[ii]+vdsprev[ii])*dr[ii];
      }
      for(i=0;i<coords->natoms;i++) dr[i] = 0.;
      myclockoff(timing,"FORCE BONDS BENDS",5);
      break;
      
      case 2: /* (torsion & 1-4) molecular */
      myclockon(timing,"FORCE TORSION ONE-FOUR",6);
      dr = simparms->ddr_tors + (1-coords->lcg)*simparms->ncg;
      vdsnow = simparms->vds_tors + (1-coords->lcg)*simparms->ncg; 
      vdsprev = simparms->vds_torsprev + (1-coords->lcg)*simparms->ncg; 
      for(i=0;i<coords->natoms;i++) vdsprev[i] = vdsnow[i];
      for(i=0;i<coords->natoms;i++) vdsnow[i] = 0.;    /* since I only use the HR part, I don' have to zero really */  
      
      fxp = coords->fxr; 
      if(coords->lcg){
        for(i=0;i<n3;i++) fxp[i] = 0.;
      }else{
        fyp = coords->fyr; fzp = coords->fzr;
        isel = coords->iselect;
        for(i=0;i<coords->nselect;i++){
          ii=isel[i]; 
          fxp[ii]=fyp[ii]=fzp[ii]=0.;
        }
      }
      
#ifndef TWO
      ftors_hr(simparms,coords);
      fonfo_hr(simparms,coords); 
      if(simparms->iextern==1) force_extern(simparms,coords,inter->ntable);
#ifdef PARA
      rec_buf = coords->scr_rec;
      for(i=0;i<n3;i++) rec_buf[i] = 0.;
      MPI_Allreduce(fxp,rec_buf,n3,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
      for(i=0;i<n3;i++) fxp[i] = rec_buf[i];
#endif
#endif
      if(coords->lcg){
        rmass(coords->natoms,coords->fxr,coords->fyr,coords->fzr,coords->amass);
      }else{
        rmassselect(coords->iselect,coords->nselect,coords->fxr,coords->fyr,coords->fzr,coords->amass);
      }
      for(i=0;i<coords->nschizo;i++){
        ii = coords->ischizo[i];
        vbuf[ii] += 0.5*(vdsnow[ii]+vdsprev[ii])*dr[ii];
      }
      for(i=0;i<coords->natoms;i++) dr[i] = 0.;
      myclockoff(timing,"FORCE TORSION ONE-FOUR",6);
      break;
      
      case 4: /* SO3  */
      fxp = coords->fxa;
      if(coords->lcg){
        md_error("errror with SO3 forces in force_hybrid() case 4");
      }else{
        fyp = coords->fya; fzp = coords->fza;
        isel = coords->inactive;
        for(i=0;i<coords->ninactive;i++){ /* let's just try the CG region for now */
          ii=isel[i];
          fxp[ii]=fyp[ii]=fzp[ii]=0.;
        }
      }
      
      fbond_so3(simparms,coords,so3,1);
      fbend_so3(simparms,coords,so3,1);
      fbondx_so3(simparms,coords,so3,1);
      ftors_so3(simparms,coords,so3,1);
      fonfo_so3(simparms,coords,so3,1); 
      break;
      
      default:
      fprintf(stderr,"ERROR: in forces!!!????\n");
      exit(1);
  }
  if(simparms->nfreeze>0) freeze_atom(simparms,coords);
  if(simparms->iensemble==3) {
    fxp = coords->vx;
    for(i=0;i<n3;i++) fxp[i] = 0.;
  }
  
#ifdef PRINT_FORCE
  printf("force iflag = %d\n",iflg);
  for(i=0;i<n3;i++) printf("force %d %g\n",i,fxp[i]);
#endif
}
/*----------------------------------------------------------------*/
void getvireal(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies,
               INTER *inter,NGBR *ngbr,SO3 *so3,int iflg)
{
  int i;
  double rvol;
  
  energies->vinter = energies->vhinter = 0.0;
  energies->velec = energies->vewald = 0.0;
  energies->vbond = energies->vbend = energies->vtors = energies->vbondx = 0.;
  energies->vonfo = energies->vonfo_e = 0.0;
  energies->vextern = energies->vextern_e = 0.0;
  energies->winter = energies->welec = energies->wewald  = 0.0;
  energies->wbond = energies->wbend = energies->wtors = energies->wbondx = 0.;
  energies->wonfo = energies->wonfo_e = 0.0;
  energies->wextern = energies->wextern_e = 0.0;
  energies->US = energies->U = energies->UI = 0.0;
  energies->W = energies->WI = 0.0;
  for(i=0;i<9;i++) {
    energies->wbondxtensor[i]=0.0;
    energies->WItensor[i]= energies->Wtensor[i]=0.0;
    energies->wbondtensor[i]= energies->wbendtensor[i]=0.0;
    energies->wonfotensor[i]= energies->wonfo_etensor[i]=0.0;
    energies->wtorstensor[i]= energies->wintertensor[i]=0.0;
    energies->welectensor[i]= energies->wewaldtensor[i]=0.0;
  }
  
  if(iflg==1) {
    getsvirter(&energies->vhinter,&energies->vinter,&energies->winter,
               energies->wintertensor,coords->lcg);
    getsvirter_e(&energies->velec,&energies->welec,energies->welectensor,coords->lcg);
  } else {
    getvirter(simparms,coords,inter,ngbr,&energies->vhinter,&energies->vinter,
              &energies->winter,energies->wintertensor);
    getvirter_e(simparms,coords,inter,ngbr,&energies->velec,&energies->welec,
                energies->welectensor,iflg);
  }
  
  getvbond(simparms,coords,
           &energies->vbond,&energies->wbond,energies->wbondtensor);
//  if((coords->lcg == 0) && (simparms->icalc_type==5)){
//    get_relat_aapos(coords->next,simparms->icglist,simparms->napcg, coords->next->ischizo,coords->next->ninactive+coords->next->nschizo);
//    energies->vbond += getvbond_so3(simparms,coords,so3,1);
//    energies->vbond += getvnonbond_so3(simparms,coords,inter,so3,ngbr->nprr,ngbr->inpr,ngbr->jnpr,0);
//    get_abs_aapos(coords->next,simparms->icglist,simparms->napcg, coords->next->ischizo,coords->next->ninactive+coords->next->nschizo);
//  }
  getvbend(simparms,coords,
           &energies->vbend,&energies->wbend,energies->wbendtensor);
  getvtors(simparms,coords,
           &energies->vtors,&energies->wtors,energies->wtorstensor);
  getvonfo(simparms,coords,
           &energies->vonfo,&energies->vonfo_e,&energies->wonfo,
           &energies->wonfo_e,energies->wonfotensor,energies->wonfo_etensor);
  getvbondx(simparms,coords,
            &energies->vbond,&energies->wbond,energies->wbondxtensor);
  
  if(simparms->iextern == 1) {
    getvextern(simparms,coords,&energies->vextern,&energies->wextern,
               &energies->vextern_e,&energies->wextern_e,inter->ntable);
  }
  
  if(coords->iperd == 3) {
    getvewald(simparms,coords,&energies->vewald,&energies->wewald,
              energies->wewaldtensor);
  }
  
  /* SO3 rotation dynamics */
  if(simparms->iso3dyn && !coords->lcg){
    energyso3(simparms,coords,energies,inter,ngbr,so3);
  }
  
#ifdef CHECK
  ptensorchk(simparms,coords,inter,ngbr,iflg);
  ngradv(simparms,coords,inter,ngbr,iflg);
#endif
  
  energies->US = (energies->vhinter + energies->velec + energies->vewald +
                  energies->vextern + energies->vextern_e);
  energies->U  = (energies->vinter  + energies->velec + energies->vewald + 
                  energies->vextern + energies->vextern_e);
  energies->W  = (energies->winter  + energies->welec + energies->wewald +
                  energies->wextern + energies->wextern_e);
  energies->UI = (energies->vbond + energies->vbend + energies->vtors + 
                  energies->vonfo + energies->vonfo_e + energies->vbondx);
  energies->WI = (energies->wbond + energies->wbend + energies->wtors + 
                  energies->wonfo + energies->wonfo_e + energies->wbondx);
  
  for(i=0;i<9;i++) {    
    energies->WItensor[i]=(energies->wbondtensor[i]+energies->wbendtensor[i]+
                           energies->wtorstensor[i]+energies->wonfotensor[i]+
                           energies->wonfo_etensor[i]+
                           energies->wbondxtensor[i]);
    energies->Wtensor[i]=(energies->wintertensor[i]+energies->welectensor[i]+
                          energies->wewaldtensor[i]);
  }
  /* add external field to the z compent of the pressure tensor */
  energies->Wtensor[8] += energies->wextern + energies->wextern_e;
  
  /* add long range correction if iperd==3 */
  if(coords->iperd == 3){
    i = 1-energies->lcg;
    rvol = 1./get_deth(coords->hmat);
    energies->W -= DIM*simparms->wlrc[i]*rvol;
    energies->Wtensor[0] -= simparms->wlrc[i]*rvol;
    energies->Wtensor[4] -= simparms->wlrc[i]*rvol;
    energies->Wtensor[8] -= simparms->wlrc[i]*rvol;
  }
  
  
#ifdef DEBUG
#ifdef KCALS
  printf("bonds = %g Kcal/mol\n",energies->vbond/KCAL);
  printf("Xbonds = %g Kcal/mol\n",energies->vbondx/KCAL);
  printf("bends = %g Kcal/mol\n",energies->vbend/KCAL);
  printf("tors  = %g Kcal/mol\n",energies->vtors/KCAL);
  printf("onfo  = %g Kcal/mol\n",energies->vonfo/KCAL);
  printf("onfo_e= %g Kcal/mol\n",energies->vonfo_e/KCAL);
  printf("inter = %g Kcal/mol\n",energies->vinter/KCAL);
  printf("elec  = %g Kcal/mol\n",energies->velec/KCAL);
  printf("recip = %g Kcal/mol\n",energies->vewald/KCAL);
#else
  printf("vhinter=%g vinter=%g velec=%g vrecip=%g\n",
         energies->vhinter,energies->vinter,energies->velec,energies->vewald);
  printf("vbond=%g vbondx=%g vbend=%g vtors=%g vonfo=%g vonfo_e=%g\n",
         energies->vbond,energies->vbondx,energies->vbend,energies->vtors,
         energies->vonfo,energies->vonfo_e);
  printf("winter=%g welec=%g wrecip=%g\n",
         energies->winter/ECONV,energies->welec/ECONV,energies->wewald/ECONV);
  printf("wbond=%g wbondx=%g wbend=%g wtors=%g wonfo = %g wonfo_e=%g\n",
         energies->wbond/ECONV,energies->wbondx/ECONV,energies->wbend/ECONV,energies->wtors/ECONV,
         energies->wonfo/ECONV,energies->wonfo_e/ECONV);
#endif
  //  exit(1);
#endif
}
/*----------------------------------------------------------------*/
void geteng(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies)
{
  int i,*isel,ndof,ndoffixed;
  double p2m,vol,pv_tensor[9],p2m_tensor[9],*vbuf;
  
  for(i=0;i<9;i++)  pv_tensor[i]= p2m_tensor[i]=0.0;
  
  vol = get_deth(coords->hmat);
  
  energies->poth  = energies->US;
  energies->poter = energies->U;
  energies->potra = energies->UI;
  
  ndoffixed = 3*simparms->natoms-simparms->ndof;
  
  if(simparms->iperd == 3){
    i = 1 - energies->lcg;
    energies->poter += simparms->vlrc[i]/vol; 
    energies->poth  += simparms->vlrc[i]/vol;
  }
  
  energies->pot_inter = energies->vinter;
  energies->pot_bond  = energies->vbond + energies->vbondx;
  energies->pot_bend  = energies->vbend;
  energies->pot_tors  = energies->vtors;
  energies->pot_onfo  = energies->vonfo;
  energies->pot_onfo_e= energies->vonfo_e;
  energies->pot_elec  = energies->velec;
  energies->pot_recip = energies->vewald;
  energies->pot_extern= energies->vextern;
  energies->pot_extern_e= energies->vextern_e;
  
  
  if(simparms->icalc_type == 5){
    if(energies->lcg){
      get_cgvel(coords,simparms->icglist,simparms->napcg,coords->ninactive+coords->nschizo,coords->ischizo);
      energies->tiout_cg = get_temp_select(coords->nselect-coords->nschizo,coords->vx,coords->vy,coords->vz,
                                           coords->amass,coords->iselect,ndoffixed);
      energies->tiout_aa = get_temp_select(coords->ninactive,coords->vx,coords->vy,coords->vz,
                                           coords->amass,coords->inactive,ndoffixed);
      get_p2mt_select(coords->nselect-coords->nschizo,coords->amass,coords->vx,coords->vy,coords->vz,
                      p2m_tensor,coords->iselect);
      p2m = p2m_tensor[0]+p2m_tensor[4]+p2m_tensor[8];
      ndof= MAX(3,3*(coords->nselect-coords->nschizo)-ndoffixed);
      energies->tiout = p2m/(double)ndof;
      energies->zkecg = 0.5*simparms->p2mbuf;
      isel = coords->iselect;
      vbuf = simparms->vcgintern;   /* add internal energy of CG units in CG region */
      energies->vcgtra = 0.;
      for(i=0;i<(coords->nselect-coords->nschizo);i++) energies->vcgtra += vbuf[isel[i]];
      vbuf = simparms->vbuf;
    }else{
      energies->tiout_aa = get_temp_select(coords->nselect-coords->nschizo,coords->vx,coords->vy,
                                           coords->vz,coords->amass,coords->iselect,ndoffixed);
      get_abs_aavel(coords->next,simparms->icglist,simparms->napcg,coords->next->iselect,
                    simparms->ncg-coords->next->nschizo);
      energies->tiout_cg = get_temp_select(coords->ninactive,coords->vx,coords->vy,coords->vz,
                                           coords->amass,coords->inactive,ndoffixed);
      get_relat_aavel(coords->next,simparms->icglist,simparms->napcg,coords->next->iselect,
                      simparms->ncg-coords->next->nschizo);
      get_p2mt_select(coords->nselect,coords->amass,coords->vx,coords->vy,coords->vz,
                      p2m_tensor,coords->iselect);
      p2m = p2m_tensor[0]+p2m_tensor[4]+p2m_tensor[8];
      ndof= MAX(3,3*coords->nselect-ndoffixed);
      energies->tiout = p2m/(double)ndof;
      vbuf = simparms->vbuf+simparms->ncg;
    }
    for(i=0;i<coords->natoms;i++) energies->vcgter += vbuf[i]; /* I can probably get rid of simparms.vbuf[natoms]*/
    for(i=0;i<coords->natoms;i++) vbuf[i]=0.;
    
    /* energies->poth += energies->vcgter; */
    energies->tiout_hr = get_temp_select(coords->nschizo,coords->vx,coords->vy,coords->vz,
                                         coords->amass,coords->ischizo,ndoffixed);
  }else{
    get_p2mt(coords->natoms,coords->amass,coords->vx,coords->vy,coords->vz,p2m_tensor);
    p2m = p2m_tensor[0]+p2m_tensor[4]+p2m_tensor[8];
    ndof= MAX(3,3*coords->natoms-ndoffixed);
    energies->tiout = p2m/(double)ndof;
  }
  energies->zke   = .5*p2m;  
  
  energies->prsi  = (p2m+energies->W+energies->WI)/(DIM*vol);
  for(i=0;i<9;i++){
    pv_tensor[i]=p2m_tensor[i]+energies->Wtensor[i]+energies->WItensor[i];
  }
  for(i=0;i<9;i++) energies->prs_tensor[i] = pv_tensor[i]/vol;
  
#ifdef DEBUG
  printf("pressure components %g %g %g %g,",p2m,energies->W,energies->WI,vol);
  printf(" pressure now = %g == %g\n",energies->prsi,
         (pv_tensor[0]+pv_tensor[4]+pv_tensor[8])/(3.*vol));
#endif
  
}
/*--------------------------------------------------------------------*/

void freeze_atom(SIMPARMS *simparms,COORDS *coords)
{
  int i,j;
  
  for(i=0;i<simparms->nfreeze;i++){
    j = coords->ifreeze[i];
    coords->vx[j] = coords->vy[j] = coords->vz[j] = 0.;
    coords->fxs[j] = coords->fys[j] = coords->fzs[j] = 0.;
    coords->fxl[j] = coords->fyl[j] = coords->fzl[j] = 0.;
    coords->fxa[j] = coords->fya[j] = coords->fza[j] = 0.;
    coords->fxr[j] = coords->fyr[j] = coords->fzr[j] = 0.;
    coords->fxt[j] = coords->fyt[j] = coords->fzt[j] = 0.;
  }
}
/*--------------------------------------------------------------------*/
void get_p2mt(int natoms,double amass[],double vx[],double vy[],double vz[],
              double p2m_tensor[9])
{
  int i;
  double am;
  for(i=0;i<9;i++) p2m_tensor[i] = 0.;
#include "vectorize.h"
  for(i=0;i<natoms;i++){
    am = amass[i];
    
    p2m_tensor[0] += am*(vx[i]*vx[i]);
    p2m_tensor[4] += am*(vy[i]*vy[i]);
    p2m_tensor[8] += am*(vz[i]*vz[i]);
    
    p2m_tensor[1] += am*(vy[i]*vx[i]);
    p2m_tensor[2] += am*(vz[i]*vx[i]);
    p2m_tensor[5] += am*(vz[i]*vy[i]);
  }
  p2m_tensor[3] = p2m_tensor[1];
  p2m_tensor[6] = p2m_tensor[2];
  p2m_tensor[7] = p2m_tensor[5];
}

/*--------------------------------------------------------------------*/

void get_p2mt_select(int nselect,double amass[],double vx[],double vy[],double vz[],
                     double p2m_tensor[9],int *iselect)
{
  int i,j;
  double am;
  for(i=0;i<9;i++) p2m_tensor[i] = 0.;
#include "vectorize.h"
  for(j=0;j<nselect;j++){
    i = iselect[j];
    am = amass[i];
    
    p2m_tensor[0] += am*(vx[i]*vx[i]);
    p2m_tensor[4] += am*(vy[i]*vy[i]);
    p2m_tensor[8] += am*(vz[i]*vz[i]);
    
    p2m_tensor[1] += am*(vy[i]*vx[i]);
    p2m_tensor[2] += am*(vz[i]*vx[i]);
    p2m_tensor[5] += am*(vz[i]*vy[i]);
  }
  
  p2m_tensor[3] = p2m_tensor[1];
  p2m_tensor[6] = p2m_tensor[2];
  p2m_tensor[7] = p2m_tensor[5];
}

/*--------------------------------------------------------------------*/

void get_cgpot_intern(SIMPARMS *simparms, COORDS *coords, INTER *inter){
  
  int i,icg,nsel,*isel;
  double *vcgtra=simparms->vcgintern,*fxp;
  double vbond=0.,vbend=0.,vbondx=0.,vtors=0.,vonfo=0.,vonfo_e=0.;
  
  /* Internal Kinetic Energy */
  nsel = coords->next->nselect - coords->next->nschizo;
  isel = coords->next->iselect;
  get_abs_aavel(coords->next,simparms->icglist,simparms->napcg,isel,nsel);
  for(i=0;i<nsel;i++){
    icg = isel[i];
    simparms->p2mbuf += get_inherent_p2m(simparms->napcg[icg],simparms->icglist[icg],
                                         coords->vx,coords->vy,coords->vz,coords->amass);
  }
  get_relat_aavel(coords->next,simparms->icglist,simparms->napcg,isel,nsel);
  
  for(i=0;i<simparms->ncg;i++) vcgtra[i] = 0.;
  
  /* Since I have no selection of inactive interactions, the energies (and forces) are  */
  /* computed for all particles. All not part of the CGR will of course be averwritten. */
    
  /* Internal Bond, Bend and Cross Bond Energy */
  if(coords->fxa!=NULL){
    fxp = coords->fxa;
    for(i=0;i<3*coords->natoms;i++) fxp[i] = 0.;  
    vbond = getvbond_intern(simparms,coords);
    vbend = getvbend_intern(simparms,coords);
    vbondx = getvbondx_intern(simparms,coords);
    rmass(coords->natoms,coords->fxa,coords->fya,coords->fza,coords->amass);
  }
  
  /* Internal Torsion and OneFour Energy */
  if(coords->fxr!=NULL){
    fxp = coords->fxr;
    for(i=0;i<3*coords->natoms;i++) fxp[i] = 0.;  
    vtors = getvtors_intern(simparms,coords);
    getvonfo_intern(simparms,coords,&vonfo,&vonfo_e);
    rmass(coords->natoms,coords->fxr,coords->fyr,coords->fzr,coords->amass);
  }
  
  printf("Calculating initial internal energies of all CG units (%d in CGR):\n",nsel);
  printf(" - Ekin       = %12.6lf [kcal/mol]\n",0.5*simparms->p2mbuf/KCAL);
  printf(" - Vbond      = %12.6lf [kcal/mol]\n",vbond/KCAL);
  printf(" - Vbend      = %12.6lf [kcal/mol]\n",vbend/KCAL);
  printf(" - Vbondx     = %12.6lf [kcal/mol]\n",vbondx/KCAL);
  printf(" - Vtors      = %12.6lf [kcal/mol]\n",vtors/KCAL);
  printf(" - Vonefour   = %12.6lf [kcal/mol]\n",vonfo/KCAL);
  printf(" - Vonefour_e = %12.6lf [kcal/mol]\n",vonfo_e/KCAL);
}

/*--------------------------------------------------------------------*/

