/* 
 Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore
 
 Dr. Preston B. Moore
 Associate Director, Center for Molecular Modeling (CMM)
 University of Pennsylvania, Department of Chemistry, Box 188 
 231 S. 34th St. Philadelphia, PA 19104-6323 USA
 EMAIL: moore@cmm.chem.upenn.edu  
 WWW: http://www.cmm.upenn.edu/~moore
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 */

/* subroutine to do up the verlist */

/*----------------------------------------------------------------*/

#include "md.h"

/* #define DEBUG */
/* #define VERBOSE */

/*----------------------------ngbr_list.c proto's-----------------------*/

void ngbr_list(int,int *,int *,int,COORDS *,INTER *,NGBR *,int *,int *,int);
void ngbr_onelist(int,int *,int *,int,COORDS *,INTER *,NGBR *,int *,int *,int *,int *,int *,int);
void cnt_ngbr_list(int,int *,int *,int,COORDS *,INTER *,int *,int *,int);
void cnt_ngbr_onelist(int,int *,int *,int,COORDS *,INTER *,int *,int *,int *,int *,int);
void count_vdw(SIMPARMS *,COORDS *,INTER *,NGBR *,int *,int *,int *,int *);

/*-----------------------------------------------------------------------*/
/* compute for each CG-CG pair interaction type the maximum cutoff for   */
/* atomistic interactions in that pair including the distance of the     */
/* atoms from the CG centers of mass                                     */
#define BLOBSIZE 2.5      /* max distance of atom from CG center of mass */
void get_aarmax_rblob(SIMPARMS *simparms,INTER *inter){

  int iat, jat, icg, jcg, itype, jtype, k, it;
  int *iaatypes = inter->next->itype;   /* list of AA atomtypes */
  int *icgtypes = inter->itype;     /* list of atomtypes */
  int ntype = inter->ntype;         /* number of interaction types */
  int **map = inter->map;           /* matrix of interaction types */
  int **aamap = inter->next->map;   /* matrix of interaction types */
  double radius, rmax;

  /* allocate and add -1 element set to zero so we can skip checking null interactions */
  inter->rmax_rblob  = (double *)cmalloc((ntype+1)*sizeof(double));
  inter->rmax_rblob++;
  inter->rmax_rblob[-1] = 0.0;

  /* find a pair of CG atoms for each interaction type */
  for(it=0; it<ntype; it++){
    k = 0;
    while( map[0][k] != it ) k++;
    itype = k / simparms->ncgtypes;
    jtype = k % simparms->ncgtypes;
    
    /* find atomistic interactions in each CG-CG pair */
    icg = jcg = 0;
    rmax = 0.;
    while( icgtypes[icg] != itype ) icg++; 
    while( icgtypes[jcg] != jtype ) jcg++;
    
    for(iat=0; iat<simparms->napcg[icg]; iat++){
      for(jat=0; jat<simparms->napcg[jcg]; jat++){
        k = aamap[iaatypes[iat]][iaatypes[jat]];
        if(k == -1){
          radius = 0.;
        }else{
          radius = inter->next->rmaxl_skin2[k];    
          if(radius > rmax) rmax = radius;
        }
      }
    }
    radius = sqrt(rmax) + 2.0*BLOBSIZE;           /* use general BLOBSIZE for now */
    inter->rmax_rblob[it] = radius * radius;
  }
  
  
  /* some feedback */
  printf("Computing atomistic VdW cutoff between CG blobs (blobsize is set to %lf)\n",BLOBSIZE);
  for(it=0; it<ntype; it++)
    printf("  CG VdW type[%d]: rmaxl_skin2= %lf, rmax_rblob= %lf\n",it,inter->rmaxl_skin2[it],inter->rmax_rblob[it]);


  return;
}

/*-----------------------------------------------------------------------*/

void get_ngbr(SIMPARMS *simparms,COORDS *coords,INTER *inter,NGBR *ngbr)
{
  int npso,npto,noffs,nofft;  /* moved from static int to ngbr struct to allow multi ngbr lists*/
  int npao,npro,noffa,noffr;  /* CG-atomistic list and SO3 Rotational Dynamics list */
  int i,j,k,ii,jj,icell,jcell,mem_change,mem_add;
  int *inp,*jnp,*excl,nexcl,*itype,*map;
  int ioffs,iofft,ioffa,ioffr,icell_old;
  int *ibreak,nbreak,*ilist,*jlist,iat,jat,icg,jcg,ipair;
  int ncg, napcg, *isel,*inp_cg,*jnp_cg,npr_cg;
  int *iaacg=simparms->iaacg;
  long ibegin,iend;
  double rdis2,*dp,*x;
  LINE line;
  
  (ngbr->update)++;
  npso = ngbr->npso;
  npto = ngbr->npto;
  noffs = ngbr->noffs;
  nofft = ngbr->nofft;
  
  npao = ngbr->npao;
  npro = ngbr->npro;
  noffa = ngbr->noffa;
  noffr = ngbr->noffr;
  mem_change = simparms->mem_bytes;
  mem_add = MAX(simparms->nlen,coords->natoms);
  
#ifdef DEBUG
  printf("finding neighbors! \n");
#endif
  
  /* first time we come in here */
  if(ngbr->update == 0){
    allocate_lists(coords->natoms,simparms->nlen,ngbr,coords,&i);
    simparms->mem_bytes += i;
    
    rdis2 = 0;
    for(i=0;i<inter->ntype;i++){
      if(inter->rmaxl_skin2[i]>rdis2) rdis2 = inter->rmaxl_skin2[i];
    }
    if(ngbr->onelist == 2){
      /* using CG list only, consider also the max atomistic cutoff */
      get_aarmax_rblob(simparms,inter);
      for(i=0;i<inter->ntype;i++){
        if(inter->rmax_rblob[i]>rdis2) rdis2 = inter->rmax_rblob[i];
      }
    }
    rdis2 = sqrt(rdis2)+inter->skin_ter;
//    if( (simparms->nocell_all==1) && (ngbr->onelist!=0) ){ /* if onelink also make AA celllist for electrostatics */
    if( simparms->nocell_all==1 ){
      set_link(simparms,ngbr,coords->hmat,rdis2,
               sqrt(inter->rmaxl_skin2_e)+inter->skin_ter);
    }
    coords->cell_old = cbrt(get_deth(coords->hmat));
    
    /* note that when ngbr->onelist==0 (AA ngbrlist made from CG list) I am not calling count_vdw... */
    switch(ngbr->ilist){
      case 1: /* ngbr list */
        npso = npto = npao = npro = 0;
        if( ngbr->onelist != 0 ) count_vdw(simparms,coords,inter,ngbr,&npso,&npto,&npao,&npro);
        npso = 1.2*npso + 1;
        npto = 1.2*npto + 1;   /* add safety margin */
        ngbr->inps = (int *)cmalloc(npso*sizeof(int));
        ngbr->jnps = (int *)cmalloc(npso*sizeof(int));
        ngbr->inpt = (int *)cmalloc(npto*sizeof(int));
        ngbr->jnpt = (int *)cmalloc(npto*sizeof(int));        
        simparms->mem_bytes += 2*npso*sizeof(int);
        simparms->mem_bytes += 2*npto*sizeof(int);
        
        noffs = nofft = coords->natoms+1;
        ngbr->offs= (int *)cmalloc(noffs*sizeof(int));
        ngbr->offt= (int *)cmalloc(nofft*sizeof(int));
        simparms->mem_bytes += noffs*sizeof(int);
        simparms->mem_bytes += nofft*sizeof(int);

//        if(ngbr->onelist == 2){
          npao = 1.2*npao + 1;
          npro = 1.2*npro + 1;
          ngbr->inpa = (int *)cmalloc(npao*sizeof(int));
          ngbr->jnpa = (int *)cmalloc(npao*sizeof(int));
          ngbr->inpr = (int *)cmalloc(npro*sizeof(int));
          ngbr->jnpr = (int *)cmalloc(npro*sizeof(int));
          simparms->mem_bytes += 2*npao*sizeof(int);
          simparms->mem_bytes += 2*npro*sizeof(int);

          noffa = noffr = 1;
          if(ngbr->onelist == 2) noffa = noffr = coords->natoms+1;
          ngbr->offa= (int *)cmalloc(noffa*sizeof(int));
          ngbr->offr= (int *)cmalloc(noffr*sizeof(int));
          simparms->mem_bytes += noffa*sizeof(int);
          simparms->mem_bytes += noffr*sizeof(int);
//        }
        break;
      case 2: /* link list only */
        noffs = nofft = simparms->natoms+mem_add;
        ngbr->offs= (int *)cmalloc(noffs*sizeof(int));
        ngbr->offt= (int *)cmalloc(nofft*sizeof(int));
        simparms->mem_bytes += noffs*sizeof(int);
        simparms->mem_bytes += nofft*sizeof(int);
        break;
      case 0:  /* nolist */
        noffs = nofft = 2;
        ngbr->offs= (int *)cmalloc(noffs*sizeof(int));
        ngbr->offt= (int *)cmalloc(nofft*sizeof(int));
        simparms->mem_bytes += noffs*sizeof(int);
        simparms->mem_bytes += nofft*sizeof(int);
        break;
    }
#ifdef DEBUG
    printf("allocated lists %d for first time noffs = %d, nofft = %d\n",
           ngbr->ilist,noffs,nofft);
#endif
  }      /* end of first time we come in here */
  
  
  ngbr->nprs = ngbr->nprt = ngbr->npra = ngbr->nprr = 0;
  ngbr->offs[0] = ngbr->offt[0] = ngbr->offa[0] = ngbr->offr[0] = 0;
  ioffs = iofft = ioffa = ioffr = 0;
  itype = inter->itype;
  
  /* no lists */
  switch(ngbr->ilist){
    case 0:                 /* nolist */
      break;
    case 2:                /* link list */
      if(fabs(coords->cell_old-cbrt(get_deth(coords->hmat))) >
         (inter->skin_ter/2.)){
        rdis2 = 0;
        for(i=0;i<inter->ntype;i++){
          if(inter->rmaxl_skin2[i]>rdis2) rdis2 = inter->rmaxl_skin2[i];
        }
        rdis2 = sqrt(rdis2)+inter->skin_ter;
        if(simparms->nocell_all==1){
          set_link(simparms,ngbr,coords->hmat,rdis2,
                   sqrt(inter->rmaxl_skin2_e)+inter->skin_ter);
        }
        coords->cell_old = cbrt(get_deth(coords->hmat));
      }
      break;
      case 1: /* verlist */
      ibreak = coords->ibreak;
      
      simparms->mem_bytes -= ((noffs+nofft+noffa+noffr) + 2*(npso+npto+npao+npro))*sizeof(int);
      
      inp = cmalloc(2*simparms->nlen*sizeof(int));
      jnp = inp + simparms->nlen;
      
      if(simparms->nocell_all==0){
        /* decompose i so that they is roughly the same number of
         i-j pairs on each processors */
        
        decomp_trig((long)coords->natoms,simparms->size,simparms->rank,
                    &ibegin,&iend);
        
        nbreak = k = 0;
        ibreak[0] = 0;
        for(i=ibegin;i<iend;i++){
          excl = &(inter->exclude[i][0]);
          nexcl = *(excl);
          if(npso < ngbr->nprs + mem_add){
            npso += mem_add;
            ngbr->inps = (int *)realloc(ngbr->inps,npso*sizeof(int));
            ngbr->jnps = (int *)realloc(ngbr->jnps,npso*sizeof(int));
          }
          if(npto < ngbr->nprt + mem_add){
            npto += mem_add;
            ngbr->inpt = (int *)realloc(ngbr->inpt,npto*sizeof(int));
            ngbr->jnpt = (int *)realloc(ngbr->jnpt,npto*sizeof(int));
          }
          
          map = inter->map[itype[i]];
          for(j=i+1;j<coords->natoms;j++){
            if(nexcl == j){ 
              nexcl = *(++excl);
            } else if(map[itype[j]] == -1){
              /* skip null interactions */
            } else {
              inp[k] = i;
              jnp[k] = j;
              k++;
              if(k==simparms->nlen){
                ibreak[1] = k+1; /* don't set any breakpoints */
                ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                          &ioffs,&iofft,simparms->ivol);
                nbreak = k=0;
              }
            }
          }
          if(ibreak[nbreak] != k) ibreak[++nbreak] = k;
          ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                    &ioffs,&iofft,simparms->ivol);
          if(ngbr->offs[ioffs]!=ngbr->nprs) ngbr->offs[++ioffs] = ngbr->nprs; 
          if(ngbr->offt[iofft]!=ngbr->nprt) ngbr->offt[++iofft] = ngbr->nprt;
          nbreak = k = 0;
        }
      } else {
        if(fabs(coords->cell_old-cbrt(get_deth(coords->hmat))) >
           (inter->skin_ter/2.)){
          rdis2 = 0;
          for(i=0;i<inter->ntype;i++){
            if(inter->rmaxl_skin2[i]>rdis2) rdis2 = inter->rmaxl_skin2[i];
          }
          rdis2 = sqrt(rdis2)+inter->skin_ter;
          set_link(simparms,ngbr,coords->hmat,rdis2,
                   sqrt(inter->rmaxl_skin2_e)+inter->skin_ter);
          coords->cell_old = cbrt(get_deth(coords->hmat));
        }
        
        ngbr->offs[0] = ngbr->offt[0] = ngbr->offa[0] = ngbr->offr[0] = 0;
        icell_old = -1;
        ibreak[0] = 0;
        nbreak = 0;
        k = 0;
        
        if(ngbr->onelist == 0){   /* use CG list to build AA list */
          /* first add AA interactions within each CG blob that belongs to the AR+HR */
          isel = coords->next->ischizo;
          ncg = coords->next->ninactive + coords->next->nschizo;
          for(i=0; i<ncg; i++){
            icg = isel[i];
            ilist = simparms->icglist[icg];
            napcg = simparms->napcg[icg];

            /* check offset memory */
            mem_add = MAX(simparms->nlen,napcg*napcg);
            if(noffs < ioffs + mem_add){
              noffs += mem_add;
              ngbr->offs = (int *)realloc(ngbr->offs,noffs*sizeof(int));
            }
            if(nofft < iofft + mem_add){
              nofft += mem_add;
              ngbr->offt = (int *)realloc(ngbr->offt,nofft*sizeof(int));
            }
            /* check neighbor list  memory */
            if(npso < ngbr->nprs + mem_add){
              npso += mem_add;
              ngbr->inps = (int *)realloc(ngbr->inps,npso*sizeof(int));
              ngbr->jnps = (int *)realloc(ngbr->jnps,npso*sizeof(int));
            }
            if(npto < ngbr->nprt + mem_add){
              npto += mem_add;
              ngbr->inpt = (int *)realloc(ngbr->inpt,npto*sizeof(int));
              ngbr->jnpt = (int *)realloc(ngbr->jnpt,npto*sizeof(int));
            }
            for(iat=0;iat<napcg-1;++iat){
              ii = ilist[iat];
              map = inter->map[itype[ii]];
              for(jat=iat+1;jat<napcg;++jat){
                jj = ilist[jat];
                if(map[itype[jj]] == -1){
                  /* skip null interaction */
                } else if(!exclij(ii,jj,inter->nexclude,inter->exclude)){
                  inp[k] = ii;
                  jnp[k] = jj;
                  k++;
                  if(k==simparms->nlen){
                    ibreak[++nbreak] = k;
                    ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                              &ioffs,&iofft,simparms->ivol);
                    nbreak = k = 0;
                  }
                }
              } /* loop over atoms jat in blob icg  */              
            } /* loop over atoms iat in blob icg  */
          } /* loop over all blobs icg for atomistic dynamics in AR+HR */
          /* now loop over all CG-CG pairs for atomistic dynamics in AR+HR */
          inp_cg = ngbr->next->inpa;
          jnp_cg = ngbr->next->jnpa;
          npr_cg = ngbr->next->npra;
          for(ipair=0;ipair<npr_cg;ipair++){
            icg = inp_cg[ipair];
            jcg = jnp_cg[ipair];
            ilist = simparms->icglist[icg];
            jlist = simparms->icglist[jcg];
            
            /* check offset memory */
            mem_add = MAX(simparms->nlen,simparms->napcg[icg]*simparms->napcg[jcg]);
            if(noffs < ioffs + mem_add){
              noffs += mem_add;
              ngbr->offs = (int *)realloc(ngbr->offs,noffs*sizeof(int));
            }
            if(nofft < iofft + mem_add){
              nofft += mem_add;
              ngbr->offt = (int *)realloc(ngbr->offt,nofft*sizeof(int));
            }
            /* check neighbor list  memory */
            if(npso < ngbr->nprs + mem_add){
              npso += mem_add;
              ngbr->inps = (int *)realloc(ngbr->inps,npso*sizeof(int));
              ngbr->jnps = (int *)realloc(ngbr->jnps,npso*sizeof(int));
            }
            if(npto < ngbr->nprt + mem_add){
              npto += mem_add;
              ngbr->inpt = (int *)realloc(ngbr->inpt,npto*sizeof(int));
              ngbr->jnpt = (int *)realloc(ngbr->jnpt,npto*sizeof(int));
            }
            for(iat=0;iat<simparms->napcg[icg];++iat){
              ii = ilist[iat];
              map = inter->map[itype[ii]];
              for(jat=0;jat<simparms->napcg[jcg];++jat){
                jj = jlist[jat];
                if(map[itype[jj]] == -1){
                  /* skip null interaction */
                } else if(!exclij(ii,jj,inter->nexclude,inter->exclude)){
                  inp[k] = ii;
                  jnp[k] = jj;
                  k++;
                  if(k==simparms->nlen){
                    ibreak[++nbreak] = k;
                    ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                              &ioffs,&iofft,simparms->ivol);
                    nbreak = k = 0;
                  }
                }
              } /* loop over atoms jat in blob jcg  */              
            } /* loop over atoms iat in blob icg  */
          } /* loop over all CG-CG pairs for atomistic dynamics in AR+HR */
          if(ibreak[nbreak] != k) ibreak[++nbreak] = k;
          ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                  &ioffs,&iofft,simparms->ivol);
          /* now loop over all CG-CG pairs for SO3 Rotational Dynamics in HR+CGR */
          k = 0;
          inp_cg = ngbr->next->inpr;
          jnp_cg = ngbr->next->jnpr;
          npr_cg = ngbr->next->nprr;
          for(ipair=0;ipair<npr_cg;ipair++){
            icg = inp_cg[ipair];
            jcg = jnp_cg[ipair];
            ilist = simparms->icglist[icg];
            jlist = simparms->icglist[jcg];
            
            /* check neighbor list memory */
            if(npro < k + mem_add){
              npro += mem_add;
              ngbr->inpr = (int *)realloc(ngbr->inpr,npro*sizeof(int));
              ngbr->jnpr = (int *)realloc(ngbr->jnpr,npro*sizeof(int));
            }
            for(iat=0;iat<simparms->napcg[icg];++iat){
              ii = ilist[iat];
              map = inter->map[itype[ii]];
              for(jat=0;jat<simparms->napcg[jcg];++jat){
                jj = jlist[jat];
                if(map[itype[jj]] == -1){
                  /* skip null interaction */
                } else if(!exclij(ii,jj,inter->nexclude,inter->exclude)){
                  ngbr->inpr[k] = ii;
                  ngbr->jnpr[k] = jj;
                  k++;
                }
              } /* loop over atoms jat in blob jcg  */              
            } /* loop over atoms iat in blob icg  */
          } /* loop over all CG-CG pairs for SO3 Rotational Dynamics */
          ngbr->nprr = k;

        }else if(ngbr->onelist == 2){   /* make initial CG lists for hybrid MD */
          get_lnklist(coords->natoms,ngbr->ncells,ngbr->hlist,ngbr->lnklist,
                    simparms->iperd,coords->hmat,coords->px,coords->py,coords->pz,
                    &simparms->iaacg[(1-coords->lcg)*simparms->ncg],ngbr->lcg);

          for(i=0;i<ngbr->npaircell;i++){
            icell = ngbr->icell[i];
            jcell = ngbr->jcell[i];
            
            ii = ngbr->hlist[icell];
            
            while(ii!=-1){
              if(icell==jcell){
                jj = ngbr->lnklist[ii];
              } else {
                jj = ngbr->hlist[jcell];
              }
              /* check offset memory */
              if(noffs < ioffs + mem_add){
                noffs += mem_add;
                ngbr->offs = (int *)realloc(ngbr->offs,noffs*sizeof(int));
              }
              if(nofft < iofft + mem_add){
                nofft += mem_add;
                ngbr->offt = (int *)realloc(ngbr->offt,nofft*sizeof(int));
              }
              if(noffa < ioffa + mem_add){
                noffa += mem_add;
                ngbr->offa = (int *)realloc(ngbr->offa,noffa*sizeof(int));
              }
              if(noffr < ioffr + mem_add){
                noffr += mem_add;
                ngbr->offr = (int *)realloc(ngbr->offr,noffr*sizeof(int));
              }
              /* check neighbor list  memory */
              if(npso < ngbr->nprs + mem_add){
                npso += mem_add;
                ngbr->inps = (int *)realloc(ngbr->inps,npso*sizeof(int));
                ngbr->jnps = (int *)realloc(ngbr->jnps,npso*sizeof(int));
              }
              if(npto < ngbr->nprt + mem_add){
                npto += mem_add;
                ngbr->inpt = (int *)realloc(ngbr->inpt,npto*sizeof(int));
                ngbr->jnpt = (int *)realloc(ngbr->jnpt,npto*sizeof(int));
              }
              if(npao < ngbr->npra + mem_add){
                npao += mem_add;
                ngbr->inpa = (int *)realloc(ngbr->inpa,npao*sizeof(int));
                ngbr->jnpa = (int *)realloc(ngbr->jnpa,npao*sizeof(int));
              }
              if(npro < ngbr->nprr + mem_add){
                npro += mem_add;
                ngbr->inpr = (int *)realloc(ngbr->inpr,npro*sizeof(int));
                ngbr->jnpr = (int *)realloc(ngbr->jnpr,npro*sizeof(int));
              }
              while(jj!=-1){
                inp[k] = ii;
                jnp[k] = jj;
                k++;
                if(k==simparms->nlen){
                  ibreak[++nbreak] = k;
                  ngbr_onelist(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                            simparms->iaacg,&ioffs,&iofft,&ioffa,&ioffr,simparms->ivol);
                  nbreak = k = 0;
                }
                jj = ngbr->lnklist[jj];
              } /* end while(jj!=-1) */
              
              ii = ngbr->lnklist[ii];
              if((icell==jcell) && (ibreak[nbreak]!=k)) ibreak[++nbreak] = k;
            } /* end while ii != -1 */
            if(ibreak[nbreak]!=k && icell_old!=icell) ibreak[++nbreak] = k;
            icell_old = icell;
          } /* for(ncell_pair) */
          if(ibreak[nbreak] != k) ibreak[++nbreak] = k;
          ngbr_onelist(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                  simparms->iaacg,&ioffs,&iofft,&ioffa,&ioffr,simparms->ivol);
          ngbr->offr[++(ioffr)] = ngbr->npra; 
          ngbr->offa[++(ioffa)] = ngbr->nprr;

        }else{ /* if not onelist==0 or onelist==2 */
          get_lnklist(coords->natoms,ngbr->ncells,ngbr->hlist,ngbr->lnklist,
                    simparms->iperd,coords->hmat,coords->px,coords->py,coords->pz,
                    &simparms->iaacg[(1-coords->lcg)*simparms->ncg],ngbr->lcg);

          for(i=0;i<ngbr->npaircell;i++){
            icell = ngbr->icell[i];
            jcell = ngbr->jcell[i];
            
            ii = ngbr->hlist[icell];
            
            while(ii!=-1){
              if(icell==jcell){
                jj = ngbr->lnklist[ii];
              } else {
                jj = ngbr->hlist[jcell];
              }
              /* check offset memory */
              if(noffs < ioffs + mem_add){
                noffs += mem_add;
                ngbr->offs = (int *)realloc(ngbr->offs,noffs*sizeof(int));
              }
              if(nofft < iofft + mem_add){
                nofft += mem_add;
                ngbr->offt = (int *)realloc(ngbr->offt,nofft*sizeof(int));
              }
              /* check neighbor list  memory */
              if(npso < ngbr->nprs + mem_add){
                npso += mem_add;
                ngbr->inps = (int *)realloc(ngbr->inps,npso*sizeof(int));
                ngbr->jnps = (int *)realloc(ngbr->jnps,npso*sizeof(int));
              }
              if(npto < ngbr->nprt + mem_add){
                npto += mem_add;
                ngbr->inpt = (int *)realloc(ngbr->inpt,npto*sizeof(int));
                ngbr->jnpt = (int *)realloc(ngbr->jnpt,npto*sizeof(int));
              }
              map = inter->map[itype[ii]];
              while(jj!=-1){
                if(map[itype[jj]] == -1){
                  /* skip null interaction */
                } else if(!exclij(ii,jj,inter->nexclude,inter->exclude)){
                  /* if hybrid CG-CG: exlude AA-AA, but not CG-AA */
                  if( !coords->lcg || (iaacg[ii]+iaacg[jj] != 0) ){ /* add all regions for now */
                    inp[k] = ii;
                    jnp[k] = jj;
                    k++;
                    if(k==simparms->nlen){
                      ibreak[++nbreak] = k;
                      ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                                &ioffs,&iofft,simparms->ivol);
                      nbreak = k = 0;
                    }
                  }
                }
                jj = ngbr->lnklist[jj];
              } /* end while(jj!=-1) */
              
              ii = ngbr->lnklist[ii];
              if((icell==jcell) && (ibreak[nbreak]!=k)) ibreak[++nbreak] = k;
            } /* end while ii != -1 */
            if(ibreak[nbreak]!=k && icell_old!=icell) ibreak[++nbreak] = k;
            icell_old = icell;
          } /* for(ncell_pair) */
          if(ibreak[nbreak] != k) ibreak[++nbreak] = k;
          ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                  &ioffs,&iofft,simparms->ivol);
        } /* if onelist or not */
        
      } /* if(simparms->ilist==1) */
      ngbr->offs[++(ioffs)] = ngbr->nprs; 
      ngbr->offt[++(iofft)] = ngbr->nprt;
      
#ifdef VERBOSE
      for(i=0;i<coords->natoms;i++){
        printf("offset %d %d %d\n",i,ngbr->offs[i],ngbr->offt[i]);
      }
      printf("Short - interactions\n");
      for(i=0;i<noffs;i++){
        printf("%d %d - %d\n",i,ngbr->inps[i],ngbr->jnps[i]);
      }
      printf("Total - interactions\n");
      for(i=0;i<nofft;i++){
        printf("%d %d - %d\n",i,ngbr->inpt[i],ngbr->jnpt[i]);
      }
#endif
      
      simparms->mem_bytes += 2*npso*sizeof(int);
      simparms->mem_bytes += 2*npto*sizeof(int);       
      simparms->mem_bytes += noffs*sizeof(int);
      simparms->mem_bytes += nofft*sizeof(int);
      
      simparms->mem_bytes += 2*npao*sizeof(int);
      simparms->mem_bytes += 2*npro*sizeof(int);       
      simparms->mem_bytes += noffa*sizeof(int);
      simparms->mem_bytes += noffr*sizeof(int);
      
      free(inp);
      break;
  }
  
  if(simparms->rank==0 && (mem_change-simparms->mem_bytes)!= 0 ){
    md_stdout("Van der Waals Neighbor list memory changed!");
    sprintf(line,"Bytes of memory allocated so far = %d",
            simparms->mem_bytes);
    md_stdout(line);
  }
  
  get_ngbr_e(simparms,coords,inter,ngbr);
  
  /* save old positions so that we know when to update */
#ifdef DEBUG
  printf("allocated lists noffs = %d, nofft = %d\n",ngbr->nprs,ngbr->nprt);
  if(simparms->rank==0) md_stdout("Storing old positions for lists");
#endif
  dp = inter->dxn;  x  = coords->px;
  k = 3*coords->natoms;
  ngbr->npso = npso;
  ngbr->npto = npto;
  ngbr->noffs = noffs;
  ngbr->nofft = nofft;

  ngbr->npao = npao;
  ngbr->npro = npro;
  ngbr->noffa = noffa;
  ngbr->noffr = noffr;
#include "vectorize.h"
  for(i=0;i<k;i++) dp[i] = x[i];
}
/*----------------------------------------------------------------*/
void ngbr_list(int nlen,int *inp,int *jnp,int iperd,
               COORDS *coords,INTER *inter,NGBR *ngbr,
               int *ioffs,int *iofft,int ivol)
{
  int j,intr,in,jn,k,*itype,**map;
  int nprs,nprt,*ips,*jps,*ipt,*jpt;
  double *xdis,*ydis,*zdis,*rdis2,*rs2,*rl2;
  double *x,*y,*z;
  
  xdis = coords->xdis;  ydis = coords->ydis;  zdis = coords->zdis;  
  x = coords->px;  y = coords->py;  z = coords->pz;
  itype = inter->itype;
  map = inter->map;
  ips = ngbr->inps;  jps = ngbr->jnps;
  ipt = ngbr->inpt;  jpt = ngbr->jnpt;
  
  rdis2 = (double *)malloc(3*nlen*sizeof(double));
  rs2 = rdis2 + nlen;
  rl2 = rdis2 + 2*nlen;
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    in = inp[j];    jn = jnp[j];
    xdis[j] = x[in] - x[jn];
    ydis[j] = y[in] - y[jn];
    zdis[j] = z[in] - z[jn];
  }
  
  period(nlen,xdis,ydis,zdis,coords->hmat,coords->hmati,iperd,ivol);
  
  k = 1;
  
#include "vectorize.h"
  for(j=0;j<nlen;j++){    
    rdis2[j] = xdis[j]*xdis[j]+ydis[j]*ydis[j]+zdis[j]*zdis[j];	    
    intr = map[itype[inp[j]]][itype[jnp[j]]];
    rs2[j] = inter->rmaxs_skin2[intr];
    rl2[j] = inter->rmaxl_skin2[intr];
  }
  nprs = ngbr->nprs;
  nprt = ngbr->nprt;
  for(j=0;j<nlen;j++){    
    /* use the fact that the long range force is ALWAYS longer */
    if( rdis2[j] < rl2[j]){
      ipt[nprt] = inp[j]; 
      jpt[nprt] = jnp[j]; 
      nprt++;
      if( rdis2[j] < rs2[j]) {
        ips[nprs] = inp[j];
        jps[nprs] = jnp[j]; 
        nprs++;
      }
    }
    if((j+1)==coords->ibreak[k]){
      if(ngbr->offs[*ioffs]!=nprs) ngbr->offs[++(*ioffs)] = nprs; 
      if(ngbr->offt[*iofft]!=nprt) ngbr->offt[++(*iofft)] = nprt;
      k++;
    }
  }
  ngbr->nprs = nprs;
  ngbr->nprt = nprt;
  free(rdis2);
}
/*----------------------------------------------------------------*/
void ngbr_onelist(int nlen,int *inp,int *jnp,int iperd,
               COORDS *coords,INTER *inter,NGBR *ngbr,
               int *iaacg,int *ioffs,int *iofft,int *ioffa,int *ioffr,int ivol)
{
  int icg, jcg, j,intr,in,jn,k,*itype,**map;
  int nprs,nprt,npra,nprr,*ips,*jps,*ipt,*jpt,*ipr,*jpr,*ipa,*jpa;
  double *xdis,*ydis,*zdis,*rdis2,*rs2,*rl2,*rso32,*raa2;
  double *x,*y,*z;
  
  xdis = coords->xdis;  ydis = coords->ydis;  zdis = coords->zdis;  
  x = coords->px;  y = coords->py;  z = coords->pz;
  itype = inter->itype;
  map = inter->map;
  ips = ngbr->inps;  jps = ngbr->jnps;
  ipt = ngbr->inpt;  jpt = ngbr->jnpt;
  ipr = ngbr->inpr;  jpr = ngbr->jnpr;
  ipa = ngbr->inpa;  jpa = ngbr->jnpa;
  
  rdis2 = (double *)malloc(5*nlen*sizeof(double));
  rs2 = rdis2 + nlen;
  rl2 = rdis2 + 2*nlen;
  rso32 = rdis2 + 3*nlen;
  raa2 = rdis2 + 4*nlen;
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    in = inp[j];    jn = jnp[j];
    xdis[j] = x[in] - x[jn];
    ydis[j] = y[in] - y[jn];
    zdis[j] = z[in] - z[jn];
  }
  
  period(nlen,xdis,ydis,zdis,coords->hmat,coords->hmati,iperd,ivol);
  
  k = 1;
  
#include "vectorize.h"
  /* null interactions are not yet excluded */
  /* instead the cutoff is set to zero (i.e. rmax_skin[-1] = 0) */
  for(j=0;j<nlen;j++){    
    rdis2[j] = xdis[j]*xdis[j]+ydis[j]*ydis[j]+zdis[j]*zdis[j];	    
    intr = map[itype[inp[j]]][itype[jnp[j]]];
    rs2[j] = inter->rmaxs_skin2[intr];
    rl2[j] = inter->rmaxl_skin2[intr];
 /*   rso32[j] = inter->rmax_so3[intr];  */   /* use cutoff for normal AA-AA interactions for now */
    rso32[j] = inter->rmax_rblob[intr]; 
    raa2[j] = inter->rmax_rblob[intr];
  }
  nprs = ngbr->nprs;  /* short range CG list */
  nprt = ngbr->nprt;  /* long range CG list */
  nprr = ngbr->nprr;  /* CG SO3 Rotational Dynamics list */
  npra = ngbr->npra;  /* CG atomistic list */
  for(j=0;j<nlen;j++){    
    /* add to CG list? */
    if( rdis2[j] < rl2[j] ){
      icg = inp[j];
      jcg = jnp[j];
      if( iaacg[icg] + iaacg[jcg] != 0 ){ /* add unless both in AR-AR */
        if(!exclij(icg,jcg,inter->nexclude,inter->exclude)){
          ipt[nprt] = inp[j]; 
          jpt[nprt] = jnp[j]; 
          nprt++;
          if( rdis2[j] < rs2[j]){
            ips[nprs] = inp[j];
            jps[nprs] = jnp[j]; 
            nprs++;
          }
        }
      }
    }
    /* add to AA list? */
    if( rdis2[j] < raa2[j] ){
      icg = inp[j];
      jcg = jnp[j];
      if( (iaacg[icg] != 2) && (iaacg[jcg] != 2) ){ /* add unless one or both are in CGR */
        ipa[npra] = inp[j]; 
        jpa[npra] = jnp[j]; 
        npra++;
      }
    }
    /* add to SO3 RotDyn list? */
    if( rdis2[j] < rso32[j] ){
      icg = inp[j];
      jcg = jnp[j];
      if( iaacg[icg] + iaacg[jcg] != 0 ){ /* add unless both are in AR */
        ipr[nprr] = inp[j];
        jpr[nprr] = jnp[j];
        nprr++;
      }
    }
    if((j+1)==coords->ibreak[k]){
      if(ngbr->offs[*ioffs]!=nprs) ngbr->offs[++(*ioffs)] = nprs; 
      if(ngbr->offt[*iofft]!=nprt) ngbr->offt[++(*iofft)] = nprt;
      if(ngbr->offa[*ioffa]!=npra) ngbr->offa[++(*ioffa)] = npra; 
      if(ngbr->offr[*ioffr]!=nprr) ngbr->offr[++(*ioffr)] = nprr;
      k++;
    }
  }
  ngbr->nprs = nprs;
  ngbr->nprt = nprt;
  ngbr->npra = npra;
  ngbr->nprr = nprr;
  free(rdis2);
}
/*----------------------------------------------------------------*/
void ngbr_liste(int nlen,int *inp,int *jnp,int iperd,
                COORDS *coords,INTER *inter,NGBR *ngbr,int ivol)
{
  int j,in,jn,nprte,nprse;
  double *xdis,*ydis,*zdis,rdis2;
  double *x,*y,*z;
  
  xdis = coords->xdis;  ydis = coords->ydis;  zdis = coords->zdis;  
  x = coords->px;  y = coords->py;  z = coords->pz;
  
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    in = inp[j];    jn = jnp[j];
    xdis[j] = x[in] - x[jn];
    ydis[j] = y[in] - y[jn];
    zdis[j] = z[in] - z[jn];
  }
  
  period(nlen,xdis,ydis,zdis,coords->hmat,coords->hmati,iperd,ivol);
  
  nprse = ngbr->nprse;
  nprte = ngbr->nprte;
  
  for(j=0;j<nlen;j++){
    rdis2 = xdis[j]*xdis[j]+ydis[j]*ydis[j]+zdis[j]*zdis[j];
    
    /* use the fact that the long range force is ALWAYS longer */
    if( rdis2 < inter->rmaxl_skin2_e){
      ngbr->inpte[nprte] = inp[j]; 
      ngbr->jnpte[nprte] = jnp[j]; 
      nprte++;
      if(rdis2 < inter->rmaxs_skin2_e){
        ngbr->inpse[nprse] = inp[j]; 
        ngbr->jnpse[nprse] = jnp[j]; 
        nprse++;
      }
    }
  }
  ngbr->nprte = nprte;
  ngbr->nprse = nprse;
}
/*----------------------------------------------------------------*/
void cnt_ngbr_list(int nlen,int *inp,int *jnp,int iperd,
                   COORDS *coords,INTER *inter,int *ns,int *nt,int ivol)
{
  int j,intr,in,jn,nst,ntt;
  double *xdis,*ydis,*zdis,rdis2;
  double *x,*y,*z;
  
  xdis = coords->xdis;  ydis = coords->ydis;  zdis = coords->zdis;  
  x = coords->px;  y = coords->py;  z = coords->pz;
  
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    in = inp[j];    jn = jnp[j];
    xdis[j] = x[in] - x[jn];
    ydis[j] = y[in] - y[jn];
    zdis[j] = z[in] - z[jn];
  }
  
  period(nlen,xdis,ydis,zdis,coords->hmat,coords->hmati,iperd,ivol);
  
  nst = ntt = 0.;
  for(j=0;j<nlen;j++){    
    rdis2 = xdis[j]*xdis[j]+ydis[j]*ydis[j]+zdis[j]*zdis[j];
    intr = inter->map[inter->itype[inp[j]]][inter->itype[jnp[j]]];
    if( rdis2 < inter->rmaxs_skin2[intr]) nst++;
    if( rdis2 < inter->rmaxl_skin2[intr]) ntt++;
  }	
  (*ns) += nst;
  (*nt) += ntt; 
}
/*----------------------------------------------------------------*/
void cnt_ngbr_onelist(int nlen,int *inp,int *jnp,int iperd,
                   COORDS *coords,INTER *inter,int *ns,int *nt,int *na,int *nr,int ivol)
{
  int j,intr,in,jn,nst,ntt,nat,nrt;
  double *xdis,*ydis,*zdis,rdis2;
  double *x,*y,*z;
  
  xdis = coords->xdis;  ydis = coords->ydis;  zdis = coords->zdis;  
  x = coords->px;  y = coords->py;  z = coords->pz;
  
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    in = inp[j];    jn = jnp[j];
    xdis[j] = x[in] - x[jn];
    ydis[j] = y[in] - y[jn];
    zdis[j] = z[in] - z[jn];
  }
  
  period(nlen,xdis,ydis,zdis,coords->hmat,coords->hmati,iperd,ivol);
  
  nst = ntt = nat = nrt = 0.;
  for(j=0;j<nlen;j++){    
    rdis2 = xdis[j]*xdis[j]+ydis[j]*ydis[j]+zdis[j]*zdis[j];
    intr = inter->map[inter->itype[inp[j]]][inter->itype[jnp[j]]];
    if( rdis2 < inter->rmaxl_skin2[intr]){
      if(!exclij(inp[j],jnp[j],inter->nexclude,inter->exclude)){
        ntt++;
        if( rdis2 < inter->rmaxs_skin2[intr]) nst++;
      }
    }
    if( rdis2 < inter->rmax_rblob[intr]) nat++;
    if( rdis2 < inter->rmax_rblob[intr]) nrt++;
/*    if( rdis2 < inter->rmax_rso3[intr]) nrt++; */ /* use other atomistic cutoff for now */
  }	
  (*ns) += nst;
  (*nt) += ntt;
  (*na) += nat;
  (*nr) += nrt; 
}
/*----------------------------------------------------------------*/
void cnt_ngbr_liste(int nlen,int *inp,int *jnp,int iperd,
                    COORDS *coords,INTER *inter,int *ns,int *nt,int ivol)
{
  int j;
  double *xdis,*ydis,*zdis,rdis2;
  double *x,*y,*z;
  
  xdis = coords->xdis;  ydis = coords->ydis;  zdis = coords->zdis;  
  x = coords->px;  y = coords->py;  z = coords->pz;
  
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    xdis[j] = x[inp[j]] - x[jnp[j]];
    ydis[j] = y[inp[j]] - y[jnp[j]];
    zdis[j] = z[inp[j]] - z[jnp[j]];
  }
  
  period(nlen,xdis,ydis,zdis,coords->hmat,coords->hmati,iperd,ivol);
  
  for(j=0;j<nlen;j++){
    rdis2 = xdis[j]*xdis[j]+ydis[j]*ydis[j]+zdis[j]*zdis[j];
    
    /* use the fact that the long range force is ALWAYS longer */
    if(rdis2 < inter->rmaxl_skin2_e){
      (*nt)++;
      if(rdis2 < inter->rmaxs_skin2_e) (*ns)++;
    }
  }
}
/*----------------------------------------------------------------*/
void count_vdw(SIMPARMS *simparms,COORDS *coords,INTER *inter,NGBR *ngbr,
               int *npso,int *npto,int *npao,int *npro)
{
  int i,j,k,ii,jj,icell,jcell;
  int *inp,*jnp,*excl,nexcl,*map,*itype;
  int *iaacg=simparms->iaacg;
  long ibegin,iend;
  
  itype = inter->itype;
  if(simparms->rank==0) md_stdout("Counting vdw pairs");
  
  if(simparms->nocell_all==0){
    *npso = *npto = 0;
    inp = coords->indx;
    jnp = coords->jndx;
    
    /* decompose i so that they is roughly the same number of
     i-j pairs on each processors */
    
    decomp_trig((long)coords->natoms,simparms->size,simparms->rank,
                &ibegin,&iend);
    
    k = 0;
    for(i=ibegin;i<iend;i++){   /* I could probably save some time in hybrid MD by using the atom selection */
      excl = &(inter->exclude[i][0]);
      nexcl = *(excl);
      map = inter->map[itype[i]];
      for(j=i+1;j<simparms->natoms;j++){ /* same here (Bernd) */
        if(nexcl == j){ 
          nexcl = *(++excl);
        } else if(map[itype[j]] == -1){
          /* exclude null interactions */
        } else {
          inp[k] = i;
          jnp[k] = j;
          k++;
          if(k==simparms->nlen){
            cnt_ngbr_list(k,inp,jnp,simparms->iperd,
                          coords,inter,npso,npto,simparms->ivol);
            k=0;
          }
        }
      }
      cnt_ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,npso,npto,
                    simparms->ivol);
      k=0;
    }
  } else if(ngbr->onelist == 2){
    get_lnklist(coords->natoms,ngbr->ncells,ngbr->hlist,ngbr->lnklist,
                simparms->iperd,coords->hmat,coords->px,coords->py,coords->pz,
                &simparms->iaacg[(1-coords->lcg)*simparms->ncg],ngbr->lcg);
    inp = coords->indx;
    jnp = coords->jndx;
    k = 0;
    for(i=0;i<ngbr->npaircell;i++){
      icell = ngbr->icell[i];
      jcell = ngbr->jcell[i];
      ii = ngbr->hlist[icell];
      while(ii!=-1){
        if(icell==jcell){
          jj = ngbr->lnklist[ii];
        } else {
          jj = ngbr->hlist[jcell];
        }
        while(jj!=-1){
          inp[k] = ii;
          jnp[k] = jj;
          k++;            
          if(k==simparms->nlen){
            cnt_ngbr_onelist(k,inp,jnp,simparms->iperd,coords,inter,
                              npso,npto,npao,npro,simparms->ivol);
            k = 0;
          }
          jj = ngbr->lnklist[jj];
        } /* end while(jj!=-1) */
        ii = ngbr->lnklist[ii];
      } 
    }
    cnt_ngbr_onelist(k,inp,jnp,simparms->iperd,coords,inter,npso,npto,
                  npao,npro,simparms->ivol);
  } else {
    get_lnklist(coords->natoms,ngbr->ncells,ngbr->hlist,ngbr->lnklist,
                simparms->iperd,coords->hmat,coords->px,coords->py,coords->pz,
                &simparms->iaacg[(1-coords->lcg)*simparms->ncg],ngbr->lcg);
    inp = coords->indx;
    jnp = coords->jndx;
    k = 0;
    for(i=0;i<ngbr->npaircell;i++){
      icell = ngbr->icell[i];
      jcell = ngbr->jcell[i];
      ii = ngbr->hlist[icell];
      while(ii!=-1){
        if(icell==jcell){
          jj = ngbr->lnklist[ii];
        } else {
          jj = ngbr->hlist[jcell];
        }
        map = inter->map[itype[ii]];
        while(jj!=-1){
          if(map[itype[jj]] == -1){
          } else if(!exclij(ii,jj,inter->nexclude,inter->exclude)){
            /* if hybrid CG-CG with separate lists: exlude AA-AA, but not CG-AA */
            if( !coords->lcg || (iaacg[ii]+iaacg[jj] != 0) || ngbr->onelist==2 ){
            /* for now put all pairs in pairlist */
              inp[k] = ii;
              jnp[k] = jj;
              k++;            
              if(k==simparms->nlen){
                cnt_ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,
                              npso,npto,simparms->ivol);
                k = 0;
              }
            }
          }
          jj = ngbr->lnklist[jj];
        } /* end while(jj!=-1) */
        ii = ngbr->lnklist[ii];
      } 
    }
    cnt_ngbr_list(k,inp,jnp,simparms->iperd,coords,inter,npso,npto,
                  simparms->ivol);
  }
}
/*-----------------------------------------------------------------------*/
void allocate_lists(int natoms,int nlen,NGBR *ngbr,COORDS *coords,int *mem)
{
  int ncell3;
  
  *mem=0;
//  if((ngbr->ilist!=0) && (ngbr->onelist != 0)){ /* need AA celllist for electrostatics also if onelist */
  if( ngbr->ilist!=0 ){
    ncell3 = ngbr->ncells*ngbr->ncells*ngbr->ncells;
    ngbr->hlist = (int *)cmalloc(ncell3*sizeof(int));
    ngbr->lnklist = (int *)cmalloc(natoms*sizeof(int));
    *mem += ncell3*sizeof(int);
    *mem += natoms*sizeof(int);
  }
  
  coords->scr_buf=(double *)cmalloc(6*natoms*sizeof(double));
  coords->scr_rec=coords->scr_buf + 3*natoms;
  *mem += 6*natoms*sizeof(double);
  
  coords->xdis = (double *)cmalloc(3*nlen*sizeof(double));
  coords->ydis = coords->xdis +   nlen;
  coords->zdis = coords->xdis + 2*nlen;
  *mem += 3*nlen*sizeof(double);
  
  /* the two is so that we don't need an if statement to see if 
   we are going over the end of the array (it could be as much as 2) */
  coords->ibreak = (int *)cmalloc((nlen+2)*sizeof(int));
  *mem += (nlen+2)*sizeof(int);
  
  coords->indx = (int *)cmalloc(5*nlen*sizeof(int));
  coords->jndx = coords->indx +   nlen;
  coords->ityp = coords->indx + 2*nlen;
  coords->jtyp = coords->indx + 3*nlen;
  coords->intr = coords->indx + 4*nlen;
  *mem += 5*nlen*sizeof(int);
  
//  if(ngbr->onelist==2){  /* if onelist: also AA part of system */
//    coords = coords->next; ngbr = ngbr->next;
//    ngbr->npso = ngbr->npto = 0;
//    ngbr->noffs = ngbr->nofft = noffs = coords->natoms+1;
//    ngbr->offs= (int *)cmalloc(noffs*sizeof(int));
//    ngbr->offt= (int *)cmalloc(noffs*sizeof(int));
//    *mem += 2*noffs*sizeof(int);
//    
//    ngbr->inps = (int *)cmalloc(sizeof(int));
//    ngbr->jnps = (int *)cmalloc(sizeof(int));
//    ngbr->inpt = (int *)cmalloc(sizeof(int));
//    ngbr->jnpt = (int *)cmalloc(sizeof(int));
//    
//    natoms = coords->natoms;
//    coords->xdis = (double *)cmalloc(3*nlen*sizeof(double)); /* remove next 3 lines when ngbr_list is only done CG */
//    coords->ydis = coords->xdis +   nlen;
//    coords->zdis = coords->xdis + 2*nlen;
//    *mem += 3*nlen*sizeof(double);
//    coords->scr_buf=(double *)cmalloc(6*natoms*sizeof(double));
//    coords->scr_rec=coords->scr_buf + 3*natoms;
//    *mem += 6*natoms*sizeof(double);
//    
//    
//    coords->ibreak = (int *)cmalloc((nlen+2)*sizeof(int));
//    *mem += (nlen+2)*sizeof(int);  
//    coords->indx = (int *)cmalloc(2*nlen*sizeof(int));
//    coords->jndx = coords->indx + nlen;
//    *mem += 2*nlen*sizeof(int);
//    coords = coords->next; ngbr = ngbr->next;
//  }
}

/*-----------------------------------------------------------------------*/
void free_lists(NGBR *ngbr,COORDS *coords)
{
  if(ngbr->ilist!=0){
    free(ngbr->hlist);
    free(ngbr->lnklist);
    free(ngbr->icell); free(ngbr->jcell);
    free(ngbr->icelle); free(ngbr->jcelle);
  }
  
  free(coords->scr_buf);
  free(coords->xdis);
  free(coords->ibreak);
  free(coords->indx);
  
  if(ngbr->ilist==1){
    free(ngbr->inps); free(ngbr->jnps);
    free(ngbr->inpt); free(ngbr->jnpt);
    free(ngbr->inpse); free(ngbr->jnpse);
    free(ngbr->inpte); free(ngbr->jnpte);
  }  
  
  free(ngbr->offs);
  free(ngbr->offt);
  free(ngbr->offse);
}

/*-----------------------------------------------------------------------*/
/* get all AA-AA VdW pairs within and between atoms of CG sites icg and jcg */
/* we could save some time if we would know that there are no AA-AA VdW  */
/* interactions within a CG unit...                                      */
/* - also, this implementation still tests all AA-AA pair distances in   */
/*   ngbr_list(), whereas we could instead just test the CG-CG distance. */
void expand_into_aalist(int icg, int jcg, SIMPARMS *simparms, INTER *inter,
                        COORDS *coords, NGBR *ngbr, int *inp, int *jnp, int *k, int *ioffs, int *iofft, int *nbreak){
  
  int i,j,ii,jj, *map, *itype,*ibreak;
  int *ilist,*jlist;
  int npso, npto, noffs, nofft,mem_add;
  int nlen = simparms->nlen;
  
  npso = ngbr->npso;
  npto = ngbr->npto;
  noffs = ngbr->noffs;
  nofft = ngbr->nofft;
  mem_add = MAX(simparms->nlen,coords->natoms);
  
  ilist = simparms->icglist[icg];
  jlist = simparms->icglist[jcg];
  itype = inter->itype;
  ibreak = coords->ibreak;
  
  /* check offset memory */
  if(noffs < *ioffs + mem_add){
    noffs += mem_add;
    ngbr->offs = (int *)realloc(ngbr->offs,noffs*sizeof(int));
    ngbr->noffs = noffs;
  }
  if(nofft < *iofft + mem_add){
    nofft += mem_add;
    ngbr->offt = (int *)realloc(ngbr->offt,nofft*sizeof(int));
    ngbr->nofft = nofft;
  }
  /* check neighbor list  memory */
  if(npso < ngbr->nprs + mem_add){
    npso += mem_add;
    ngbr->inps = (int *)realloc(ngbr->inps,npso*sizeof(int));
    ngbr->jnps = (int *)realloc(ngbr->jnps,npso*sizeof(int));
    ngbr->npso = npso;
  }
  if(npto < ngbr->nprt + mem_add){
    npto += mem_add;
    ngbr->inpt = (int *)realloc(ngbr->inpt,npto*sizeof(int));
    ngbr->jnpt = (int *)realloc(ngbr->jnpt,npto*sizeof(int));
    ngbr->npto = npto;
  }
  
  /* AA pairs within ICG */
  if(icg==jcg){
    for(i=0;i<simparms->napcg[icg]-1;++i){
      ii=ilist[i];
      map = inter->map[itype[ii]];
      for(j=i+1;j<simparms->napcg[icg];++j){
        jj=ilist[j];
        if(map[itype[jj]] != -1){
          if(!exclij(ii,jj,inter->nexclude,inter->exclude)){
            inp[*k] = ii;
            jnp[*k] = jj;
            (*k)++;
            if(*k==nlen){
              ibreak[++*nbreak] = *k;
              ngbr_list(*k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                        ioffs,iofft,simparms->ivol);
              *nbreak = *k = 0;
            }
          }
        }	  
      }
    }
  }else{
    /* AA pairs between ICG and JCG */  
    for(i=0;i<simparms->napcg[icg];++i){
      ii=ilist[i];
      map = inter->map[itype[ii]];
      for(j=0;j<simparms->napcg[jcg];++j){
        jj=jlist[j];
        if(map[itype[jj]] != -1){
          if(!exclij(ii,jj,inter->nexclude,inter->exclude)){
            inp[*k] = ii;
            jnp[*k] = jj;
            (*k)++;
            if(*k==nlen){
              ibreak[++*nbreak] = *k;
              ngbr_list(*k,inp,jnp,simparms->iperd,coords,inter,ngbr,
                        ioffs,iofft,simparms->ivol);
              *nbreak = *k = 0;
            }
          }
        }
      }
    }
  }
  
}
