/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "md.h"

#define NHILL_ALLOC 1000
  
static int nhills_alloc=0;

#define FACTOR .50
#define DELTAV .2

void add_hills(COLVAR *colvar, int istep)
{
  int icol,mbytes,nhills;
  ++(colvar->nhills);
  nhills = (colvar->nhills);

  if(nhills > nhills_alloc){
    if(nhills_alloc ==0 ){
      nhills_alloc = NHILL_ALLOC;
      mbytes = nhills_alloc*sizeof(double);
      colvar->t_hilldepth =(double *)malloc(mbytes);
      colvar->istep=(int *)malloc(sizeof(int)*nhills_alloc);
      for(icol=0;icol<colvar->ncolvar;++icol){
	colvar->t_hillwidth[icol] = (double *)malloc(mbytes);
	colvar->t_pcolvar[icol] =  (double *)malloc(mbytes);
      }
    } else {   
      nhills_alloc += NHILL_ALLOC;
      
      mbytes = nhills_alloc*sizeof(double);
      colvar->t_hilldepth =(double *)realloc(colvar->t_hilldepth,mbytes);
      colvar->istep=(int *)realloc(colvar->istep,sizeof(int)*nhills_alloc);
      for(icol=0;icol<colvar->ncolvar;++icol){
	colvar->t_hillwidth[icol] = 
	  (double *)realloc(colvar->t_hillwidth[icol],mbytes);
	colvar->t_pcolvar[icol] = 
	  (double *)realloc(colvar->t_pcolvar[icol],mbytes);
      }
    }
  }
  colvar->istep[nhills-1]=istep;
  colvar->t_hilldepth[nhills-1] = colvar->hilldepth;
  for(icol=0;icol<colvar->ncolvar;++icol){
    colvar->t_hillwidth[icol][nhills-1]=colvar->hillwidth[icol];
    colvar->t_pcolvar[icol][nhills-1]=colvar->pcolvarlast[icol];
  }
}

/*-----------------------------------------------------------------------*/
int cv_displacement(COLVAR *colvar)
{
  int icol;
  double dv,sdv2;
  int nhills = colvar->nhills;
  sdv2=0;

  /* DO WE TAKE DIFFERENT HILLWIDTHS INTO ACCOUNT ?? */
  /* we do seem to have a different mindwidth for each colvar... */

  if(nhills<1) {
    for(icol=0;icol<colvar->ncolvar;++icol){
      colvar->pcolvarlast[icol] = colvar->pcolvar[icol];
    }
    return 1;
  }
  for(icol=0;icol<colvar->ncolvar;++icol){
    dv = colvar->pcolvar[icol] - colvar->pcolvarlast[icol];
    dv /= ( colvar->hillwidth[icol] * colvar->mindwidth[icol] );
    sdv2 += dv*dv;
  }
  if(sdv2 > 1.0 ) {
    for(icol=0;icol<colvar->ncolvar;++icol){
      colvar->pcolvarlast[icol] = colvar->pcolvar[icol];
    }
    return 1;
  }
  return 0;
}

/*-----------------------------------------------------------------------*/
int tunehillplacement(COLVAR *colvar)
{
  static int istart=0;
  static int itime=0;
  int iflag = 0;
  int icol;
  double dv,sign;
  static double avgtemp;

  if(istart==0){
    for(icol=0;icol<colvar->ncolvar;++icol){
      colvar->vcolvarlast[icol] = 0.0;
      colvar->pcolvarlast[icol] = colvar->pcolvar[icol];
      colvar->pcolvarmin[icol] = colvar->pcolvar[icol];
      colvar->pcolvarmax[icol] = colvar->pcolvar[icol];
      colvar->pcolvarmean[icol] = 0.0;
      colvar->pcolvarstd[icol]  = 0.0;
    }
    istart=1;
    itime = 0;
    avgtemp = 0.0;
    return 0;
  }  

  ++itime;

  for(icol=0;icol<colvar->ncolvar;++icol){
    colvar->pcolvarmean[icol] += colvar->pcolvar[icol];
    colvar->pcolvarstd[icol]  += colvar->pcolvar[icol]*colvar->pcolvar[icol];

    sign = COPY_SIGN(1.0,colvar->vcolvar[icol]*colvar->vcolvarlast[icol]);
    if(sign < 0){
      if(colvar->vcolvar[icol] > 0.0){
	colvar->pcolvarmin[icol] = MIN(colvar->pcolvarmin[icol],colvar->pcolvar[icol]);
	if(colvar->pcolvarmin[icol] == colvar->pcolvarlast[icol]){
	  colvar->pcolvarmin[icol] = colvar->pcolvar[icol];
	}
      } else {
	colvar->pcolvarmax[icol] = MAX(colvar->pcolvarmax[icol],colvar->pcolvar[icol]);
	if(colvar->pcolvarmax[icol] == colvar->pcolvarlast[icol]){
	  colvar->pcolvarmax[icol] = colvar->pcolvar[icol];
	}
      }
    }
    avgtemp += colvar->amass[icol]*colvar->vcolvar[icol]*colvar->vcolvar[icol];
    colvar->vcolvarlast[icol] = colvar->vcolvar[icol];
  }

  /* have we set all the mins and maxs */
  iflag = 1;
  for(icol=0;icol<colvar->ncolvar;++icol){
    if(colvar->pcolvarmax[icol] == colvar->pcolvarlast[icol] ||
       colvar->pcolvarmin[icol] == colvar->pcolvarlast[icol]){
      iflag = 0;
      break;
    }
  }

  if(iflag==1){
    for(icol=0;icol<colvar->ncolvar;++icol){
      if( (avgtemp/(double)(itime*colvar->ncolvar))*FACTOR < colvar->temp[icol]){
	iflag=0;
	break;
      }
    }
  }

  if(iflag==1){ /* we have a good temp and max and min off ALL variables */
    
    avgtemp /= (double)(itime*colvar->ncolvar);
    
    colvar->hilldepth = avgtemp*FACTOR;  
    avgtemp = sqrt(1.0-FACTOR);
 
    for(icol=0;icol<colvar->ncolvar;++icol){
      colvar->vcolvar[icol] *= avgtemp;
    }

    for(icol=0;icol<colvar->ncolvar;++icol){
      colvar->pcolvarstd[icol] /= (double)itime;
      colvar->pcolvarmean[icol] /= (double)itime;
      dv = sqrt(colvar->pcolvarstd[icol] - colvar->pcolvarmean[icol]*colvar->pcolvarmean[icol]);
      
      colvar->hillwidth[icol]   = dv;
      
      dv = .5*(colvar->pcolvarmax[icol]-colvar->pcolvarmin[icol]);
      
      /* colvar->hillwidth[icol]   = dv*M_SQRT1_2; old method using min and max
	 dv += colvar->pcolvarmin[icol];
      */
      
      colvar->pcolvarlast[icol] = colvar->pcolvarmean[icol];
      colvar->pcolvarmax[icol]  = colvar->pcolvarmean[icol];
      colvar->pcolvarmin[icol]  = colvar->pcolvarmean[icol];
      colvar->pcolvarmean[icol] = 0.0;
      colvar->pcolvarstd[icol]  = 0.0;
    }
    
    avgtemp = 0.0;
    itime = 0;
  }

  return iflag;
}

/*-----------------------------------------------------------------------*/
int forcehillplacement(COLVAR *colvar)
{
  static int istart=0;
  static int itime=0;
  int iflag = 0;
  int icol;
  double dx,sign,dv;
  static double t1,avgt;

  if(istart==0){
    for(icol=0;icol<colvar->ncolvar;++icol){
      colvar->vcolvar[icol] = sqrt(colvar->temp[icol]/colvar->amass[icol]);
      colvar->pcolvarlast[icol] = colvar->pcolvar[icol];
      colvar->pcolvarmin[icol] = colvar->pcolvar[icol];
      colvar->pcolvarmax[icol] = colvar->pcolvar[icol];
      colvar->pcolvarmean[icol] = 0.0;
      colvar->pcolvarstd[icol]  = 0.0;
      t1=0;
      avgt=0.0;
    }
    istart=1;
    itime = 0;
    return 0;
  }  

  ++itime;
  
  /* addapt the velocity of the collective variables */
  for(icol=0;icol<colvar->ncolvar;++icol){
    colvar->vcolvarlast[icol] = colvar->vcolvar[icol];
    t1=0.5*colvar->fk[icol]*dx*dx;
    avgt= (4*avgt+t1)/5.0;
    dx = colvar->pcolvar[icol]-colvar->pistcolvar[icol];
    sign = COPY_SIGN(1.0,dx);
    if(avgt > colvar->temp[icol]){
      colvar->vcolvar[icol] -= DELTAV*sign*colvar->temp[icol]/colvar->amass[icol];
    }
  }

  /* check for minimum and maximum pcolvar */
  for(icol=0;icol<colvar->ncolvar;++icol){
    colvar->pcolvarmean[icol] += colvar->pcolvar[icol];
    colvar->pcolvarstd[icol]  += colvar->pcolvar[icol]*colvar->pcolvar[icol];

    sign = COPY_SIGN(1.0,colvar->vcolvar[icol]*colvar->vcolvarlast[icol]);
    if(sign < 0){
      if(colvar->vcolvar[icol] > 0.0){
	colvar->pcolvarmin[icol] = MIN(colvar->pcolvarmin[icol],colvar->pcolvar[icol]);
      } else {
	colvar->pcolvarmax[icol] = MAX(colvar->pcolvarmax[icol],colvar->pcolvar[icol]);
      }
    }
  }

  /* have we set all the mins and maxs */
  iflag = 1;
  for(icol=0;icol<colvar->ncolvar;++icol){
    if(colvar->pcolvarmax[icol] == colvar->pcolvarlast[icol] ||
       colvar->pcolvarmin[icol] == colvar->pcolvarlast[icol]){
      iflag = 0;
      break;
    }
  }
 
  if(iflag==1){ /* we have max and min of ALL variables */
    
    colvar->hilldepth = colvar->temp[0];  

    for(icol=0;icol<colvar->ncolvar;++icol){
      colvar->pcolvarstd[icol] /= (double)itime;
      colvar->pcolvarmean[icol] /= (double)itime;
      dv = sqrt(colvar->pcolvarstd[icol] - colvar->pcolvarmean[icol]*colvar->pcolvarmean[icol]);
      
      colvar->hillwidth[icol]   = dv;
      
      dv = .5*(colvar->pcolvarmax[icol]-colvar->pcolvarmin[icol]);
      
      /* colvar->hillwidth[icol]   = dv*M_SQRT1_2; old method using min and max
	 dv += colvar->pcolvarmin[icol];
      */
      
      colvar->pcolvarlast[icol] = colvar->pcolvarmean[icol];
      colvar->pcolvarmax[icol]  = colvar->pcolvarmean[icol];
      colvar->pcolvarmin[icol]  = colvar->pcolvarmean[icol];
      colvar->pcolvarmean[icol] = 0.0;
      colvar->pcolvarstd[icol]  = 0.0;
    }
    
    itime = 0;
  }

  return iflag;
}

/*-----------------------------------------------------------------------*/
void hills(SIMPARMS *simparms,COORDS *coords)
{
  int icol;
  double *t_hilldepth  = coords->colvar.t_hilldepth;
  double **t_hillwidth  = coords->colvar.t_hillwidth;
  double **t_pcolvar    = coords->colvar.t_pcolvar;
  double *f_hills = coords->colvar.f_hills;
  double *pcolvar = coords->colvar.pcolvar;
  int ncolvar = coords->colvar.ncolvar;
  int nhills = coords->colvar.nhills;
  int *lhills = coords->colvar.lhills;
  int ihill,dt;
  double *dp, dp2;
  double ecolvar = 0;
  double gauss;

  /* shall we add a hill */
  ihill=0;
  dt = simparms->istep - coords->colvar.lasthill;
  if(coords->colvar.ltunehills==0){
    if(dt > coords->colvar.maxstephill){
      md_warning("Maxstep hill reached");
      for(icol=0;icol<ncolvar;++icol){
	coords->colvar.pcolvarlast[icol] = coords->colvar.pcolvar[icol];
      }
      ihill = 1;
    } else if(dt > coords->colvar.minstephill){
      ihill = cv_displacement(&(coords->colvar));
    }
  } else if(coords->colvar.ltunehills==1){
    ihill = tunehillplacement(&(coords->colvar));
  } else {
    ihill = forcehillplacement(&(coords->colvar));
  }

  if(ihill==1) {
   add_hills(&(coords->colvar),simparms->istep);
   coords->colvar.lasthill = simparms->istep;
  }

  /* compute forces from hills on colvars */
  /* todo:add parallel gauss correction and hilldepth tuning              */

  dp = (double *)malloc(ncolvar*sizeof(double));

  for(icol=0;icol<ncolvar;++icol){
    f_hills[icol]=0.0;
  }
  for(ihill=0;ihill<nhills;++ihill){
    dp2=0.;
    for(icol=0;icol<ncolvar;++icol){      
      if(lhills[icol]) {
	dp[icol]=(pcolvar[icol]-t_pcolvar[icol][ihill])/t_hillwidth[icol][ihill];
	dp2 += dp[icol]*dp[icol];
      }
    }
    gauss = -t_hilldepth[ihill]*exp(-0.5*dp2);
    for(icol=0;icol<ncolvar;++icol){
      if(lhills[icol]) {
	f_hills[icol] -= gauss*dp[icol]/t_hillwidth[icol][ihill];
      }
    }
    ecolvar += gauss;
  }
  
  for(icol=0;icol<ncolvar;++icol){
    coords->colvar.fcolvar[icol] += f_hills[icol];
  }
  
  free(dp);
  return;
}
