/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore
#ifdef PRINTFUNCTION

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* subroutines dealing with bonded interactions
   search_bond_base - searches the data base and sets up initial vectors
   fbond - gets the force of the bond
   getvbond - gets the potential energy of the bond
   bonded   - check to see if to atoms are bonded
*/

#define DEBUGOFF
#define PRINTFUNCTIONOFF
#define PRINTTORSIONOF

#include "md.h"

/* #define PRINT_COLVAR */

/* rescale colvar velocities to target temperature  */
/* if temperature larger than temp+thrmmass         */
void scale_colvar(COLVAR *colvar)
{
  int icol;
  double t_var;

  for(icol=0;icol<colvar->ncolvar;++icol){
    t_var = colvar->amass[icol]*colvar->vcolvar[icol]*colvar->vcolvar[icol];
    if(t_var > colvar->temp[icol]+colvar->thrmmass[icol]){
      colvar->vcolvar[icol] = COPY_SIGN(sqrt(colvar->temp[icol]/colvar->amass[icol]),colvar->vcolvar[icol]); 
    }
  }
}

/*---------------------------------------------------------------------*/
double torsion(double *a,double *b,double *c,double *d,double *sign)
{
  double dx1,dy1,dz1,dx2,dy2,dz2,dx3,dy3,dz3,norm,angle;
  double vx1,vy1,vz1,vx2,vy2,vz2,vx3,vy3,vz3,factor,ac;

  /* distance between centers of mass */
  dx1 = a[0]-b[0];
  dy1 = a[1]-b[1];
  dz1 = a[2]-b[2];
  dx2 = c[0]-b[0];
  dy2 = c[1]-b[1];
  dz2 = c[2]-b[2];

  /*normalize */
  norm=sqrt(dx1*dx1+dy1*dy1+dz1*dz1);
  dx1/=norm; dy1/=norm; dz1/=norm;
  norm=sqrt(dx2*dx2+dy2*dy2+dz2*dz2);
  dx2/=norm; dy2/=norm; dz2/=norm;
  /* vector product */
  vx1= dy1*dz2 - dz1*dy2;
  vy1= dz1*dx2 - dx1*dz2;
  vz1= dx1*dy2 - dy1*dx2;
  /*normalize */
  norm=sqrt(vx1*vx1+vy1*vy1+vz1*vz1);
  vx1/=norm; vy1/=norm; vz1/=norm;

  /* distance between centers of mass */
  dx3 = d[0]-c[0];
  dy3 = d[1]-c[1];
  dz3 = d[2]-c[2];
  dx2 = b[0]-c[0];
  dy2 = b[1]-c[1];
  dz2 = b[2]-c[2];

    /*normalize */
  norm=sqrt(dx2*dx2+dy2*dy2+dz2*dz2);
  dx2/=norm; dy2/=norm; dz2/=norm;
  norm=sqrt(dx3*dx3+dy3*dy3+dz3*dz3);
  dx3/=norm; dy3/=norm; dz3/=norm;
  /* vector product */
  vx2= dy2*dz3 - dz2*dy3;
  vy2= dz2*dx3 - dx2*dz3;
  vz2= dx2*dy3 - dy2*dx3;
  /*normalize */
  norm=sqrt(vx2*vx2+vy2*vy2+vz2*vz2);
  vx2/=norm; vy2/=norm; vz2/=norm;

  /*scaler product */
  ac = vx1*vx2 + vy1*vy2 + vz1*vz2;
  /* vector product */
  vx3= vy1*vz2 - vz1*vy2;
  vy3= vz1*vx2 - vx1*vz2;
  vz3= vx1*vy2 - vy1*vx2;

  dx1 = c[0] - b[0];
  dy1 = c[1] - b[1];
  dz1 = c[2] - b[2];
  
  factor = dx1*vx3 + dy1*vy3  + dz1*vz3;
  *sign=1.0;
  /* BERND BUG FIX 29 January 2009 */  
 if( (ac>=1.0) || (ac<=-1.0) ){
    angle = 0.0;
  }else{
    angle = acos(ac);
  }
  if(factor<0.0){
    angle=-angle;
    *sign=-1.0;
  }

  return angle;
}

/*---------------------------------------------------------------------*/
void ddtorsion(double *ddangle,double *a,double *b,double *c,double *d,double sign){

  double ax,ay,az,bx,by,bz,cx,cy,cz;
  double dx,dy,dz,dl,dl2,ex,ey,ez,el,el2;
  double cphi,dudcos=0.;
  double g,dre,erd,r11;
  double dfx1,dfy1,dfz1,dfx2,dfy2,dfz2,dfx3,dfy3,dfz3,dfx4,dfy4,dfz4;
  double dgx1,dgy1,dgz1,dgx2,dgy2,dgz2,dgx3,dgy3,dgz3,dgx4,dgy4,dgz4;

    /* Derivatives  */

    ax = a[0] - b[0];
    ay = a[1] - b[1];
    az = a[2] - b[2];
    
    bx = b[0] - c[0];
    by = b[1] - c[1];
    bz = b[2] - c[2];

    cx = c[0] - d[0];
    cy = c[1] - d[1];
    cz = c[2] - d[2];

    dx = ay*bz - az*by;
    dy = az*bx - ax*bz;
    dz = ax*by - ay*bx;
    dl2= dx*dx+dy*dy+dz*dz;
    dl = sqrt(dl2);

    ex=by*cz - bz*cy;
    ey=bz*cx - bx*cz;
    ez=bx*cy - by*cx;
    el2=ex*ex+ey*ey+ez*ez;
    el=sqrt(el2);

    cphi=(dx*ex+dy*ey+dz*ez)/(dl*el);
    dudcos  = -1./sqrt(1.-cphi*cphi);
    if(cphi>=1.0 || cphi <= -1.0) dudcos   = 0.; 

    g=dl*el;
    dre=dl/el;
    erd=el/dl;
    r11=1./g;

    dfx1 = by*ez - bz*ey;
    dfy1 = bz*ex - bx*ez;
    dfz1 = bx*ey - by*ex;
    
    dfx2 = (cy*dz-cz*dy+(az+bz)*ey-(ay+by)*ez);
    dfy2 = (cz*dx-cx*dz+(ax+bx)*ez-(az+bz)*ex);
    dfz2 = (cx*dy-cy*dx+(ay+by)*ex-(ax+bx)*ey);
    
    dfx3 = (ay*ez-az*ey+(bz+cz)*dy-(by+cy)*dz);
    dfy3 = (az*ex-ax*ez+(bx+cx)*dz-(bz+cz)*dx);
    dfz3 = (ax*ey-ay*ex+(by+cy)*dx-(bx+cx)*dy);
    
    dfx4 = by*dz - bz*dy;
    dfy4 = bz*dx - bx*dz;
    dfz4 = bx*dy - by*dx;

    dgx1 = (by*dz - bz*dy)*erd;
    dgy1 = (bz*dx - bx*dz)*erd;
    dgz1 = (bx*dy - by*dx)*erd;

    dgx2 = ((dy*(az+bz)-dz*(ay+by))*erd+(cy*ez-cz*ey)*dre);
    dgy2 = ((dz*(ax+bx)-dx*(az+bz))*erd+(cz*ex-cx*ez)*dre);
    dgz2 = ((dx*(ay+by)-dy*(ax+bx))*erd+(cx*ey-cy*ex)*dre);
    
    dgx3 = ((ay*dz-az*dy)*erd+(ey*(bz+cz)-ez*(by+cy))*dre);
    dgy3 = ((az*dx-ax*dz)*erd+(ez*(bx+cx)-ex*(bz+cz))*dre);
    dgz3 = ((ax*dy-ay*dx)*erd+(ex*(by+cy)-ey*(bx+cx))*dre);
    
    dgx4 = (by*ez - bz*ey)*dre;
    dgy4 = (bz*ex - bx*ez)*dre;
    dgz4 = (bx*ey - by*ex)*dre;
    
    dudcos *= r11*sign;

    ddangle[0] = dudcos*(dfx1 - cphi*dgx1);
    ddangle[1] = dudcos*(dfy1 - cphi*dgy1);
    ddangle[2] = dudcos*(dfz1 - cphi*dgz1);
    ddangle[3] = dudcos*(dfx2 - cphi*dgx2);
    ddangle[4] = dudcos*(dfy2 - cphi*dgy2);
    ddangle[5] = dudcos*(dfz2 - cphi*dgz2);
    ddangle[6] = dudcos*(dfx3 - cphi*dgx3);
    ddangle[7] = dudcos*(dfy3 - cphi*dgy3);
    ddangle[8] = dudcos*(dfz3 - cphi*dgz3);
    ddangle[9] = dudcos*(dfx4 - cphi*dgx4);
    ddangle[10]= dudcos*(dfy4 - cphi*dgy4);
    ddangle[11]= dudcos*(dfz4 - cphi*dgz4);

    return;
}

/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
void ddtorsion2(double *ddangle,double *a,double *b,double *c,double *d,double sign){
#define epsilon 1.0e-12

  int i;
  double t1,t2,t3,t4,t5,t6,t7,t8,t9,cost;
  double caa,cab,cbb,cac,cbc,ccc,dab,dbc,ddd,ac;

    t1 = a[0] - b[0];
    t2 = a[1] - b[1];
    t3 = a[2] - b[2];
    t4 = b[0] - c[0];
    t5 = b[1] - c[1];
    t6 = b[2] - c[2];
    t7 = c[0] - d[0];
    t8 = c[1] - d[1];
    t9 = c[2] - d[2];

    caa = t1*t1+t2*t2+t3*t3;
    cbb = t4*t4+t5*t5+t6*t6;
    ccc = t7*t7+t8*t8+t9*t9;
    cab = t1*t4+t2*t5+t3*t6;
    cac = t1*t7+t2*t8+t3*t9;
    cbc = t4*t7+t5*t8+t6*t9;
    dab = caa*cbb-cab*cab;
    dbc = cbb*ccc-cbc*cbc;
    if((dab<epsilon)||(dbc<epsilon)){
      for(i=0;i<12;++i) ddangle[i]=0.0;
    } else{
      ddd = -1.0/sqrt(dab*dbc);
      ac = -(cab*cbc-cac*cbb)/sqrt(dab*dbc);
      if(abs(ac)>0.99999999) ac=0.99999999; 
      cost= 1.0/sqrt(1.0-ac*ac)*sign;
      ddd = 1.0/sqrt(dab*dbc)*cost;
      
      ddangle[0] = -ddd*(cbc*t4-cbb*t7-(cab*cbc-cac*cbb)/dab*(cbb*t1-cab*t4));
      ddangle[1] = -ddd*(cbc*t5-cbb*t8-(cab*cbc-cac*cbb)/dab*(cbb*t2-cab*t5));
      ddangle[2] = -ddd*(cbc*t6-cbb*t9-(cab*cbc-cac*cbb)/dab*(cbb*t3-cab*t6));
      ddangle[3] = -ddd*(cbc*t1-cbc*t4+cab*t7+cbb*t7-2*cac*t4-(cab*cbc-cac*cbb)/dbc*(ccc*t4-cbc*t7)-(cab*cbc-cac*cbb)/dab*(caa*t4-cbb*t1-cab*t1+cab*t4));
      ddangle[4] = -ddd*(cbc*t2-cbc*t5+cab*t8+cbb*t8-2*cac*t5-(cab*cbc-cac*cbb)/dbc*(ccc*t5-cbc*t8)-(cab*cbc-cac*cbb)/dab*(caa*t5-cbb*t2-cab*t2+cab*t5));
      ddangle[5] = -ddd*(cbc*t3-cbc*t6+cab*t9+cbb*t9-2*cac*t6-(cab*cbc-cac*cbb)/dbc*(ccc*t6-cbc*t9)-(cab*cbc-cac*cbb)/dab*(caa*t6-cbb*t3-cab*t3+cab*t6));
      ddangle[6] = -ddd*(-cbc*t1+cab*t4-cab*t7-cbb*t1+2.0*cac*t4-(cab*cbc-cac*cbb)/dbc*(cbb*t7-ccc*t4-cbc*t4+cbc*t7)-(cab*cbc-cac*cbb)/dab*(-caa*t4+cab*t1));
      ddangle[7] = -ddd*(-cbc*t2+cab*t5-cab*t8-cbb*t2+2.0*cac*t5-(cab*cbc-cac*cbb)/dbc*(cbb*t8-ccc*t5-cbc*t5+cbc*t8)-(cab*cbc-cac*cbb)/dab*(-caa*t5+cab*t2));
      ddangle[8] = -ddd*(-cbc*t3+cab*t6-cab*t9-cbb*t3+2.0*cac*t6-(cab*cbc-cac*cbb)/dbc*(cbb*t9-ccc*t6-cbc*t6+cbc*t9)-(cab*cbc-cac*cbb)/dab*(-caa*t6+cab*t3));
      ddangle[9]= -ddd*(-cab*t4+cbb*t1-(cab*cbc-cac*cbb)/dbc*(-cbb*t7+cbc*t4));
      ddangle[10]= -ddd*(-cab*t5+cbb*t2-(cab*cbc-cac*cbb)/dbc*(-cbb*t8+cbc*t5));
      ddangle[11]= -ddd*(-cab*t6+cbb*t3-(cab*cbc-cac*cbb)/dbc*(-cbb*t9+cbc*t6));
    }
    return;
}

/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/* forces on atoms due to spring of cv icol                            */
double fcolvar_torsionswitches(int icol,SIMPARMS *simparms,COORDS *coords,int iforce)
{
  int i,iff;
  /* long ibegin,iend; */
  double angle,sign,dangle,xyz1[3],xyz2[3],xyz3[3],xyz4[3];
  double dcdr;
  double func,dfunc,sum,dsum,num,den,ren,rem;
  double ddangle[12];
  static int maxnsize=0;
  static double *ffx,*ffy,*ffz;


  int ntorsions = coords->colvar.natom[icol][0];
  int *iatom1 = coords->colvar.atom[icol][0];  /* atom 1 of torsion i */
  int *iatom2 = coords->colvar.atom[icol][1];  /* atom 2 of torsion i */
  int *iatom3 = coords->colvar.atom[icol][2];  /* atom 3 of torsion i */
  int *iatom4 = coords->colvar.atom[icol][3];  /* atom 4 of torsion i */
  double fk = coords->colvar.fk[icol];

  double *r0 = coords->colvar.switch_r[icol];
  double rc = coords->colvar.switch_rc[icol];
  int n = coords->colvar.switch_n[icol];
  int m = coords->colvar.switch_m[icol];

  double *px = coords->px;
  double *py = coords->py;
  double *pz = coords->pz;

  double *fx = coords->fxa;
  double *fy = coords->fya;
  double *fz = coords->fza;

  double *f_spring = coords->colvar.f_spring;


  /* doc -1 because we don't include ourselves */
  iff = 4*coords->colvar.natom[icol][0];
  if(iff>maxnsize){
    if(maxnsize==0){
      maxnsize=iff;
      ffx = (double *)malloc(iff*3*sizeof(double));
      ffy = ffx+iff;
      ffz = ffy+iff;
    } else {
      maxnsize=iff;
      ffx = (double *)realloc(ffx,iff*3*sizeof(double));
      ffy = ffx + iff;
      ffz = ffy + iff;
    }
  }

  /* if(coords->colvar.ndim[icol] !=2) error(); */
  /* sum of switch functions of torsion angles     */
  /*                   (1-((t-r0)/rc)^n            */
  /*          c =  SUM ----------------            */
  /* 	               (1-((t-r0)/rc)^m            */

  sum=0.0;
  iff=0;
  for(i=0;i<ntorsions;++i){
    xyz1[0]=px[iatom1[i]]; xyz1[1]=py[iatom1[i]]; xyz1[2]=pz[iatom1[i]];
    xyz2[0]=px[iatom2[i]]; xyz2[1]=py[iatom2[i]]; xyz2[2]=pz[iatom2[i]];
    xyz3[0]=px[iatom3[i]]; xyz3[1]=py[iatom3[i]]; xyz3[2]=pz[iatom3[i]];
    xyz4[0]=px[iatom4[i]]; xyz4[1]=py[iatom4[i]]; xyz4[2]=pz[iatom4[i]];
    angle=torsion(xyz1,xyz2,xyz3,xyz4,&sign);
#ifdef PRINTFUNCTION
    nstep=100;
    step=2*M_PI/nstep;
    angle=-1.0*M_PI-step;
    for(j=0;j<nstep+1;++j){
      angle+=step;
#endif
    dangle=angle-r0[i];
    dangle += -2.0*M_PI*(int)(dangle/M_PI);     /* apply PBC's */
#ifdef PRINTTORSION
    printf("%8.4lf  ",angle);
#endif
    /*    if( (angle < r0[i]) || (angle > (r0[i]+M_PI-2*rc)) ){
          if(angle > (r0[i]+M_PI-2*rc )) dangle=angle-r0[i]-M_PI+2*rc;  */
      ren= pow((dangle)/rc,n);
      rem= pow((dangle)/rc,m);
      den=1.- ren;
      num=1.- rem;
      sum+=den/num;
      
      /* store derivatives since we already computed most of it anyway */
      ddtorsion(ddangle,xyz1,xyz2,xyz3,xyz4,sign);
      dcdr = (den*m*rem - num*n*ren)/(num*num*dangle);
#ifdef DEBUG
      printf("A= %8.4lf, S= %8.4lf, F= %8.4lf  \n",angle*180.0/M_PI,den/num,dcdr);
#endif
      /*      printf("%2d torsion= %lf, switch= %lf, sum= %lf dcdr= %lf",i,angle,den/num,sum,dcdr);  */
      ffx[iff] = dcdr * ddangle[0];
      ffy[iff] = dcdr * ddangle[1];
      ffz[iff] = dcdr * ddangle[2];
      ffx[iff+1] = dcdr * ddangle[3];
      ffy[iff+1] = dcdr * ddangle[4];
      ffz[iff+1] = dcdr * ddangle[5];
      ffx[iff+2] = dcdr * ddangle[6];
      ffy[iff+2] = dcdr * ddangle[7];
      ffz[iff+2] = dcdr * ddangle[8];
      ffx[iff+3] = dcdr * ddangle[9];
      ffy[iff+3] = dcdr * ddangle[10];
      ffz[iff+3] = dcdr * ddangle[11];
      iff+=4;
/*     }else{ */
/*       sum+=1.0; */
/*       dcdr=0.0; */
/* #ifdef DEBUG */
/*       printf("A= %8.4lf, S= %8.4lf, F= %8.4lf  \n",angle,1.0,dcdr); */
/* #endif */
/*       ffx[iff] = 0.0; ffy[iff] = 0.0; ffz[iff] = 0.0; */
/*       ffx[iff+1] = 0.0; ffy[iff+1] = 0.0; ffz[iff+1] = 0.0; */
/*       ffx[iff+2] = 0.0; ffy[iff+2] = 0.0; ffz[iff+2] = 0.0; */
/*       ffx[iff+3] = 0.0; ffy[iff+3] = 0.0; ffz[iff+3] = 0.0; */
/*       iff+=4; */
/*     } */
#ifdef PRINTFUNCTION
    }
    printf("successfully printed torsion switch function\n");
    printf("n= %d,  m=%d,  rc= %lf, r0[0]= %lf\n",n,m,rc,r0[0]);
    exit(1);
#endif

  }
#ifdef DEBUG2
  printf("\n");
#endif
#ifdef PRINTTORSION
  printf("  ToRsIoN\n");
#endif

  /* WHAT ABOUT THE PERIODICY OF TORSION ANGLES ???? */

  dsum = sum - coords->colvar.pcolvar[icol];  
  func = .5*fk*dsum*dsum;        /* energy from spring */
  dfunc = fk*dsum;               /* force from spring  */

#ifdef DEBUG2
  printf("\n>> angle=%lf, pcolvar=%lf, dangle=%lf, dfunc=%lf\n",angle,coords->colvar.pcolvar[icol],dangle,dfunc);
#endif
  if(iforce && (simparms->rank==0)){          /* change the forces on the atoms */
    iff=0;
    for(i=0;i<ntorsions;++i){
#ifdef DEBUG2
      printf(">> %d %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf\n",iatom1[i],ffx[iff],ffy[iff],ffz[iff],fx[iatom1[i]],fy[iatom1[i]],fz[iatom1[i]]);
      printf(">> %d %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf\n",iatom2[i],ffx[iff+1],ffy[iff+1],ffz[iff+1],fx[iatom2[i]],fy[iatom2[i]],fz[iatom2[i]]);
      printf(">> %d %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf\n",iatom3[i],ffx[iff+2],ffy[iff+2],ffz[iff+2],fx[iatom3[i]],fy[iatom3[i]],fz[iatom3[i]]);
      printf(">> %d %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf\n\n",iatom4[i],ffx[iff+3],ffy[iff+3],ffz[iff+3],fx[iatom4[i]],fy[iatom4[i]],fz[iatom4[i]]);
#endif
      fx[iatom1[i]] -= dfunc*ffx[iff];
      fy[iatom1[i]] -= dfunc*ffy[iff];
      fz[iatom1[i]] -= dfunc*ffz[iff];
      fx[iatom2[i]] -= dfunc*ffx[iff+1];
      fy[iatom2[i]] -= dfunc*ffy[iff+1];
      fz[iatom2[i]] -= dfunc*ffz[iff+1];
      fx[iatom3[i]] -= dfunc*ffx[iff+2];
      fy[iatom3[i]] -= dfunc*ffy[iff+2];
      fz[iatom3[i]] -= dfunc*ffz[iff+2];
      fx[iatom4[i]] -= dfunc*ffx[iff+3];
      fy[iatom4[i]] -= dfunc*ffy[iff+3];
      fz[iatom4[i]] -= dfunc*ffz[iff+3];
      iff+=4;
    }
  }
    
  coords->colvar.pistcolvar[icol] = sum;
  f_spring[icol] = dfunc;
  coords->colvar.fcolvar[icol] += f_spring[icol];

  return func;
}

/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
/* forces on atoms due to spring of cv icol                            */
double fcolvar_sumofcoordnums(int icol,SIMPARMS *simparms,COORDS *coords,int iforce)
{
  int i,j,ii,jj,iff;
  /* long ibegin,iend; */
  double dx,dy,dz,r,dcdr;
  double func,dfunc,sum,num,den,ren,rem,total,dtot;

  int nicolvar;
  int njcolvar;
  int *icolvar;
  int *jcolvar;
  double fk = coords->colvar.fk[icol];

  double *r0 = coords->colvar.switch_r[icol];
  int n = coords->colvar.switch_n[icol];
  int m = coords->colvar.switch_m[icol];

  double *px = coords->px;
  double *py = coords->py;
  double *pz = coords->pz;

  double *fx = coords->fxa;
  double *fy = coords->fya;
  double *fz = coords->fza;

  double *f_spring = coords->colvar.f_spring;

  static int maxnsize;
  static double *ffx,*ffy,*ffz;

  /* max possible is natoms, we need to reduce this if possible */
  nicolvar = coords->colvar.natom[icol][0];
  if(iforce == 0){
    iff=0;
    for(i=0;i<nicolvar;++i){
      njcolvar = coords->colvar.natom[icol][i+1];
      for(j=0;j<njcolvar;++j){
	iff++;
      }
    }
    if(iff>maxnsize){
      if(maxnsize==0){
	maxnsize=iff;
	ffx = (double *)malloc(iff*3*sizeof(double));
	ffy = ffx+iff;
	ffz = ffy+iff;
      } else {
	maxnsize=iff;
	ffx = (double *)realloc(ffx,iff*3*sizeof(double));
	ffy = ffx + iff;
	ffz = ffy + iff;
      }
    }
  }


  /* sum of switch functions               */
  /*                      (1-(r12/r0)^n    */
  /*          c = SUM SUM -------------    */
  /* 	                  (1-(r12/r0)^m    */

  nicolvar = coords->colvar.natom[icol][0];
  icolvar = coords->colvar.atom[icol][0];
  total=0.0;
  iff=0;
  for(i=0;i<nicolvar;++i){
    njcolvar = coords->colvar.natom[icol][i+1];
    jcolvar = coords->colvar.atom[icol][i+1];
    sum=0.0;
    for(j=0;j<njcolvar;++j){
      ii = icolvar[i];
      jj = jcolvar[j];
      dx=px[ii]-px[jj];
      dy=py[ii]-py[jj];
      dz=pz[ii]-pz[jj];
      r=sqrt(dx*dx + dy*dy + dz*dz);
      ren= pow(r/r0[i],n);
      rem= pow(r/r0[i],m);
      den=1.- ren;
      num=1.- rem;
      sum+=den/num;
      /*      printf("%2d %2d Distance between %3d and %3d is %6.3lf, switch=%6.3lf, coordnum=%6.3lf, tot=%6.3lf\n",i,j,ii,jj,r,den/num,sum,total); */

      dcdr = (den*m*rem - num*n*ren)/(r*r*num*num);
      ffx[iff] = dcdr*dx;
      ffy[iff] = dcdr*dy;
      ffz[iff] = dcdr*dz;
      iff++;
    }
    total+=sum;
  }



  dtot = total - coords->colvar.pcolvar[icol];
  func = .5*fk*dtot*dtot;       /* energy from spring */
  dfunc = fk*dtot;              /* force from spring  */
    
  if(iforce && (simparms->rank==0)){          /* change the forces on the atoms */
    iff=0;
    for(i=0;i<nicolvar;++i){
      for(j=0;j<njcolvar;++j){
        ii = icolvar[i];
        jj = jcolvar[j];
        fx[ii] -= dfunc*ffx[iff];
        fy[ii] -= dfunc*ffy[iff];
        fz[ii] -= dfunc*ffz[iff];
        fx[jj] += dfunc*ffx[iff];
        fy[jj] += dfunc*ffy[iff];
        fz[jj] += dfunc*ffz[iff];
	iff++;
      }
    }
  }

  coords->colvar.pistcolvar[icol] = total;
  f_spring[icol] = dfunc;
  coords->colvar.fcolvar[icol] += f_spring[icol];

  return func;
}
/*---------------------------------------------------------------------*/
/* forces on atoms due to spring of cv icol                            */
double fcolvar_sumofswitches(int icol,SIMPARMS *simparms,COORDS *coords,int iforce)
{
  int i,ii,jj;
  /* long ibegin,iend; */
  double dx,dy,dz,r,dcdr;
  double func,dfunc,sum,dsum,num,den,ren,rem;
  static double *ffx,*ffy,*ffz;

  /* if(coords->colvar.ndim[icol] !=2) error(); */

  int nicolvar = coords->colvar.natom[icol][0];
  int *icolvar = coords->colvar.atom[icol][0];
  int *jcolvar = coords->colvar.atom[icol][1];
  double fk = coords->colvar.fk[icol];

  double *r0 = coords->colvar.switch_r[icol];
  int n = coords->colvar.switch_n[icol];
  int m = coords->colvar.switch_m[icol];

  double *px = coords->px;
  double *py = coords->py;
  double *pz = coords->pz;

  double *fx = coords->fxa;
  double *fy = coords->fya;
  double *fz = coords->fza;

  double *f_spring = coords->colvar.f_spring;

  /* BERND BUG FIX 29 January 2009 */  

  /* first time: allocate some workspace */
  if(iforce == 0){
    ffx = (double *)malloc(nicolvar*3*sizeof(double));
    ffy = ffx+nicolvar;
    ffz = ffy+nicolvar;
  }

  /* sum of switch functions            */
  /*                   (1-(r12/r0)^n    */
  /*          c =  SUM -------------    */
  /* 	               (1-(r12/r0)^m    */

  sum=0.0;
  for(i=0;i<nicolvar;++i){
    ii = icolvar[i];
    jj = jcolvar[i];
    dx=px[ii]-px[jj];
    dy=py[ii]-py[jj];
    dz=pz[ii]-pz[jj];
    r=sqrt(dx*dx + dy*dy + dz*dz);
    ren= pow(r/r0[i],n);
    rem= pow(r/r0[i],m);
    den=1.- ren;
    num=1.- rem;
    sum+=den/num;
    /*    printf("%2d Distance between %3d and %3d is %6.3lf, switch=%6.3lf, sum=%6.3lf\n",i,ii,jj,r,den/num,sum);  */

    dcdr = (den*m*rem - num*n*ren)/(r*r*num*num);
    ffx[i] = dcdr*dx;
    ffy[i] = dcdr*dy;
    ffz[i] = dcdr*dz;
  }

  dsum = sum - coords->colvar.pcolvar[icol];
  func = .5*fk*dsum*dsum;       /* energy from spring */
  dfunc = fk*dsum;              /* force from spring  */
    
  if(iforce && (simparms->rank==0)){          /* change the forces on the atoms */
    for(i=0;i<nicolvar;++i){
      ii = icolvar[i];
      jj = jcolvar[i];
      fx[ii] -= dfunc*ffx[i];
      fy[ii] -= dfunc*ffy[i];
      fz[ii] -= dfunc*ffz[i];
      fx[jj] += dfunc*ffx[i];
      fy[jj] += dfunc*ffy[i];
      fz[jj] += dfunc*ffz[i];
    }
  }

  coords->colvar.pistcolvar[icol] = sum;
  f_spring[icol] = dfunc;
  coords->colvar.fcolvar[icol] += f_spring[icol];

  return func;
}

/*---------------------------------------------------------------------*/
/* forces on atoms due to spring of cv icol                            */
double fcolvar_dzplane(int icol,SIMPARMS *simparms,COORDS *coords,int iforce)
{
  int i,j,ii,jj;
  /* long ibegin,iend; */
  double r,dr;
  double zcomi,zcomj,tmassi,tmassj,mass;
  double *amass = coords->amass;
  double func,dfunc;

  /* if(coords->colvar.ndim[icol] !=2) error(); */

  int nicolvar = coords->colvar.natom[icol][0];  /* i= com of lipid bilayer */
  int *icolvar = coords->colvar.atom[icol][0];
  int njcolvar = coords->colvar.natom[icol][1];  /* j= object in lipid bilayer */
  int *jcolvar = coords->colvar.atom[icol][1];
  double fk = coords->colvar.fk[icol];

  double *pz = coords->pz;
  double *fz = coords->fza;

  double *f_spring = coords->colvar.f_spring;
  /*decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);*/

  zcomi = 0.0;
  zcomj = 0.0;
  tmassi = tmassj = 0.0;

  /* center of mass of atoms from lipid bilayer */
  for(i=0;i<nicolvar;++i){
    ii = icolvar[i];
    mass = amass[ii];
    zcomi += mass*pz[ii];
    tmassi += mass;
  }
  zcomi /= tmassi;

  /* center of mass of object in or above lipid bilayer */
  for(j=0;j<njcolvar;++j){
    jj = jcolvar[j];
    mass = amass[jj];
    zcomj += mass*pz[jj];
    tmassj += mass;
  }
  zcomj /= tmassj;
  
  /* heigth difference between centers of mass */ 
  r = zcomi - zcomj;
  dr = r - coords->colvar.pcolvar[icol];

  func = .5*fk*dr*dr;
  dfunc = fk*dr;
    
  /* change the force */
  if(iforce && (simparms->rank==0)){
    for(i=0;i<nicolvar;++i){
      ii = icolvar[i];
      fz[ii] -= dfunc*amass[ii]/tmassi;
    }
    for(i=0;i<njcolvar;++i){
      jj = jcolvar[i];
      fz[jj] += dfunc*amass[jj]/tmassj;
    }
  }

  coords->colvar.pistcolvar[icol] = r;
  f_spring[icol] = dfunc;
  coords->colvar.fcolvar[icol] += f_spring[icol];
  return func;
}

/*---------------------------------------------------------------------*/
/* forces on atoms due to spring of cv icol                            */
double fcolvar_angle(int icol,SIMPARMS *simparms,COORDS *coords,int iforce)
{
  int i,j,ii,jj,k,kk;
  /* long ibegin,iend; */
  double dx1,dy1,dz1;
  double dx2,dy2,dz2;
  double caa,cab,cbb,ac,angle,dangle,factor;
  double xcomi,ycomi,zcomi,xcomj,ycomj,zcomj,tmassi,tmassj,tmassk,mass;
  double xcomk,ycomk,zcomk;
  double *amass = coords->amass;
  double func,dfunc,dthdcos;
  double ddangle[9];

  /* if(coords->colvar.ndim[icol] !=2) error(); */

  int nicolvar = coords->colvar.natom[icol][0];
  int *icolvar = coords->colvar.atom[icol][0];
  int njcolvar = coords->colvar.natom[icol][1];
  int *jcolvar = coords->colvar.atom[icol][1];
  int nkcolvar = coords->colvar.natom[icol][2];
  int *kcolvar = coords->colvar.atom[icol][2];
  double fk = coords->colvar.fk[icol];

  double *px = coords->px;
  double *py = coords->py;
  double *pz = coords->pz;

  double *fx = coords->fxa;
  double *fy = coords->fya;
  double *fz = coords->fza;

  double *f_spring = coords->colvar.f_spring;
  /*decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);*/
  xcomi = ycomi = zcomi = 0.0;
  xcomj = ycomj = zcomj = 0.0;
  xcomk = ycomk = zcomk = 0.0;
  tmassi = tmassj = tmassk = 0.0;

  /* center of mass on one end of the angle  */
  for(i=0;i<nicolvar;++i){
    ii = icolvar[i];
    mass = amass[ii];
    xcomi += mass*px[ii];
    ycomi += mass*py[ii];
    zcomi += mass*pz[ii];
    tmassi += mass;
  }
  xcomi /= tmassi;
  ycomi /= tmassi;
  zcomi /= tmassi;

  /* center of mass on middle end of the angle */
  for(j=0;j<njcolvar;++j){
    jj = jcolvar[j];
    mass = amass[jj];
    xcomj += mass*px[jj];
    ycomj += mass*py[jj];
    zcomj += mass*pz[jj];
    tmassj += mass;
  }
  xcomj /= tmassj;
  ycomj /= tmassj;
  zcomj /= tmassj;
  
  /* center of mass on other end of the angle */
  for(k=0;k<nkcolvar;++k){
    kk = kcolvar[k];
    mass = amass[kk];
    xcomk += mass*px[kk];
    ycomk += mass*py[kk];
    zcomk += mass*pz[kk];
    tmassk += mass;
  }
  xcomk /= tmassk;
  ycomk /= tmassk;
  zcomk /= tmassk;
  
  /* distance between centers of mass */
  dx1 = xcomi - xcomj;  
  dy1 = ycomi - ycomj;  
  dz1 = zcomi - zcomj;
  dx2 = xcomj - xcomk;  
  dy2 = ycomj - ycomk;  
  dz2 = zcomj - zcomk;

  /* compute angle */
  caa = dx1*dx1+dy1*dy1+dz1*dz1;
  cab = dx1*dx2+dy1*dy2+dz1*dz2;
  cbb = dx2*dx2+dy2*dy2+dz2*dz2;
  ac = cab/sqrt(caa*cbb);
  if(ac<-1.0) ac=-1.0;
  if(ac>1.0) ac=1.0;
  angle = acos(-ac);

  /* Don't use p.d.b's! because the angle runs to PI, not 2PI */
  /* have to use refelctive boundaries instead                */
  /* factor = M_PI*floor(coords->colvar.pcolvar[icol]/M_PI);  */
  /* coords->colvar.pcolvar[icol] -= factor;                  */
  dangle = angle - coords->colvar.pcolvar[icol];

  func = .5*fk*dangle*dangle;
  dfunc = fk*dangle;
    
  /* change the force */
  if(iforce && (simparms->rank==0)){

  /* derivatives   */
    dthdcos = -1./sqrt(1.-ac*ac);
    factor = -1.0*dthdcos*dfunc/sqrt(caa*cbb);
    ddangle[0] =  factor*(cab/caa*dx1-dx2);   /* atom 1 */
    ddangle[1] =  factor*(cab/caa*dy1-dy2);
    ddangle[2] =  factor*(cab/caa*dz1-dz2);
    ddangle[3] = -factor*(cab/caa*dx1-cab/cbb*dx2+dx1-dx2);  /* atom 2 */
    ddangle[4] = -factor*(cab/caa*dy1-cab/cbb*dy2+dy1-dy2);
    ddangle[5] = -factor*(cab/caa*dz1-cab/cbb*dz2+dz1-dz2);
    ddangle[6] = -factor*(cab/cbb*dx2-dx1);   /* atom 3 */
    ddangle[7] = -factor*(cab/cbb*dy2-dy1);
    ddangle[8] = -factor*(cab/cbb*dz2-dz1);
 
    for(i=0;i<nicolvar;++i){
      ii = icolvar[i];
      fx[ii] += ddangle[0]*amass[ii]/tmassi;
      fy[ii] += ddangle[1]*amass[ii]/tmassi;
      fz[ii] += ddangle[2]*amass[ii]/tmassi;
    }
    for(i=0;i<njcolvar;++i){
      jj = jcolvar[i];
      fx[jj] += ddangle[3]*amass[jj]/tmassj;
      fy[jj] += ddangle[4]*amass[jj]/tmassj;
      fz[jj] += ddangle[5]*amass[jj]/tmassj;
    }
    for(i=0;i<nkcolvar;++i){
      kk = kcolvar[i];
      fx[kk] += ddangle[6]*amass[kk]/tmassk;
      fy[kk] += ddangle[7]*amass[kk]/tmassk;
      fz[kk] += ddangle[8]*amass[kk]/tmassk;
    }
  }

  coords->colvar.pistcolvar[icol] = angle;
  f_spring[icol] = dfunc;
  coords->colvar.fcolvar[icol] += f_spring[icol];
  return func;
}
/*---------------------------------------------------------------------*/
/* forces on atoms due to spring of cv icol                            */
double fcolvar_zangle(int icol,SIMPARMS *simparms,COORDS *coords,int iforce)
{
  int i,j,ii,jj;
  /* long ibegin,iend; */
  double dx1,dy1,dz1;
  double dx2,dy2,dz2;
  double caa,cab,cbb,ac,angle,dangle,factor;
  double xcomi,ycomi,zcomi,xcomj,ycomj,zcomj,tmassi,tmassj,mass;
  double xcomk,ycomk,zcomk;
  double *amass = coords->amass;
  double func,dfunc,dthdcos;
  double ddangle[6];

  /* if(coords->colvar.ndim[icol] !=2) error(); */

  int nicolvar = coords->colvar.natom[icol][0];
  int *icolvar = coords->colvar.atom[icol][0];
  int njcolvar = coords->colvar.natom[icol][1];
  int *jcolvar = coords->colvar.atom[icol][1];
  double fk = coords->colvar.fk[icol];

  double *px = coords->px;
  double *py = coords->py;
  double *pz = coords->pz;

  double *fx = coords->fxa;
  double *fy = coords->fya;
  double *fz = coords->fza;

  double *f_spring = coords->colvar.f_spring;
  /*decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);*/
  xcomi = ycomi = zcomi = 0.0;
  xcomj = ycomj = zcomj = 0.0;
  xcomk = ycomk = zcomk = 0.0;
  tmassi = tmassj = 0.0;

  /* center of mass on one end of the angle  */
  for(i=0;i<nicolvar;++i){
    ii = icolvar[i];
    mass = amass[ii];
    xcomi += mass*px[ii];
    ycomi += mass*py[ii];
    zcomi += mass*pz[ii];
    tmassi += mass;
  }
  xcomi /= tmassi;
  ycomi /= tmassi;
  zcomi /= tmassi;

  /* center of mass on middle end of the angle */
  for(j=0;j<njcolvar;++j){
    jj = jcolvar[j];
    mass = amass[jj];
    xcomj += mass*px[jj];
    ycomj += mass*py[jj];
    zcomj += mass*pz[jj];
    tmassj += mass;
  }
  xcomj /= tmassj;
  ycomj /= tmassj;
  zcomj /= tmassj;
  
  /* center of mass on other end of the angle */
  xcomk = xcomj;
  ycomk = ycomj;
  zcomk = zcomj - 1000.0;
  
  /* distance between centers of mass */
  dx1 = xcomi - xcomj;  
  dy1 = ycomi - ycomj;  
  dz1 = zcomi - zcomj;
  dx2 = xcomj - xcomk;  
  dy2 = ycomj - ycomk;  
  dz2 = zcomj - zcomk;

  /* compute angle */
  caa = dx1*dx1+dy1*dy1+dz1*dz1;
  cab = dx1*dx2+dy1*dy2+dz1*dz2;
  cbb = dx2*dx2+dy2*dy2+dz2*dz2;
  ac = cab/sqrt(caa*cbb);
  if(ac<-1.0) ac=-1.0;
  if(ac>1.0) ac=1.0;
  angle = acos(-ac);

  dangle = angle - coords->colvar.pcolvar[icol];
  func = .5*fk*dangle*dangle;
  dfunc = fk*dangle;
    
  /* change the force */
  if(iforce && (simparms->rank==0)){

  /* derivatives   */
    dthdcos = -1./sqrt(1.-ac*ac);
    factor = -1.0*dthdcos*dfunc/sqrt(caa*cbb);
    ddangle[0] =  factor*(cab/caa*dx1-dx2);   /* atom 1 */
    ddangle[1] =  factor*(cab/caa*dy1-dy2);
    ddangle[2] =  factor*(cab/caa*dz1-dz2);
    ddangle[3] =  -ddangle[0];   /* atom 2 */
    ddangle[4] =  -ddangle[1];
    ddangle[5] =  -ddangle[2];
/*     ddangle[3] = -factor*(cab/caa*dx1-cab/cbb*dx2+dx1-dx2);  /\* atom 2 *\/ */
/*     ddangle[4] = -factor*(cab/caa*dy1-cab/cbb*dy2+dy1-dy2);  */
/*     ddangle[5] = -factor*(cab/caa*dz1-cab/cbb*dz2+dz1-dz2);  */
 
    for(i=0;i<nicolvar;++i){
      ii = icolvar[i];
      fx[ii] += ddangle[0]*amass[ii]/tmassi;
      fy[ii] += ddangle[1]*amass[ii]/tmassi;
      fz[ii] += ddangle[2]*amass[ii]/tmassi;
    }
    for(i=0;i<njcolvar;++i){
      jj = jcolvar[i];
      fx[jj] += ddangle[3]*amass[jj]/tmassj;
      fy[jj] += ddangle[4]*amass[jj]/tmassj;
      fz[jj] += ddangle[5]*amass[jj]/tmassj;
    }
  }

  coords->colvar.pistcolvar[icol] = angle;
  f_spring[icol] = dfunc;
  coords->colvar.fcolvar[icol] += f_spring[icol];
  return func;
}

/*---------------------------------------------------------------------*/
/* forces on atoms due to spring of cv icol                            */
double fcolvar_torsion(int icol,SIMPARMS *simparms,COORDS *coords,int iforce)
{
#define epsilon 1.0e-12
  int i,j,ii,jj,k,kk,l,ll;
  /* long ibegin,iend; */
  double angle,dangle,jump,sign;
  double xyz1[3],xyz2[3],xyz3[3],xyz4[3];
  double tmassi,tmassj,tmassk,tmassl,mass;
  double *amass = coords->amass;
  double func,dfunc;
  double ddangle[12];

  /* if(coords->colvar.ndim[icol] !=2) error(); */

  int nicolvar = coords->colvar.natom[icol][0];
  int *icolvar = coords->colvar.atom[icol][0];
  int njcolvar = coords->colvar.natom[icol][1];
  int *jcolvar = coords->colvar.atom[icol][1];
  int nkcolvar = coords->colvar.natom[icol][2];
  int *kcolvar = coords->colvar.atom[icol][2];
  int nlcolvar = coords->colvar.natom[icol][3];
  int *lcolvar = coords->colvar.atom[icol][3];
  double fk = coords->colvar.fk[icol];

  double *px = coords->px;
  double *py = coords->py;
  double *pz = coords->pz;

  double *fx = coords->fxa;
  double *fy = coords->fya;
  double *fz = coords->fza;

  double *f_spring = coords->colvar.f_spring;
  /*decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);*/

  for(i=0;i<3;++i){
    xyz1[i]=0;
    xyz2[i]=0;
    xyz3[i]=0;
    xyz4[i]=0;
  }
  tmassi = tmassj = tmassk = tmassl = 0.0;

  /* center of mass on one end of the angle  */
  for(i=0;i<nicolvar;++i){
    ii = icolvar[i];
    mass = amass[ii];
    xyz1[0] += mass*px[ii];
    xyz1[1] += mass*py[ii];
    xyz1[2] += mass*pz[ii];
    tmassi += mass;
  }
   xyz1[0] /= tmassi;
   xyz1[1] /= tmassi;
   xyz1[2] /= tmassi;

  /* center of mass on middle end of the angle */
  for(j=0;j<njcolvar;++j){
    jj = jcolvar[j];
    mass = amass[jj];
    xyz2[0] += mass*px[jj];
    xyz2[1] += mass*py[jj];
    xyz2[2] += mass*pz[jj];
    tmassj += mass;
  }
  xyz2[0] /= tmassj;
  xyz2[1] /= tmassj;
  xyz2[2] /= tmassj;
  
  /* center of mass on other end of the angle */
  for(k=0;k<nkcolvar;++k){
    kk = kcolvar[k];
    mass = amass[kk];
    xyz3[0] += mass*px[kk];
    xyz3[1] += mass*py[kk];
    xyz3[2] += mass*pz[kk];
    tmassk += mass;
  }
  xyz3[0] /= tmassk;
  xyz3[1] /= tmassk;
  xyz3[2] /= tmassk;
  /* center of mass on other end of the angle */
  for(l=0;l<nlcolvar;++l){
    ll = lcolvar[l];
    mass = amass[ll];
    xyz4[0] += mass*px[ll];
    xyz4[1] += mass*py[ll];
    xyz4[2] += mass*pz[ll];
    tmassl += mass;
  }
  xyz4[0] /= tmassl;
  xyz4[1] /= tmassl;
  xyz4[2] /= tmassl;
 
  angle=torsion(xyz1,xyz2,xyz3,xyz4,&sign);

  /* Use p.d.b's */
  if(iforce) {
    jump = coords->colvar.pistcolvar[icol]-angle;
    if(abs(jump) > 2.0){
      coords->colvar.pcolvar[icol] -= COPY_SIGN(2.0*M_PI,jump);
    }
  }

  dangle = angle - coords->colvar.pcolvar[icol];

  func = .5*fk*dangle*dangle;
  dfunc = fk*dangle;
    
  /* change the force */
  /*  printf("\n>> angle=%lf, pcolvar=%lf, dangle=%lf, dfunc=%lf\n",angle,coords->colvar.pcolvar[icol],dangle,dfunc); */
  if(iforce && (simparms->rank==0)){
    ddtorsion(ddangle,xyz1,xyz2,xyz3,xyz4,sign);
    for(i=0;i<nicolvar;++i){
      ii = icolvar[i];
      /*      printf(">> %d %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf\n",ii,ddangle[0],ddangle[1],ddangle[2],fx[ii],fy[ii],fz[ii]);  */
      fx[ii] -= dfunc*ddangle[0]*amass[ii]/tmassi;
      fy[ii] -= dfunc*ddangle[1]*amass[ii]/tmassi;
      fz[ii] -= dfunc*ddangle[2]*amass[ii]/tmassi;
   }
    for(i=0;i<njcolvar;++i){
      jj = jcolvar[i];
      /*      printf(">> %d %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf\n",jj,ddangle[3],ddangle[4],ddangle[5],fx[jj],fy[jj],fz[jj]); */
      fx[jj] -= dfunc*ddangle[3]*amass[jj]/tmassj;
      fy[jj] -= dfunc*ddangle[4]*amass[jj]/tmassj;
      fz[jj] -= dfunc*ddangle[5]*amass[jj]/tmassj;
    }
    for(i=0;i<nkcolvar;++i){
      kk = kcolvar[i];
      /*      printf(">> %d %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf\n",kk,ddangle[6],ddangle[7],ddangle[8],fx[kk],fy[kk],fz[kk]);  */
      fx[kk] -= dfunc*ddangle[6]*amass[kk]/tmassk;
      fy[kk] -= dfunc*ddangle[7]*amass[kk]/tmassk;
      fz[kk] -= dfunc*ddangle[8]*amass[kk]/tmassk;
    }
    for(i=0;i<nlcolvar;++i){
      ll = lcolvar[i];
      /*      printf(">> %d %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf %12.4lf\n",ll,ddangle[9],ddangle[10],ddangle[11],fx[ll],fy[ll],fz[ll]);  */
      fx[ll] -= dfunc*ddangle[9]*amass[ll]/tmassl;
      fy[ll] -= dfunc*ddangle[10]*amass[ll]/tmassl;
      fz[ll] -= dfunc*ddangle[11]*amass[ll]/tmassl;
    }
  }

  coords->colvar.pistcolvar[icol] = angle;
  f_spring[icol] = dfunc;
  coords->colvar.fcolvar[icol] += f_spring[icol];
  return func;
}

/*---------------------------------------------------------------------*/
/* forces on atoms due to spring of cv icol                            */
double fcolvar_bond(int icol,SIMPARMS *simparms,COORDS *coords,int iforce)
{
  int i,j,ii,jj;
  /* long ibegin,iend; */
  double dx,dy,dz,r,dr;
  double xcomi,ycomi,zcomi,xcomj,ycomj,zcomj,tmassi,tmassj,mass;
  double *amass = coords->amass;
  double func,dfunc;

  /* if(coords->colvar.ndim[icol] !=2) error(); */

  int nicolvar = coords->colvar.natom[icol][0];
  int *icolvar = coords->colvar.atom[icol][0];
  int njcolvar = coords->colvar.natom[icol][1];
  int *jcolvar = coords->colvar.atom[icol][1];
  double fk = coords->colvar.fk[icol];

  double *px = coords->px;
  double *py = coords->py;
  double *pz = coords->pz;

  double *fx = coords->fxa;
  double *fy = coords->fya;
  double *fz = coords->fza;

  double *f_spring = coords->colvar.f_spring;
  /*decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);*/

  xcomi = ycomi = zcomi = 0.0;
  xcomj = ycomj = zcomj = 0.0;
  tmassi = tmassj = 0.0;

  /* center of mass on one end of the bond */
  for(i=0;i<nicolvar;++i){
    ii = icolvar[i];
    mass = amass[ii];
    xcomi += mass*px[ii];
    ycomi += mass*py[ii];
    zcomi += mass*pz[ii];
    tmassi += mass;
  }
  xcomi /= tmassi;
  ycomi /= tmassi;
  zcomi /= tmassi;

  /* center of mass on other end of the bond */
  for(j=0;j<njcolvar;++j){
    jj = jcolvar[j];
    mass = amass[jj];
    xcomj += mass*px[jj];
    ycomj += mass*py[jj];
    zcomj += mass*pz[jj];
    tmassj += mass;
  }
  xcomj /= tmassj;
  ycomj /= tmassj;
  zcomj /= tmassj;
  
  /* distance between centers of mass */
  dx = xcomi - xcomj;  
  dy = ycomi - ycomj;  
  dz = zcomi - zcomj;

  r = sqrt(dx*dx+dy*dy+dz*dz);
  dr = r - coords->colvar.pcolvar[icol];

  func = .5*fk*dr*dr;
  dfunc = fk*dr;
    
  /* change the force */
  if(iforce && (simparms->rank==0)){
    for(i=0;i<nicolvar;++i){
      ii = icolvar[i];
      fx[ii] -= dfunc*dx*amass[ii]/(r*tmassi);
      fy[ii] -= dfunc*dy*amass[ii]/(r*tmassi);
      fz[ii] -= dfunc*dz*amass[ii]/(r*tmassi);
    }
    for(i=0;i<njcolvar;++i){
      jj = jcolvar[i];
      fx[jj] += dfunc*dx*amass[jj]/(r*tmassj);
      fy[jj] += dfunc*dy*amass[jj]/(r*tmassj);
      fz[jj] += dfunc*dz*amass[jj]/(r*tmassj);
    }
  }

  coords->colvar.pistcolvar[icol] = r;
  f_spring[icol] = dfunc;
  coords->colvar.fcolvar[icol] += f_spring[icol];
  return func;
}
/*---------------------------------------------------------------------*/
void fcolvar_cntr(SIMPARMS *simparms,COORDS *coords,int iforce)
{              /* if iforce=0 forces on atoms are not applied    */
               /* for initialization called from input/colvar.c  */
  int ncolvar = coords->colvar.ncolvar;
  int *type = coords->colvar.type;
  int *lhills = coords->colvar.lhills;
  int *lconst_vel = coords->colvar.lconst_vel;
  int lthills = 0;
  int icol;
  double dxmax,dxmin,dx;
  char line[MAXLINELEN];

  if(ncolvar==0) return;
  /* get forces on atoms from harmonic springs with colvars */

  for(icol=0;icol<ncolvar;++icol) coords->colvar.fcolvar[icol] = 0.0;

  for(icol=0;icol<ncolvar;++icol){
    switch(type[icol]) {
    case 0: /* distance */
      coords->colvar.ecolvar += fcolvar_bond(icol,simparms,coords,iforce);
      break;
    case 1: /* angle */
      coords->colvar.ecolvar += fcolvar_angle(icol,simparms,coords,iforce);
      break;
    case 2: /* torsion */
      coords->colvar.ecolvar += fcolvar_torsion(icol,simparms,coords,iforce);
      break;
    case 4: /* sum of coordination numbers */
      coords->colvar.ecolvar += fcolvar_sumofcoordnums(icol,simparms,coords,iforce);
      break;
    case 5: /* dzplane */
      coords->colvar.ecolvar += fcolvar_dzplane(icol,simparms,coords,iforce);
      break;
    case 6: /* sumofswitches */
      coords->colvar.ecolvar += fcolvar_sumofswitches(icol,simparms,coords,iforce);
      break;
    case 7: /* torsionswitches */
      coords->colvar.ecolvar += fcolvar_torsionswitches(icol,simparms,coords,iforce);
      break;
    case 8: /* zangle */
      coords->colvar.ecolvar += fcolvar_zangle(icol,simparms,coords,iforce);
      break;
    default:
      md_error("in meta_dyn colvar type not known!");
    }      
    if(lhills[icol]){
      lthills=1;
      if(coords->colvar.ltunehills==2) coords->colvar.fcolvar[icol] = 0.0;
    }
    if(lconst_vel[icol]) coords->colvar.fcolvar[icol] = 0.0;

    /* Reset pcolvar to pistcolvar and zero velocities and forces */
    if(iforce==0){
      if(coords->colvar.pcolvar[icol] == RARE){
	coords->colvar.pcolvar[icol] = coords->colvar.pistcolvar[icol];
      }
      if(coords->colvar.lconst_vel[icol]==0){
	coords->colvar.vcolvar[icol] = 0.0;
      }
      coords->colvar.fcolvar[icol] = 0.0;
      coords->colvar.f_hills[icol] = 0.0;
      coords->colvar.f_spring[icol] = 0.0;
    }
  }

  /* put hills and get forces on colvars from hills */
  if(lthills) hills(simparms,coords);

  /* scale velocities if we are too hot */
  scale_colvar(&coords->colvar);

  /* set reflective boundaries */

  for(icol=0;icol<ncolvar;++icol){
    dxmin = coords->colvar.pcolvar[icol] - coords->colvar.min_val[icol];
    dxmax = coords->colvar.pcolvar[icol] - coords->colvar.max_val[icol];
    dx = MIN(fabs(dxmax),fabs(dxmin));
    if( dx > (coords->colvar.max_val[icol]-coords->colvar.min_val[icol]) ){
      sprintf(line,"Boundaries for coll.var. %d to close together.",icol);
      md_warning(line);
      coords->colvar.pcolvar[icol]=coords->colvar.max_val[icol];
      coords->colvar.vcolvar[icol]=0.0;
      dxmin=0.0;
      dxmax=0.0;
    }
    if(dxmin < 0.0 ){
      coords->colvar.pcolvar[icol] -= 2.0*dxmin ;
      coords->colvar.vcolvar[icol] *= -1.0 ;

/* Adding a hill at the boundary is not a good idea if the colvar  */
/* diffuses slowly. This should be turned into an optional feature */
/*      if(lthills){
	for(icol=0;icol<ncolvar;++icol){
	  coords->colvar.pcolvarlast[icol] = coords->colvar.pcolvar[icol];
	}
	add_hills(&(coords->colvar),simparms->istep);
	coords->colvar.lasthill = simparms->istep;
      }
*/
    }else if( dxmax > 0.0 ){
      coords->colvar.pcolvar[icol] -= 2.0*dxmax ;
      coords->colvar.vcolvar[icol] *= -1.0 ;

/*
      if(lthills){
	for(icol=0;icol<ncolvar;++icol){
	  coords->colvar.pcolvarlast[icol] = coords->colvar.pcolvar[icol];
	}
	add_hills(&(coords->colvar),simparms->istep);
	coords->colvar.lasthill = simparms->istep;
      }
*/
    }
  }
  
  /* change force to acel */

  for(icol=0;icol<ncolvar;++icol){
    coords->colvar.fcolvar[icol] /= coords->colvar.amass[icol];
  }
}
/*---------------------------------------------------------------------*/


