/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/* routine to get the forces from an interaction list */

#define NOERROR
#include "md.h"
#define NOPRINT_INTERACTIONSWAPS

#define RESCALE_CGBONDS
#define RCGBOND 3.71   /* 3.71 have to make this an input variable */

static double spotl[2],spothl[2],sprsl[2],sprslt[2][9],spotlso3;


/* #define VERBOSE */
/* #define PRINT_VDW_ENERGIES */

void zero_potl(int icg)
{
  int i;
  for(i=0;i<9;i++) sprslt[icg][i] = 0;
  spotl[icg]= spothl[icg] = sprsl[icg] = 0.;
}
/*-----------------------------------------------------------------------*/
void zero_potlso3()
{
  spotlso3= 0.;
}
/*-----------------------------------------------------------------------*/
void force_npol(SIMPARMS *simparms,COORDS *coords,int nlen,int *indx,int *jndx,
		int nbreak,int *ibreak,INTER *inter,int iflg)
{
  int natoms,iperd,ivol,natoms2,icg;
  int j,k,in,ni,nj,jer,ierror,ntable,ntm4;
  int *ityp,*jtyp,*itype,**imap,*ifhr,nswp;
  double swv,f,p,fm1,f0,f1,dis2,fcc,spotlt,spothlt;
  double *xdis,*ydis,*zdis,*fxds,*fyds,*fzds,**sw,*dvt,*vt;
  double *x,*y,*z,*hmat,*scr,*rec, *fhr;
  double *faacgp, *faacg, *dsp, *ds, *vdsl, *vdslp, *pot;
  double *rmax2,*dx2tab,**dvtab,*rminh2,*rmint2,**vtab;
  double hmati[9];
  LINE line;
#ifdef CUBIC
  double f2;
#endif
#ifdef FIXHYBRIDSWAPS
	int nswptot = inter->nswptot;
	int nswpmax = inter->nswpmax;
	int *iatswp = inter->iatswp;
	int *jatswp = inter->jatswp;
	nswp = inter->nswp;
	iswp = inter->iswp;
	double *swppot = inter->swppot;
	double *dr = simparms->ddr_lrf + (1-coords->lcg)*simparms->ncg;
	double *drnow = simparms->dr + (1-coords->lcg)*simparms->ncg;
#endif	

  nswp = 0;
  x = coords->px;  y = coords->py;  z = coords->pz; hmat = coords->hmat;
  icg = coords->lcg;
  scr = coords->scr_buf; rec = coords->scr_rec;
  natoms= coords->natoms;
  iperd = simparms->iperd;
  ivol  = simparms->ivol;
  natoms2 = 2*natoms;
  spotlt = spothlt = 0.;
  rmax2 = NULL;
  ntable= inter->ntable;
  ntm4  = ntable-4;
  itype = inter->itype;
  sw    = inter->switchs;
  imap  = inter->map;
  rminh2= inter->rminl_cut2;
  /* rmaxh2= inter->rmaxs_cut2; if you want a long range force short cutoff */
  rmint2= inter->rmins_cut2;
  dx2tab= inter->dx2tab;
  dvtab = inter->dvtab;
  vtab  = inter->vtab;
  faacg = simparms->faacg + (1-coords->lcg)*simparms->ncg;
  faacgp = simparms->faacgprev + (1-coords->lcg)*simparms->ncg;
  ds = simparms->dfaacg + (1-coords->lcg)*simparms->ncg;
  dsp = simparms->dfaacgprev + (1-coords->lcg)*simparms->ncg;
  vdsl = simparms->vds_lrf + (1-coords->lcg)*simparms->ncg;
  vdslp = simparms->vds_lrfprev + (1-coords->lcg)*simparms->ncg;

  /* switch to long pair or short pair */
  switch(iflg){
  case 1:
    rmax2 = inter->rmaxs_cut2;  
    break;
  case 0:
    rmax2 = inter->rmaxl_cut2; 
    break;
  default:
    md_error("ERROR: in inter-molecular force routine?!?!?");
  }

  jer = -1; swv = 1.;  ierror=0;
  gethinv9(hmat,hmati);
  ityp = (int *)cmalloc(3*nlen*sizeof(int));
  jtyp = ityp +   nlen;
  ifhr = ityp + 2*nlen;
  xdis = (double *)cmalloc(8*nlen*sizeof(double));
  ydis = xdis +   nlen;
  zdis = xdis + 2*nlen;
  fxds = xdis + 3*nlen;  fyds = xdis + 4*nlen;  fzds = xdis + 5*nlen;
  fhr = xdis + 6*nlen;
  pot = xdis + 7*nlen;

#ifdef VERBOSE
  for(j=0;j<nlen;j++)printf("interaction %d pair %d - %d\n",j,indx[j],jndx[j]);
#endif

#include "vectorize.h"
  for(j=0;j<nlen;j++){
    ityp[j] = itype[indx[j]]; 
    jtyp[j] = itype[jndx[j]];
  }
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    ni = indx[j]; nj = jndx[j];
    xdis[j] = x[ni] - x[nj];
    ydis[j] = y[ni] - y[nj];
    zdis[j] = z[ni] - z[nj];
  }

  if(simparms->icalc_type == 5){
    if(coords->lcg){
#include "vectorize.h"
			for(j=0;j<nlen;j++){
				ni = indx[j]; nj = jndx[j];
#ifdef PRINT_INTERACTIONSWAPS
				if( (faacg[ni]>faacg[nj]) != (faacgp[ni]>faacgp[nj]) ){
					if( (faacg[ni]==0) || (faacg[ni]==1)){
					}else{
						iswp[nswp++]=j;
						printf("OOPS: interaction swap\n");
					}
				}
#endif
				ni = ifhr[j] = (faacg[ni]>faacg[nj])? ni:nj;
				fhr[j] = faacg[ni];
			}
    }else{
#include "vectorize.h"
      for(j=0;j<nlen;j++){
				ni = indx[j]; nj = jndx[j];
#ifdef PRINT_INTERACTIONSWAPS
				if( (faacg[ni]>faacg[nj]) != (faacgp[ni]>faacgp[nj]) ){
					if( (faacg[ni]==0) || (faacg[ni]==1)){
					}else{
						iswp[nswp++]=j;
						printf("OOPS: interaction swap\n");
					}
				}
#endif
				ni = ifhr[j] = (faacg[ni]<faacg[nj])? ni:nj;
				fhr[j] = faacg[ni];
			}
    }
  }else{
#include "vectorize.h"
		for(j=0;j<nlen;j++){
			fhr[j]=1.;
			ifhr[j]=0;
		}
  }
  period(nlen,xdis,ydis,zdis,hmat,hmati,iperd,ivol);
  
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    in = imap[ityp[j]][jtyp[j]];
    dis2 = xdis[j]*xdis[j]+ydis[j]*ydis[j]+zdis[j]*zdis[j];

    if(dis2<rmax2[in]){
      p = (dis2-rmint2[in])*dx2tab[in];
      k = (int)p;
      if(k<1){
#ifdef NOERROR
	ierror++;
#endif
	jer = j;k=1;p=1.;
      }
      
      p = p-(double)k;
      dvt = dvtab[in];
      fm1 = dvt[k-1]; f0=dvt[k]; f1=dvt[k+1];
#ifdef CUBIC
      f2  = dvt[k+2];
      f=f0+(p/2)*((2*f1-2*fm1/3-f0-f2/3)+
		  p*((fm1-2*f0+f1)+p*(f0-fm1/3-f1+f2/3)));
#else
      f = f0 + .5*p*(f1-fm1 + p*(fm1-2.*f0+f1));
#endif
      f /= sqrt(dis2);
			f *= fhr[j];  /* hybrid switch added here BERND */
      if(dis2>rminh2[in]){
        dvt = sw[in];
        fm1 = dvt[k-1];	f0=dvt[k];  f1=dvt[k+1];
#ifdef CUBIC
        f2  = dvt[k+2];
        swv = f0+(p/2)*((2*f1-2*fm1/3-f0-f2/3)+
           p*((fm1-2*f0+f1)+p*(f0-fm1/3-f1+f2/3)));
#else
        swv = f0 + .5*p*(f1-fm1 + p*(fm1-2.*f0+f1));
#endif
      } else {
        swv = 1.;
      }      
#ifdef VERBOSE 
      printf("force[%d] = %f (%f %d %f %d)\n",j,f,swv,k,p,in);
#endif
      if(iflg==0){
        sprsl[icg] -= dis2*f;
        sprslt[icg][0] -= xdis[j]*xdis[j]*f;
        sprslt[icg][1] -= ydis[j]*xdis[j]*f;
        sprslt[icg][2] -= zdis[j]*xdis[j]*f;
        sprslt[icg][4] -= ydis[j]*ydis[j]*f;
        sprslt[icg][5] -= zdis[j]*ydis[j]*f;
        sprslt[icg][8] -= zdis[j]*zdis[j]*f;
        
        xdis[j] *= f;
        ydis[j] *= f;  
        zdis[j] *= f;
        fxds[j] = xdis[j]*swv;
        fyds[j] = ydis[j]*swv;
        fzds[j] = zdis[j]*swv;
        
        vt = vtab[in];
        fm1=vt[k-1];	f0 =vt[k];	f1 =vt[k+1];
#ifdef CUBIC
        f2  = vt[k+2];
        fcc = f0+(p/2)*((2*f1-2*fm1/3-f0-f2/3)+
           p*((fm1-2*f0+f1)+p*(f0-fm1/3-f1+f2/3)));
#else
        fcc = f0 + .5*p*(f1-fm1 + p*(fm1-2.*f0+f1));
#endif
				pot[j] = fcc-vt[ntm4];
        spotlt  += fcc*fhr[j];
        spothlt += (fcc-vt[ntm4])*fhr[j];
#ifdef VERBOSE 
        printf("potential[%d] = %f\n",j,fcc/KCAL);
#endif
      } else {
        f *= swv;
        xdis[j] *= f;
        ydis[j] *= f;  
        zdis[j] *= f;
      }
    } else {
      xdis[j] = 0.;      ydis[j] = 0.;      zdis[j] = 0.;
      fxds[j] = 0.;      fyds[j] = 0.;      fzds[j] = 0.;
	  pot[j] = 0.;
    }
  }
  if(iflg==0){
    sprslt[icg][3] = sprslt[icg][1];
    sprslt[icg][6] = sprslt[icg][2];
    sprslt[icg][7] = sprslt[icg][5];
  }
  spotl[icg] += spotlt;
  spothl[icg] += spothlt;
  
  /* fix swapped interactions in healing region, add correction */
#ifdef FIXHYBRIDSWAPS
  for(i=0;i<nswp;i++){
		j = iswp[i];
		ni = ifhr[j];
		nj = (ni==indx[j])? jndx[j]:indx[j];
		for(k=0;k<nswpmax;k++) if((ni==jatswp[k])&&(nj==iatswp[k])) break;
		if(k==nswpmax) md_error("Did not find previous interaction in swap correction");
		simparms->vbuf[ni] -= 0.5*( swppot[k]*dsp[nj]*(drnow[ni]-drnow[nj]) +
																pot[j]*ds[ni]*(drnow[ni]+dr[ni]-drnow[nj]-dr[nj]) );
  }
	/* store each scaled interaction */
	if(nswptot+nlen>nswpmax){
		nswpmax = inter->nswpmax += nlen;
		printf("Max number of VdW interactions stored for swap correction increased to %d\n",nswpmax);
		inter->iatswp = (int *)realloc(inter->iatswp,nswpmax*sizeof(int));
		inter->jatswp = (int *)realloc(inter->jatswp,nswpmax*sizeof(int));
		inter->swppot = (double *)realloc(inter->swppot,nswpmax*sizeof(double));
	}
  for(j=0;j<nlen;j++){
		if((fhr[j]!=0)&&(fhr[j]!=1)){
			iatswp[nswptot] = ifhr[j]; 
			jatswp[nswptot] = (ifhr[j]==indx[j])? jndx[j]:indx[j];
			swppot[nswptot] = pot[j];
			nswptot++;
		}
	}
#endif  
  
  /* recursion use breakpoints (offsets) to vectorize inner loop */
  for(k=0;k<nbreak;k++){
#ifdef VERBOSE
    printf("break point %d from %d to %d\n",k,ibreak[k],ibreak[k+1]);
#endif
#include "vectorize.h"
    for(j=ibreak[k];j<ibreak[k+1];j++){
      ni=indx[j];
      nj=jndx[j];
#ifdef VERBOSE
      printf("adding force pairs %d %d-%d\n",j,indx[j],jndx[j]);
#endif
      scr[ni        ] -= xdis[j];
      scr[ni+natoms ] -= ydis[j];
      scr[ni+natoms2] -= zdis[j];
      scr[nj        ] += xdis[j];
      scr[nj+natoms ] += ydis[j];
      scr[nj+natoms2] += zdis[j];
      if(iflg==0){
        rec[ni        ] -= fxds[j];
        rec[ni+natoms ] -= fyds[j];
        rec[ni+natoms2] -= fzds[j];
        rec[nj        ] += fxds[j];
        rec[nj+natoms ] += fyds[j];
        rec[nj+natoms2] += fzds[j];
	    in=ifhr[j];
	    vdsl[in] += pot[j]*ds[in];
      }
    }
  }

  /* free temporary vectors */
  free(ityp);
  free(xdis);
  
#ifdef NOERROR
  if(ierror){
    sprintf(line,"%d distance out of range in force routine\n",ierror);
    sprintf(line,"%s (ie atoms %d and %d)",line,indx[jer],jndx[jer]);
    md_warning(line);
  }
#endif
}
/*-----------------------------------------------------------------------*/
void vinter_npol(int nlen,int *indx,int *jndx,int *itype,
		 double *x,double *y,double *z,double *hmat,int iperd,
		 int **imap,double *rmin_cut2,double *rmax_cut2,
		 double *dx2tab,int ntable,
		 double **vtab,double **dvtab,double *spot,double *spoth,
		 double *sprs,double sprst[9],int ivol)
{
  int i,j,k,*ityp,*jtyp,intr,ierror,jer,in,jn;
  double *xdis,*ydis,*zdis,dis2,p;
  double fm1,f0,f1,f,spotl,spothl,sprsl;
  double sprslt[9];
  double hmati[9];
  LINE line;
#ifdef CUBIC
  double f2;
#endif

  jer = -1;
  ierror = 0;
  gethinv9(hmat,hmati);
  ityp = (int *)cmalloc(2*nlen*sizeof(int));
  jtyp = ityp + nlen;
  xdis = (double *)cmalloc(3*nlen*sizeof(double));
  ydis = xdis +   nlen;
  zdis = xdis + 2*nlen;
  spotl = spothl = sprsl = 0.0;
  for(i=0;i<9;i++)  sprslt[i]=0.0;

#include "vectorize.h"
  for(j=0;j<nlen;j++){
    ityp[j] = itype[indx[j]]; 
    jtyp[j] = itype[jndx[j]];
  }
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    in = indx[j] ; jn = jndx[j];
    xdis[j] = x[in] - x[jn];
    ydis[j] = y[in] - y[jn];
    zdis[j] = z[in] - z[jn];
  }
  
  period(nlen,xdis,ydis,zdis,hmat,hmati,iperd,ivol);
      
#include "vectorize.h"
  for(j=0;j<nlen;j++){
    intr = imap[ityp[j]][jtyp[j]];
    dis2 = xdis[j]*xdis[j]+ydis[j]*ydis[j]+zdis[j]*zdis[j];
    
    if(dis2<rmax_cut2[intr]){
      p = (dis2-rmin_cut2[intr])*dx2tab[intr];
      k = (int)p;
      if(k<1) {
#ifdef NOERROR
	ierror= 1; 
#endif
	jer = j;k = 1; p=1.;
      }
      
      p = p-(double)k;
      fm1=vtab[intr][k-1];
      f0 =vtab[intr][k];
      f1 =vtab[intr][k+1];
      
#ifdef CUBIC
      f2  = vtab[intr][k+2];
      f=f0+(p/2)*((2*f1-2*fm1/3-f0-f2/3)+
		  p*((fm1-2*f0+f1)+p*(f0-fm1/3-f1+f2/3)));
#else
      f = f0 + .5*p*(f1-fm1 + p*(fm1-2.*f0+f1));
#endif
      spotl  += f;
      spothl += f-vtab[intr][ntable-4];
      
#ifdef PRINT_VDW_ENERGIES
      printf("VDW %d-%d = %g\n",indx[j],jndx[j],f);
#endif
      fm1 = dvtab[intr][k-1];
      f0  = dvtab[intr][k];
      f1  = dvtab[intr][k+1];
#ifdef CUBIC
      f2  = dvtab[intr][k+2];
      f=f0+(p/2)*((2*f1-2*fm1/3-f0-f2/3)+
		  p*((fm1-2*f0+f1)+p*(f0-fm1/3-f1+f2/3)));
#else
      f = f0 + .5*p*(f1-fm1 + p*(fm1-2.*f0+f1));
#endif
      f /= sqrt(dis2);
      sprsl -= dis2*f;
      sprslt[0] -= xdis[j]*xdis[j]*f;
      sprslt[1] -= ydis[j]*xdis[j]*f;
      sprslt[2] -= zdis[j]*xdis[j]*f;

      sprslt[3] -= xdis[j]*ydis[j]*f;
      sprslt[4] -= ydis[j]*ydis[j]*f;
      sprslt[5] -= zdis[j]*ydis[j]*f;

      sprslt[6] -= xdis[j]*zdis[j]*f;
      sprslt[7] -= ydis[j]*zdis[j]*f;
      sprslt[8] -= zdis[j]*zdis[j]*f;

#ifdef VERBOSE 
      printf("distance %d %g %d %g %g\n",j,sqrt(dis2),intr,
	     sqrt(rmax_cut2[intr]),f);
#endif
    }
  }
#ifdef NOERROR
  if(ierror){
    sprintf(line,"distance out of range in energy routine\n");
    sprintf(line,"%s between atoms %d and %d)",line,indx[jer],jndx[jer]);
    md_warning(line);
  }
#endif

  *spot += spotl;
  *spoth += spothl;
  *sprs += sprsl;
  for(i=0;i<9;i++)  sprst[i] += sprslt[i];

  /* free temporary vectors */
  free(ityp);
  free(xdis);

#ifdef VERBOSE
  printf("energies %g %g %g\n",*spot,*spoth,*sprs);
#endif

}
/*-----------------------------------------------------------------------*/
void getsvirter(double *vhinter,double *vinter,double *winter,
		double wintert[9], int icg)
{
  int i;
#ifdef PARA
  double sprst[12],sprlt[12];

  for(i=0;i<9;i++) sprst[i] = sprslt[icg][i];
  sprst[9] = spothl[icg];
  sprst[10] = spotl[icg];
  sprst[11] = sprsl[icg];
  MPI_Allreduce(sprst,sprlt,12,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  for(i=0;i<9;i++) sprslt[icg][i] = sprlt[i];
  spothl[icg] = sprlt[9];
  spotl[icg] = sprlt[10];
  sprsl[icg] = sprlt[11];
#endif

  *vhinter += spothl[icg];
  *vinter  += spotl[icg];
  *winter  += sprsl[icg];
  for(i=0;i<9;i++) wintert[i] += sprslt[icg][i];
}

/*-----------------------------------------------------------------------*/
void getsvirterso3(double *vinter)
{
  *vinter  = spotlso3;
}

/*-----------------------------------------------------------------------*/
/* non-bonding AA repulsion in healing region                            */
void fnonbond_so3(SIMPARMS *simparms,COORDS *coords,INTER *inter,SO3 *so3,
    int nbonds,int *indx,int *jndx,int iregion)
{
  int i,j,ii,nb;
  int icg, jcg, *icgunit, lcoupling;
  int iperd, ivol;
  long ibegin,iend;
  double dx,dy,dz,r,rcg,fr,dr,fhr,*faacg,*ds,*vds,pot,spotlt;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz, dcgx, dcgy, dcgz;
  double px1, py1, pz1, px2, py2, pz2;
  double tx1, ty1, tz1, tx2, ty2, tz2;
  double *fx,*fy,*fz,*fcgx,*fcgy,*fcgz;
  double (*tor)[3];
  double *hmat, hmati[9];

  iperd = simparms->iperd;
  ivol  = simparms->ivol;
  hmat = coords->hmat;
  gethinv9(hmat,hmati);
  spotlt = 0.;

  lcoupling = (simparms->iso3dyn == 1);

  px = coords->px;   py = coords->py;   pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  fx = coords->fxt;  fy = coords->fyt;  fz = coords->fzt;
  fcgx = coords->next->fxt;   fcgy = coords->next->fyt;   fcgz = coords->next->fzt;

//  if( iregion == 0){
//    /* coarse grain region selection */
//    nbonds = coords->bonds.nsel_so3cgr;
//    iselect = coords->bonds.isel_so3cgr;
//  }else{
//    /* healing region selection */
//    nbonds = coords->bonds.nsel_so3hr;
//    iselect = coords->bonds.isel_so3hr;		
//  }
//  ibond = coords->bonds.ibond2;
//  jbond = coords->bonds.jbond2;
//  idxbond = coords->bonds.idxbond2;
//  itypbond = coords->bonds.itypbond;
//  eqbond = coords->bonds.eqbond;
//  fkbond = coords->bonds.fkbond;
//  bmorse = coords->bonds.bmorse;
  icgunit = simparms->icgunit;
  faacg = simparms->faacg + simparms->ncg;
  ds = simparms->dfaacg + simparms->ncg;
  vds = simparms->vds_intra + simparms->ncg; 
  tor = so3->tor;
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
  
  fr = 0.;
  for(nb=ibegin;nb<iend;nb++){ 
    i = indx[nb]; j = jndx[nb]; 
//    idx = idxbond[ib];

    px1 = px[i]; py1 = py[i]; pz1 = pz[i]; 
    px2 = px[j]; py2 = py[j]; pz2 = pz[j];
    icg = icgunit[i]; jcg = icgunit[j];
    
    /* renormalize for distance between CG beads */
    dcgx = pcgx[icg] - pcgx[jcg];  
    dcgy = pcgy[icg] - pcgy[jcg];
    dcgz = pcgz[icg] - pcgz[jcg];		
    period(1,&dcgx,&dcgy,&dcgz,hmat,hmati,iperd,ivol);
    rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);

#ifdef RESCALE_CGBONDS
    dr = RCGBOND/rcg;
#else
    dr = 1.0;
#endif

    dx = dcgx * dr + px1 - px2; 
    dy = dcgy * dr + py1 - py2; 
    dz = dcgz * dr + pz1 - pz2; 
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-5.0;
//    dr = r-eqbond[idx];
    
//    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
//    case 0:
      fr = 500.*dr/r;
//      fr = fkbond[idx]*dr/r;
      pot = 0.5*500.*dr*dr;
//      pot = 0.5*fkbond[idx]*dr*dr;
//      break;
//    case 1:
//      mb = exp(-bmorse[idx]*dr);
//      pot = fkbond[idx]*(1-mb)*(1-mb);
//      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
//      break;
//    }

    ii = (faacg[i]<faacg[j])? i:j;
    fhr = 1. - faacg[ii]; 
    vds[ii] -= pot*ds[ii];
    spotlt += pot*fhr;
    
    fr *= fhr;
    dx *= fr; dy *= fr; dz *= fr;
    fx[i] -= dx;  fy[i] -= dy;  fz[i] -= dz;
    fx[j] += dx;  fy[j] += dy;  fz[j] += dz;

    /* now compute coupling component on the CG blobs */
    tx1 = pz1*dy - py1*dz;   ty1 = px1*dz - pz1*dx;   tz1 = py1*dx - px1*dy;
    tx2 = py2*dz - pz2*dy;   ty2 = pz2*dx - px2*dz;   tz2 = px2*dy - py2*dx;

    tor[icg][0]+=tx1;    tor[icg][1]+=ty1;    tor[icg][2]+=tz1;
    tor[jcg][0]+=tx2;    tor[jcg][1]+=ty2;    tor[jcg][2]+=tz2;

    /* == COMcounterforce plan IN HEALING AND ATOMISTIC REGION to maintain rotational component only == */
    /* an if-statement for every atom in every bond?? should be improved */
    if( simparms->iaacg[icg] < 2){
      fcgx[icg] += dx;  fcgy[icg] += dy;  fcgz[icg] += dz;
    }
    if( simparms->iaacg[jcg] < 2){
      fcgx[jcg] -= dx;  fcgy[jcg] -= dy;  fcgz[jcg] -= dz;
    }
    /* ============================================ */

    if(lcoupling){
      tx1+=tx2;  ty1+=ty2;  tz1+=tz2;
      r = 1./rcg/rcg;
      dx = (dcgy*tz1 - dcgz*ty1) * r;
      dy = (dcgz*tx1 - dcgx*tz1) * r;
      dz = (dcgx*ty1 - dcgy*tx1) * r;
      
      fcgx[icg] += dx;  fcgy[icg] += dy;  fcgz[icg] += dz;
      fcgx[jcg] -= dx;  fcgy[jcg] -= dy;  fcgz[jcg] -= dz;
    }
  }
  spotlso3 += spotlt;
}
/*-----------------------------------------------------------------------*/
/* force from so3 non-bonds in healing region for CG blob icg                */
void fnonbond_so3hr_CGR2HR(SIMPARMS *simparms,COORDS *coords,int icgnow)
{
#define MAXBONDS 200
#define RCUTOFF 10.
  int i,j,ii,jj,nb,nbonds,ibond[MAXBONDS],jbond[MAXBONDS];
  int icg, jcg, *icgunit, inow, jnow, jcgnow, *iaacg;
  int iperd, ivol;
  long ibegin,iend;
  double dx,dy,dz,r,rcg,fr,dr,amass;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz, dcgx, dcgy, dcgz;
  double px1, py1, pz1, px2, py2, pz2;
  double *fx,*fy,*fz,*fcgx,*fcgy,*fcgz;
//  double *eqbond,*fkbond;
  double *hmat, hmati[9];

  iperd = simparms->iperd;
  ivol  = simparms->ivol;
  hmat = coords->hmat;
  gethinv9(hmat,hmati);
  
  /* why do I use here fxa whereas in  fnonbond_so3 I use fxt ??  Have to figure this out */

  px = coords->px;   py = coords->py;   pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  fcgx = coords->next->fxa;   fcgy = coords->next->fya;   fcgz = coords->next->fza;

  /* healing region selection */
  /* first create list of bonds */
  nbonds = 0;
  for(icg=0; icg<coords->next->natoms-1; icg++){
    for(jcg=icg+1; jcg<coords->next->natoms; jcg++){
      dcgx = pcgx[icg] - pcgx[jcg];  
      dcgy = pcgy[icg] - pcgy[jcg];
      dcgz = pcgz[icg] - pcgz[jcg];
      rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);
      period(1,&dcgx,&dcgy,&dcgz,hmat,hmati,iperd,ivol);
      rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);
      if(rcg < RCUTOFF){
        for(i=0; i<simparms->napcg[icg]; i++){
          ii = simparms->icglist[icg][i];
          if(simparms->itype[ii+simparms->ncg] > 4){  /* HAL2 or HAL3 */
            for(j=0; j<simparms->napcg[jcg]; j++){
              jj = simparms->icglist[jcg][j];
              if(simparms->itype[jj+simparms->ncg] > 4){
                ibond[nbonds] = ii;
                jbond[nbonds] = jj;
                nbonds++;
              }
            }
          }
        }
      }
    } 
  }

//   nbonds = coords->bonds.nsel_so3hr;
//  iselect = coords->bonds.isel_so3hr;	
//  ibond = coords->bonds.ibond2;
//  jbond = coords->bonds.jbond2;
//  idxbond = coords->bonds.idxbond2;
//  itypbond = coords->bonds.itypbond;
//  eqbond = coords->bonds.eqbond;
//  fkbond = coords->bonds.fkbond;
//  bmorse = coords->bonds.bmorse;
  icgunit = simparms->icgunit;
  iaacg = simparms->iaacg;
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
  
  fr = 0.;
  for(nb=ibegin;nb<iend;nb++){
//    ib = iselect[nb];
    i = ibond[nb]; j = jbond[nb];
    icg = icgunit[i]; jcg = icgunit[j];
    if(icg==icgnow){
      inow = i;
      jnow = j;
      jcgnow = jcg;
    }else{
      if(jcg!=icgnow) continue;
      inow = j;
      jnow = i;
      jcgnow = icg;
    }

    px1 = px[inow]-pcgx[icgnow];  py1 = py[inow]-pcgy[icgnow];  pz1 = pz[inow]-pcgz[icgnow];
    if(iaacg[jcgnow]==2){
      px2 = px[jnow];  py2 = py[jnow];  pz2 = pz[jnow];
    }else{
      px2 = px[jnow]-pcgx[jcgnow];  py2 = py[jnow]-pcgy[jcgnow];  pz2 = pz[jnow]-pcgz[jcgnow];
    }
//    idx = idxbond[ib];
    
    /* renormalize for distance between CG beads */
    dcgx = pcgx[icgnow] - pcgx[jcgnow];  
    dcgy = pcgy[icgnow] - pcgy[jcgnow];
    dcgz = pcgz[icgnow] - pcgz[jcgnow];		
    period(1,&dcgx,&dcgy,&dcgz,hmat,hmati,iperd,ivol);
    rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);

    dr = RCGBOND/rcg;

    dx = dcgx * dr + px1 - px2; 
    dy = dcgy * dr + py1 - py2; 
    dz = dcgz * dr + pz1 - pz2; 
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-5.0;
//    dr = r-eqbond[idx];
    
//    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
//    case 0:
      fr = 500.*dr/r;
//      fr = fkbond[idx]*dr/r;
//      break;
//    case 1:
//      mb = exp(-bmorse[idx]*dr);
//      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
//      break;
//    }

    dx *= fr; dy *= fr; dz *= fr;
    fx[inow] -= dx;  fy[inow] -= dy;  fz[inow] -= dz;

    /* == COMcounterforce == */
    amass = 1./coords->next->amass[icgnow];
    fcgx[icgnow] += dx*amass;  fcgy[icgnow] += dy*amass;  fcgz[icgnow] += dz*amass;
  } 
}

/*-----------------------------------------------------------------------*/
/* force from so3 non-bonds in healing region for CG blob icg                */
void fnonbond_so3hr_HR2CGR(SIMPARMS *simparms,COORDS *coords,SO3 *so3,int icgnow)
{
  int i,j,ii,jj,nb,nbonds,ibond[MAXBONDS],jbond[MAXBONDS];
  int icg, jcg, *icgunit, inow, jnow, jcgnow, *iaacg;
  int iperd, ivol;
  long ibegin,iend;
  double dx,dy,dz,r,rcg,fr,dr,amass;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz, dcgx, dcgy, dcgz;
  double px1, py1, pz1, px2, py2, pz2;
  double *fx,*fy,*fz,*fcgx,*fcgy,*fcgz;
  double *hmat, hmati[9];

  iperd = simparms->iperd;
  ivol  = simparms->ivol;
  hmat = coords->hmat;
  gethinv9(hmat,hmati);
  
  px = coords->px;   py = coords->py;   pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  fcgx = coords->next->fxa;   fcgy = coords->next->fya;   fcgz = coords->next->fza;

//  ibond = coords->bonds.ibond2;
//  jbond = coords->bonds.jbond2;
//  idxbond = coords->bonds.idxbond2;
//  itypbond = coords->bonds.itypbond;
//  eqbond = coords->bonds.eqbond;
//  fkbond = coords->bonds.fkbond;
//  bmorse = coords->bonds.bmorse;
  icgunit = simparms->icgunit;
  iaacg = simparms->iaacg;
    
  /* first create list of bonds */
  nbonds = 0;
  for(icg=0; icg<coords->next->natoms-1; icg++){
    for(jcg=icg+1; jcg<coords->next->natoms; jcg++){
      dcgx = pcgx[icg] - pcgx[jcg];  
      dcgy = pcgy[icg] - pcgy[jcg];
      dcgz = pcgz[icg] - pcgz[jcg];
      rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);
      period(1,&dcgx,&dcgy,&dcgz,hmat,hmati,iperd,ivol);
      rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);
      if(rcg < RCUTOFF){
        for(i=0; i<simparms->napcg[icg]; i++){
          ii = simparms->icglist[icg][i];
          if(simparms->itype[ii+simparms->ncg] > 4){  /* HAL2 or HAL3 */
            for(j=0; j<simparms->napcg[jcg]; j++){
              jj = simparms->icglist[jcg][j];
              if(simparms->itype[jj+simparms->ncg] > 4){
                ibond[nbonds] = ii;
                jbond[nbonds] = jj;
                nbonds++;
              }
            }
          }
        }
      }
    } 
  }
  
    
    
  fr = 0.;
  /* healing region selection plus CG region selection (unfortunately both are needed) */
//  nbonds = coords->bonds.nsel_so3cgr;
//  iselect = coords->bonds.isel_so3cgr;
//  for(iregion=0;iregion<2;++iregion){
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
  for(nb=ibegin;nb<iend;nb++){
//    ib = iselect[nb];
    i = ibond[nb]; j = jbond[nb];
    icg = icgunit[i]; jcg = icgunit[j];
    if(icg==icgnow){
      inow = i;
      jnow = j;
      jcgnow = jcg;
    }else{
      if(jcg!=icgnow) continue;
      inow = j;
      jnow = i;
      jcgnow = icg;
    }

    px1 = px[inow];  py1 = py[inow];  pz1 = pz[inow];
    if(iaacg[jcgnow]==2){
      px2 = px[jnow];  py2 = py[jnow];  pz2 = pz[jnow];
    }else{
      px2 = px[jnow]-pcgx[jcgnow];  py2 = py[jnow]-pcgy[jcgnow];  pz2 = pz[jnow]-pcgz[jcgnow];
    }
//    idx = idxbond[ib];
    
    /* renormalize for distance between CG beads */
    dcgx = pcgx[icgnow] - pcgx[jcgnow];  
    dcgy = pcgy[icgnow] - pcgy[jcgnow];
    dcgz = pcgz[icgnow] - pcgz[jcgnow];		
    period(1,&dcgx,&dcgy,&dcgz,hmat,hmati,iperd,ivol);
    rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);

    dr = RCGBOND/rcg;

    dx = dcgx * dr + px1 - px2; 
    dy = dcgy * dr + py1 - py2; 
    dz = dcgz * dr + pz1 - pz2; 
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-5.0;
//    dr = r-eqbond[idx];
    
//    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
//    case 0:
      fr = 500.*dr/r;
//      fr = fkbond[idx]*dr/r;
//      break;
//    case 1:
//      mb = exp(-bmorse[idx]*dr);
//      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
//      break;
//    }

    dx *= fr; dy *= fr; dz *= fr;

    /* == subtract previous computed COMcounterforce  == */
    amass = 1./coords->next->amass[icgnow];
    fcgx[icgnow] -= dx*amass;  fcgy[icgnow] -= dy*amass;  fcgz[icgnow] -= dz*amass;
  }
//  /* now look for SO3 bonds in healing region */
//  nbonds = coords->bonds.nsel_so3hr;
//  iselect = coords->bonds.isel_so3hr;
//  } 
}

/*---------------------------------------------------------------*/
double getvnonbond_so3(SIMPARMS *simparms,COORDS *coords,INTER *inter,SO3 *so3,
    int nbonds,int *indx,int *jndx,int iregion)
{
  int i,j,nb,*itypbond,*idxbond;
  int icg, jcg, *icgunit;
  int iperd, ivol;
  long ibegin,iend;
  double dx,dy,dz,rcg,r,dr,fhr,tpot;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz,*faacg;
  double px1, py1, pz1, px2, py2, pz2;
  double *eqbond,*fkbond,*bmorse;
  double *hmat, hmati[9];

  iperd = simparms->iperd;
  ivol  = simparms->ivol;
  hmat = coords->hmat;
  gethinv9(hmat,hmati);
    
  px = coords->px;   py = coords->py;   pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;

  idxbond = coords->bonds.idxbond2;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  faacg = simparms->faacg + (1-coords->lcg)*simparms->ncg;
  icgunit = simparms->icgunit; 
  tpot = 0.;

  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);  

  for(nb=ibegin;nb<iend;nb++){
    i = indx[nb]; j = jndx[nb]; 
//    idx = idxbond[nb];

    px1 = px[i]; py1 = py[i]; pz1 = pz[i]; 
    px2 = px[j]; py2 = py[j]; pz2 = pz[j];
    icg = icgunit[i]; jcg = icgunit[j];
    
    /* renormalize for distance between CG beads */
    dx = pcgx[icg] - pcgx[jcg];  
    dy = pcgy[icg] - pcgy[jcg];
    dz = pcgz[icg] - pcgz[jcg];		
    period(1,&dx,&dy,&dz,hmat,hmati,iperd,ivol);
    rcg = sqrt(dx*dx+dy*dy+dz*dz);
    
#ifdef RESCALE_CGBONDS
    dr = RCGBOND/rcg;
#else
    dr = 1.0;
#endif

    dx = dx * dr + px1 - px2; 
    dy = dy * dr + py1 - py2; 
    dz = dz * dr + pz1 - pz2; 
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r - 5.0;
//    fk = fkbond[idx];
    
    fhr = 1. - MIN(faacg[i],faacg[j]);

    tpot += 0.5*500.*dr*dr*fhr;
    
  }
  return tpot;  
}
