/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/*---------------------------------------------------------------*/
/* subroutines dealing with bended interactions
   search_bond_base - searches the data base and sets up initial vectors
   fbend - gets the force of the bend
   getvbend - gets the potential energy of the bend
   bended   - check to see if to atoms are involved in a bend
*/

#include "md.h"
#define RCGBOND 3.71   /* 3.71 have to make this an input variable */

#define PRESSURE /* calculate pressure but it is = 0? */

static int iben,jben,kben;
static double eqben,fkben;

static double fact_quart(double x)
{
  int i;
  double result=1.0;
  for(i=2;i<=x;i++) result*=(double)i;
  return(result);	
}

static double poly_quart( double s_theta, int order )
{
  int i;
  double result=1.0;
  double fqi;
  
  for(i=1;i<=order;i++){
    fqi = fact_quart(i);
    result+=(fact_quart(2*i)/(pow(2, 2*i)*fqi*fqi*(2*i+1)))*pow(s_theta, 2*i);
  }
  return(result);
}

/*-----------------------------------------------------------------------*/
int bend_pair(int iatom,int jatom,BENDS *bends)
{
  int i,j,itemp,nb;

  if(iatom>jatom){ itemp = iatom;iatom = jatom; jatom = itemp; }

  for(nb=0;nb<bends->nbends;nb++){
    i = bends->ibend[nb]; j = bends->jbend[nb];
    if(i>j){itemp = i;i = j; j = itemp; }
    if(iatom == i && jatom == j) return 1;

    i = bends->ibend[nb]; j = bends->kbend[nb];
    if(i>j){itemp = i;i = j; j = itemp; }
    if(iatom == i && jatom == j) return 1;

    i = bends->jbend[nb]; j = bends->kbend[nb];
    if(i>j){itemp = i;i = j;j = itemp;}
    if(iatom == i && jatom == j) return 1;
  }
  return 0;
}
/*-----------------------------------------------------------------------*/
void exclude_bends(int *nexclude,int **exclude,BENDS *bends)
{
  int i,j,nb;

  for(nb=0;nb<bends->nbends;nb++){
    if(bends->ibend[nb]<bends->jbend[nb]){
      i=bends->ibend[nb];j=bends->jbend[nb];
    }else {
      j=bends->ibend[nb];i=bends->jbend[nb];
    }
    insert(exclude[i],&nexclude[i],j);

/* JCS if the bendtype is not 0 (harmonic) don't exclude the 1-3 interaction */
/*
    printf("bend type %i\n",bends->ibendtype[nb]);
    fflush(stdout);
*/
/*
      if(bends->ibend[nb]<bends->kbend[nb]){
        i=bends->ibend[nb];j=bends->kbend[nb];
      }else {
        j=bends->ibend[nb];i=bends->kbend[nb];
      }
      insert(exclude[i],&nexclude[i],j);
    printf("i j %i %i \n",i,j);
*/
    if(bends->ibendtype[nb] == 0) {
      if(bends->ibend[nb]<bends->kbend[nb]){
        i=bends->ibend[nb];j=bends->kbend[nb];
      }else {
        j=bends->ibend[nb];i=bends->kbend[nb];
      }
      insert(exclude[i],&nexclude[i],j);
    }

    if(bends->jbend[nb]<bends->kbend[nb]){
      i=bends->jbend[nb];j=bends->kbend[nb];
    } else {
      j=bends->jbend[nb];i=bends->kbend[nb];
    }
    insert(exclude[i],&nexclude[i],j);
  }
}

/*-----------------------------------------------------------------------*/
/* add bend to selection if the three atoms belong an active AA/CG region */

int select_bends(int *iaacg,BENDS *bends,int iflg)
{
  int k, nsel, nsel2, ib, *isel;
  int *ibend, *jbend, *kbend;
  
  nsel = 0;
	switch(iflg){
		case 0:
      ibend = bends->ibend;
      jbend = bends->jbend;
      kbend = bends->kbend;
      isel = bends->iselect;
			for(ib=0;ib<bends->nbends;ib++){ /* select: non of the atoms is inactive */
				if((iaacg[ibend[ib]]!=0)&&(iaacg[jbend[ib]]!=0)&&(iaacg[kbend[ib]]!=0) ){
          isel[nsel++]=ib;
				} 
			}
			bends->nselect = nsel;
			break;
		case 1:
      ibend = bends->ibend;
      jbend = bends->jbend;
      kbend = bends->kbend;
      isel = bends->iselect;
			for(ib=0;ib<bends->nbends;ib++){ /* select: at least one atom is not inactive */
				if((iaacg[ibend[ib]]+iaacg[jbend[ib]]+iaacg[kbend[ib]] != 0) ){
					isel[nsel++]=ib;
				} 
			}
			bends->nselect = nsel;
			break;
		case 2:
      ibend = bends->ibend2;
      jbend = bends->jbend2;
      kbend = bends->kbend2;
      isel = bends->isel_so3cgr;
      nsel2 = bends->nbends2;
			for(ib=0;ib<bends->nbends2;ib++){  /* CG=0 HR=1 AA=2 */
        k = iaacg[ibend[ib]] + iaacg[jbend[ib]] + iaacg[kbend[ib]];
        if( k == 0){
          isel[nsel++]=ib;
        }else if( k != 6 ){
          isel[--nsel2]=ib;
        } 
			}
      bends->nsel_so3cgr = nsel;
      bends->nsel_so3hr = bends->nbends2 - nsel2;
      bends->isel_so3hr = isel + nsel2;
      nsel += bends->nsel_so3hr;
      break;
		}

	return nsel;
}

/*------------------------------------------------------------------------*/

void fbend(SIMPARMS *simparms,COORDS *coords)
{
  int i,j,k,itype,nb,nbends,*ibend,*jbend,*kbend,*tybend,*idxbend,ifunc_type;
  long ibegin,iend,nbt,inow;
  double adotb,r11,costh,th,dudth,dthdcos,fact,dth;
  double dcdx1,dcdy1,dcdz1,dcdx2,dcdy2,dcdz2,dcdx3,dcdy3,dcdz3;
  double ax,ay,az,bx,by,bz,al,bl;
  double *px,*py,*pz,*fx,*fy,*fz,*eqbend,*fkbend;
  double *fx1,*fy1,*fz1,*fx2,*fy2,*fz2,*fx3,*fy3,*fz3;
  double sinth, a, b, c, t1;
	
  px = coords->px;  py = coords->py;  pz = coords->pz;
  fx = coords->fxa; fy = coords->fya; fz = coords->fza;
  nbends = coords->bends.nbends;  
	idxbend= coords->bends.idxbend;
  ibend = coords->bends.ibend;
  jbend = coords->bends.jbend;
  kbend = coords->bends.kbend;
  tybend = coords->bends.ibendtype;
  eqbend = coords->bends.eqbend;  fkbend = coords->bends.fkbend;
	
  decomp1d((long)nbends,simparms->size,simparms->rank,&ibegin,&iend);
	
  nbt = iend-ibegin;
  fx1 = (double *)cmalloc(9*nbt*sizeof(double));
  fy1 = fx1 + nbt;
  fz1 = fx1 + 2*nbt;
  fx2 = fx1 + 3*nbt;  fy2 = fx1 + 4*nbt;  fz2 = fx1 + 5*nbt;
  fx3 = fx1 + 6*nbt;  fy3 = fx1 + 7*nbt;  fz3 = fx1 + 8*nbt;
  
#include "vectorize.h"
	/*
	 printf("n bends ibegin %i iend %i\n",ibegin,iend);
	 */
  fact = 0.;
  dudth = 0.;
	
  for(inow=0,nb=ibegin;nb<iend;nb++,inow++){
    i = ibend[nb];
    j = jbend[nb];
    k = kbend[nb];
    itype = idxbend[nb];
    ifunc_type = tybend[nb];
		/*
		 printf("inow %i itype %i ifunc_type %i\n",inow,itype,ifunc_type);
		 */
		
    ax = px[i] - px[j];
    ay = py[i] - py[j];
    az = pz[i] - pz[j];
    al = ax*ax+ay*ay+az*az;
		
    bx = px[k] - px[j];
    by = py[k] - py[j];
    bz = pz[k] - pz[j];
    bl = bx*bx+by*by+bz*bz;
		
    adotb = ax*bx + ay*by + az*bz;
    r11 = 1./sqrt(al*bl);
    costh = adotb*r11;
    switch(ifunc_type) {
			case 0: /* original harmonic */
    /* 10 Nov 2009 FIX BERND */
        if(costh > 1.){
          costh =1.;
        }else if(costh < -1.){
          costh = -1.;
        }
//				if(fabs(costh) >= 1.){
//					th = 1.;
//					dth = th - eqbend[itype];
//					dthdcos = -1.;
//					dudth = 0.;
//				} else {
					th = acos(costh);
					dth = th - eqbend[itype];      
					dudth   = fkbend[itype]*dth;
          if(dth == 0.){
            dthdcos = 0.0;
          }else{
            dthdcos = -1./sqrt(1.-costh*costh);
          }
//				}
				fact = dudth*dthdcos;
				break;
			case 1: /* cosine */
				if(costh > 1.) costh =1.;
				else if(costh < -1.) costh = -1.;
					/* th = acos(costh); */
					fact =  fkbend[itype];
				break;
			case 2: /* quartic */
				if(costh > 1.) costh =1.;
				else if(costh < -1.) costh = -1.;
					th = acos(costh);
				
				sinth = sin(th-M_PI);
				
				if(sinth > 1.) sinth =1.;
				else if(sinth < -1.) sinth = -1.;
					if(fabs(th-M_PI) < 0.01) t1 =  -1. * poly_quart(sinth, 20);
				else t1 = -(th-M_PI)/sinth;
				
				c = (eqbend[itype]-M_PI)*(eqbend[itype]-M_PI);
				b = c - (th-M_PI)*(th-M_PI);
				a = 0.5 * fkbend[itype] / c;
				fact = a * b * t1;      
				break;
    }
    dcdx1 = costh*ax/al - bx*r11;
    dcdy1 = costh*ay/al - by*r11;
    dcdz1 = costh*az/al - bz*r11;
    dcdx3 = costh*bx/bl - ax*r11;
    dcdy3 = costh*by/bl - ay*r11;
    dcdz3 = costh*bz/bl - az*r11;
    dcdx2 = -dcdx1-dcdx3;
    dcdy2 = -dcdy1-dcdy3;
    dcdz2 = -dcdz1-dcdz3;
		
    fx1[inow] = fact*dcdx1;
    fy1[inow] = fact*dcdy1;
    fz1[inow] = fact*dcdz1;
    fx2[inow] = fact*dcdx2;
    fy2[inow] = fact*dcdy2;
    fz2[inow] = fact*dcdz2;
    fx3[inow] = fact*dcdx3;
    fy3[inow] = fact*dcdy3;
    fz3[inow] = fact*dcdz3;
  }
	
	
  for(inow=0,nb=ibegin;nb<iend;nb++,inow++){
    i = ibend[nb];    j = jbend[nb];    k = kbend[nb];
    fx[i] += fx1[inow];    fy[i] += fy1[inow];    fz[i] += fz1[inow];
    fx[j] += fx2[inow];    fy[j] += fy2[inow];    fz[j] += fz2[inow];
    fx[k] += fx3[inow];    fy[k] += fy3[inow];    fz[k] += fz3[inow];
  }
  free(fx1);
}

/*------------------------------------------------------------------------*/

void fbend_so3(SIMPARMS *simparms,COORDS *coords,SO3 *so3,int iregion)
{
  int i,ib,j,k,itype,nb,nbends,*ibend,*jbend,*kbend,*tybend,*idxbend,ifunc_type;
	int *iselect,icg,jcg,kcg,*icgunit,imaxcg,lcoupling;
  long ibegin,iend,inow;
  double adotb,r11,costh,th,dudth,dthdcos,fact,dth;
  double dcdx1,dcdy1,dcdz1,dcdx2,dcdy2,dcdz2,dcdx3,dcdy3,dcdz3;
  double ax,ay,az,bx,by,bz,al,bl;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz,*fx,*fy,*fz,*eqbend,*fkbend;
  double fx1,fy1,fz1,fx2,fy2,fz2,fx3,fy3,fz3,*fcgx,*fcgy,*fcgz;
  double px1,py1,pz1,px2,py2,pz2,px3,py3,pz3,tx1,ty1,tz1,tx2,ty2,tz2,tx3,ty3,tz3;
  double dcgx, dcgy, dcgz, rcg;
  double sinth, a, b, c, t1, dr;
  double pot, *faacg, *ds, *vds, (*tor)[3];
	
  lcoupling = (simparms->iso3dyn == 1);

  px = coords->px;  py = coords->py;  pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  fx = coords->fxt; fy = coords->fyt; fz = coords->fzt;
  fcgx = coords->next->fxt;   fcgy = coords->next->fyt;   fcgz = coords->next->fzt;
  
  if( iregion == 0){
    /* coarse grain region selection */
    nbends = coords->bends.nsel_so3cgr;
    iselect = coords->bends.isel_so3cgr;
  }else{
    /* healing region selection */
    nbends = coords->bends.nsel_so3hr;
    iselect = coords->bends.isel_so3hr;		
  }
	idxbend= coords->bends.idxbend2;
  ibend = coords->bends.ibend2;
  jbend = coords->bends.jbend2;
  kbend = coords->bends.kbend2;
  tybend = coords->bends.ibendtype;
  eqbend = coords->bends.eqbend;  
	fkbend = coords->bends.fkbend;
  icgunit = simparms->icgunit;
  faacg = simparms->faacg + simparms->ncg;
  ds = simparms->dfaacg + simparms->ncg;
  vds = simparms->vds_intra + simparms->ncg; 
  tor = so3->tor;
	
  decomp1d((long)nbends,simparms->size,simparms->rank,&ibegin,&iend);
	  
#include "vectorize.h"
	/*
	 printf("n bends ibegin %i iend %i\n",ibegin,iend);
	 */
  fact = 0.;
  dudth = 0.;
	
  for(inow=0,ib=ibegin;ib<iend;ib++,inow++){
    nb = iselect[ib];
    i = ibend[nb];    j = jbend[nb];    k = kbend[nb];
    itype = idxbend[nb];
    ifunc_type = tybend[nb];
    px1 = px[i]; py1 = py[i]; pz1 = pz[i]; 
    px2 = px[j]; py2 = py[j]; pz2 = pz[j];
    px3 = px[k]; py3 = py[k]; pz3 = pz[k];
		/*
		 printf("inow %i itype %i ifunc_type %i\n",inow,itype,ifunc_type);
		 */
		icg = icgunit[i]; 
		jcg = icgunit[j];   /* NB we assume that j shares a CG unit wit either i or k    */
    kcg = icgunit[k];   /* I should check somewhere in the beginning that icg != kcg */
		if(icg == jcg){
			dcgx = pcgx[kcg] - pcgx[icg];
			dcgy = pcgy[kcg] - pcgy[icg];
			dcgz = pcgz[kcg] - pcgz[icg];
			rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);
			dr = RCGBOND/rcg;
			bx = dcgx * dr; 
			by = dcgy * dr; 
			bz = dcgz * dr;
			bx += px3 - px2;
			by += py3 - py2;
			bz += pz3 - pz2;
			bl = bx*bx+by*by+bz*bz;
			ax = px1 - px2;
			ay = py1 - py2;
			az = pz1 - pz2;
			al = ax*ax+ay*ay+az*az;
		}else{
			dcgx = pcgx[icg] - pcgx[kcg];
			dcgy = pcgy[icg] - pcgy[kcg];
			dcgz = pcgz[icg] - pcgz[kcg];
			rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);
			dr = RCGBOND/rcg;
			ax = dcgx * dr; 
			ay = dcgy * dr; 
			az = dcgz * dr;
			ax += px1 - px2;
			ay += py1 - py2;
			az += pz1 - pz2;
			al = ax*ax+ay*ay+az*az;
			bx = px3 - px2;
			by = py3 - py2;
			bz = pz3 - pz2;
			bl = bx*bx+by*by+bz*bz;
      dcgx *= -1.;
      dcgy *= -1.;
      dcgz *= -1.;
		}
    imaxcg = (faacg[i]<faacg[k])? i:k;
		
    adotb = ax*bx + ay*by + az*bz;
    r11 = 1./sqrt(al*bl);
    costh = adotb*r11;
    switch(ifunc_type) {
			case 0: /* original harmonic */
      /* 10 Nov 2009 FIX BERND */
        if(costh > 1.){
          costh =1.;
        }else if(costh < -1.){
          costh = -1.;
        }
//				if(fabs(costh) >= 1.){
//					th = 1.;
//					dth = th - eqbend[itype];
//					dthdcos = -1.;
//					dudth = 0.;
//          pot = .5*fkbend[itype]*dth*dth;
//				} else {
					th = acos(costh);
					dth = th - eqbend[itype];      
					dudth   = fkbend[itype]*dth;
          if(dth == 0.){
            dthdcos = 0.0;
          }else{
            dthdcos = -1./sqrt(1.-costh*costh);
          }
          pot = .5*dudth*dth;
//				}
				fact = dudth*dthdcos;
				break;
			case 1: /* cosine */
				if(costh > 1.) costh =1.;
				else if(costh < -1.) costh = -1.;
					/* th = acos(costh); */
					fact =  fkbend[itype];
          pot = fact * (1.+costh);
				break;
			case 2: /* quartic */
				if(costh > 1.) costh =1.;
				else if(costh < -1.) costh = -1.;
					th = acos(costh);
				
				sinth = sin(th-M_PI);
				
				if(sinth > 1.) sinth =1.;
				else if(sinth < -1.) sinth = -1.;
					if(fabs(th-M_PI) < 0.01) t1 =  -1. * poly_quart(sinth, 20);
				else t1 = -(th-M_PI)/sinth;
				
				c = (eqbend[itype]-M_PI)*(eqbend[itype]-M_PI);
				b = c - (th-M_PI)*(th-M_PI);
				a = 0.5 * fkbend[itype] / c;
				fact = a * b * t1;
        pot = 0.25*a * b*b;
				break;
    }
    
    dcdx1 = costh*ax/al - bx*r11;
    dcdy1 = costh*ay/al - by*r11;
    dcdz1 = costh*az/al - bz*r11;
    dcdx3 = costh*bx/bl - ax*r11;
    dcdy3 = costh*by/bl - ay*r11;
    dcdz3 = costh*bz/bl - az*r11;
    dcdx2 = -dcdx1-dcdx3;
    dcdy2 = -dcdy1-dcdy3;
    dcdz2 = -dcdz1-dcdz3;
		
    fact *= (1.0 - faacg[imaxcg]);
    vds[imaxcg] -= pot*ds[imaxcg];
    
    fx1 = fact*dcdx1;
    fy1 = fact*dcdy1;
    fz1 = fact*dcdz1;
    fx2 = fact*dcdx2;
    fy2 = fact*dcdy2;
    fz2 = fact*dcdz2;
    fx3 = fact*dcdx3;
    fy3 = fact*dcdy3;
    fz3 = fact*dcdz3;

    fx[i] += fx1;    fy[i] += fy1;    fz[i] += fz1;
    fx[j] += fx2;    fy[j] += fy2;    fz[j] += fz2;
    fx[k] += fx3;    fy[k] += fy3;    fz[k] += fz3;
    
    /* now compute coupling component on the CG blobs */
    tx1 = py1*fz1 - pz1*fy1;   ty1 = pz1*fx1 - px1*fz1;   tz1 = px1*fy1 - py1*fx1;
    tx2 = py2*fz2 - pz2*fy2;   ty2 = pz2*fx2 - px2*fz2;   tz2 = px2*fy2 - py2*fx2;
    tx3 = py3*fz3 - pz3*fy3;   ty3 = pz3*fx3 - px3*fz3;   tz3 = px3*fy3 - py3*fx3;

    tor[icg][0]+=tx1;    tor[icg][1]+=ty1;    tor[icg][2]+=tz1;
    tor[jcg][0]+=tx2;    tor[jcg][1]+=ty2;    tor[jcg][2]+=tz2;
    tor[kcg][0]+=tx3;    tor[kcg][1]+=ty3;    tor[kcg][2]+=tz3;

    /* == COMcounterforce plan IN HEALING AND ATOMISTIC REGION == */
    /* an if-statement for every atom in every bend?? should be improved */
    if( simparms->iaacg[icg] < 2){
      fcgx[icg] -= fx1;  fcgy[icg] -= fy1;  fcgz[icg] -= fz1;
    }
    if( simparms->iaacg[jcg] < 2){
      fcgx[jcg] -= fx2;  fcgy[jcg] -= fy2;  fcgz[jcg] -= fz2;
    }
    if( simparms->iaacg[kcg] < 2){
      fcgx[kcg] -= fx3;  fcgy[kcg] -= fy3;  fcgz[kcg] -= fz3;
    }
    /* ============================================ */


    if(lcoupling){
      tx1+=tx2;  ty1+=ty2;  tz1+=tz2;
      tx1+=tx3;  ty1+=ty3;  tz1+=tz3;
      dr = 1./rcg/rcg;
      fx1 = (dcgy*tz1 - dcgz*ty1) * dr;
      fy1 = (dcgz*tx1 - dcgx*tz1) * dr;
      fz1 = (dcgx*ty1 - dcgy*tx1) * dr;
      
      fcgx[icg] -= fx1;  fcgy[icg] -= fy1;  fcgz[icg] -= fz1;
      fcgx[kcg] += fx1;  fcgy[kcg] += fy1;  fcgz[kcg] += fz1;
    }
        
  }
  
}

/*-----------------------------------------------------------------------*/

void fbend_hr(SIMPARMS *simparms,COORDS *coords)
{
  int i,j,k,ii,itype,ib,nb,nbends,*ibend,*jbend,*kbend,*tybend,*idxbend;
  int ifunc_type,*iselect,*icg,lhybrid;
  long ibegin,iend,nbt,inow;
  double adotb,r11,costh,th,dudth,dthdcos,fact,dth,fhr,*faacg,*ds,pot,*vcg,*vds;
  double dcdx1,dcdy1,dcdz1,dcdx2,dcdy2,dcdz2,dcdx3,dcdy3,dcdz3;
  double ax,ay,az,bx,by,bz,al,bl;
  double *px,*py,*pz,*fx,*fy,*fz,*eqbend,*fkbend;
  double *fx1,*fy1,*fz1,*fx2,*fy2,*fz2,*fx3,*fy3,*fz3;
  double sinth, a, b, c, t1;
 
  px = coords->px;  py = coords->py;  pz = coords->pz;
  fx = coords->fxa; fy = coords->fya; fz = coords->fza;
  nbends = coords->bends.nselect;
  iselect = coords->bends.iselect;
  idxbend= coords->bends.idxbend;
  ibend = coords->bends.ibend;
  jbend = coords->bends.jbend;
  kbend = coords->bends.kbend;
  tybend = coords->bends.ibendtype;
  eqbend = coords->bends.eqbend;  fkbend = coords->bends.fkbend;
  faacg = simparms->faacg + (1-coords->lcg)*simparms->ncg;
  ds = simparms->dfaacg + (1-coords->lcg)*simparms->ncg;
  icg = simparms->icgunit;
  vcg = simparms->vcgintern;
  vds = simparms->vds_intra + (1-coords->lcg)*simparms->ncg; 
  lhybrid=(simparms->icalc_type == 5);

  decomp1d((long)nbends,simparms->size,simparms->rank,&ibegin,&iend);

  nbt = iend-ibegin;
  fx1 = (double *)cmalloc(9*nbt*sizeof(double));
  fy1 = fx1 + nbt;
  fz1 = fx1 + 2*nbt;
  fx2 = fx1 + 3*nbt;  fy2 = fx1 + 4*nbt;  fz2 = fx1 + 5*nbt;
  fx3 = fx1 + 6*nbt;  fy3 = fx1 + 7*nbt;  fz3 = fx1 + 8*nbt;
  
#include "vectorize.h"
/*
  printf("n bends ibegin %i iend %i\n",ibegin,iend);
*/
  fact = 0.;
  dudth = 0.;

  for(inow=0,ib=ibegin;ib<iend;ib++,inow++){
    nb = iselect[ib];
    i = ibend[nb];
    j = jbend[nb];
    k = kbend[nb];
    itype = idxbend[nb];
    ifunc_type = tybend[nb];
/*
    printf("inow %i itype %i ifunc_type %i\n",inow,itype,ifunc_type);
*/

    ax = px[i] - px[j];
    ay = py[i] - py[j];
    az = pz[i] - pz[j];
    al = ax*ax+ay*ay+az*az;

    bx = px[k] - px[j];
    by = py[k] - py[j];
    bz = pz[k] - pz[j];
    bl = bx*bx+by*by+bz*bz;

    adotb = ax*bx + ay*by + az*bz;
    r11 = 1./sqrt(al*bl);
    costh = adotb*r11;
    switch(ifunc_type) {
    case 0: /* original harmonic */
    /* 10 Nov 2009 FIX BERND */
      if(costh > 1.){
        costh =1.;
      }else if(costh < -1.){
        costh = -1.;
      }
//      if(fabs(costh) >= 1.){
//	th = 1.;
//	dth = th - eqbend[itype];
//	dthdcos = -1.;
//	dudth = 0.;
//	pot = .5*fkbend[itype]*dth*dth;
 //     } else {
      th = acos(costh);
      dth = th - eqbend[itype];      
      dudth   = fkbend[itype]*dth;
      if(dth == 0.){
        dthdcos = 0.0;
      }else{
        dthdcos = -1./sqrt(1.-costh*costh);
      }
      pot = .5*dudth*dth;
//      }
      fact = dudth*dthdcos;
      break;
    case 1: /* cosine */
      if(costh > 1.) costh =1.;
      else if(costh < -1.) costh = -1.;
      /* th = acos(costh); */
      fact =  fkbend[itype];
      pot = fact * (1.+costh);
      break;
    case 2: /* quartic */
      if(costh > 1.) costh =1.;
      else if(costh < -1.) costh = -1.;
      th = acos(costh);
      
      sinth = sin(th-M_PI);
      
      if(sinth > 1.) sinth =1.;
      else if(sinth < -1.) sinth = -1.;
      if(fabs(th-M_PI) < 0.01) t1 =  -1. * poly_quart(sinth, 20);
      else t1 = -(th-M_PI)/sinth;
      
      c = (eqbend[itype]-M_PI)*(eqbend[itype]-M_PI);
      b = c - (th-M_PI)*(th-M_PI);
      a = 0.5 * fkbend[itype] / c;
      fact = a * b * t1;      
      pot = 0.25*a * b*b;
      break;
    }
	
	/* healing region switch function added here BERND */
	if(lhybrid){
	  if(coords->lcg){    /* CG-CG bonds are always scaled */
		ii = (faacg[i]>faacg[j])? i:j;
		ii = (faacg[ii]>faacg[k])? ii:k;
	    fhr = faacg[ii];
		vds[ii] += pot*ds[ii];
	  }else{              /* only AA-AA-AA bends over different CG atoms are scaled */
		ii = icg[i];
	    if((ii==icg[j]) && (ii==icg[k])){
          vcg[ii] += pot;		  
	      fhr = 1.;
	    }else{
		  ii = (faacg[i]<faacg[j])? i:j;
		  ii = (faacg[ii]<faacg[k])? ii:k;
		  fhr = faacg[ii]; 
		  vds[ii] += pot*ds[ii];
	    }
	  }
	}else{
	  fhr=1.;
	}	
	
    dcdx1 = costh*ax/al - bx*r11;
    dcdy1 = costh*ay/al - by*r11;
    dcdz1 = costh*az/al - bz*r11;
    dcdx3 = costh*bx/bl - ax*r11;
    dcdy3 = costh*by/bl - ay*r11;
    dcdz3 = costh*bz/bl - az*r11;
    dcdx2 = -dcdx1-dcdx3;
    dcdy2 = -dcdy1-dcdy3;
    dcdz2 = -dcdz1-dcdz3;

    fact *= fhr;
    fx1[inow] = fact*dcdx1;
    fy1[inow] = fact*dcdy1;
    fz1[inow] = fact*dcdz1;
    fx2[inow] = fact*dcdx2;
    fy2[inow] = fact*dcdy2;
    fz2[inow] = fact*dcdz2;
    fx3[inow] = fact*dcdx3;
    fy3[inow] = fact*dcdy3;
    fz3[inow] = fact*dcdz3;
  }

 
  for(inow=0,ib=ibegin;ib<iend;ib++,inow++){
    nb = iselect[ib];
    i = ibend[nb];    j = jbend[nb];    k = kbend[nb];
    fx[i] += fx1[inow];    fy[i] += fy1[inow];    fz[i] += fz1[inow];
    fx[j] += fx2[inow];    fy[j] += fy2[inow];    fz[j] += fz2[inow];
    fx[k] += fx3[inow];    fy[k] += fy3[inow];    fz[k] += fz3[inow];
  }
  free(fx1);
}

/*-----------------------------------------------------------------------*/
void getvbend(SIMPARMS *simparms,COORDS *coords,
	      double *UI,double *WI,double WItensor[9])
{
  int i,j,k,itype,ib,nb,*iselect,ifunc_type, *tybend, *icg,lhybrid;
  long ibegin,iend;
  double adotb,r11,costh,th,dth,fhr,*faacg;
  double ax,ay,az,bx,by,bz,al,bl;
  double spot,sprs,spotl,sprsl;
  double *px,*py,*pz;
  double *eqbend,*fkbend;
  int nbends,*ibend,*jbend,*kbend,*idxbend;
  double sprstensor[9],sprstl[9];
#ifdef PRESSURE
  double dudth,dthdcos,fact,dcdx1,dcdy1,dcdz1,dcdx3,dcdy3,dcdz3;
  double dcdx2,dcdy2,dcdz2;
#endif
  double t_coef, b, c;

  px = coords->px;
  py = coords->py;
  pz = coords->pz;
  nbends = coords->bends.nselect;
  iselect = coords->bends.iselect;
  ibend = coords->bends.ibend;
  jbend = coords->bends.jbend;
  kbend = coords->bends.kbend;
  idxbend= coords->bends.idxbend;
  eqbend = coords->bends.eqbend;
  fkbend = coords->bends.fkbend;
  tybend = coords->bends.ibendtype;
  faacg = simparms->faacg + (1-coords->lcg)*simparms->ncg;
  icg = simparms->icgunit;    
  lhybrid=(simparms->icalc_type == 5);

  for(i=0;i<9;i++) sprstensor[i]=0.0;

  spot = sprs = 0.;
  decomp1d((long)nbends,simparms->size,simparms->rank,&ibegin,&iend);

#include "vectorize.h"
  for(ib=ibegin;ib<iend;ib++){
    nb = iselect[ib];
    i = ibend[nb];
    j = jbend[nb];
    k = kbend[nb];
    itype = idxbend[nb];
    ifunc_type = tybend[nb];
    
	/* healing region switch function added here BERND */
	if(lhybrid){
	  if(coords->lcg){    /* CG-CG bonds are always scaled */
	    fhr = MAX(faacg[i],MAX(faacg[j],faacg[k]));
	  }else{              /* only AA-AA-AA bends over different CG atoms are scaled */
	    if((icg[i]==icg[j]) && (icg[i]==icg[k])){
	      fhr = 1.;
	    }else{
	      fhr = MIN(faacg[i],MIN(faacg[j],faacg[k]));
		}
	  }
	}else{
	  fhr = 1.;
	}

    ax = px[i] - px[j];
    ay = py[i] - py[j];
    az = pz[i] - pz[j];
    al = ax*ax+ay*ay+az*az;

    bx = px[k] - px[j];
    by = py[k] - py[j];
    bz = pz[k] - pz[j];
    bl = bx*bx+by*by+bz*bz;

    adotb = ax*bx + ay*by + az*bz;
    r11 = 1./sqrt(al*bl);
    costh = adotb*r11;
    
    costh = MIN(1.,costh);
    costh = MAX(-1.,costh);
    switch(ifunc_type) {
    case 0: /* original harmonic */
    /* 10 Nov 2009 FIX BERND */
      if(costh > 1.){
        costh =1.;
      }else if(costh < -1.){
        costh = -1.;
      }
//      if(fabs(costh) >= 1.){
//	th = 1.;
//	dth = th - eqbend[itype];
//	spot += .5*fkbend[itype]*dth*dth*fhr;
#ifdef PRESSURE
//	dthdcos = -1.;
//	dudth = 0.;
#endif
//      } else {
	th = acos(costh);
	dth = th - eqbend[itype];
	spot += .5*fkbend[itype]*dth*dth*fhr;
#ifdef PRESSURE
	dthdcos = -1.;
	dudth   = fkbend[itype]*dth;
#endif
      if(dth == 0.){
        dthdcos = 0.0;
      }else{
        dthdcos = -1./sqrt(1.-costh*costh);
      }
//      }
      break;
    case 1: /* cosine */
      if(costh > 1.) costh =1.;
      else if(costh < -1.) costh = -1.;
      /* th = acos(costh); */      
      spot += fkbend[itype] * (1.+costh) * fhr;
      dudth = - fkbend[itype] * sqrt(1. - costh*costh);
      dthdcos = -1./sqrt(1.-costh*costh);
      break;
    case 2: /* quartic */
      if(costh > 1.) costh =1.;
      else if(costh < -1.) costh = -1.;
      
      th = acos(costh);      
      c = (eqbend[itype]-M_PI)*(eqbend[itype]-M_PI);
      b = c - (th-M_PI)*(th-M_PI);
      /* a = 0.5 * fkbend[itype] / c; */
      t_coef =  0.25 * (fkbend[itype] * 0.5 / c);
      spot += t_coef * b*b * fhr;
      dudth =  4. * t_coef * b * (th-M_PI);
      dthdcos = -1./sqrt(1.-costh*costh);
      break;
    }
    fact = dudth*dthdcos*fhr;
#ifdef PRESSURE
    dcdx1 = costh*ax/al - bx*r11;
    dcdy1 = costh*ay/al - by*r11;
    dcdz1 = costh*az/al - bz*r11;
    dcdx2 = -costh*(ax/al+bx/bl)+(ax+bx)*r11;
    dcdy2 = -costh*(ay/al+by/bl)+(ay+by)*r11;
    dcdz2 = -costh*(az/al+bz/bl)+(az+bz)*r11;
    dcdx3 = costh*bx/bl - ax*r11;
    dcdy3 = costh*by/bl - ay*r11;
    dcdz3 = costh*bz/bl - az*r11;
    sprs += fact*(ax*dcdx1+ay*dcdy1+az*dcdz1-bx*dcdx3-by*dcdy3-bz*dcdz3);

    sprstensor[0] += fact*(px[i]*dcdx1+px[j]*dcdx2+px[k]*dcdx3);
    sprstensor[1] += fact*(py[i]*dcdx1+py[j]*dcdx2+py[k]*dcdx3);
    sprstensor[2] += fact*(pz[i]*dcdx1+pz[j]*dcdx2+pz[k]*dcdx3);

    sprstensor[3] += fact*(px[i]*dcdy1+px[j]*dcdy2+px[k]*dcdy3);
    sprstensor[4] += fact*(py[i]*dcdy1+py[j]*dcdy2+py[k]*dcdy3);
    sprstensor[5] += fact*(pz[i]*dcdy1+pz[j]*dcdy2+pz[k]*dcdy3);

    sprstensor[6] += fact*(px[i]*dcdz1+px[j]*dcdz2+px[k]*dcdz3);
    sprstensor[7] += fact*(py[i]*dcdz1+py[j]*dcdz2+py[k]*dcdz3);
    sprstensor[8] += fact*(pz[i]*dcdz1+pz[j]*dcdz2+pz[k]*dcdz3);

#endif
  }
#ifdef PARA
  MPI_Allreduce(&spot,&spotl,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  MPI_Allreduce(&sprs,&sprsl,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  MPI_Allreduce(sprstensor,sprstl,9,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
#else
  spotl = spot;
  sprsl = sprs;
  for(i=0;i<9;i++) sprstl[i] = sprstensor[i];
#endif

  *UI += spotl;  
  *WI += sprsl;
  for(i=0;i<9;i++) WItensor[i] += sprstl[i];
}
/*-----------------------------------------------------------------------*/
double getvbend_so3(SIMPARMS *simparms,COORDS *coords,int iregion)
{
  int i,j,k,itype,ib,nb,*iselect,ifunc_type, *tybend, *icgunit, icg, jcg;
  long ibegin,iend;
  double adotb,r11,costh,th,dth,fhr,*faacg;
  double ax,ay,az,bx,by,bz,al,bl,dr;
  double tpot;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz;
  double *eqbend,*fkbend;
  int nbends,*ibend,*jbend,*kbend,*idxbend;
  double t_coef, b, c;
	
  px = coords->px;  py = coords->py;  pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  if( iregion == 0){
    /* coarse grain region selection */
    nbends = coords->bends.nsel_so3cgr;
    iselect = coords->bends.isel_so3cgr;
  }else{
    /* healing region selection */
    nbends = coords->bends.nsel_so3hr;
    iselect = coords->bends.isel_so3hr;		
  }
  ibend = coords->bends.ibend2;
  jbend = coords->bends.jbend2;
  kbend = coords->bends.kbend2;
  idxbend= coords->bends.idxbend2;
  eqbend = coords->bends.eqbend;
  fkbend = coords->bends.fkbend;
  tybend = coords->bends.ibendtype;
  faacg = simparms->faacg + simparms->ncg;
  icgunit = simparms->icgunit;    
	tpot = 0.;

  decomp1d((long)nbends,simparms->size,simparms->rank,&ibegin,&iend);
	
#include "vectorize.h"
  for(ib=ibegin;ib<iend;ib++){
    nb = iselect[ib];
    i = ibend[nb];
    j = jbend[nb];
    k = kbend[nb];
    itype = idxbend[nb];
    ifunc_type = tybend[nb];
    
		fhr = 1.;

		icg = icgunit[i]; 
		jcg = icgunit[j];
		if(icg == jcg){
			jcg=icgunit[k];								/* check at initialization that icg!=jcg */
			bx = pcgx[jcg] - pcgx[icg];
			by = pcgy[jcg] - pcgy[icg];
			bz = pcgz[jcg] - pcgz[icg];
			bl = sqrt(bx*bx+by*by+bz*bz);
			dr = RCGBOND/bl;
			bx *= dr; 
			by *= dr; 
			bz *= dr;
			bx += px[k] - px[j];
			by += py[k] - py[j];
			bz += pz[k] - pz[j];
			bl = bx*bx+by*by+bz*bz;
			ax = px[i] - px[j];
			ay = py[i] - py[j];
			az = pz[i] - pz[j];
			al = ax*ax+ay*ay+az*az;			
		}else{
			ax = pcgx[icg] - pcgx[jcg];
			ay = pcgy[icg] - pcgy[jcg];
			az = pcgz[icg] - pcgz[jcg];
			al = sqrt(ax*ax+ay*ay+az*az);
			dr = RCGBOND/al;
			ax *= dr; 
			ay *= dr; 
			az *= dr;
			ax += px[i] - px[j];
			ay += py[i] - py[j];
			az += pz[i] - pz[j];
			al = ax*ax+ay*ay+az*az;
			bx = px[k] - px[j];
			by = py[k] - py[j];
			bz = pz[k] - pz[j];
			bl = bx*bx+by*by+bz*bz;			
		}
				
    fhr = 1. - MIN(faacg[i],faacg[k]);
    adotb = ax*bx + ay*by + az*bz;
    r11 = 1./sqrt(al*bl);
    costh = adotb*r11;
    
    costh = MIN(1.,costh);
    costh = MAX(-1.,costh);
    switch(ifunc_type) {
			case 0: /* original harmonic */
      /* 10 Nov 2009 FIX BERND */
        if(costh > 1.){
          costh =1.;
        }else if(costh < -1.){
          costh = -1.;
        }
//				if(fabs(costh) >= 1.){
//					th = 1.;
//					dth = th - eqbend[itype];
//					tpot += .5*fkbend[itype]*dth*dth*fhr;
//				} else {
					th = acos(costh);
					dth = th - eqbend[itype];
					tpot += .5*fkbend[itype]*dth*dth*fhr;
//				}
				break;
			case 1: /* cosine */
				if(costh > 1.) costh =1.;
				else if(costh < -1.) costh = -1.;
					/* th = acos(costh); */      
					tpot += fkbend[itype] * (1.+costh) * fhr;
				break;
			case 2: /* quartic */
				if(costh > 1.) costh =1.;
				else if(costh < -1.) costh = -1.;
					
					th = acos(costh);      
				c = (eqbend[itype]-M_PI)*(eqbend[itype]-M_PI);
				b = c - (th-M_PI)*(th-M_PI);
				/* a = 0.5 * fkbend[itype] / c; */
				t_coef =  0.25 * (fkbend[itype] * 0.5 / c);
				tpot += t_coef * b*b * fhr;
				break;
    }
  }
	return tpot;
}
/*-----------------------------------------------------------------------*/
double getvbend_intern(SIMPARMS *simparms,COORDS *coords)
{
  int i,j,k,itype,ib,ifunc_type, *tybend, *icg,iat;
  long ibegin,iend,nbt,inow;
  double adotb,r11,costh,th,dudth,dthdcos,fact,dth;
  double a,dcdx1,dcdy1,dcdz1,dcdx2,dcdy2,dcdz2,dcdx3,dcdy3,dcdz3;
  double ax,ay,az,bx,by,bz,al,bl;
  double *spot,sinth,t1,pot,tpot;
  double *fx,*fy,*fz,*px,*py,*pz;
  double *fx1,*fy1,*fz1,*fx2,*fy2,*fz2,*fx3,*fy3,*fz3;
  double *eqbend,*fkbend;
  int nbends,*ibend,*jbend,*kbend,*idxbend,*iaacg;

  double t_coef, b, c;

  px = coords->px;  py = coords->py; pz = coords->pz;
  fx = coords->fxa; fy = coords->fya; fz = coords->fza;
  nbends = coords->bends.nbends;
  ibend = coords->bends.ibend;
  jbend = coords->bends.jbend;
  kbend = coords->bends.kbend;
  idxbend= coords->bends.idxbend;
  eqbend = coords->bends.eqbend;
  fkbend = coords->bends.fkbend;
  tybend = coords->bends.ibendtype;
  iaacg = simparms->iaacg;
  icg = simparms->icgunit;
  spot = simparms->vcgintern;
  tpot = 0.;

  decomp1d((long)nbends,simparms->size,simparms->rank,&ibegin,&iend);

  nbt = iend-ibegin;
  fx1 = (double *)cmalloc(9*nbt*sizeof(double));
  fy1 = fx1 + nbt;
  fz1 = fx1 + 2*nbt;
  fx2 = fx1 + 3*nbt;  fy2 = fx1 + 4*nbt;  fz2 = fx1 + 5*nbt;
  fx3 = fx1 + 6*nbt;  fy3 = fx1 + 7*nbt;  fz3 = fx1 + 8*nbt;
  for(ib=0;ib<9*nbt;ib++) fx1[ib]=0.;

#include "vectorize.h"
  fact = 0.;
  dudth = 0.;

  for(inow=0,ib=ibegin;ib<iend;ib++,inow++){
    i = ibend[ib];
    j = jbend[ib];
    k = kbend[ib];
	iat = icg[i];
	if((iat!=icg[j]) || (iat!=icg[k])) continue;
	if(iaacg[iat]==0) continue;
    itype = idxbend[ib];
    ifunc_type = tybend[ib];
    
    ax = px[i] - px[j];
    ay = py[i] - py[j];
    az = pz[i] - pz[j];
    al = ax*ax+ay*ay+az*az;

    bx = px[k] - px[j];
    by = py[k] - py[j];
    bz = pz[k] - pz[j];
    bl = bx*bx+by*by+bz*bz;

    adotb = ax*bx + ay*by + az*bz;
    r11 = 1./sqrt(al*bl);
    costh = adotb*r11;
    
    costh = MIN(1.,costh);
    costh = MAX(-1.,costh);
    switch(ifunc_type) {
    case 0: /* original harmonic */
    /* 10 Nov 2009 FIX BERND */
      if(costh > 1.){
        costh =1.;
      }else if(costh < -1.){
        costh = -1.;
      }
//      if(fabs(costh) >= 1.){
//	th = 1.;
//	dth = th - eqbend[itype];
//	dthdcos = -1.;
//	dudth = 0.;
//	pot = .5*fkbend[itype]*dth*dth;
//	tpot += pot;
//	spot[iat] += pot;
//      } else {
      th = acos(costh);
      dth = th - eqbend[itype];
      dudth   = fkbend[itype]*dth;
      if(dth == 0.){
        dthdcos = 0.0;
      }else{
        dthdcos = -1./sqrt(1.-costh*costh);
      }
      pot = .5*dudth*dth;
      tpot += pot;
      spot[iat] += pot;
//      }
      fact = dudth*dthdcos;	  
      break;
    case 1: /* cosine */
      if(costh > 1.) costh =1.;
      else if(costh < -1.) costh = -1.;
      fact =  fkbend[itype];
      pot = fact * (1.+costh);
	  tpot += pot;
	  spot[iat] += pot;
      break;
    case 2: /* quartic */
      if(costh > 1.) costh =1.;
      else if(costh < -1.) costh = -1.;      
      th = acos(costh);      
      sinth = sin(th-M_PI);
      
      if(sinth > 1.) sinth =1.;
      else if(sinth < -1.) sinth = -1.;
      if(fabs(th-M_PI) < 0.01) t1 =  -1. * poly_quart(sinth, 20);
      else t1 = -(th-M_PI)/sinth;
      c = (eqbend[itype]-M_PI)*(eqbend[itype]-M_PI);
      b = c - (th-M_PI)*(th-M_PI);
      /* a = 0.5 * fkbend[itype] / c; */
      a = 0.5 * fkbend[itype] / c;
      t_coef =  0.25 * a;
      fact = a * b * t1;      
      pot = t_coef * b*b ;
	  tpot += pot;
	  spot[iat] += pot;
      break;
    }
	
    dcdx1 = costh*ax/al - bx*r11;
    dcdy1 = costh*ay/al - by*r11;
    dcdz1 = costh*az/al - bz*r11;
    dcdx3 = costh*bx/bl - ax*r11;
    dcdy3 = costh*by/bl - ay*r11;
    dcdz3 = costh*bz/bl - az*r11;
    dcdx2 = -dcdx1-dcdx3;
    dcdy2 = -dcdy1-dcdy3;
    dcdz2 = -dcdz1-dcdz3;

    fx1[inow] = fact*dcdx1;
    fy1[inow] = fact*dcdy1;
    fz1[inow] = fact*dcdz1;
    fx2[inow] = fact*dcdx2;
    fy2[inow] = fact*dcdy2;
    fz2[inow] = fact*dcdz2;
    fx3[inow] = fact*dcdx3;
    fy3[inow] = fact*dcdy3;
    fz3[inow] = fact*dcdz3;
		
  }
 
  for(inow=0,ib=ibegin;ib<iend;ib++,inow++){
    i = ibend[ib];    j = jbend[ib];    k = kbend[ib];
    fx[i] += fx1[inow];    fy[i] += fy1[inow];    fz[i] += fz1[inow];
    fx[j] += fx2[inow];    fy[j] += fy2[inow];    fz[j] += fz2[inow];
    fx[k] += fx3[inow];    fy[k] += fy3[inow];    fz[k] += fz3[inow];
  }
  return tpot;
}
/*-----------------------------------------------------------------------*/
double getvbend_intern_CG(SIMPARMS *simparms,COORDS *coords,int icg)
{
  int i,j,k,itype,ib,ifunc_type, *tybend, *ilist;
  double adotb,r11,costh,th,dudth,dthdcos,fact,dth;
  double a,dcdx1,dcdy1,dcdz1,dcdx2,dcdy2,dcdz2,dcdx3,dcdy3,dcdz3;
  double ax,ay,az,bx,by,bz,al,bl;
  double sinth,t1,pot,tpot;
  double *fx,*fy,*fz,*px,*py,*pz;
  double *eqbend,*fkbend;
  int *ibend,*jbend,*kbend,*idxbend;
	
  double t_coef, b, c;
	
  px = coords->px;  py = coords->py; pz = coords->pz;
  fx = coords->fxa; fy = coords->fya; fz = coords->fza;
  ibend = coords->bends.ibend;
  jbend = coords->bends.jbend;
  kbend = coords->bends.kbend;
  idxbend= coords->bends.idxbend;
  eqbend = coords->bends.eqbend;
  fkbend = coords->bends.fkbend;
  tybend = coords->bends.ibendtype;
  ilist = coords->bends.icgunit;
//  spot = simparms->vcgintern;
  tpot = 0.;
	
	
#include "vectorize.h"
  fact = 0.;
  dudth = 0.;
	
  for(ib=ilist[icg];ib<ilist[icg+1];ib++){
    i = ibend[ib];
    j = jbend[ib];
    k = kbend[ib];
    itype = idxbend[ib];
    ifunc_type = tybend[ib];
    
    ax = px[i] - px[j];
    ay = py[i] - py[j];
    az = pz[i] - pz[j];
    al = ax*ax+ay*ay+az*az;
		
    bx = px[k] - px[j];
    by = py[k] - py[j];
    bz = pz[k] - pz[j];
    bl = bx*bx+by*by+bz*bz;
		
    adotb = ax*bx + ay*by + az*bz;
    r11 = 1./sqrt(al*bl);
    costh = adotb*r11;
    
    costh = MIN(1.,costh);
    costh = MAX(-1.,costh);
    switch(ifunc_type) {
			case 0: /* original harmonic */
      /* 10 Nov 2009 FIX BERND */
        if(costh > 1.){
          costh =1.;
        }else if(costh < -1.){
          costh = -1.;
        }
//				if(fabs(costh) >= 1.){
//					th = 1.;
//					dth = th - eqbend[itype];
//					dthdcos = -1.;
//					dudth = 0.;
//					pot = .5*fkbend[itype]*dth*dth;
//					tpot += pot;
////					spot[iat] += pot;
//				} else {
					th = acos(costh);
					dth = th - eqbend[itype];
					dudth   = fkbend[itype]*dth;
          if(dth == 0.){
            dthdcos = 0.0;
          }else{
            dthdcos = -1./sqrt(1.-costh*costh);
          }
					pot = .5*dudth*dth;
					tpot += pot;
//					spot[iat] += pot;
//				}
				fact = dudth*dthdcos;	  
				break;
			case 1: /* cosine */
				if(costh > 1.) costh =1.;
				else if(costh < -1.) costh = -1.;
					fact =  fkbend[itype];
				pot = fact * (1.+costh);
				tpot += pot;
//				spot[iat] += pot;
				break;
			case 2: /* quartic */
				if(costh > 1.) costh =1.;
				else if(costh < -1.) costh = -1.;      
					th = acos(costh);      
				sinth = sin(th-M_PI);
				
				if(sinth > 1.) sinth =1.;
				else if(sinth < -1.) sinth = -1.;
					if(fabs(th-M_PI) < 0.01) t1 =  -1. * poly_quart(sinth, 20);
				else t1 = -(th-M_PI)/sinth;
				c = (eqbend[itype]-M_PI)*(eqbend[itype]-M_PI);
				b = c - (th-M_PI)*(th-M_PI);
				/* a = 0.5 * fkbend[itype] / c; */
				a = 0.5 * fkbend[itype] / c;
				t_coef =  0.25 * a;
				fact = a * b * t1;      
				pot = t_coef * b*b ;
				tpot += pot;
//				spot[iat] += pot;
				break;
    }
		
    dcdx1 = costh*ax/al - bx*r11;
    dcdy1 = costh*ay/al - by*r11;
    dcdz1 = costh*az/al - bz*r11;
    dcdx3 = costh*bx/bl - ax*r11;
    dcdy3 = costh*by/bl - ay*r11;
    dcdz3 = costh*bz/bl - az*r11;
    dcdx2 = -dcdx1-dcdx3;
    dcdy2 = -dcdy1-dcdy3;
    dcdz2 = -dcdz1-dcdz3;

//    printf("BEND: %d %d %d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf\n",ib,i,j,k,fact*dcdx1,fact*dcdy1,fact*dcdz1,fact*dcdx2,fact*dcdy2,fact*dcdz2,fact*dcdx3,fact*dcdy3,fact*dcdz3);

    fx[i] += fact*dcdx1;    fy[i] += fact*dcdy1;    fz[i] += fact*dcdz1;
    fx[j] += fact*dcdx2;    fy[j] += fact*dcdy2;    fz[j] += fact*dcdz2;
    fx[k] += fact*dcdx3;    fy[k] += fact*dcdy3;    fz[k] += fact*dcdz3;
  }
  return tpot;
}
/*------------------------------------------------------------------------*/
void fcbend(COORDS *coords,double **fcmat)
{
  int i,j,k,ii,jj,nb,itype;
  double ax,ay,az,al2,ral2,ral4,bx,by,bz,bl2,rbl2,rbl4;
  double adotb,r11,r33,costh,th,costh2;
  double dthdcos,dth,dudth,du2dth2,dth2cos2,du2dcos2,dudcos;
  double dcdxi,dcdyi,dcdzi,dcdxj,dcdyj,dcdzj,dcdxk,dcdyk,dcdzk;
  double dxx,dyy,dzz,dxy,dxz,dyz,dyx,dzx,dzy;
  double dc2dxx,dc2dyy,dc2dzz,dc2dxy,dc2dxz,dc2dyx,dc2dyz,dc2dzx,dc2dzy;
  double dabdxi,dabdyi,dabdzi,dabdxj,dabdyj,dabdzj,dabdxk,dabdyk,dabdzk;
  double drrdxi,drrdyi,drrdzi,drrdxj,drrdyj,drrdzj,drrdxk,drrdyk,drrdzk;
  double *px,*py,*pz;
  double *eqbend,*fkbend;
  int nbends,*ibend,*jbend,*kbend,*idxbend;

  px = coords->px;
  py = coords->py;
  pz = coords->pz;
  nbends = coords->bends.nbends;
  ibend = coords->bends.ibend;
  jbend = coords->bends.jbend;
  kbend = coords->bends.kbend;
  idxbend= coords->bends.idxbend;
  eqbend = coords->bends.eqbend;
  fkbend = coords->bends.fkbend;

  for(nb=0;nb<nbends;nb++){
    
    i = ibend[nb];
    j = jbend[nb];
    k = kbend[nb];
    itype = idxbend[nb];

    ax = px[i] - px[j];
    ay = py[i] - py[j];
    az = pz[i] - pz[j];
    al2 = ax*ax+ay*ay+az*az;
    ral2=1./al2;
    ral4 = ral2*ral2;
  
    bx = px[k] - px[j];
    by = py[k] - py[j];
    bz = pz[k] - pz[j];
    bl2=bx*bx+by*by+bz*bz;

    rbl2=1./bl2;
    rbl4=rbl2*rbl2;
    
    adotb = ax*bx + ay*by + az*bz;
    r11 = 1./sqrt(al2*bl2);
    r33 = r11*r11*r11;
    costh = adotb*r11;
    if(costh <= -1.){
      costh = -1.;
      th = M_PI;
      dthdcos = -1.;
    } else {
      th = acos(costh);
      dthdcos = -1./sqrt(1.-costh*costh);
    }
    costh2 = costh*costh;
    
    /* d f[g[h[x]]]/ dx = f'' g' g' h' h' + f' g'' h' h' + f' g' h'' */
    
    dth = th - eqbend[itype];
    dudth   = fkbend[itype]*dth;
    du2dth2 = fkbend[itype];
    dthdcos = -1./sqrt(1.-costh2);
    dth2cos2= -costh/pow(1-costh2,3./2.);
    du2dcos2= (du2dth2*dthdcos*dthdcos+dudth*dth2cos2);
    dudcos  = dudth*dthdcos;
    
    dabdxi = bx;     dabdyi = by;     dabdzi = bz;
    dabdxj = -bx-ax; dabdyj = -by-ay; dabdzj = -bz-az;
    dabdxk = ax;     dabdyk = ay;     dabdzk = az;
    
    drrdxi = -ax*ral2*r11;    drrdyi = -ay*ral2*r11;   drrdzi = -az*ral2*r11;
    drrdxj = (ax*ral2 + bx*rbl2)*r11;
    drrdyj = (ay*ral2 + by*rbl2)*r11;
    drrdzj = (az*ral2 + bz*rbl2)*r11;
    drrdxk = -bx*rbl2*r11;    drrdyk = -by*rbl2*r11;   drrdzk = -bz*rbl2*r11;
    
    dcdxi = adotb*drrdxi + dabdxi*r11;
    dcdyi = adotb*drrdyi + dabdyi*r11;
    dcdzi = adotb*drrdzi + dabdzi*r11;
    dcdxj = adotb*drrdxj + dabdxj*r11;
    dcdyj = adotb*drrdyj + dabdyj*r11;
    dcdzj = adotb*drrdzj + dabdzj*r11;
    dcdxk = adotb*drrdxk + dabdxk*r11;
    dcdyk = adotb*drrdyk + dabdyk*r11;
    dcdzk = adotb*drrdzk + dabdzk*r11;
    
    /* d f[x,y]h[x,y]/dxy=df/dx dh/dy+dh/dx df/dy+h df2/dxy+f dh2/dxy*/
    /*  calculate force constant matrix interact partical 1-1  */
  
    dc2dxx=(2.*dabdxi*drrdxi+adotb*(3.*ax*ax*r11*ral4 - r11*ral2));
    dc2dyy=(2.*dabdyi*drrdyi+adotb*(3.*ay*ay*r11*ral4 - r11*ral2));
    dc2dzz=(2.*dabdzi*drrdzi+adotb*(3.*az*az*r11*ral4 - r11*ral2));
    
    dc2dxy=(dabdxi*drrdyi+drrdxi*dabdyi+adotb*3.*ax*ay*r11*ral4);
    dc2dxz=(dabdxi*drrdzi+drrdxi*dabdzi+adotb*3.*ax*az*r11*ral4);
    dc2dyz=(dabdyi*drrdzi+drrdyi*dabdzi+adotb*3.*ay*az*r11*ral4);
    
    dc2dyx= dc2dxy;
    dc2dzx= dc2dxz;
    dc2dzy= dc2dyz;
    
    dxx = du2dcos2*dcdxi*dcdxi + dudcos*dc2dxx;
    dxy = du2dcos2*dcdxi*dcdyi + dudcos*dc2dxy;
    dxz = du2dcos2*dcdxi*dcdzi + dudcos*dc2dxz;
    dyx = du2dcos2*dcdyi*dcdxi + dudcos*dc2dyx;
    dyy = du2dcos2*dcdyi*dcdyi + dudcos*dc2dyy;
    dyz = du2dcos2*dcdyi*dcdzi + dudcos*dc2dyz;
    dzx = du2dcos2*dcdzi*dcdxi + dudcos*dc2dzx;
    dzy = du2dcos2*dcdzi*dcdyi + dudcos*dc2dzy;
    dzz = du2dcos2*dcdzi*dcdzi + dudcos*dc2dzz;
    
    ii = 3*i-1;

    fcmat[ii+1][ii+1] += dxx;
    fcmat[ii+1][ii+2] += dxy;
    fcmat[ii+1][ii+3] += dxz;
    fcmat[ii+2][ii+1] += dyx;
    fcmat[ii+2][ii+2] += dyy;
    fcmat[ii+2][ii+3] += dyz;
    fcmat[ii+3][ii+1] += dzx;
    fcmat[ii+3][ii+2] += dzy;
    fcmat[ii+3][ii+3] += dzz;
    
    /*  calculate force constant matrix for interact 1-2 */
    
    dc2dxx=(dabdxi*drrdxj+drrdxi*dabdxj-r11+
            adotb*(-3.*ax*ax*r11*ral4-ax*bx*r33+r11*ral2));
    dc2dyy=(dabdyi*drrdyj+drrdyi*dabdyj-r11+
            adotb*(-3.*ay*ay*r11*ral4-ay*by*r33+r11*ral2));
    dc2dzz=(dabdzi*drrdzj+drrdzi*dabdzj-r11+
            adotb*(-3.*az*az*r11*ral4-az*bz*r33+r11*ral2));

    dc2dxy=(dabdxi*drrdyj+drrdxi*dabdyj+adotb*
            (-3.*ax*ay*r11*ral4-ax*by*r33));
    dc2dyx=(dabdyi*drrdxj+drrdyi*dabdxj+adotb*
            (-3.*ay*ax*r11*ral4-ay*bx*r33));
    dc2dxz=(dabdxi*drrdzj+drrdxi*dabdzj+adotb*
            (-3.*ax*az*r11*ral4-ax*bz*r33));
    dc2dzx=(dabdzi*drrdxj+drrdzi*dabdxj+adotb*
            (-3.*az*ax*r11*ral4-az*bx*r33));
    dc2dyz=(dabdyi*drrdzj+drrdyi*dabdzj+adotb*
            (-3.*ay*az*r11*ral4-ay*bz*r33));
    dc2dzy=(dabdzi*drrdyj+drrdzi*dabdyj+adotb*
            (-3.*az*ay*r11*ral4-az*by*r33));
    
    dxx = du2dcos2*dcdxi*dcdxj + dudcos*dc2dxx;
    dxy = du2dcos2*dcdxi*dcdyj + dudcos*dc2dxy;
    dxz = du2dcos2*dcdxi*dcdzj + dudcos*dc2dxz;
    dyx = du2dcos2*dcdyi*dcdxj + dudcos*dc2dyx;
    dyy = du2dcos2*dcdyi*dcdyj + dudcos*dc2dyy;
    dyz = du2dcos2*dcdyi*dcdzj + dudcos*dc2dyz;
    dzx = du2dcos2*dcdzi*dcdxj + dudcos*dc2dzx;
    dzy = du2dcos2*dcdzi*dcdyj + dudcos*dc2dzy;
    dzz = du2dcos2*dcdzi*dcdzj + dudcos*dc2dzz;
    
    ii=3*i-1;
    jj=3*j-1;

    fcmat[ii+1][jj+1] += dxx;
    fcmat[ii+1][jj+2] += dxy;
    fcmat[ii+1][jj+3] += dxz;
    fcmat[ii+2][jj+1] += dyx;
    fcmat[ii+2][jj+2] += dyy;
    fcmat[ii+2][jj+3] += dyz;
    fcmat[ii+3][jj+1] += dzx;
    fcmat[ii+3][jj+2] += dzy;
    fcmat[ii+3][jj+3] += dzz;

    fcmat[jj+1][ii+1] += dxx;
    fcmat[jj+2][ii+1] += dxy;
    fcmat[jj+3][ii+1] += dxz;
    fcmat[jj+1][ii+2] += dyx;
    fcmat[jj+2][ii+2] += dyy;
    fcmat[jj+3][ii+2] += dyz;
    fcmat[jj+1][ii+3] += dzx;
    fcmat[jj+2][ii+3] += dzy;
    fcmat[jj+3][ii+3] += dzz;
    
    /*  calculate force constant matrix for interact 1-3 */
    
    dc2dxx=(dabdxi*drrdxk+drrdxi*dabdxk+r11+adotb*(ax*bx*r33));
    dc2dyy=(dabdyi*drrdyk+drrdyi*dabdyk+r11+adotb*(ay*by*r33));
    dc2dzz=(dabdzi*drrdzk+drrdzi*dabdzk+r11+adotb*(az*bz*r33));
    
    dc2dxy=(dabdxi*drrdyk+drrdxi*dabdyk+adotb*ax*by*r33);
    dc2dxz=(dabdxi*drrdzk+drrdxi*dabdzk+adotb*ax*bz*r33);
    dc2dyz=(dabdyi*drrdzk+drrdyi*dabdzk+adotb*ay*bz*r33);
    
    dc2dyx=(dabdyi*drrdxk+drrdyi*dabdxk+adotb*ay*bx*r33);
    dc2dzx=(dabdzi*drrdxk+drrdzi*dabdxk+adotb*az*bx*r33);
    dc2dzy=(dabdzi*drrdyk+drrdzi*dabdyk+adotb*az*by*r33);
    
    dxx = du2dcos2*dcdxi*dcdxk + dudcos*dc2dxx;
    dxy = du2dcos2*dcdxi*dcdyk + dudcos*dc2dxy;
    dxz = du2dcos2*dcdxi*dcdzk + dudcos*dc2dxz;
    dyx = du2dcos2*dcdyi*dcdxk + dudcos*dc2dyx;
    dyy = du2dcos2*dcdyi*dcdyk + dudcos*dc2dyy;
    dyz = du2dcos2*dcdyi*dcdzk + dudcos*dc2dyz;
    dzx = du2dcos2*dcdzi*dcdxk + dudcos*dc2dzx;
    dzy = du2dcos2*dcdzi*dcdyk + dudcos*dc2dzy;
    dzz = du2dcos2*dcdzi*dcdzk + dudcos*dc2dzz;
    
    ii=3*i-1;
    jj=3*k-1;
    
    fcmat[ii+1][jj+1] += dxx;
    fcmat[ii+1][jj+2] += dxy;
    fcmat[ii+1][jj+3] += dxz;
    fcmat[ii+2][jj+1] += dyx;
    fcmat[ii+2][jj+2] += dyy;
    fcmat[ii+2][jj+3] += dyz;
    fcmat[ii+3][jj+1] += dzx;
    fcmat[ii+3][jj+2] += dzy;
    fcmat[ii+3][jj+3] += dzz;

    fcmat[jj+1][ii+1] += dxx;
    fcmat[jj+2][ii+1] += dxy;
    fcmat[jj+3][ii+1] += dxz;
    fcmat[jj+1][ii+2] += dyx;
    fcmat[jj+2][ii+2] += dyy;
    fcmat[jj+3][ii+2] += dyz;
    fcmat[jj+1][ii+3] += dzx;
    fcmat[jj+2][ii+3] += dzy;
    fcmat[jj+3][ii+3] += dzz;

    /*  calculate force constant matrix for interact 2-2 */
    
    dc2dxx=(2.*dabdxj*drrdxj+2.*r11+adotb*
            (3.*ax*ax*r11*ral4+3.*bx*bx*r11*rbl4+
             2.*ax*bx*r33-r11*ral2-r11*rbl2));
    dc2dyy=(2.*dabdyj*drrdyj+2.*r11+adotb*
            (3.*ay*ay*r11*ral4+3.*by*by*r11*rbl4+
             2.*ay*by*r33-r11*ral2-r11*rbl2));
    dc2dzz=(2.*dabdzj*drrdzj+2.*r11+adotb*
            (3.*az*az*r11*ral4+3.*bz*bz*r11*rbl4+
             2.*az*bz*r33-r11*ral2-r11*rbl2));
    
    dc2dxy=(dabdxj*drrdyj+drrdxj*dabdyj+adotb*
            (3.*ax*ay*r11*ral4+3.*bx*by*r11*rbl4+
             (ax*by+ay*bx)*r33));
    dc2dxz=(dabdxj*drrdzj+drrdxj*dabdzj+adotb*
            (3.*ax*az*r11*ral4+3.*bx*bz*r11*rbl4+
             (ax*bz+az*bx)*r33));
    dc2dyz=(dabdyj*drrdzj+drrdyj*dabdzj+adotb*
            (3.*ay*az*r11*ral4+3.*by*bz*r11*rbl4+
             (ay*bz+az*by)*r33));
    
    dc2dyx=dc2dxy;
    dc2dzx=dc2dxz;
    dc2dzy=dc2dyz;

    
    dxx = du2dcos2*dcdxj*dcdxj + dudcos*dc2dxx;
    dxy = du2dcos2*dcdxj*dcdyj + dudcos*dc2dxy;
    dxz = du2dcos2*dcdxj*dcdzj + dudcos*dc2dxz;
    dyx = du2dcos2*dcdyj*dcdxj + dudcos*dc2dyx;
    dyy = du2dcos2*dcdyj*dcdyj + dudcos*dc2dyy;
    dyz = du2dcos2*dcdyj*dcdzj + dudcos*dc2dyz;
    dzx = du2dcos2*dcdzj*dcdxj + dudcos*dc2dzx;
    dzy = du2dcos2*dcdzj*dcdyj + dudcos*dc2dzy;
    dzz = du2dcos2*dcdzj*dcdzj + dudcos*dc2dzz;
    
    ii=3*j-1;

    fcmat[ii+1][ii+1] += dxx;
    fcmat[ii+1][ii+2] += dxy;
    fcmat[ii+1][ii+3] += dxz;
    fcmat[ii+2][ii+1] += dyx;
    fcmat[ii+2][ii+2] += dyy;
    fcmat[ii+2][ii+3] += dyz;
    fcmat[ii+3][ii+1] += dzx;
    fcmat[ii+3][ii+2] += dzy;
    fcmat[ii+3][ii+3] += dzz;
    
    /*  calculate force constant matrix for interact 2-3 */
    
    dc2dxx=(dabdxj*drrdxk+drrdxj*dabdxk-r11+
            adotb*(-3.*bx*bx*r11*rbl4-ax*bx*r33+r11*rbl2));
    dc2dyy=(dabdyj*drrdyk+drrdyj*dabdyk-r11+
            adotb*(-3.*by*by*r11*rbl4-ay*by*r33+r11*rbl2));
    dc2dzz=(dabdzj*drrdzk+drrdzj*dabdzk-r11+
            adotb*(-3.*bz*bz*r11*rbl4-az*bz*r33+r11*rbl2));
   
    dc2dxy=(dabdxj*drrdyk+drrdxj*dabdyk+adotb*(-ax*by*r33-3.*bx*by*r11*rbl4));
    dc2dxz=(dabdxj*drrdzk+drrdxj*dabdzk+adotb*(-ax*bz*r33-3.*bx*bz*r11*rbl4));
    dc2dyz=(dabdyj*drrdzk+drrdyj*dabdzk+adotb*(-ay*bz*r33-3.*by*bz*r11*rbl4));
    dc2dyx=(dabdyj*drrdxk+drrdyj*dabdxk+adotb*(-ay*bx*r33-3.*by*bx*r11*rbl4));
    dc2dzx=(dabdzj*drrdxk+drrdzj*dabdxk+adotb*(-az*bx*r33-3.*bz*bx*r11*rbl4));
    dc2dzy=(dabdzj*drrdyk+drrdzj*dabdyk+adotb*(-az*by*r33-3.*bz*by*r11*rbl4));
    
    dxx = du2dcos2*dcdxj*dcdxk + dudcos*dc2dxx;
    dxy = du2dcos2*dcdxj*dcdyk + dudcos*dc2dxy;
    dxz = du2dcos2*dcdxj*dcdzk + dudcos*dc2dxz;
    dyx = du2dcos2*dcdyj*dcdxk + dudcos*dc2dyx;
    dyy = du2dcos2*dcdyj*dcdyk + dudcos*dc2dyy;
    dyz = du2dcos2*dcdyj*dcdzk + dudcos*dc2dyz;
    dzx = du2dcos2*dcdzj*dcdxk + dudcos*dc2dzx;
    dzy = du2dcos2*dcdzj*dcdyk + dudcos*dc2dzy;
    dzz = du2dcos2*dcdzj*dcdzk + dudcos*dc2dzz;

    ii=3*j-1;
    jj=3*k-1;

    fcmat[ii+1][jj+1] += dxx;
    fcmat[ii+1][jj+2] += dxy;
    fcmat[ii+1][jj+3] += dxz;
    fcmat[ii+2][jj+1] += dyx;
    fcmat[ii+2][jj+2] += dyy;
    fcmat[ii+2][jj+3] += dyz;
    fcmat[ii+3][jj+1] += dzx;
    fcmat[ii+3][jj+2] += dzy;
    fcmat[ii+3][jj+3] += dzz;
    
    fcmat[jj+1][ii+1] += dxx;
    fcmat[jj+2][ii+1] += dxy;
    fcmat[jj+3][ii+1] += dxz;
    fcmat[jj+1][ii+2] += dyx;
    fcmat[jj+2][ii+2] += dyy;
    fcmat[jj+3][ii+2] += dyz;
    fcmat[jj+1][ii+3] += dzx;
    fcmat[jj+2][ii+3] += dzy;
    fcmat[jj+3][ii+3] += dzz;

    /*  calculate force constant matrix for interact 3-3 */
    
    dc2dxx=(2.*dabdxk*drrdxk+adotb*(3.*bx*bx*r11*rbl4 - r11*rbl2));
    dc2dyy=(2.*dabdyk*drrdyk+adotb*(3.*by*by*r11*rbl4 - r11*rbl2));
    dc2dzz=(2.*dabdzk*drrdzk+adotb*(3.*bz*bz*r11*rbl4 - r11*rbl2));
    dc2dxy=(dabdxk*drrdyk+drrdxk*dabdyk+adotb*3.*bx*by*r11*rbl4);
    dc2dxz=(dabdxk*drrdzk+drrdxk*dabdzk+adotb*3.*bx*bz*r11*rbl4);
    dc2dyz=(dabdyk*drrdzk+drrdyk*dabdzk+adotb*3.*by*bz*r11*rbl4);
    dc2dyx= dc2dxy;
    dc2dzx= dc2dxz;
    dc2dzy= dc2dyz;
    
    dxx = du2dcos2*dcdxk*dcdxk + dudcos*dc2dxx;
    dxy = du2dcos2*dcdxk*dcdyk + dudcos*dc2dxy;
    dxz = du2dcos2*dcdxk*dcdzk + dudcos*dc2dxz;
    dyx = du2dcos2*dcdyk*dcdxk + dudcos*dc2dyx;
    dyy = du2dcos2*dcdyk*dcdyk + dudcos*dc2dyy;
    dyz = du2dcos2*dcdyk*dcdzk + dudcos*dc2dyz;
    dzx = du2dcos2*dcdzk*dcdxk + dudcos*dc2dzx;
    dzy = du2dcos2*dcdzk*dcdyk + dudcos*dc2dzy;
    dzz = du2dcos2*dcdzk*dcdzk + dudcos*dc2dzz;

    ii=3*k-1;
    
    fcmat[ii+1][ii+1] += dxx;
    fcmat[ii+1][ii+2] += dxy;
    fcmat[ii+1][ii+3] += dxz;
    fcmat[ii+2][ii+1] += dyx;
    fcmat[ii+2][ii+2] += dyy;
    fcmat[ii+2][ii+3] += dyz;
    fcmat[ii+3][ii+1] += dzx;
    fcmat[ii+3][ii+2] += dzy;
    fcmat[ii+3][ii+3] += dzz;
  }
}
/*------------------------------------------------------------------------*/
void fcbendn(COORDS *coords,double **fcmat)
{
  int nb,itype;
  for(nb=0;nb<coords->bends.nbends;nb++){
    
    iben = coords->bends.ibend[nb];
    jben = coords->bends.jbend[nb];
    kben = coords->bends.kbend[nb];
    itype = coords->bends.idxbend[nb];

    eqben = coords->bends.eqbend[itype];
    fkben = coords->bends.fkbend[itype];

    ngradv2(coords->px,coords->py,coords->pz,iben,iben,potbend,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,iben,jben,potbend,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,iben,kben,potbend,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,jben,iben,potbend,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,jben,jben,potbend,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,jben,kben,potbend,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,kben,iben,potbend,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,kben,jben,potbend,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,kben,kben,potbend,fcmat);
  }
}
/*---------------------------------------------------------------------*/
double potbend(double *px,double *py,double *pz)
{
  double ax,ay,az,al2,bx,by,bz,bl2,adotb,r11,costh,th,dth,spot;

  ax = px[iben] - px[jben];
  ay = py[iben] - py[jben];
  az = pz[iben] - pz[jben];
  al2 = ax*ax+ay*ay+az*az;

  
  bx = px[kben] - px[jben];
  by = py[kben] - py[jben];
  bz = pz[kben] - pz[jben];
  bl2=bx*bx+by*by+bz*bz;
  
  adotb = ax*bx + ay*by + az*bz;
  r11 = 1./sqrt(al2*bl2);
  costh = adotb*r11;
  if(costh <= -1.){
    costh = -1.;
    th = M_PI;
  } else {
    th = acos(costh);
  }
  dth = th - eqben;
  spot = .5*fkben*dth*dth;

  return spot;
}
/*---------------------------------------------------------------------*/
