/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

   /* compute positions of the coarse grain sites */

#include "md.h"  
#define MIN(A,B) ((A<B)?(A):(B))

#define DEBUG
#define SWITCH2
 
/*------------------------------------------------------------------*/
  /* allocate memory for AA/HR/CG regions */
  /* ToDo: 1) only allocate for hybrid MD. 2) bookkeep memory. 3) reduce array lengths */
void allocate_hybrid_lists(SIMPARMS *simparms){
  int i;
  int natoms = simparms->natoms;
  int ncg = simparms->ncg;

  simparms->iaacg = (int *)cmalloc(natoms*sizeof(int));
  simparms->iaacg_b4 = (int *)cmalloc(ncg*sizeof(int));
  simparms->faacg = (double *)cmalloc(natoms*sizeof(double));
  simparms->faacgprev = (double *)cmalloc(natoms*sizeof(double));
  simparms->dfaacg = (double *)cmalloc(natoms*sizeof(double));
  simparms->dfaacgprev = (double *)cmalloc(natoms*sizeof(double));
  
  simparms->dr = (double *)cmalloc(natoms*sizeof(double));
  simparms->ddr = (double *)cmalloc(natoms*sizeof(double));
  simparms->ddr_intra = (double *)cmalloc(natoms*sizeof(double));
  simparms->ddr_tors = (double *)cmalloc(natoms*sizeof(double));
  simparms->ddr_lrf = (double *)cmalloc(natoms*sizeof(double));
  
  simparms->vbuf= (double *)cmalloc(natoms*sizeof(double));
  simparms->vds_tors = (double *)cmalloc(natoms*sizeof(double));
  simparms->vds_torsprev = (double *)cmalloc(natoms*sizeof(double));
  simparms->vds_intra = (double *)cmalloc(natoms*sizeof(double));
  simparms->vds_intraprev = (double *)cmalloc(natoms*sizeof(double));
  simparms->vds_lrfprev = (double *)cmalloc(natoms*sizeof(double));
  simparms->vds_lrf = (double *)cmalloc(natoms*sizeof(double));
  simparms->vds_lrfprev = (double *)cmalloc(natoms*sizeof(double));
  
  simparms->vcgintern = (double *)cmalloc((ncg+1)*sizeof(double));
  simparms->vcgintern++ ;  /* use vcgintern[-1] as trashbin */
  
  for(i=0; i<ncg; i++) simparms->iaacg[i] = 0;  /* default is whole system is AA region */
  for(i=0; i<ncg; i++) simparms->iaacg_b4[i] = 0;  
  for(i=ncg; i<natoms; i++) simparms->iaacg[i] = 1;
  for(i=0; i<natoms; i++) simparms->vds_intraprev[i] = simparms->vds_lrfprev[i] = 
    simparms->vds_torsprev[i] = 0.;
    
  return;
}
/*------------------------------------------------------------------*/
/* fill out iaacg for AA part using CG part                   */
void set_aaselection(int ncg, int natoms, int *icgunit, int *iaacg){
  int iaa, naa=natoms-ncg;

  for(iaa=0;iaa<naa;++iaa) 
	iaacg[ncg+iaa] = 2 - iaacg[icgunit[iaa]];
}

/*-------------------------------------------------------------------*/
/* Compute centers of mass of AA atoms and compare with CG positions */
/* and adjust AA positions accordingly                               */
/* Can also be used for velocities instead of positions              */
void check_aa_com(double *pcgx,double *paax,double *aamass,
	double *cgmass,int **icglist, int *napcg, int naa, int ncg, int iflg){

  int i,iat,icg,*ilist;
  double px,py,pz,dx,dy,dz,dxmax,dymax,dzmax;;
  double *pcgy,*pcgz,*paay,*paaz;
  double tmass;
  LINE line;

  pcgy = pcgx+ncg;  pcgz = pcgy+ncg;
  paay = paax+naa;  paaz = paay+naa;
  dxmax = dymax = dzmax = 0.0;
  
  if(iflg==0){
    printf("Comparing AA centers of mass to CG positions...\n");
  }else{
    printf("Comparing AA centers of mass to CG velocities...\n");
  }
    
  for(icg=0;icg<ncg;icg++){
	tmass=0.0;
	px = py = pz = 0.0;
	ilist = icglist[icg];
    for(i=0;i<napcg[icg];i++){
      iat = ilist[i];
	  px += aamass[iat]*paax[iat ];
	  py += aamass[iat]*paay[iat];
	  pz += aamass[iat]*paaz[iat];
	  tmass += aamass[iat];
	}
	px /= tmass;
	py /= tmass;
	pz /= tmass;
	dx = pcgx[icg] - px;
	dy = pcgy[icg] - py;
	dz = pcgz[icg] - pz;
    for(i=0;i<napcg[icg];i++){
      iat = ilist[i];
	  paax[iat] += dx;
	  paay[iat] += dy;
	  paaz[iat] += dz;
	}
/*	printf("%d | %lf %lf %lf | %lf %lf %lf |  %lf %lf %lf \n",
		icg,px,pcgx[icg],dx,py,pcgy[icg],dy,pz,pcgz[icg],dz); */
    dx = fabs(dx); dy = fabs(dy); dz = fabs(dz);
	if(dx>dxmax) dxmax = dx; 
	if(dy>dymax) dymax = dy; 
	if(dz>dzmax) dzmax = dz; 
/* if CG mass is 0.1 set it to the sum of AA masses */
	if(abs(cgmass[icg]-0.1*MCONV) < 0.00001 ){
	  cgmass[icg]=tmass;
	}else if(abs(cgmass[icg]-tmass) > 0.00001){
	  sprintf(line," Mass of CG atom no.%d (=%12.8f) is not equal to that of sum of atoms (=%12.8f)!!!",
		i,cgmass[icg]/MCONV,tmass/MCONV);
	  md_warning(line);
	}
  }
  if((dxmax>0.0)||(dymax>0.0)||(dzmax>0)){
    if(iflg==0){
      sprintf(line,"Translating AA positions to match c.o.m. with CG positions\nLargest AA displacements in x, y and z: %lf %lf %lf\n",dxmax,dymax,dzmax);
	}else{
      sprintf(line,"Translating AA velocities to match c.o.m. with CG velocities\nLargest AA displacements in x, y and z: %lf %lf %lf\n",dxmax,dymax,dzmax);
	}
	md_warning(line);
	printf("%s\n",line);
  }
}

/*------------------------------------------------------------------*/
/* Compute CG atom positions as c.o.m. of their AA atoms  */
void get_cgcom(SIMPARMS *simparms,COORDS *caa,COORDS *ccg)
{

  int i,j,ii,ncg;
  double *amass = caa->amass;
  double *px,*py,*pz,tmass;
  double *pcgx,*pcgy,*pcgz;

  px = caa->px;  py = caa->py;  pz = caa->pz;
  pcgx = ccg->px;  pcgy = ccg->py;  pcgz = ccg->pz;
  ncg = simparms->ncg;

  for(i=0;i<ncg;i++){
/*    if(simparms->iaacg[i]==0){ */
      tmass=0.0;
      pcgx[i] = pcgy[i] = pcgz[i] = 0.0;
      for(j=0;j<simparms->napcg[i];j++){
        ii=simparms->icglist[i][j];
        pcgx[i] += amass[ii]*px[ii];
        pcgy[i] += amass[ii]*py[ii];
        pcgz[i] += amass[ii]*pz[ii];
        tmass += amass[ii];
      }
	  /* if CG mass is zero set it to the sum of AA masses */
      if(abs(ccg->amass[i]-0.1*MCONV) < 0.00001 ) ccg->amass[i]=tmass;
      pcgx[i] /= tmass;
      pcgy[i] /= tmass;
      pcgz[i] /= tmass;
	}
/*  }  */
}

/*------------------------------------------------------------------*/
/* Compute CG atom velocities as c.o.m. of their AA atoms  */
void get_cgvelcom(SIMPARMS *simparms,COORDS *caa,COORDS *ccg)
{

  int i,j,ii,ncg;
  double *amass = caa->amass;
  double *vx,*vy,*vz,tmass;
  double *vcgx,*vcgy,*vcgz;

  vx = caa->vx;  vy = caa->vy;  vz = caa->vz;
  vcgx = ccg->vx;  vcgy = ccg->vy;  vcgz = ccg->vz;
  ncg = simparms->ncg;

  for(i=0;i<ncg;i++){
/*    if(simparms->iaacg[i]==0){ */
      vcgx[i] = vcgy[i] = vcgz[i] = 0.0;
      for(j=0;j<simparms->napcg[i];j++){
        ii=simparms->icglist[i][j];
        vcgx[i] += amass[ii]*vx[ii];
        vcgy[i] += amass[ii]*vy[ii];
        vcgz[i] += amass[ii]*vz[ii];
      }
	  /* if CG mass is zero set it to the sum of AA masses */
      tmass = ccg->amass[i];
      vcgx[i] /= tmass;
      vcgy[i] /= tmass;
      vcgz[i] /= tmass;
	}
/*  }  */
}

/*------------------------------------------------------------------*/
/* Compute non-active CG atom position as c.o.m. of its AA atoms    */
void get_cgpos(COORDS *coords,int **icglist, int *napcg, int nselect, int *iselect){

  int i,j,iat,icg;
  double *pcgx,*pcgy,*pcgz, *paax,*paay,*paaz;
  double tmass, *amass=coords->next->amass;

  pcgx = coords->px; pcgy = coords->py;  pcgz = coords->pz;
  paax = coords->next->px; paay = coords->next->py;  paaz = coords->next->pz;
  
  for(i=0;i<nselect;i++){
	icg = iselect[i];
	tmass=0.0;
	pcgx[icg] = pcgy[icg] = pcgz[icg] = 0.0;
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
	  pcgx[icg] += amass[iat]*paax[iat];
	  pcgy[icg] += amass[iat]*paay[iat];
	  pcgz[icg] += amass[iat]*paaz[iat];
	  tmass += amass[iat];
	}
	pcgx[icg] /= tmass;
	pcgy[icg] /= tmass;
	pcgz[icg] /= tmass;
  }
}


/*------------------------------------------------------------------ */
/* Compute non-active CG atom velocities as c.o.m. of their AA atoms */
/* void get_cgvel(SIMPARMS *simparms,COORDS *caa, COORDS *ccg) */
void get_cgvel(COORDS *coords,int **icglist, int *napcg, int nselect, int *iselect){
{

  int i,j,iat,icg;
  double *amass = coords->next->amass;
  double *vaax,*vaay,*vaaz,tmass;
  double *vcgx,*vcgy,*vcgz;

  vaax = coords->next->vx;  vaay = coords->next->vy;  vaaz = coords->next->vz;
  vcgx = coords->vx;  vcgy = coords->vy;  vcgz = coords->vz;

  for(i=0;i<nselect;i++){
	icg = iselect[i];
	tmass=0.0;
	vcgx[icg] = vcgy[icg] = vcgz[icg] = 0.0;
      for(j=0;j<napcg[icg];j++){
        iat=icglist[icg][j];
        vcgx[icg] += amass[iat]*vaax[iat];
        vcgy[icg] += amass[iat]*vaay[iat];
        vcgz[icg] += amass[iat]*vaaz[iat];
        tmass += amass[iat];
      }  
      vcgx[icg] /= tmass;
      vcgy[icg] /= tmass;
      vcgz[icg] /= tmass;
	}
  } 
}


/*------------------------------------------------------------------ */
/* subtract CG unit position from its (inactive) AA positions        */
  void get_relat_aapos(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *pcgx,*pcgy,*pcgz, *paax,*paay,*paaz;
  
  pcgx = coords->px; pcgy = coords->py;  pcgz = coords->pz;
  paax = coords->next->px; paay = coords->next->py;  paaz = coords->next->pz;
  
  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      paax[iat] -= pcgx[icg];
      paay[iat] -= pcgy[icg];
      paaz[iat] -= pcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* add selected CG unit position to its (inactive) AA positions      */
  void get_abs_aapos(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *pcgx,*pcgy,*pcgz, *paax,*paay,*paaz;
  
  pcgx = coords->px; pcgy = coords->py;  pcgz = coords->pz;
  paax = coords->next->px; paay = coords->next->py;  paaz = coords->next->pz;

  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      paax[iat] += pcgx[icg];
      paay[iat] += pcgy[icg];
      paaz[iat] += pcgz[icg];
		}
  }
} 
/*------------------------------------------------------------------ */
/* subtract selected CG unit velocities from its (inactive) AA velocities */
  void get_relat_aavel(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *vcgx,*vcgy,*vcgz, *vaax,*vaay,*vaaz;
  
  vcgx = coords->vx; vcgy = coords->vy;  vcgz = coords->vz;
  vaax = coords->next->vx; vaay = coords->next->vy;  vaaz = coords->next->vz;
  
  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      vaax[iat] -= vcgx[icg];
      vaay[iat] -= vcgy[icg];
      vaaz[iat] -= vcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* add selected CG unit velocities to its (inactive) AA velocities   */
  void get_abs_aavel(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *vcgx,*vcgy,*vcgz, *vaax,*vaay,*vaaz;
  
  vcgx = coords->vx; vcgy = coords->vy;  vcgz = coords->vz;
  vaax = coords->next->vx; vaay = coords->next->vy;  vaaz = coords->next->vz;

  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      vaax[iat] += vcgx[icg];
      vaay[iat] += vcgy[icg];
      vaaz[iat] += vcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* set relative AA velocities to zero but maintain centr of mass motion  */
  void zero_aavel(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *vcgx,*vcgy,*vcgz, *vaax,*vaay,*vaaz;
  
  vcgx = coords->vx; vcgy = coords->vy;  vcgz = coords->vz;
  vaax = coords->next->vx; vaay = coords->next->vy;  vaaz = coords->next->vz;

  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      vaax[iat] = vcgx[icg];
      vaay[iat] = vcgy[icg];
      vaaz[iat] = vcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* set AA velocities to CG velocities                                */
  void distribute_cgvel_over_aa(COORDS *coords,int **icglist, int *napcg){
  
  int i,j,iat,icg;
  int *iselect;
  double *vcgx,*vcgy,*vcgz, *vaax,*vaay,*vaaz;
  
  iselect = coords->iselect;
  vcgx = coords->vx; vcgy = coords->vy;  vcgz = coords->vz;
  vaax = coords->next->vx; vaay = coords->next->vy;  vaaz = coords->next->vz;

  for(i=0;i<coords->nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      vaax[iat] = vcgx[icg];
      vaay[iat] = vcgy[icg];
      vaaz[iat] = vcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* compute AA velocities from CG accelerations for iselect AA units     */
  void distribute_cgaccel(double *vaax,double dt,int ncg,int naa,int *icgunit,int *iselect,
							int nselect,double *acgx){
  
  int i,iaa,icg;
  double *acgy,*acgz,*vaay,*vaaz;
  
  acgy = acgx + ncg;  acgz = acgy + ncg;  
  vaay = vaax + naa;  vaaz = vaay + naa;  
  
  for(i=0;i<nselect;i++){
	iaa = iselect[i];
	icg = icgunit[iaa];
	vaax[iaa] += dt*acgx[icg];
	vaay[iaa] += dt*acgy[icg];
	vaaz[iaa] += dt*acgz[icg];
  }
} 

/*------------------------------------------------------------------ */
/* set AA accelerations to CG accelerations                          */
  void distribute_cgaccel_over_aa(COORDS *coords,int **icglist,int *napcg,double *acgx,double *aaax){
						
  int i,j,iat,icg;  
  int *iselect;  
  double *acgy,*acgz,*aaay,*aaaz; 
									
  iselect = coords->iselect;      
  acgy = acgx + coords->natoms;    acgz = acgy + coords->natoms;          
  aaay = aaax + coords->next->natoms;  aaaz = aaay + coords->next->natoms;   
									
  for(i=0;i<coords->nselect;i++){  
	icg = iselect[i];             
    for(j=0;j<napcg[icg];j++){  
      iat = icglist[icg][j];  
      aaax[iat] = acgx[icg];  
      aaay[iat] = acgy[icg];  
      aaaz[iat] = acgz[icg];  
	}  
  }  
 }   

/*------------------------------------------------------------------ */
/*  select atoms that are in an active AA/CG region                  */
/*  nselect: number of active atoms; iselect points to start of list */
/*  nschizo: number of HR active atoms; ischizo points to list (2nd part)  */
/*  ninactive: number of inactive atoms; inactive points to list (3rd part)*/
int select_atoms(int *iaacg, int *ndof, COORDS *coords)
{

  int i,ii; 
  int nsel,nunsel;
  int *iselect=coords->iselect;
  
  nsel=0;
  nunsel=coords->natoms-1;
  for(i=0;i<coords->natoms;i++){
    if(iaacg[i]!=0){
	  iselect[nsel++]=i;
	}else{
	  iselect[nunsel--]=i;
	}
  }
  coords->inactive = coords->iselect+nsel;
  /* *ndof -= 3*(coords->nselect - nsel); */
  coords->nselect = nsel;
  coords->ninactive = coords->natoms - nsel;
  
  coords->ischizo = coords->inactive;       /*points to list of HR atoms */
  coords->nschizo = 0;
  i=0;
  while(i<nsel){
    ii = iselect[i];
    if(iaacg[ii]==1){
	  swap_active2healing(coords,ii,i);
	  nsel--;
	}else{
	  i++;
	}
  }

  /* put HR atoms in second part of active selection */
	  
  
  return coords->nselect;
}

/*------------------------------------------------------------------ */
/*  swap atom from active region to healing region            */
void swap_active2healing(COORDS *coords, int iatom, int ifind)
{
  int idummy, iat;
  int *iselect=coords->iselect;
  int ilast = coords->nselect-coords->nschizo-1;
  
  if(ifind<0){
    for(iat=0;iat<coords->nselect;iat++)
      if(iselect[iat]==iatom) break;
  }else{
    iat = ifind;
  }
  
  idummy = iselect[ilast];
  iselect[ilast] = iatom;
  iselect[iat] = idummy;
  /* coords->nselect--; */   /* coords->nselect is total number of active atoms inc. HR)  */
  coords->nschizo++;
  coords->ischizo--;
}


/*------------------------------------------------------------------ */
/*  swap atom from inactive region to active AA/CG region            */
void activate_atom(COORDS *coords, int iatom, int ifind)
{
  int idummy, iat;
  int *inactive=coords->inactive;
  
  if(ifind<0){
    for(iat=0;iat<coords->ninactive;iat++)
      if(inactive[iat]==iatom) break;
  }else{
    iat = ifind;
  }
  
  idummy = inactive[0];
  inactive[0] = iatom;
  inactive[iat] = idummy;
  coords->nselect++;
  coords->ninactive--;
  inactive++;
}

/*------------------------------------------------------------------ */
/*  swap atom from active region to inactive AA/CG region            */
void deactivate_atom(COORDS *coords, int iatom, int ifind)
{
  int idummy, iat;
  int *iselect=coords->iselect;
  int natoms = coords->natoms;
  
  if(ifind<0){
    for(iat=0;iat<coords->nselect;iat++)
      if(iselect[iat]==iatom) break;
  }else{
    iat = ifind;
  }
  
  idummy = iselect[natoms];
  iselect[natoms] = iatom;
  iselect[iat] = idummy;
  coords->nselect--;
  coords->ninactive++;
  coords->inactive--;
}

/*------------------------------------------------------------------ */
/*  select atoms that are in an active AA/CG region                  */
void distribute_force_over_aa(COORDS *coords,int **icglist,int *napcg,double *fcgx,double *faax)
{
  int i,j,iat,icg;
  int nselect=coords->nselect;
  int *iselect=coords->iselect;
  double fact;
  double *fcgy, *fcgz, *faay, *faaz;
/*  double *px=coords->next->px, *py=coords->next->py, *pz=coords->next->pz; */
  double *aamass=coords->next->amass,  *cgmass=coords->amass;

  fcgy = fcgx + coords->natoms; fcgz = fcgy + coords->natoms;             
  faay = faax + coords->next->natoms; faaz = faay + coords->next->natoms; 
  
  for(i=0;i<nselect;i++){
	icg = iselect[i];
	for(j=0;j<napcg[icg];j++){
	  iat = icglist[icg][j];
	  fact = aamass[iat]/cgmass[icg];
	  faax[iat] = fact*fcgx[icg]; 
	  faay[iat] = fact*fcgy[icg]; 
	  faaz[iat] = fact*fcgz[icg]; 
    }
  }
}

/*------------------------------------------------------------------   */
/*  compute switch function faacg for CG units for which iaacg==1 (HR) */
/*  set to 0 (if in AA region ) or 1 (if in CG region ) otherwise      */
/*	Switch function1: S(r) = (r-R1)^2 * (3R2-R1-2r) / (R2-R1)^3    	   */
/*	Switch function2: S(r) = 1- (4(r-R2)^5 + 5(R2-R1)(r-R2)^4) / (R2-R1)^5  */
/*	Switch function3: S(r) = 1- [(4(r-R2)^5 + 5(R2-R1)(r-R2)^4) / (R2-R1)^5]^2  */
void set_switchfunction(SIMPARMS *simparms,COORDS *coords)
{
  int i,j,ii,idir,iperd,ivol,*iaa;
  int *iaacg = simparms->iaacg;
  int ierr;
  double pcgx,pcgy,pcgz,dx,dy,dz,dr2,dr2min,r1_hr,r2_hr,*faa,*dfaa;
  double z,tnat_cg,tnat_aa,dum,*ptr;

#ifdef SWITCH2
  double z5,dum4;
#elif defined(SWITCH3)
  double z10,dum4,piece;
#endif
  double *faacg,*dfaacg;
  double *drnow;
  LINE line;
  
   
  ierr=0;
  iperd = simparms->iperd;
  ivol = simparms->ivol;
  r1_hr = simparms->rmin_hr;
  r2_hr = simparms->rmax_hr;
  z = r2_hr - r1_hr;
#ifdef SWITCH1
  z = z*z*z;
#elif defined(SWITCH2)
  z5 = z*z; 
  z5 *= z5; 
  z5 *= z;
#elif defined(SWITCH3)
  z10 = z*z; 
  z10 *= z10; 
  z10 *= z;
  z10 *= z10;
#else
#error "No switch function defined"
#endif
  idir = simparms->itype_hr-1;
  
  /* rotate pointers */
  ptr = simparms->dr;
  drnow = simparms->dr = simparms->ddr;
  simparms->ddr = ptr;
  ptr = simparms->dfaacg;
  dfaacg = simparms->dfaacg = simparms->dfaacgprev;
  simparms->dfaacgprev = ptr;
  ptr = simparms->faacg;
  faacg = simparms->faacg = simparms->faacgprev;
  simparms->faacgprev = ptr;
     
  switch(simparms->itype_hr){
  case -1: /* no healing region */
    for(i=0;i<simparms->ncg;i++){
      if(iaacg[i]==2){
	faacg[i] = 1.;
	dfaacg[i] = 0.;
      }else if(iaacg[i]==0){
	faacg[i] = 0.;
	dfaacg[i] = 0.;
      }else{
	md_error("Hoe kan dat nou??");
      }
    }
  break;  
  case 0:   /* healing region is sphere shape */
    for(i=0;i<simparms->ncg;i++){
      if(iaacg[i]==2){
	faacg[i] = 1.;
	dfaacg[i] = 0.;
	drnow[i] = 0.;
      }else if(iaacg[i]==0){
	faacg[i] = 0.;
	dfaacg[i] = 0.;
	drnow[i] = 0.;
      }else{
	/* compute switchfunction */
	/* loop over units that determine aaregion and find closest one */
	dr2min=999999.;
	for(j=0;j<simparms->naaregion;j++){
	  ii=simparms->iaaregion[j];
	  pcgx=coords->px[ii]; pcgy=coords->py[ii]; pcgz=coords->pz[ii];
	  dx=coords->px[i]-pcgx; dy=coords->py[i]-pcgy; dz=coords->pz[i]-pcgz;
	  period(1,&dx,&dy,&dz,coords->hmat,coords->hmati,iperd,ivol); /* could be vectorized here */
	  dr2 = dx*dx+dy*dy+dz*dz;
	  dr2min=MIN(dr2,dr2min);
	}
	dr2min = sqrt(dr2min);
	if(dr2min < r1_hr){			/* inner skin */
	  faacg[i] = 0.;
	  dfaacg[i] = 0.;
	  drnow[i] = r1_hr;
	}else if(dr2min > r2_hr){	/* outher skin */
	  faacg[i] = 1.;
	  dfaacg[i] = 0.;
	  drnow[i] = r2_hr;
	}else{							/* actual healing region */
	  if(faacg[i] == -1) ierr++;
#ifdef SWITCH1
	  dum = dr2min-r1_hr;
	  faacg[i] = dum*dum*(3.0*r2_hr - r1_hr - dr2min*2.)/z;
	  dfaacg[i] = 6.*dum*(r2_hr-dr2min)/z;
#elif defined(SWITCH2)
	  dum = dr2min - r2_hr;
	  dum4 = dum*dum; 
	  dum4 *= dum4;
	  dfaacg[i] = -20*(dum4 + z*dum*dum*dum)/z5;
	  faacg[i] = 1-(4*dum4*dum + 5*z*dum4)/z5;
#elif defined(SWITCH3)
	  dum = dr2min - r2_hr;
	  dum4 = dum*dum; 
	  dum4 *= dum4;
	  piece = 4*dum4*dum + 5*z*dum4;
	  dfaacg[i] = -40*piece*(dum4 + z*dum*dum*dum)/z10;
	  faacg[i] = 1-(piece*piece)/z10;
#endif
	  drnow[i] = dr2min;
	  /* faacg[i] = 0.5; DEBUG TEST */
	}
      }
    }
    break;
  case 1:   /* healing region is yx-slice shape */
  case 2:   /* healing region is xz-slice shape */
  case 3:   /* healing region is yz-slice shape */
    
    for(i=0;i<simparms->ncg;i++){
      if(iaacg[i]==2){
	faacg[i] = 1.;
	dfaacg[i] = 0.;
	drnow[i] = 0.;
      }else if(iaacg[i]==0){
	faacg[i] = 0.;
	dfaacg[i] = 0.;
	drnow[i] = 0.;
      }else{
	/* compute switchfunction */
	/* loop over units that determine aaregion and find closest one */
	dr2min=999999.;
	for(j=0;j<simparms->naaregion;j++){
	  ii=simparms->iaaregion[j];
	  pcgx=coords->px[ii]; pcgy=coords->py[ii]; pcgz=coords->pz[ii];
	  if(idir==0){
	    dx=coords->px[i]-pcgx;
	  }else if(idir==1){ 
	    dx=coords->py[i]-pcgy; 
	  }else{
	    dx=coords->pz[i]-pcgz;
	  }
	  period(1,&dx,&dx,&dx,coords->hmat,coords->hmati,iperd,ivol); /* could be vectorized here */
	  dx*=dx; 
	  dr2min = MIN(dr2min,dx);
	}
	dr2min = sqrt(dr2min);
	if(dr2min < r1_hr){			/* inner skin */
	  faacg[i] = 0.;
	  dfaacg[i] = 0.;
	  drnow[i] = r1_hr;
	}else if(dr2min > r2_hr){	/* outher skin */
	  faacg[i] = 1.;
	  dfaacg[i] = 0.;
	  drnow[i] = r2_hr;
	}else{							/* actual healing region */
	  if(faacg[i] == -1) ierr++;
#ifdef SWITCH1
	  dum = dr2min-r1_hr;
	  faacg[i] = dum*dum*(3.0*r2_hr - r1_hr - dr2min*2.)/z;
	  dfaacg[i] = 6.*dum*(r2_hr-dr2min)/z;
#elif defined(SWITCH2)
	  dum = dr2min - r2_hr;
	  dum4 = dum*dum; 
	  dum4 *= dum4;
	  dfaacg[i] = -20*(dum4 + z*dum*dum*dum)/z5;
	  faacg[i] = 1-(4*dum4*dum + 5*z*dum4)/z5;
#elif defined(SWITCH3)
	  dum = dr2min - r2_hr;
	  dum4 = dum*dum; 
	  dum4 *= dum4;
	  piece = 4*dum4*dum + 5*z*dum4;
	  dfaacg[i] = -40*piece*(dum4 + z*dum*dum*dum)/z10;
	  faacg[i] = 1-(piece*piece)/z10;
#endif
	  drnow[i] = dr2min;
	}
      }
    }
    break;
    
  case 4:   /* whole system is hybrid healing region */
    /*	z = simparms->switch_fmin + simparms->switch_df*simparms->istep;
	for(i=0;i<simparms->ncg;i++){
	if(iaacg[i]==1){
	faacg[i] = z; 
	dfaacg[i] = simparms->switch_df;
	drnow[i] = 1.;
	}else{
	faacg[i] = 1.; 
	dfaacg[i] = 0.;
	drnow[i] = 0.;
	}
	}
	break;
    */
  case 5:   /* healing region is a cylinder in the x-direction */
  case 6:   /* healing region is a cylinder in the y-direction */
  case 7:   /* healing region is a cylinder in the z-direction */    
    for(i=0;i<simparms->ncg;i++){
      if(iaacg[i]==2){
	faacg[i] = 1.;
	dfaacg[i] = 0.;
	drnow[i] = 0.;
      }else if(iaacg[i]==0){
	faacg[i] = 0.;
	dfaacg[i] = 0.;
	drnow[i] = 0.;
      }else{
	/* compute switchfunction */
	/* loop over units that determine aaregion and find closest one */
	dr2min=999999.;
	for(j=0;j<simparms->naaregion;j++){
	  ii=simparms->iaaregion[j];
	  pcgx=coords->px[ii]; pcgy=coords->py[ii]; pcgz=coords->pz[ii];
	  dx=coords->px[i]-pcgx;
	  dy=coords->py[i]-pcgy;
	  period(1,&dx,&dy,&dx,coords->hmat,coords->hmati,iperd,ivol); /* could be vectorized here */
	  dr2 = dx*dx + dy*dy;
	  dr2min = MIN(dr2min,dr2);
	}
	dr2min = sqrt(dr2min);
	if(dr2min < r1_hr){			/* inner skin */
	  faacg[i] = 0.;
	  dfaacg[i] = 0.;
	  drnow[i] = r1_hr;
	}else if(dr2min > r2_hr){	/* outher skin */
	  faacg[i] = 1.;
	  dfaacg[i] = 0.;
	  drnow[i] = r2_hr;
	}else{							/* actual healing region */
	  if(faacg[i] == -1) ierr++;
#ifdef SWITCH1
	  dum = dr2min-r1_hr;
	  faacg[i] = dum*dum*(3.0*r2_hr - r1_hr - dr2min*2.)/z;
	  dfaacg[i] = 6.*dum*(r2_hr-dr2min)/z;
#elif defined(SWITCH2)
	  dum = dr2min - r2_hr;
	  dum4 = dum*dum; 
	  dum4 *= dum4;
	  dfaacg[i] = -20*(dum4 + z*dum*dum*dum)/z5;
	  faacg[i] = 1-(4*dum4*dum + 5*z*dum4)/z5;
#elif defined(SWITCH3)
	  dum = dr2min - r2_hr;
	  dum4 = dum*dum; 
	  dum4 *= dum4;
	  piece = 4*dum4*dum + 5*z*dum4;
	  dfaacg[i] = -40*piece*(dum4 + z*dum*dum*dum)/z10;
	  faacg[i] = 1-(piece*piece)/z10;
#endif
	  drnow[i] = dr2min;
	}
      }
    }
    break;
  default:
    md_error("Unknown shape type in set_switchfunction");
  }
  
  if(ierr>0){
    sprintf(line,"  %d particles jumped over the HR skin, increase hr_skin (hr_skin=%g)!",
	    ierr,simparms->rskin_hr);
    md_warning(line);
  }
  
  /* if HybridSwitch multipl with current switch value */
  if(simparms->iHybridSwitch){
    z = simparms->switch_fmin + simparms->switch_df*simparms->istep;
    for(i=0;i<simparms->ncg;i++){
      if(iaacg[i] != 2){
        faacg[i] *= z;
        dfaacg[i] *= z;
      }
    }
  }
  
  /* set faacg for AA atoms */
  tnat_cg = tnat_aa = 0.;
  faa = faacg + simparms->ncg;
  dfaa = dfaacg + simparms->ncg;
  ptr = drnow + simparms->ncg;
  for(i=0;i<simparms->ncg;i++){
    tnat_cg += faacg[i];
    z = 1.0 - faacg[i];
    iaa = simparms->icglist[i];
    for(j=0;j<simparms->napcg[i];j++){
      ii=iaa[j];
      faa[ii]=z;
      dfaa[ii]=-1.*dfaacg[i];
      ptr[ii]=drnow[i];
      tnat_aa += z;
    }
  }
  
  ptr = simparms->ddr;
  for(i=0;i<simparms->natoms;i++){
    ptr[i] -= drnow[i];
    simparms->ddr_lrf[i] += ptr[i];
    simparms->ddr_intra[i] += ptr[i];
    simparms->ddr_tors[i] += ptr[i];
  }
  
  simparms->fnatoms[0] = tnat_cg;	/* NO LONGER USED...REMOVE FRACTIONAL DOFs */
  simparms->fnatoms[1] = tnat_aa;
  /* printf("Number of fractional atoms is CG=%lf, AA= %lf\n",tnat_cg,tnat_aa); */
}


/*------------------------------------------------------------------ */
/*  separate CG units over AA (iaacg[i]=0), HR (iaacg[i]=1) and      */
/*  CG (iaacg[i]=2) regions                                          */
/*  - returns 1 when atoms changed region and 0 otherwise            */
int get_healingregion(SIMPARMS *simparms,COORDS *coords,SO3 *so3,int ifirst)
{

  int i,j,ii,naa,nhr,iperd,ivol;
  int *iaacg,*iprev;
  double pcgx,pcgy,pcgz,dx,dy,dz,dr2,xmin2,xmax2,*pos;
  
  iperd = simparms->iperd;
  ivol = simparms->ivol;
  naa = nhr = 0;
  iaacg = simparms->iaacg;
  iprev = simparms->iaacg_b4;
  for(i=0;i<simparms->ncg;i++) iaacg[i]=2; /* active (CG) region */
  for(i=0;i<simparms->naaregion;i++) iaacg[simparms->iaaregion[i]]=0;

  switch(simparms->itype_hr){  
  case -1:   /* no healing region */
  break;
  case 0:   /* healing region is sphere shape */
  xmin2 = MAX(0.,simparms->rmin_hr-simparms->rskin_hr);
  xmin2*=xmin2;
  xmax2 = simparms->rmax_hr+simparms->rskin_hr;
  xmax2*=xmax2;
  for(i=0;i<simparms->naaregion;i++){
    ii=simparms->iaaregion[i];
    pcgx=coords->px[ii]; pcgy=coords->py[ii]; pcgz=coords->pz[ii];
    for(j=0;j<simparms->ncg;j++){
      if(iaacg[j]==0) continue;
      dx=coords->px[j]-pcgx; dy=coords->py[j]-pcgy; dz=coords->pz[j]-pcgz;
      period(1,&dx,&dy,&dz,coords->hmat,coords->hmati,iperd,ivol); /* could be vectorized here */
      dr2 = dx*dx+dy*dy+dz*dz;
      if(dr2 < xmax2){
	iaacg[j]=1;        /* healing region */
	nhr++;
	if(dr2 < xmin2){
	  iaacg[j]=0;      /* AA region */
	  naa++;
	}
      }
    }
  }	
  nhr-=naa;
  naa+=simparms->naaregion;
  break;
  case 1:   /* healing region is yz-slice shape */
  case 2:   /* healing region is xz-slice shape */
  case 3:   /* healing region is yz-slice shape */
    pos = coords->px+simparms->ncg*(simparms->itype_hr-1);
    xmin2 = MAX(0.,simparms->rmin_hr-simparms->rskin_hr);
    xmin2*=xmin2;
    xmax2 = simparms->rmax_hr+simparms->rskin_hr;
    xmax2*=xmax2;
    for(i=0;i<simparms->naaregion;i++){
      ii=simparms->iaaregion[i];
      pcgx=pos[ii];
      for(j=0;j<simparms->ncg;j++){
	if(iaacg[j]==0) continue;
	dx=pos[j]-pcgx;
	period(1,&dx,&dx,&dx,coords->hmat,coords->hmati,iperd,ivol); /* could be vectorized here */
	dx*=dx;
	if(dx < xmax2) {  /* healing region */
	  iaacg[j]=1;
	  nhr++;
	  if(dx < xmin2){
	    iaacg[j]=0;      /* AA region */
	    naa++;
	  }
	}
      }
    }
    nhr-=naa;
    naa+=simparms->naaregion;
    break;
  case 4:   /* selection has some hybrid value, the rest of the system is CG */
    for(i=0;i<simparms->naaregion;i++) iaacg[simparms->iaaregion[i]]=1;
    naa = 0;
    nhr = simparms->naaregion;
    break;
  case 5:   /* healing region is a cylinder in the x-direction */
  case 6:   /* healing region is a cylinder in the y-direction */
  case 7:   /* healing region is a cylinder in the z-direction */
    xmin2 = MAX(0.,simparms->rmin_hr-simparms->rskin_hr);
    xmin2*=xmin2;
    xmax2 = simparms->rmax_hr+simparms->rskin_hr;
    xmax2*=xmax2;
    for(i=0;i<simparms->naaregion;i++){
      ii=simparms->iaaregion[i];
      pcgx=coords->px[ii];
      pcgy=coords->py[ii];
      for(j=0;j<simparms->ncg;j++){
	if(iaacg[j]==0) continue;
	dx= coords->px[j]-pcgx;
	dy= coords->py[j]-pcgy;
	period(1,&dx,&dy,&dx,coords->hmat,coords->hmati,iperd,ivol); /* could be vectorized here */
	dr2 = dx*dx + dy*dy;
	if(dr2 < xmax2) {  /* healing region */
	  iaacg[j]=1;
	  nhr++;
	  if(dr2 < xmin2){
	    iaacg[j]=0;      /* AA region */
	    naa++;
	  }
	}
      }
    }
    nhr-=naa;
    naa+=simparms->naaregion;
    break;
  default:
    md_error("Unknown shape type in get_healingregion");
  }
  
  if(ifirst==1){
    for(i=0;i<simparms->ncg;i++) iprev[i]=iaacg[i];
    return 1;
  }else{
    for(i=0;i<simparms->ncg;i++){
      if(iprev[i] != iaacg[i]) {
        printf("Change in number of CG sites in regions of CG=%d HR=%d AA=%d\n",simparms->ncg-naa-nhr,nhr,naa);
        return 1;
      }
    }
  }
  return 0;
}

/* Transform particles that have changed region                      */
/*------------------------------------------------------------------ */
void transform_particle(SIMPARMS *simparms,COORDS *coords,SO3 *so3){
  int i,k,kk,iperd,ivol;
  int *iaacg,*iprev;
	double v,ve;
  LINE line;  
  
  iperd = simparms->iperd;
  ivol = simparms->ivol;
  iaacg = simparms->iaacg;
  iprev = simparms->iaacg_b4;

  for(i=0;i<simparms->ncg;i++){
    if(iprev[i]==iaacg[i]){
      continue;
    }else if((iprev[i]==0) && (iaacg[i]==1)){ /* AA->HR */
      get_cgpos(coords,simparms->icglist,simparms->napcg,1,&i);  /* are these 2 necessary? */
      get_cgvel(coords,simparms->icglist,simparms->napcg,1,&i);
      simparms->faacg[i]=-1.;
    }else if((iprev[i]==2) && (iaacg[i]==1)){ /* CG->HR */
      get_abs_aapos(coords,simparms->icglist,simparms->napcg,&i,1);
      get_abs_aavel(coords,simparms->icglist,simparms->napcg,&i,1);
    /* idea below to reset AA vel to zero in CGR does not actually work */
//      zero_aavel(coords,simparms->icglist,simparms->napcg,&i,1); 
      simparms->p2mbuf -= get_inherent_p2m(simparms->napcg[i],simparms->icglist[i],
      coords->next->vx,coords->next->vy,coords->next->vz,coords->next->amass);
      simparms->p2mbuf += get_so3_ekin(so3->omega[i],so3->mit[i]);
      for(k=0;k<simparms->napcg[i];k++){
        kk=simparms->icglist[i][k];
        coords->next->fxa[kk]=0;
        coords->next->fya[kk]=0;
        coords->next->fza[kk]=0;
        coords->next->fxr[kk]=0;
        coords->next->fyr[kk]=0;
        coords->next->fzr[kk]=0;
      }
      getvbond_intern_CG(simparms,coords->next,i);
      getvbend_intern_CG(simparms,coords->next,i);
      getvtors_intern_CG(simparms,coords->next,i);
      getvonfo_intern_CG(simparms,coords->next,&v,&ve,i);
      getvbondx_intern_CG(simparms,coords->next,i);
    
      fbond_so3hr_CGR2HR(simparms,coords->next,i); /* recompute this end of the so3 interaction */
//      fnonbond_so3hr_CGR2HR(simparms,coords->next,i); /* recompute this end of the so3 interaction */

      for(k=0;k<simparms->napcg[i];k++){
        kk=simparms->icglist[i][k];
        coords->next->fxa[kk] /= coords->next->amass[kk];
        coords->next->fya[kk] /= coords->next->amass[kk];
        coords->next->fza[kk] /= coords->next->amass[kk];
        coords->next->fxr[kk] /= coords->next->amass[kk];
        coords->next->fyr[kk] /= coords->next->amass[kk];
        coords->next->fzr[kk] /= coords->next->amass[kk];
      }
      simparms->faacg[i]=-1.;
    }else if((iprev[i]==1) && (iaacg[i]==0)){ /* HR->AA */
    }else if((iprev[i]==1) && (iaacg[i]==2)){ /* HR->CG */
      get_cgvel(coords,simparms->icglist,simparms->napcg,1,&i);	  
      get_relat_aapos(coords,simparms->icglist,simparms->napcg,&i,1);
      simparms->p2mbuf += get_inherent_p2m(simparms->napcg[i],simparms->icglist[i],
      coords->next->vx,coords->next->vy,coords->next->vz,coords->next->amass);
      get_relat_aavel(coords,simparms->icglist,simparms->napcg,&i,1);
      if(simparms->iso3dyn){    /* recompute the angular velocity from the atomic velocities */
        getOmegaInit_CG(simparms,coords->next,so3,i);                         /* this routine inverts MiT */
        getAlphaInit(so3,1,&i);                                               /* this routine also inverts MiT !! */
        simparms->p2mbuf -= get_so3_ekin(so3->omega[i],so3->mit[i]);

        fbond_so3hr_HR2CGR(simparms,coords->next,so3,i); /* recompute this end of the so3 force */
//        fnonbond_so3hr_HR2CGR(simparms,coords->next,so3,i); /* recompute this end of the so3 force */
      }
    }else{
      sprintf(line,"CG no.%d moved from region %d to region %d; increase HR skin?",i,iprev[i],iaacg[i]);
      md_warning(line);
    }
#ifdef DEBUG
    printf("CG particle %d moved from region %d to region %d (AA=0; HR=1; CG=2)\n",i,iprev[i],iaacg[i]);
#endif
    iprev[i] = iaacg[i];
  }
}

/*------------------------------------------------------------------ */
/* select active interactions for the AA/CG regions */
void select_hybridregions(SIMPARMS *simparms, COORDS *coords, INTER *inter, NGBR *ngbr, SO3 *so3, int iprint,int ifirst){

  int natoms, ncg, nbonds, nbends, nbondxs, n14s, ntors, ncharge,ichange,ntherm;
	int nbonds2,nbends2,ntors2,n14s2,nbondxs2;
  char line[MAXLINELEN];
  
  ichange = get_healingregion(simparms,coords,so3,ifirst);
//  set_switchfunction(simparms,coords);
//  if( (ichange) && (!ifirst)) transform_particle(simparms,coords,so3);
  if(ichange==1){
    ncg = coords->natoms = simparms->ncg;
    coords->next->natoms = simparms->naa;
    natoms = select_atoms(simparms->iaacg,&simparms->ndof,coords);
    set_aaselection(simparms->ncg,simparms->natoms,simparms->icgunit,simparms->iaacg);
    natoms += select_atoms(&simparms->iaacg[ncg],&simparms->ndof,coords->next);
    if(ifirst) select_ewald(coords);
    nbonds = select_bonds(simparms->iaacg,&coords->bonds,1);
    nbonds += select_bonds(&simparms->iaacg[ncg],&coords->next->bonds,0);
    nbends = select_bends(simparms->iaacg,&coords->bends,1);     
    nbends += select_bends(&simparms->iaacg[ncg],&coords->next->bends,0);     
    nbondxs = select_bondxs(simparms->iaacg,&coords->bondxs,1);     
    nbondxs += select_bondxs(&simparms->iaacg[ncg],&coords->next->bondxs,0);     
    n14s = select_onfos(simparms->iaacg,&coords->onfo,1);  
    n14s += select_onfos(&simparms->iaacg[ncg],&coords->next->onfo,0);  
    ntors = select_tors(simparms->iaacg,&coords->tors,1);
    ntors += select_tors(&simparms->iaacg[ncg],&coords->next->tors,0);
    ncharge = select_charges(simparms->iaacg,&coords->coul,coords->qch,coords->iperd);
    ncharge += select_charges(&simparms->iaacg[ncg],&coords->next->coul,
		coords->next->qch,coords->next->iperd);
		ntherm = select_therm(coords);
		ntherm += select_therm(coords->next);
		update_vlrc(inter,coords->natoms,coords->iselect,
								simparms->ncgtypes,simparms->vlrc,simparms->wlrc,coords->iperd);  
		update_vlrc(inter->next,coords->next->natoms,coords->next->iselect,
								simparms->ntypes-simparms->ncgtypes,&simparms->vlrc[1],&simparms->wlrc[1],coords->next->iperd);
    nbonds2 = select_bonds(&simparms->iaacg[ncg],&coords->next->bonds,2);
    nbends2 = select_bends(&simparms->iaacg[ncg],&coords->next->bends,2);     
    nbondxs2 = select_bondxs(&simparms->iaacg[ncg],&coords->next->bondxs,2);     
    n14s2 = select_onfos(&simparms->iaacg[ncg],&coords->next->onfo,2);  
    ntors2 = select_tors(&simparms->iaacg[ncg],&coords->next->tors,2);
  }

  if( (ichange) && (!ifirst)) transform_particle(simparms,coords,so3);
  set_switchfunction(simparms,coords);

  if((iprint==1) && (simparms->rank==0)){
	md_stdout("Hybrid CG/AA MD interaction selection:");
	md_stdout("           Active  CG      AA    CG+AA    SO3 | Total CG     AA    CG+AA    SO3");
	sprintf(line," Atoms      : %7d %7d %7d       - | %7d %7d %7d       -",
	  coords->nselect,coords->next->nselect,natoms,
	  simparms->ncg,simparms->naa,simparms->natoms);
	md_stdout(line);
	sprintf(line," Bonds      : %7d %7d %7d %7d | %7d %7d %7d %7d", 
	  coords->bonds.nselect,coords->next->bonds.nselect,nbonds,nbonds2,
	  coords->bonds.nbonds,coords->next->bonds.nbonds,
	  coords->bonds.nbonds+coords->next->bonds.nbonds,
					coords->next->bonds.nbonds2);
	md_stdout(line);
	sprintf(line," Bends      : %7d %7d %7d %7d | %7d %7d %7d %7d", 
	  coords->bends.nselect,coords->next->bends.nselect,nbends,nbends2,
	  coords->bends.nbends,coords->next->bends.nbends,
	  coords->bends.nbends+coords->next->bends.nbends,
					coords->next->bends.nbends2);
	md_stdout(line);
	sprintf(line," Cross-bonds: %7d %7d %7d %7d | %7d %7d %7d %7d", 
	  coords->bondxs.nselect,coords->next->bondxs.nselect,nbondxs,nbondxs2,
	  coords->bondxs.nbondxs,coords->next->bondxs.nbondxs,
	  coords->bondxs.nbondxs+coords->next->bondxs.nbondxs,
					coords->next->bondxs.nbondxs2);
	md_stdout(line);
	sprintf(line," One-fours  : %7d %7d %7d %7d | %7d %7d %7d %7d", 
	  coords->onfo.nselect,coords->next->onfo.nselect, n14s,n14s2,
	  coords->onfo.n14s,coords->next->onfo.n14s,
	  coords->onfo.n14s+coords->next->onfo.n14s,
					coords->next->onfo.n14s2);
	md_stdout(line);
	sprintf(line," Torsions   : %7d %7d %7d %7d | %7d %7d %7d %7d", 
	  coords->tors.nselect,coords->next->tors.nselect,ntors,ntors2,
	  coords->tors.ntorss,coords->next->tors.ntorss,
	  coords->tors.ntorss+coords->next->tors.ntorss,
					coords->next->tors.ntorss2);
	md_stdout(line);
	sprintf(line," Charges    : %7d %7d %7d       - | %7d %7d %7d       -", 
	coords->coul.nselect,coords->next->coul.nselect,ncharge,
	coords->coul.ncharge,coords->next->coul.ncharge,
	coords->coul.ncharge+coords->next->coul.ncharge);
	md_stdout(line);
	sprintf(line," Thermostats: %7d %7d %7d       - | %7d %7d %7d       -", 
	coords->ntselect,coords->next->ntselect,ntherm,
	coords->ntherm,coords->next->ntherm,
	coords->ntherm+coords->next->ntherm);
	md_stdout(line);
/*	sprintf(line,"  %d (CG=%d/AA=%d) out of %d (CG=%d/AA=%d) elec corrections",necorr,	*/
/*	coords->coul.nselect_ecorr,coords->next->coul.nselect_ecorr,						*/
/*	coords->coul.necorr+coords->next->coul.necorr,coords->coul.necorr,					*/
/*	coords->next->coul.necorr);															*/
/*	md_stdout(line);																	*/
	if(coords->iperd) md_stdout("  Ewald summation done at the coarse grain level");
	if(coords->next->iperd) md_stdout(" Ewald summation done at the all atom level");
	sprintf(line,"Updated CG long range correction factor V = %g K A^3, W = %g K A^3\nUpdated AA long range correction factor V = %g K A^3, W = %g K A^3",
	    simparms->vlrc[0],simparms->wlrc[0],simparms->vlrc[1],simparms->wlrc[1]);
    md_stdout(line);
  }

}

