/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

   /* compute positions of the coarse grain sites */

#include "md.h"  

#define DEBUG

/*------------------------------------------------------------------*/
/* fill out iaacg for AA part using CG part                   */
void set_aaselection(int ncg, int natoms, int *icgunit, int *iaacg){
  int iaa, naa=natoms-ncg;

  for(iaa=0;iaa<naa;++iaa) 
	iaacg[ncg+iaa] = 2 - iaacg[icgunit[iaa]];
}

/*-------------------------------------------------------------------*/
/* Compute centers of mass of AA atoms and compare with CG positions */
/* and adjust AA positions accordingly                               */
void check_aa_com(double *pcgx,double *paax,double *aamass,int **icglist, int *napcg, int naa, int ncg){

  int i,iat,icg;
  double px,py,pz,dx,dy,dz,dxmax,dymax,dzmax;;
  double *pcgy,*pcgz,*paay,*paaz;
  double tmass;
  LINE line;

  pcgy = pcgx+ncg;  pcgz = pcgy+ncg;
  paay = paax+naa;  paaz = paay+naa;
  dxmax = dymax = dzmax = 0.0;
  
  printf("Comparing AA centers of mass to CG positions...\n");
  
  for(icg=0;icg<ncg;icg++){
	tmass=0.0;
	px = py = pz = 0.0;
    for(i=0;i<napcg[icg];i++){
      iat = icglist[icg][i];
	  px += aamass[iat]*paax[iat ];
	  py += aamass[iat]*paay[iat];
	  pz += aamass[iat]*paaz[iat];
	  tmass += aamass[iat];
	}
	px /= tmass;
	py /= tmass;
	pz /= tmass;
	dx = pcgx[icg] - px;
	dy = pcgy[icg] - py;
	dz = pcgz[icg] - pz;
    for(i=0;i<napcg[icg];i++){
      iat = icglist[icg][i];
	  paax[iat] += dx;
	  paay[iat] += dy;
	  paaz[iat] += dz;
	}
/*	printf("%d | %lf %lf %lf | %lf %lf %lf |  %lf %lf %lf \n",
		icg,px,pcgx[icg],dx,py,pcgy[icg],dy,pz,pcgz[icg],dz); */
    dx = fabs(dx); dy = fabs(dy); dz = fabs(dz);
	if(dx>dxmax) dxmax = dx; 
	if(dy>dymax) dymax = dy; 
	if(dz>dzmax) dzmax = dz; 
  }
  if((dx>0.0)||(dy>0.0)||(dz>0)){
    sprintf(line,"Translating AA positions to match c.o.m. with CG positions\nLargest AA displacements in x, y and z: %lf %lf %lf\n",dxmax,dymax,dzmax);
	md_warning(line);
	printf("%s\n",line);
  }
}


/*------------------------------------------------------------------*/
/* Compute non-active CG atom position as c.o.m. of their AA atoms  */
void get_cgcom(SIMPARMS *simparms,COORDS *caa,COORDS *ccg)
{

  int i,j,ii,ncg;
  double *amass = caa->amass;
  double *px,*py,*pz,tmass;
  double *pcgx,*pcgy,*pcgz;

  px = caa->px;  py = caa->py;  pz = caa->pz;
  pcgx = ccg->px;  pcgy = ccg->py;  pcgz = ccg->pz;
  ncg = simparms->ncg;

  for(i=0;i<ncg;i++){
    if(simparms->iaacg[i]==0){
      tmass=0.0;
      pcgx[i] = pcgy[i] = pcgz[i] = 0.0;
      for(j=0;j<simparms->napcg[i];j++){
        ii=simparms->icglist[i][j];
        pcgx[i] += amass[ii]*px[ii];
        pcgy[i] += amass[ii]*py[ii];
        pcgz[i] += amass[ii]*pz[ii];
        tmass += amass[ii];
      }
	  /* if CG mass is zero set it to the sum of AA masses */
      if(abs(ccg->amass[i]-0.1*MCONV) < 0.00001 ) ccg->amass[i]=tmass;
      pcgx[i] /= tmass;
      pcgy[i] /= tmass;
      pcgz[i] /= tmass;
	}
  } 
}

/*------------------------------------------------------------------*/
/* Compute non-active CG atom position as c.o.m. of their AA atoms  */
void get_cgpos(COORDS *coords,int **icglist, int *napcg, int nselect, int *iselect){

  int i,j,iat,icg;
  double *pcgx,*pcgy,*pcgz, *paax,*paay,*paaz;
  double tmass, *amass=coords->next->amass;

  pcgx = coords->px; pcgy = coords->py;  pcgz = coords->pz;
  paax = coords->next->px; paay = coords->next->py;  paaz = coords->next->pz;
  
  for(i=0;i<nselect;i++){
	icg = iselect[i];
	tmass=0.0;
	pcgx[icg] = pcgy[icg] = pcgz[icg] = 0.0;
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
	  pcgx[icg] += amass[iat]*paax[iat];
	  pcgy[icg] += amass[iat]*paay[iat];
	  pcgz[icg] += amass[iat]*paaz[iat];
	  tmass += amass[iat];
	}
	pcgx[icg] /= tmass;
	pcgy[icg] /= tmass;
	pcgz[icg] /= tmass;
  }
}


/*------------------------------------------------------------------ */
/* Compute non-active CG atom velocities as c.o.m. of their AA atoms */
/* void get_cgvel(SIMPARMS *simparms,COORDS *caa, COORDS *ccg) */
void get_cgvel(COORDS *coords,int **icglist, int *napcg, int nselect, int *iselect){
{

  int i,j,iat,icg;
  double *amass = coords->next->amass;
  double *vaax,*vaay,*vaaz,tmass;
  double *vcgx,*vcgy,*vcgz;

  vaax = coords->next->vx;  vaay = coords->next->vy;  vaaz = coords->next->vz;
  vcgx = coords->vx;  vcgy = coords->vy;  vcgz = coords->vz;

  for(i=0;i<nselect;i++){
	icg = iselect[i];
	tmass=0.0;
	vcgx[icg] = vcgy[icg] = vcgz[icg] = 0.0;
      for(j=0;j<napcg[icg];j++){
        iat=icglist[icg][j];
        vcgx[icg] += amass[iat]*vaax[iat];
        vcgy[icg] += amass[iat]*vaay[iat];
        vcgz[icg] += amass[iat]*vaaz[iat];
        tmass += amass[iat];
      }  
      vcgx[icg] /= tmass;
      vcgy[icg] /= tmass;
      vcgz[icg] /= tmass;
	}
  } 
}


/*------------------------------------------------------------------ */
/* subtract CG unit position from its (inactive) AA positions        */
  void get_relat_aapos(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *pcgx,*pcgy,*pcgz, *paax,*paay,*paaz;
  
  pcgx = coords->px; pcgy = coords->py;  pcgz = coords->pz;
  paax = coords->next->px; paay = coords->next->py;  paaz = coords->next->pz;
  
  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      paax[iat] -= pcgx[icg];
      paay[iat] -= pcgy[icg];
      paaz[iat] -= pcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* add selected CG unit position to its (inactive) AA positions      */
  void get_abs_aapos(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *pcgx,*pcgy,*pcgz, *paax,*paay,*paaz;
  
  pcgx = coords->px; pcgy = coords->py;  pcgz = coords->pz;
  paax = coords->next->px; paay = coords->next->py;  paaz = coords->next->pz;

  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      paax[iat] += pcgx[icg];
      paay[iat] += pcgy[icg];
      paaz[iat] += pcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* subtract selected CG unit velocities from its (inactive) AA velocities */
  void get_relat_aavel(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *vcgx,*vcgy,*vcgz, *vaax,*vaay,*vaaz;
  
  vcgx = coords->vx; vcgy = coords->vy;  vcgz = coords->vz;
  vaax = coords->next->vx; vaay = coords->next->vy;  vaaz = coords->next->vz;
  
  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      vaax[iat] -= vcgx[icg];
      vaay[iat] -= vcgy[icg];
      vaaz[iat] -= vcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* add selected CG unit velocities to its (inactive) AA velocities   */
  void get_abs_aavel(COORDS *coords,int **icglist, int *napcg,int *iselect,int nselect){
  
  int i,j,iat,icg;
  double *vcgx,*vcgy,*vcgz, *vaax,*vaay,*vaaz;
  
  vcgx = coords->vx; vcgy = coords->vy;  vcgz = coords->vz;
  vaax = coords->next->vx; vaay = coords->next->vy;  vaaz = coords->next->vz;

  for(i=0;i<nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      vaax[iat] += vcgx[icg];
      vaay[iat] += vcgy[icg];
      vaaz[iat] += vcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* set AA velocities to CG velocities                                */
  void distribute_cgvel_over_aa(COORDS *coords,int **icglist, int *napcg){
  
  int i,j,iat,icg;
  int *iselect;
  double *vcgx,*vcgy,*vcgz, *vaax,*vaay,*vaaz;
  
  iselect = coords->iselect;
  vcgx = coords->vx; vcgy = coords->vy;  vcgz = coords->vz;
  vaax = coords->next->vx; vaay = coords->next->vy;  vaaz = coords->next->vz;

  for(i=0;i<coords->nselect;i++){
	icg = iselect[i];
    for(j=0;j<napcg[icg];j++){
      iat = icglist[icg][j];
      vaax[iat] = vcgx[icg];
      vaay[iat] = vcgy[icg];
      vaaz[iat] = vcgz[icg];
	}
  }
} 
/*------------------------------------------------------------------ */
/* compute AA velocities from CG accelerations for iselect AA units     */
  void distribute_cgaccel(double *vaax,double dt,int ncg,int naa,int *icgunit,int *iselect,
							int nselect,double *acgx){
  
  int i,iaa,icg;
  double *acgy,*acgz,*vaay,*vaaz;
  
  acgy = acgx + ncg;  acgz = acgy + ncg;  
  vaay = vaax + naa;  vaaz = vaay + naa;  
  
  for(i=0;i<nselect;i++){
	iaa = iselect[i];
	icg = icgunit[iaa];
	vaax[iaa] += dt*acgx[icg];
	vaay[iaa] += dt*acgy[icg];
	vaaz[iaa] += dt*acgz[icg];
  }
} 

/*------------------------------------------------------------------ */
/* set AA accelerations to CG accelerations                          */
  void distribute_cgaccel_over_aa(COORDS *coords,int **icglist,int *napcg,double *acgx,double *aaax){
						
  int i,j,iat,icg;  
  int *iselect;  
  double *acgy,*acgz,*aaay,*aaaz; 
									
  iselect = coords->iselect;      
  acgy = acgx + coords->natoms;    acgz = acgy + coords->natoms;          
  aaay = aaax + coords->next->natoms;  aaaz = aaay + coords->next->natoms;   
									
  for(i=0;i<coords->nselect;i++){  
	icg = iselect[i];             
    for(j=0;j<napcg[icg];j++){  
      iat = icglist[icg][j];  
      aaax[iat] = acgx[icg];  
      aaay[iat] = acgy[icg];  
      aaaz[iat] = acgz[icg];  
	}  
  }  
 }   

/*------------------------------------------------------------------ */
/*  select atoms that are in an active AA/CG region                  */
int select_atoms(int *iaacg, int *ndof, COORDS *coords)
{

  int i; 
  int nsel,nunsel;
  int *iselect=coords->iselect;
  
  nsel=0;
  nunsel=coords->natoms-1;
  for(i=0;i<coords->natoms;i++){
    if(iaacg[i]==2){
	  iselect[nsel++]=i;
	}else{
	  iselect[nunsel--]=i;
	}
  }
  coords->inactive = coords->iselect+nsel;
	  
  *ndof -= 3*(coords->nselect - nsel);
  coords->nselect = nsel;
  coords->ninactive = coords->natoms - nsel;
  
  return nsel;
}

/*------------------------------------------------------------------ */
/*  swap atom from inactive region to active AA/CG region            */
void activate_atom(COORDS *coords, int iatom, int ifind)
{
  int idummy, iat;
  int *inactive=coords->inactive;
  
  if(ifind<0){
    for(iat=0;iat<coords->ninactive;iat++)
      if(inactive[iat]==iatom) break;
  }else{
    iat = ifind;
  }
  
  idummy = inactive[0];
  inactive[0] = iatom;
  inactive[iat] = idummy;
  coords->nselect++;
  coords->ninactive--;
  inactive++;
}

/*------------------------------------------------------------------ */
/*  swap atom from active region to inactive AA/CG region            */
void deactivate_atom(COORDS *coords, int iatom, int ifind)
{
  int idummy, iat;
  int *iselect=coords->iselect;
  int natoms = coords->natoms;
  
  if(ifind<0){
    for(iat=0;iat<coords->nselect;iat++)
      if(iselect[iat]==iatom) break;
  }else{
    iat = ifind;
  }
  
  idummy = iselect[natoms];
  iselect[natoms] = iatom;
  iselect[iat] = idummy;
  coords->nselect--;
  coords->ninactive++;
  coords->inactive--;
}

/*------------------------------------------------------------------ */
/*  select atoms that are in an active AA/CG region                  */
void distribute_force_over_aa(COORDS *coords,int **icglist,int *napcg,double *fcgx,double *faax)
{
  int i,j,iat,icg;
  int nselect=coords->nselect;
  int *iselect=coords->iselect;
  double fact;
  double *fcgy, *fcgz, *faay, *faaz;
/*  double *px=coords->next->px, *py=coords->next->py, *pz=coords->next->pz; */
  double *aamass=coords->next->amass,  *cgmass=coords->amass;

  fcgy = fcgx + coords->natoms; fcgz = fcgy + coords->natoms;             
  faay = faax + coords->next->natoms; faaz = faay + coords->next->natoms; 
  
  for(i=0;i<nselect;i++){
	icg = iselect[i];
	for(j=0;j<napcg[icg];j++){
	  iat = icglist[icg][j];
	  fact = aamass[iat]/cgmass[icg];
	  faax[iat] = fact*fcgx[icg]; 
	  faay[iat] = fact*fcgy[icg]; 
	  faaz[iat] = fact*fcgz[icg]; 
    }
  }
}

/*------------------------------------------------------------------ */
/*  separate CG units over AA (iaacg[i]=0), HR (iaacg[i]=1) and      */
/*  CG (iaacg[i]=2) regions                                          */
/*  - returns 1 when atoms changed healing region and 0 otherwise    */
int get_healingregion(SIMPARMS *simparms,COORDS *coords,int ifirst)
{

  int i,j,ii,naa,nhr,iperd,ivol;
  int *iaacg,*iprev, ichange;
  double pcgx,pcgy,pcgz,dx,dy,dz,dr2,zmax2,zmin2,ymin2,ymax2,xmin2,xmax2;
  LINE line;  
  
  ichange=0;
  iperd = simparms->iperd;
  ivol = simparms->ivol;
  naa = nhr = 0;
  iaacg = simparms->iaacg;
  iprev = simparms->iaacg_b4;
  for(i=0;i<simparms->ncg;i++) iaacg[i]=2; /* active (CG) region */
  for(i=0;i<simparms->naaregion;i++) iaacg[simparms->iaaregion[i]]=0;

  switch(simparms->itype_hr){  
  case 1000:   /* healing region is sphere shape */
  xmin2 = simparms->rmin_hr[0]-simparms->rskin_hr;
  xmin2*=xmin2;
  xmax2 = simparms->rmax_hr[0]+simparms->rskin_hr;
  xmax2*=xmax2;
  for(i=0;i<simparms->naaregion;i++){
    ii=simparms->iaaregion[i];
    pcgx=coords->px[ii]; pcgy=coords->py[ii]; pcgz=coords->pz[ii];
    for(j=0;j<simparms->ncg;j++){
	  if(iaacg[j]==0) continue;
	  dx=coords->px[j]-pcgx; dy=coords->py[j]-pcgy; dz=coords->pz[j]-pcgz;
	  period(1,&dx,&dy,&dz,coords->hmat,coords->hmati,iperd,ivol); /* could be vectorized here */
	  dr2 = dx*dx+dy*dy+dz*dz;
	  if(dr2 < xmax2){
	    iaacg[j]=1;        /* healing region */
		nhr++;
		if(dr2 < xmin2){
		  iaacg[j]=0;      /* AA region */
		  naa++;
		}
	  }
    }
  }	
  break;
  case 10000:   /* healing region is slice shape */
  xmin2 = simparms->rmin_hr[0]-simparms->rskin_hr;
  ymin2 = simparms->rmin_hr[1]-simparms->rskin_hr;
  zmin2 = simparms->rmin_hr[2]-simparms->rskin_hr;
  xmin2*=xmin2; ymin2*=ymin2; zmin2*=zmin2;
  xmax2 = simparms->rmax_hr[0]+simparms->rskin_hr;
  ymax2 = simparms->rmax_hr[1]+simparms->rskin_hr;
  zmax2 = simparms->rmax_hr[2]+simparms->rskin_hr;
  xmax2*=xmax2; ymax2*=ymax2; zmax2*=zmax2;
  for(i=0;i<simparms->naaregion;i++){
    ii=simparms->iaaregion[i];
    pcgx=coords->px[ii]; pcgy=coords->py[ii]; pcgz=coords->pz[ii];
    for(j=0;j<simparms->ncg;j++){
	  if(iaacg[j]==0) continue;
      dx=coords->px[j]-pcgx; dy=coords->py[j]-pcgy; dz=coords->pz[j]-pcgz;
	  period(1,&dx,&dy,&dz,coords->hmat,coords->hmati,iperd,ivol); /* could be vectorized here */
	  dx*=dx; dy*=dy; dz*=dz;
	  if((dx < xmax2) || (dy < ymax2) || (dz < zmax2)) {  /* healing region */
		iaacg[j]=1;
		nhr++;
		if((dx < xmin2) || (dy < ymin2) || (dz < zmin2)){
		  iaacg[j]=0;      /* AA region */
		  naa++;
		}
	  }
    }
  }
  break;
  default:
    if(ifirst==3){
	  printf("DEBUG INFO get_HEALINGREGION; FIX LATER\n");
 	  iaacg[1]=0;
	  naa++; nhr++;
	}
/*    md_error("Unknown shape type in get_healingregion"); */
  }
  nhr-=naa;
  naa+=simparms->naaregion;
  
  /* Synchronize AA and CG pos and vel for particles that changed region */
  if(ifirst==0){
    for(i=0;i<simparms->ncg;i++){
      if(iprev[i]==iaacg[i]){
	    continue;
	  }else if((iprev[i]==0) && (iaacg[i]==1)){ /* AA->HR */
	    get_cgpos(coords,simparms->icglist,simparms->napcg,1,&i);
	    get_cgvel(coords,simparms->icglist,simparms->napcg,1,&i);
	  }else if((iprev[i]==2) && (iaacg[i]==1)){ /* CG->HR */
	    get_abs_aapos(coords,simparms->icglist,simparms->napcg,&i,1);
	    get_abs_aavel(coords,simparms->icglist,simparms->napcg,&i,1);
	  }else if((iprev[i]==1) && (iaacg[i]==0)){ /* HR->AA */
	  }else if((iprev[i]==1) && (iaacg[i]==2)){ /* HR->CG */
	    get_relat_aapos(coords,simparms->icglist,simparms->napcg,&i,1);
	    get_relat_aavel(coords,simparms->icglist,simparms->napcg,&i,1);
	  }else if((iprev[i]==0) && (iaacg[i]==2)){ /* AA->CG */             /* remove later */
	    get_cgpos(coords,simparms->icglist,simparms->napcg,1,&i);
	    get_cgvel(coords,simparms->icglist,simparms->napcg,1,&i);
	    get_relat_aapos(coords,simparms->icglist,simparms->napcg,&i,1);
	    get_relat_aavel(coords,simparms->icglist,simparms->napcg,&i,1);
	  }else if((iprev[i]==2) && (iaacg[i]==0)){ /* CG->AA */             /* remove later */
	    get_abs_aapos(coords,simparms->icglist,simparms->napcg,&i,1);
	    get_abs_aavel(coords,simparms->icglist,simparms->napcg,&i,1);
	  }else{
	    sprintf(line,"CG no.%d moved from region %d to region %d; increase HR skin?",i,iprev[i],iaacg[i]);
	    md_warning(line);
	  }
#ifdef DEBUG
      printf("CG particle %d moved from region %d to region %d (AA=0; HR=1; CG=2)\n",i,iprev[i],iaacg[i]);
#endif
	  iprev[i] = iaacg[i];
	  ichange = 1;
	}
  }else{
	for(i=0;i<simparms->ncg;i++) iprev[i]=iaacg[i];
  }
  printf(" Number of CG sites in regions of CG=%d HR=%d AA=%d\n",simparms->ncg-naa-nhr,nhr,naa);
  return ichange;
}

/*------------------------------------------------------------------ */
/* select active interactions for the AA/CG regions */
void select_hybridregions(SIMPARMS *simparms, COORDS *coords, INTER *inter, NGBR *ngbr, int iprint,int ifirst){

  int natoms, ncg, nbonds, nbends, nbondxs, n14s, ntors, ncharge,ichange;
  char line[MAXLINELEN];
  
  ichange = get_healingregion(simparms,coords,ifirst);
  ncg = coords->natoms = simparms->ncg;
  coords->next->natoms = simparms->naa;
  natoms = select_atoms(simparms->iaacg,&simparms->ndof,coords);
  set_aaselection(simparms->ncg,simparms->natoms,simparms->icgunit,simparms->iaacg);
  natoms += select_atoms(&simparms->iaacg[ncg],&simparms->ndof,coords->next);
  if(ifirst) select_ewald(coords);
  nbonds = select_bonds(simparms->iaacg,&coords->bonds,1);
  nbonds += select_bonds(&simparms->iaacg[ncg],&coords->next->bonds,0);
  nbends = select_bends(simparms->iaacg,&coords->bends,1);     
  nbends += select_bends(&simparms->iaacg[ncg],&coords->next->bends,0);     
  nbondxs = select_bondxs(simparms->iaacg,&coords->bondxs,1);     
  nbondxs += select_bondxs(&simparms->iaacg[ncg],&coords->next->bondxs,0);     
  n14s = select_onfos(simparms->iaacg,&coords->onfo,1);  
  n14s += select_onfos(&simparms->iaacg[ncg],&coords->next->onfo,0);  
  ntors = select_tors(simparms->iaacg,&coords->tors,1);
  ntors += select_tors(&simparms->iaacg[ncg],&coords->next->tors,0);
  ncharge = select_charges(simparms->iaacg,&coords->coul,coords->qch,coords->iperd);
  ncharge += select_charges(&simparms->iaacg[ncg],&coords->next->coul,
		coords->next->qch,coords->next->iperd);
/*  necorr = select_ecorr(simparms->iaacg,&coords->coul);              */   /* no selection but all or nothing */
/*  necorr += select_ecorr(&simparms->iaacg[ncg],&coords->next->coul); */
	update_vlrc(inter,coords->natoms,coords->iselect,
		simparms->ncgtypes,simparms->vlrc,simparms->wlrc,coords->iperd);  
	update_vlrc(inter->next,coords->next->natoms,coords->next->iselect,
		simparms->ntypes-simparms->ncgtypes,&simparms->vlrc[1],&simparms->wlrc[1],coords->next->iperd); 
		
  if(ichange){
	get_ngbr(simparms,coords,inter,ngbr);
	if(ngbr->next->onelist==0) get_ngbr(simparms,coords->next,inter->next,ngbr->next);
	force(simparms,coords,inter,ngbr,0);   /* VdW + elec long range and short range (case 0) */
    force(simparms,coords,inter,ngbr,2);   /* torsions + one-fours */
    force(simparms,coords,inter,ngbr,3);   /* bonds + bends */
	coords=coords->next; ngbr=ngbr->next; inter=inter->next;
	force(simparms,coords,inter,ngbr,0);   /* VdW + elec long range and short range (case 0) */
    force(simparms,coords,inter,ngbr,2);   /* torsions + one-fours */
    force(simparms,coords,inter,ngbr,3);   /* bonds + bends */
	coords=coords->next; ngbr=ngbr->next; inter=inter->next;
  }
    
  if((iprint==1) && (simparms->rank==0)){
	md_stdout("Hybrid CG/AA MD interaction selection:");
	md_stdout("          Active  CG      AA   CG+AA  | Total CG    AA    CG+AA");
	sprintf(line," Atoms      : %7d %7d %7d | %7d %7d %7d",
	  coords->nselect,coords->next->nselect,natoms,
	  simparms->ncg,simparms->naa,simparms->natoms);
	md_stdout(line);
	sprintf(line," Bonds      : %7d %7d %7d | %7d %7d %7d", 
	  coords->bonds.nselect,coords->next->bonds.nselect,nbonds,
	  coords->bonds.nbonds,coords->next->bonds.nbonds,
	  coords->bonds.nbonds+coords->next->bonds.nbonds);
	md_stdout(line);
	sprintf(line," Bends      : %7d %7d %7d | %7d %7d %7d", 
	  coords->bends.nselect,coords->next->bends.nselect,nbends,
	  coords->bends.nbends,coords->next->bends.nbends,
	  coords->bends.nbends+coords->next->bends.nbends);
	md_stdout(line);
	sprintf(line," Cross-bonds: %7d %7d %7d | %7d %7d %7d", 
	  coords->bondxs.nselect,coords->next->bondxs.nselect,nbondxs,
	  coords->bondxs.nbondxs,coords->next->bondxs.nbondxs,
	  coords->bondxs.nbondxs+coords->next->bondxs.nbondxs);
	md_stdout(line);
	sprintf(line," One-fours  : %7d %7d %7d | %7d %7d %7d", 
	  coords->onfo.nselect,coords->next->onfo.nselect, n14s,
	  coords->onfo.n14s,coords->next->onfo.n14s,
	  coords->onfo.n14s+coords->next->onfo.n14s);
	md_stdout(line);
	sprintf(line," Torsions   : %7d %7d %7d | %7d %7d %7d", 
	  coords->tors.nselect,coords->next->tors.nselect,ntors,
	  coords->tors.ntorss,coords->next->tors.ntorss,
	  coords->tors.ntorss+coords->next->tors.ntorss);
	md_stdout(line);
	sprintf(line," Charges    : %7d %7d %7d | %7d %7d %7d", 
	coords->coul.nselect,coords->next->coul.nselect,ncharge,
	coords->coul.ncharge,coords->next->coul.ncharge,
	coords->coul.ncharge+coords->next->coul.ncharge);
	md_stdout(line);
/*	sprintf(line,"  %d (CG=%d/AA=%d) out of %d (CG=%d/AA=%d) elec corrections",necorr,	*/
/*	coords->coul.nselect_ecorr,coords->next->coul.nselect_ecorr,						*/
/*	coords->coul.necorr+coords->next->coul.necorr,coords->coul.necorr,					*/
/*	coords->next->coul.necorr);															*/
/*	md_stdout(line);																	*/
	if(coords->iperd) md_stdout("  Ewald summation done at the coarse grain level");
	if(coords->next->iperd) md_stdout(" Ewald summation done at the all atom level");
	sprintf(line,"Updated CG long range correction factor V = %g K A^3, W = %g K A^3\nUpdated AA long range correction factor V = %g K A^3, W = %g K A^3",
	    simparms->vlrc[0],simparms->wlrc[0],simparms->vlrc[1],simparms->wlrc[1]);
    md_stdout(line);
  }

}

