/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore
   
   Dr. Preston B. Moore
   Associate Director, Center for Molecular Modeling (CMM)
   University of Pennsylvania, Department of Chemistry, Box 188 
   231 S. 34th St. Philadelphia, PA 19104-6323 USA
   EMAIL: moore@cmm.chem.upenn.edu  
   WWW: http://www.cmm.upenn.edu/~moore
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   
*/

/* compute positions of the coarse grain sites */

#include "md.h"
#define MIN(A,B) ((A<B)?(A):(B))

/* #define DEBUG */

/*------------------------------------------------------------------ */
void getcom(SIMPARMS *simparms,COORDS *caa,double cgcom[3],int icg)
{
  int j,ii;
  double *amass = caa->amass;
  double *px,*py,*pz;
  double tmass=0.0;
  px = caa->px;  py = caa->py;  pz = caa->pz;
  
  for(j=0;j<3;++j){ cgcom[j] = 0.0; }
  for(j=0;j<simparms->napcg[icg];j++){
    ii=simparms->icglist[icg][j];
    cgcom[0] += amass[ii]*px[ii];
    cgcom[1] += amass[ii]*py[ii];
    cgcom[2] += amass[ii]*pz[ii];
    tmass += amass[ii];
  }
  for(j=0;j<3;++j){
    cgcom[j] /= tmass;
  }
}

/* Compute CG Angular Momentum from the cartesian velocties */
/*------------------------------------------------------------------ */
void getLAA(SIMPARMS *simparms,COORDS *caa,SO3 *so3)
{
  
  int i,j,ii,ncg;
  double *amass = caa->amass;
  double *px,*py,*pz;
  double *vx,*vy,*vz,xvel,yvel,zvel;
  double amt,dx,dy,dz;
  double l[3];
	
  px = caa->px;  py = caa->py;  pz = caa->pz;
  vx = caa->vx;  vy = caa->vy;  vz = caa->vz;
  ncg = simparms->ncg;
  
  for(i=0;i<ncg;i++){  /* probably I only need to loop over blobs that change atomistically ?? HR only?*/
    if(simparms->napcg[i] == 1){
      l[0]=l[1]=l[2]=0.0;
    } else {
      
      l[0]=l[1]=l[2]=0.0;
      
      for(j=0;j<simparms->napcg[i];j++){
	ii=simparms->icglist[i][j];
	dx = px[ii];	dy = py[ii];	dz = pz[ii];
	xvel = vx[ii];   yvel = vy[ii];   zvel = vz[ii];
	amt = amass[ii];
	
	l[0] += amt*(dy*zvel-dz*yvel);
	l[1] += amt*(dz*xvel-dx*zvel);
	l[2] += amt*(dx*yvel-dy*xvel);
      }			
      printf("Angular momentum[%d]= %g %g %g\n",i,l[0],l[1],l[2]);
    }
  }
}

/* Compute CG Moment of inertia  */
/*------------------------------------------------------------------ */
void getMitAA(SIMPARMS *simparms,COORDS *caa,SO3 *so3)
{
  
  int i,j,ii,ncg;
  double *amass = caa->amass;
  double *px,*py,*pz;
  double amt,dx,dy,dz;
  double (*mit)[3][3];

  px = caa->px;  py = caa->py;  pz = caa->pz;
  ncg = simparms->ncg;
  mit = so3->mit; 
  
  for(i=0;i<ncg;i++){  
    if(simparms->napcg[i] == 1){
      Imat33(mit[i]);
    } else {
      
      for(j=0;j<9;++j) mit[i][0][j] = 0.0;
      
      for(j=0;j<simparms->napcg[i];j++){
	ii=simparms->icglist[i][j];
	dx = px[ii];	dy = py[ii];	dz = pz[ii];
	amt = amass[ii];
	
	mit[i][0][0] += amt*(dy*dy+dz*dz);
	mit[i][1][1] += amt*(dx*dx+dz*dz);
	mit[i][2][2] += amt*(dx*dx+dy*dy);
	
	mit[i][0][1] -= amt*dx*dy;
	mit[i][0][2] -= amt*dx*dz;
	mit[i][1][2] -= amt*dy*dz;
								
      }
      mit[i][1][0]=mit[i][0][1]; 
      mit[i][2][0]=mit[i][0][2]; 
      mit[i][2][1]=mit[i][1][2];
      
    }
#ifdef DEBUG
    fprintf(stdout,"Mit Init  %d %g %g %g  \n",
            i,so3->mit[i][0][0],so3->mit[i][1][0],so3->mit[i][2][0]);
    fprintf(stdout,"          %d %g %g %g  \n",
            i,so3->mit[i][0][1],so3->mit[i][1][1],so3->mit[i][2][1]);
    fprintf(stdout,"          %d %g %g %g  \n",
            i,so3->mit[i][0][2],so3->mit[i][1][2],so3->mit[i][2][2]);
    fprintf(stdout,"det[Mit[%d]= %g\n",i,mit[i][0][0]*mit[i][1][1]*mit[i][2][2]-
	    mit[i][0][0]*mit[i][1][2]*mit[i][2][1]+mit[i][1][0]*mit[i][2][1]*mit[i][0][2]-
	    mit[i][1][0]*mit[i][0][1]*mit[i][2][2]+mit[i][2][0]*mit[i][0][1]*mit[i][1][2]-
	    mit[i][2][0]*mit[i][0][2]*mit[i][1][1]);
#endif
  }
}
/*------------------------------------------------------------------ */
void getOmegaInit(SIMPARMS *simparms,COORDS *caa, SO3 *so3){

  int i,j,ii,ncg;
  double *amass = caa->amass;
  double *px,*py,*pz;
  double *vx,*vy,*vz;
  double amt,dx,dy,dz;
  double xvel,yvel,zvel;
  double l[3],(*mat)[3][3],matinv[3][3];

  px = caa->px;  py = caa->py;  pz = caa->pz;
  vx = caa->vx;  vy = caa->vy;  vz = caa->vz;
  ncg = simparms->ncg;
  mat=so3->mit;
  
  
  for(i=0;i<ncg;i++){
    if(simparms->napcg[i] == 1){
      for(j=0;j<3;++j) so3->omega[i][j] = 0.0;
    } else {
      
      l[0] = l[1] = l[2] = 0.0;
      
      /*    for(j=0;j<9;++j) mat[0][j] = 0.0; */
      
      for(j=0;j<simparms->napcg[i];j++){
	ii=simparms->icglist[i][j];
	
	amt = amass[ii];	
	dx = px[ii];  dy = py[ii];	dz = pz[ii];
	xvel = vx[ii];   yvel = vy[ii];   zvel = vz[ii];
	
	l[0] += amt*(dy*zvel-dz*yvel);
	l[1] += amt*(dz*xvel-dx*zvel);
	l[2] += amt*(dx*yvel-dy*xvel);
	
	/*	mat[0][0] += amt*(dy*dy+dz*dz);
	  mat[1][1] += amt*(dx*dx+dz*dz);
	  mat[2][2] += amt*(dx*dx+dy*dy);
	  
	  mat[0][1] -= amt*dx*dy;
	  mat[0][2] -= amt*dx*dz;
	  mat[1][2] -= amt*dy*dz; */
	
      }
      /*      mat[1][0]=mat[0][1]; mat[2][0]=mat[0][2]; mat[2][1]=mat[1][2];*/
#ifdef DEBUG
      printf(" L= %g %g %g\n",l[0],l[1],l[2]);
      printf(" mat= %g %g %g\n",mat[i][0][0],mat[i][0][1],mat[i][0][2]);
      printf(" mat= %g %g %g\n",mat[i][1][0],mat[i][1][1],mat[i][1][2]);
      printf(" mat= %g %g %g\n",mat[i][2][0],mat[i][2][1],mat[i][2][2]);
#endif      
      Invert33(mat[i],matinv);
      MultMatVec(matinv,l,so3->omega[i]);
      
#ifdef DEBUG
      fprintf(stdout,"Omega Init  %d %g %g %g  \n",
	      i,so3->omega[i][0],so3->omega[i][1],so3->omega[i][2]);
#endif

    }
  }
}
/*------------------------------------------------------------------ */
void getOmegaInit_CG(SIMPARMS *simparms,COORDS *caa, SO3 *so3,int icg){

  int j,ii;
  double *amass = caa->amass;
  double *px,*py,*pz;
  double *vx,*vy,*vz;
  double amt,dx,dy,dz;
  double xvel,yvel,zvel;
  double l[3],(*mat)[3][3],matinv[3][3];

  px = caa->px;  py = caa->py;  pz = caa->pz;
  vx = caa->vx;  vy = caa->vy;  vz = caa->vz;
  mat=so3->mit;
  
  
  if(simparms->napcg[icg] == 1){
    for(j=0;j<3;++j) so3->omega[icg][j] = 0.0;
  } else {
    l[0] = l[1] = l[2] = 0.0;
    for(j=0;j<simparms->napcg[icg];j++){
      ii=simparms->icglist[icg][j];	
      amt = amass[ii];	
      dx = px[ii];  dy = py[ii];	dz = pz[ii];
      xvel = vx[ii];   yvel = vy[ii];   zvel = vz[ii];
	
      l[0] += amt*(dy*zvel-dz*yvel);
      l[1] += amt*(dz*xvel-dx*zvel);
      l[2] += amt*(dx*yvel-dy*xvel);
    }
    Invert33(mat[icg],matinv);
    MultMatVec(matinv,l,so3->omega[icg]);
  }
}

/*------------------------------------------------------------------ */
void getAlphaInit(SO3 *so3,int nsel,int *isel)
{
  int i,ii;
  double t1[3],t2[3],t3[3];
  double mitinv[3][3];
  
  for(ii=0;ii<nsel;ii++){
    i=isel[ii];
    MultMatVec(so3->mit[i],so3->omega[i],t1);
    CrossProduct(so3->omega[i],t1,t2);
    VecSub(so3->tor[i],t2,t3);
    Invert33(so3->mit[i],mitinv);
    MultMatVec(mitinv,t3,so3->accel[i]);
    
#ifdef DEBUG
    fprintf(stdout,"alpha init mit %d %g %g %g  \n",
            i,so3->mit[i][0][0],so3->mit[i][0][1],so3->mit[i][0][2]);
    fprintf(stdout,"alpha init torque %d %g %g %g  \n",
            i,so3->tor[i][0],so3->tor[i][1],so3->tor[i][2]);
    fprintf(stdout,"alpha init omega %d %g %g %g  \n",
            i,so3->omega[i][0],so3->omega[i][1],so3->omega[i][2]);
    fprintf(stdout,"alpha init  %d %g %g %g  \n",
            i,so3->accel[i][0],so3->accel[i][1],so3->accel[i][2]);
#endif
  }
}

/*------------------------------------------------------------------ */
void getAlphaInc(SO3 *so3,int nsel,int *isel)
{
  int i,ii;
  double mitinv[3][3];
  
  for(ii=0;ii<nsel;ii++){
    i=isel[ii];
    Invert33(so3->mit[i],mitinv);
    MultMatVec(mitinv,so3->tor[i],so3->accel[i]);
  }
}

/*------------------------------------------------------------------ */
void UpdateOmega(SIMPARMS *simparms, SO3 *so3,int nsel,int *isel,double tau)
{
  int ii,i,j;
  double fac = simparms->scaleso3vel;
  
  for(ii=0;ii<nsel;ii++){
    i=isel[ii];
    for(j=0;j<3;++j){
      so3->omega[i][j] += tau*so3->accel[i][j];
    }
  }
  
  /* add friction*/
  for(ii=0;ii<nsel;ii++){
    i=isel[ii];
    for(j=0;j<3;++j){
      so3->omega[i][j] *= fac;
    }
  }
  
}

/*------------------------------------------------------------------ */
void ScaleOmega(double (*omega)[3],int ncg,double fac)
{
  int i,j;
  
  /* apply friction*/
  for(i=0;i<ncg;i++){
    for(j=0;j<3;++j){
      omega[i][j] *= fac;
    }
  }
  
}


/*------------------------------------------------------------------ */
void UpdateSO3Coords(SIMPARMS *simparms, COORDS *caa, SO3 *so3,double tau)
{
  /* compute the rotational update and rotate them */
  
  int i,ii,j,jj,ncg,*iselect;
  double t1[3],t2[3],m1[3][3];
  double *px,*py,*pz;
  
  //	return;
	
  ncg = caa->next->nselect - caa->next->nschizo; /* rotate only CG beads for now */
  iselect = caa->next->iselect;
  /*  tau = so3->tau; */
  px = caa->px;  py = caa->py;  pz = caa->pz;
  
  for(ii=0;ii<ncg;ii++){
    i=iselect[ii];
    
    /* calculate rotation matrix */
    for(j=0;j<3;++j){
      t1[j] = tau*so3->omega[i][j];
    }
    Rodrigues(t1,m1);
    
    for(j=0;j<simparms->napcg[i];j++){
      jj=simparms->icglist[i][j];
      t1[0] = px[jj];
      t1[1] = py[jj];
      t1[2] = pz[jj];
      
      MultMatVec(m1,t1,t2); /* rotate */
      px[jj] = t2[0];
      py[jj] = t2[1];
      pz[jj] = t2[2];
    }
  }
}


/*------------------------------------------------------------------ */
void getTorqueAA(SIMPARMS *simparms,COORDS *caa,SO3 *so3,double *fx,int n)
{
  int j,iat,icg,ncg;
  double *fy,*fz;
  double *px,*py,*pz;
  double rvec[3],fvec[3],t1[3],tortmp[3];
  ncg = simparms->ncg;
  px = caa->px;  py = caa->py;  pz = caa->pz;

  fy = fx+n;  fz = fy+n;
    
  for(icg=0;icg<ncg;icg++){ /* loop over all CG blobs */
    t1[0] = t1[1] = t1[2] = 0;
    
    for(j=0;j<simparms->napcg[icg];j++){
      iat=simparms->icglist[icg][j];
      rvec[0] = px[iat];
      rvec[1] = py[iat];
      rvec[2] = pz[iat];
      fvec[0] = fx[iat];
      fvec[1] = fy[iat];
      fvec[2] = fz[iat];
      CrossProduct(rvec,fvec,tortmp);
      VecAdd(tortmp,t1,t1);
    }
    so3->tor[icg][0] = t1[0]; 
    so3->tor[icg][1] = t1[1];    
    so3->tor[icg][2] = t1[2];
    //    printf("TOTAL %d: %g %g %g\n",icg,so3->tor[icg][0],so3->tor[icg][1],so3->tor[icg][2]);
  }
}

/*------------------------------------------------------------------ */
void DistributeAlphaAA(SIMPARMS *simparms,COORDS *caa,SO3 *so3,double *fx,int nsel,int *isel)
{
  int icg,iat,j,i,*icglist;
  double *px,*py,*pz,*fy,*fz;
  double rvec[3],avec[3];
  
  px = caa->px;  py = caa->py;  pz = caa->pz;
  fy = fx+caa->natoms;  fz = fy+caa->natoms;
  
  for(i=0;i<nsel;++i){
    icg=isel[i];
    icglist = simparms->icglist[icg];
    
    /* a = alpha x p */
         
    for(j=0;j<simparms->napcg[icg];j++){
      iat=icglist[j];
      rvec[0] = px[iat];
      rvec[1] = py[iat];
      rvec[2] = pz[iat];
      CrossProduct(so3->accel[icg],rvec,avec);
      fx[iat] =  avec[0];
      fy[iat] =  avec[1];
      fz[iat] =  avec[2];
    }
  }
}

/*------------------------------------------------------------------ */
void energyso3(SIMPARMS *simparms,COORDS *caa, ENERGIES *energies, INTER *inter, NGBR *ngbr, SO3 *so3)
{
  
  int i,icg,ncg,nsel,*isel,ndof;
  double t1[3],t2,p2m,vonfo,vonfo_e;
  ncg = simparms->ncg;

  nsel = caa->next->ninactive+caa->next->nschizo;
  isel = caa->next->ischizo;

  /* intra potential energy terms */
  get_relat_aapos(caa->next,simparms->icglist,simparms->napcg,isel,nsel);
  energies->so3_bond = getvbond_so3(simparms,caa,0); /* coarse grain region */
  energies->so3_bond += getvbond_so3(simparms,caa,1); /* healing region */
  energies->so3_bend = getvbend_so3(simparms,caa,0); /* coarse grain region */
  energies->so3_bend += getvbend_so3(simparms,caa,1); /* healing region */
  energies->so3_tors = getvtors_so3(simparms,caa,0); /* coarse grain region */
  energies->so3_tors += getvtors_so3(simparms,caa,1); /* healing region */
  getvonfo_so3(simparms,caa,0,&vonfo,&vonfo_e); /* coarse grain region */
  energies->so3_onfo = vonfo;
  energies->so3_onfo_e = vonfo_e;
  getvonfo_so3(simparms,caa,1,&vonfo,&vonfo_e); /* healing region */
  energies->so3_onfo += vonfo;
  energies->so3_onfo_e += vonfo_e;
  get_abs_aapos(caa->next,simparms->icglist,simparms->napcg,isel,nsel);
  energies->so3_potra = energies->so3_bond + energies->so3_bend + energies->so3_tors + energies->so3_onfo + energies->so3_onfo_e;
  
  /* inter potential energy term (VdW) */
  getsvirterso3(&energies->so3_poter);
  energies->so3_vzero_ngbr += so3->vzero_ngbr;
  so3->vzero_ngbr = 0.;
  energies->so3_poter += energies->so3_vzero_ngbr;
  
  /* kinetic energy */
  nsel = caa->next->nselect-caa->next->nschizo;
  isel = caa->next->iselect;
  p2m = 0.0;
  for(i=0;i<nsel;i++){  /* sum over the CG units in the CGR only */
    icg = isel[i];
    MultMatVec(so3->mit[icg],so3->omega[icg],t1);
    DotProduct(so3->omega[icg],t1,&t2);
    p2m += t2;
  }
  energies->so3_ekin = 0.5*p2m;
  ndof = 3*caa->next->nselect;   /* 3 rotational degrees per blob inside the CGR... */
  energies->so3_temp = p2m/(double)ndof;
  energies->so3_ham = energies->so3_poter + energies->so3_potra + energies->so3_ekin;

}

/*------------------------------------------------------------------ */
double get_so3_ekin(double omega[3], double mit[3][3]){
 
  double t1[3],t2;

  MultMatVec(mit,omega,t1);
  t2 = omega[0]*t1[0] + omega[1]*t1[1] + omega[2]*t1[2];
  return t2;
}

