

#include "md.h"

void mnewtfunc(double a[3],double it[3][3],double wnp[3],double tor[3],double tau,double jc[3][3],double fxnval[3],double fxnjac[3][3]){
  /* jc is the non-updating part of the jacobian */
  int i;
  double t1[3],t2[3],t3[3],t4[3],t5[3],v1[3];
  double m1[3][3], m2[3][3], m3[3][3], m4[3][3];

  for(i=0;i<3;++i){
    t1[i] = wnp[i] + 0.5*tau*a[i];
  }
  MultMatVec(it,t1,t2);
  CrossProduct(t1,t2,t3);
  VecSub(t3,tor,t4);
  MultMatVec(it,a,t5);
  VecAdd(t5,t4,fxnval);
  /* function value done, now get jacobian */
  Skew(a,m1);
  MatMult(m1,it,m2);
  MultMatVec(it,a,v1);
  Skew(v1,m3);
  MatSub(m2,m3,m4);
  MultScalarMat(0.5*tau*tau,m4);
  MatAdd(jc,m4,fxnjac);
}


void mnewt3(int ntrial,double x[3],double it[3][3], double wnp[3], double tor[3], double tau, double tolx,double tolf)
{
  /* Given an initial guess x for a root in 3 dimensions, take ntrial Newton-Raphson steps to improve the root. Stop if the root converges in either summed absolute variable increments tolx or summed absolute function values tolf  */
  int k,i,indx[3];
  double errx,errf,d ;
  double fjac[3][3],fvec[3],p[3];
  double matrix1[3][3],matrix2[3][3],matrix3[3][3],matrix4[3][3];
  double vector1[3],jc[3][3];

  Skew(wnp,matrix1);										/* create Skew(omega) */
  MatMult(matrix1,it,matrix2);					/* Skew(omega) MIT */
  MultMatVec(it,wnp,vector1);						/* MIT omega */
  Skew(vector1,matrix3);								/* Skew(MIT omega) */
  MatSub(matrix2,matrix3,matrix4);			/* Skew(omega) MIT - Skew(MIT omega) */
  MultScalarMat(0.5*tau,matrix4);
  MatAdd(it,matrix4,jc);

  for (k=0;k<ntrial;++k) {
    mnewtfunc(x,it,wnp,tor,tau,jc,fvec,fjac);
    errf=0.0;
    for (i=0;i<3;++i) errf += fabs(fvec[i]);
    if (errf <= tolf) {
      return;
    }
    for (i=0;i<3;++i) p[i] = -fvec[i];
    ludcmp33(fjac,indx,&d);
    lubksb33(fjac,indx,p);
    errx=0.0;
    for (i=0;i<3;++i) {
      errx += fabs(p[i]);
      x[i] += p[i];
    }
    if (errx <= tolx) {
      return;
    }
  }
  fprintf(stderr,"no convergence in mnewt\n");
  exit(1);
  return;
}

void getAccel(int nsel,int *isel,SO3 *so3,double tau,int ilevel){  
  int i,ii,ntrial = 10; /* max interation in mnewt3 */
  double tolx = 1.0e-5;
  double tolf = 1.0e-5;
  double (*t1)[3];
/*  double tau = so3->tau; */

  switch(ilevel){
    case 0:
      t1 = so3->tor;
      break;
    case 1:
      t1 = so3->tor_s;
      break;
  }

  for(ii=0;ii<nsel;++ii){
		i=isel[ii];
    mnewt3(ntrial,so3->accel[i],so3->mit[i],so3->omega[i],t1[i],tau,tolx,tolf);
//		printf("accel= %g %g %g\n",accel[i][0],accel[i][1],accel[i][2]); 
  }
	
}

