/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

   /* compute positions of the coarse grain sites */

#include "md.h"  

/*------------------------------------------------------------------*/
/* split the ngbr struct into two: one AA part and one CG part     */
void split_ngbr_in_two(int ncg, int naa, NGBR *ngbr){
	
  ngbr->next = (NGBR *)cmalloc(sizeof(NGBR));
  ngbr->next->next = ngbr;
  ngbr->lcg = 1;
  ngbr->next->lcg = 0;
  ngbr->next->istart = 0;
  ngbr->next->ilist = ngbr->ilist;
  ngbr->next->ncells = ngbr->ncells;
  if(ngbr->onelist==2){ /* use only one CG ngbrlist */
    ngbr->next->onelist = 0;
  }else{               /* use seperate CG and AA ngbrlists */
    ngbr->onelist = 1;
    ngbr->next->onelist = 1;
  }

  return;
}

/*------------------------------------------------------------------ */
/* split the INTER struct into two: one AA part and one CG part       */
void split_inter_in_two(int ncg, int naa, int ntype, int ncgtype, INTER *inter, int nlen){

  int i, j, k, naatype;
  int **excl, **map;
  int natoms = naa+ncg;
  
  inter->lcg = 1;
  inter->next = (INTER *)cmalloc(sizeof(INTER));
  inter->next->lcg = 0;
  inter->next->next = inter;

  inter->next->ntable = inter->ntable;
  inter->next->rheal = inter->rheal;
  inter->next->skin_ter = inter->skin_ter;
  naatype = ntype - ncgtype;
  inter->next->itype = &inter->itype[ncg];
  /* make atom types for AA start at 0 here in inter->itype */
  /* note that simparms->itype is not renumbered; i.e.      */
  /* simparms->itype[simparms->ncg]=simparms->ncgtypes;     */
  for(i=ncg;i<natoms;i++) inter->itype[i]-=ncgtype;  
  
  map = (int **)cmalloc(ncgtype*sizeof(int*));
  map[0] = (int *)cmalloc(ncgtype*ncgtype*sizeof(int));
  for(i=1;i<ncgtype;i++) map[i] = map[i-1]+ncgtype;
  inter->next->map = (int **)cmalloc(naatype*sizeof(int*));
  inter->next->map[0] = (int *)cmalloc(naatype*naatype*sizeof(int));
  for(i=1;i<naatype;i++) inter->next->map[i] = inter->next->map[i-1]+naatype;
  for(i=0;i<ncgtype;i++){
    for(j=0;j<ncgtype;j++) map[i][j] = inter->map[i][j];
  }
  k = inter->ncgtype;
  for(i=0;i<naatype;i++){
    for(j=0;j<naatype;j++){
      if( inter->map[ncgtype+i][ncgtype+j] == -1 ) {
        inter->next->map[i][j] = -1;                                  /* renumbering AA map */
      }else{
        inter->next->map[i][j] = inter->map[ncgtype+i][ncgtype+j]-k;  /* renumbering AA map */
      }
    }
  }
  free(inter->map);
  inter->map = map;
  
  /*  inter->next->map = inter->map; */
  inter->next->ntype = inter->ntype - inter->ncgtype;
  inter->ntype = inter->ncgtype;
  
  inter->next->nexclude = &inter->nexclude[ncg];
  excl = inter->next->exclude = &inter->exclude[ncg];
  for(i=0;i<naa;i++){
    for(j=0;j<inter->next->nexclude[i];j++){
      excl[i][j]-=ncg;
    }
  }
  
  inter->next->rminl_cut2 = &inter->rminl_cut2[k];
  inter->next->rmaxl_cut2 = &inter->rmaxl_cut2[k];
  inter->next->rmins_cut2 = &inter->rmins_cut2[k];
  inter->next->rmaxs_cut2 = &inter->rmaxs_cut2[k];
  inter->next->rmaxs_skin2 = &inter->rmaxs_skin2[k];
  inter->next->rmaxl_skin2 = &inter->rmaxl_skin2[k];
  inter->next->rminl_skin2 = &inter->rminl_skin2[k];
  inter->next->dx2tab = &inter->dx2tab[k];
  inter->next->vlrc = &inter->vlrc[k];
  inter->next->wlrc = &inter->wlrc[k];
  inter->next->vtab = &inter->vtab[k];
  inter->next->dvtab = &inter->dvtab[k];
  inter->next->d2vtab = &inter->d2vtab[k];
  inter->next->switchs = &inter->switchs[k];

  inter->dxn = (double *)crealloc(inter->dxn,3*ncg*sizeof(double));
  inter->dyn = inter->dxn + ncg;
  inter->dzn = inter->dxn + 2*ncg;
  inter->next->dxn = (double *)cmalloc(3*naa*sizeof(double));
  inter->next->dyn = inter->next->dxn + naa;
  inter->next->dzn = inter->next->dxn + 2*naa;

  /* electrostatic */
  inter->next->rmins_coul2 = inter->rmins_coul2;
  inter->next->rmaxs_coul2 = inter->rmaxs_coul2;
  inter->next->rminl_coul2 = inter->rminl_coul2;
  inter->next->rmaxl_coul2 = inter->rmaxl_coul2;
  inter->next->rmaxl_skin2_e = inter->rmaxl_skin2_e;
  inter->next->rminl_skin2_e = inter->rminl_skin2_e;
  inter->next->rmaxs_skin2_e = inter->rmaxs_skin2_e;
  inter->next->dx2coul = inter->dx2coul;
  inter->next->coultab = inter->coultab;
  inter->next->dcoultab = inter->dcoultab;
  inter->next->coultab_aa = inter->coultab_aa;
  inter->next->dcoultab_aa = inter->dcoultab_aa;
  inter->next->switchse = inter->switchse; 

#ifdef FIXHYBRIDSWAPS
  inter->nswp = 0;
  inter->iswp = (int *)cmalloc(nlen*sizeof(int));
  inter->nswptot = 0;
  inter->nswpmax = nlen;
  inter->iatswp = (int *)cmalloc(nlen*sizeof(int));
  inter->jatswp = (int *)cmalloc(nlen*sizeof(int));
  inter->swppot = (double *)cmalloc((nlen+1)*sizeof(double));
  inter->swppot++; /* use swppot[-1] as trashbin */
  inter->next->nswp = 0;
  inter->next->iswp = (int *)cmalloc(nlen*sizeof(int));
  inter->next->nswptot = 0;
  inter->next->nswpmax = nlen;
  inter->next->iatswp = (int *)cmalloc(nlen*sizeof(int));
  inter->next->jatswp = (int *)cmalloc(nlen*sizeof(int));
  inter->next->swppot = (double *)cmalloc((nlen+1)*sizeof(double));
  inter->next->swppot++; /* use swppot[-1] as trashbin */
#endif	
}


/*------------------------------------------------------------------ */
/* split the COUL struct into two: one AA part and one CG part       */
void split_coul_in_two(int ncg, int naa, COUL *ccg, COUL *caa){
  int i,j;
  int **excl;

  for(i=0;i<ccg->ncharge;i++)		/* find AA part */
    if(ccg->icharge[i] >= ncg) break;
  caa->ncharge = caa->nselect = ccg->ncharge - i;
  caa->icharge = &ccg->icharge[i];
  caa->iselect = &ccg->iselect[i];
  ccg->ncharge = ccg->nselect = i;
  for(i=0;i<caa->ncharge;i++) {
    caa->icharge[i] -= ncg;
    caa->iselect[i] = caa->icharge[i];
  }

  caa->kappa = ccg->kappa;
  caa->kappa2 = ccg->kappa2;
  caa->sqrt_pi = ccg->sqrt_pi;
  caa->ewald_self = ccg->ewald_self;  /* is recomputed for selection */
  caa->ewald_bkg = ccg->ewald_bkg;    /* is recomputed for selection */
  caa->falp2 = ccg->falp2;
  caa->ktot = ccg->ktot;              /* same k-space grid (or ktot=0 */
  caa->ka = ccg->ka; 
  caa->kb = ccg->kb;
  caa->kc = ccg->kc;
  caa->kup = ccg->kup;

  caa->cossc = ccg->cossc;           /* set the same for now (see ewald_setup) */
  caa->sinsc = ccg->sinsc;
  caa->helr = ccg->helr;
  caa->heli = ccg->heli;

  caa->nexclude = &ccg->nexclude[ncg];
  excl = caa->exclude = &ccg->exclude[ncg];
  for(i=0;i<naa;i++){
    for(j=0;j<caa->nexclude[i];j++){
      excl[i][j]-=ncg;
    }
  }
  
  for(i=0;i<ccg->necorr;i++)
    if(ccg->iecorr[i] >= ncg) break;		/* find AA part (make use of jecorr[i]>iecorr[i]) */
  caa->necorr = caa->nselect_ecorr = ccg->necorr - i;
  caa->iecorr = &ccg->iecorr[i];
  caa->jecorr = &ccg->jecorr[i];
  caa->iselect_ecorr = &ccg->iselect_ecorr[i];
  ccg->necorr = ccg->nselect_ecorr = i;
  for(i=0;i<caa->necorr;i++){
    caa->iecorr[i] -= ncg;
    caa->jecorr[i] -= ncg;
  }
  
  /*	used in energies/inter/md_ewald.c: void fk_ewald()
    caa->vk_ewald = ccg->
    caa->vp_ewald = ccg->
    caa->pot_ecorr = ccg->
    caa->prs_ecorr = ccg->
    caa->prs_ecorr_tensor[9] = ccg->
    caa->vp_ewald_tensor[9] = ccg->
  */	
  
  return;
}

/*------------------------------------------------------------------ */
/* split the bonds struct into two: one AA part and one CG part      */
void split_bonds_in_two(int ncg, int *icg, BONDS *bcg, BONDS *baa){
  int i,j,n,n2,k;

  n = bcg->nbonds - bcg->ncgbonds;
  baa->itypes = bcg->itypes;
  baa->nbonds = baa->nselect = n;
  baa->idxbond = (int *)cmalloc(n*sizeof(int));
  baa->ibond = (int *)cmalloc(n*sizeof(int));
  baa->jbond = (int *)cmalloc(n*sizeof(int));
  baa->iselect = (int *)cmalloc(n*sizeof(int));
  baa->ifix = bcg->ifix;
  baa->itypbond = bcg->itypbond;
  baa->eqbond = bcg->eqbond;
  baa->fkbond = bcg->fkbond;
  baa->bmorse = bcg->bmorse;

  n2 = baa->nbonds2 = bcg->nbonds2;
  baa->nsel_so3cgr = 0; /* so3 selection in CG region */
  baa->nsel_so3hr = 0; /* so3 selection in healing region */
  baa->idxbond2 = (int *)cmalloc(n2*sizeof(int));
  baa->ibond2 = (int *)cmalloc(n2*sizeof(int));
  baa->jbond2 = (int *)cmalloc(n2*sizeof(int));
  /* the HR selection (isel_so3hr) is stored in last part of the next array */
  baa->isel_so3cgr = (int *)cmalloc(n2*sizeof(int));  

  /* Order the interactions: first intraCG per CG unit, then interCG */
  
  baa->icgunit = (int *)cmalloc((ncg+1)*sizeof(int));
  /* first filter out the intraCG bonds per CG unit */
  for(i=bcg->ncgbonds;i<bcg->ncgbonds+n;++i){
    bcg->ibond[i] -= ncg;
    bcg->jbond[i] -= ncg;		
  }
  j = 0;
  for(k=0;k<ncg;k++){
    baa->icgunit[k] = j;
    for(i=bcg->ncgbonds;i<bcg->ncgbonds+n;++i){
      if((icg[bcg->ibond[i]]==k) && (icg[bcg->jbond[i]]==k)){
	baa->idxbond[j] = bcg->idxbond[i];
	baa->ibond[j] = bcg->ibond[i];
	baa->jbond[j] = bcg->jbond[i];
	j++;
      }
    }
  }
  baa->icgunit[k] = j;
  /* then add the interCG bonds */
  for(i=bcg->ncgbonds;i<bcg->ncgbonds+n;++i){
    if( icg[bcg->ibond[i]] != icg[bcg->jbond[i]] ){
      baa->idxbond[j] = bcg->idxbond[i];
      baa->ibond[j] = bcg->ibond[i];
      baa->jbond[j++] = bcg->jbond[i];
    }
  }
  
  for(i=0;i<n2;i++){
    baa->idxbond2[i] = bcg->idxbond2[i];
    baa->ibond2[i] = bcg->ibond2[i]-ncg;
    baa->jbond2[i] = bcg->jbond2[i]-ncg;
  }
  
  n=bcg->ncgbonds;
  bcg->nbonds = bcg->nselect = n;
  if(n==0) n=1;
  bcg->idxbond = (int *)crealloc(bcg->idxbond,n*sizeof(int));
  bcg->ibond = (int *)crealloc(bcg->ibond,n*sizeof(int));
  bcg->jbond = (int *)crealloc(bcg->jbond,n*sizeof(int));
  if(bcg->iselect!=NULL) 
    bcg->iselect = (int *)crealloc(bcg->iselect,n*sizeof(int));
  
  /* free SO3 stuff in CG struct */
  bcg->nsel_so3cgr = bcg->nsel_so3hr = 0;
  if(bcg->nbonds2>0){
    free(bcg->isel_so3cgr);
    bcg->isel_so3cgr = bcg->isel_so3hr = NULL;
    free(bcg->idxbond2);
    free(bcg->ibond2);
    free(bcg->jbond2);
    bcg->nbonds2 = 0;
  }
  
  return; 
}

/*------------------------------------------------------------------ */
/* split the bends struct into two: one AA part and one CG part     */
void split_bends_in_two(int ncg, int *icg, BENDS *bcg, BENDS *baa){
  int i,j,n,n2,k;

  n = bcg->nbends - bcg->ncgbends;
  baa->itypes = bcg->itypes;
  baa->nbends = baa->nselect = n;
  baa->idxbend = (int *)cmalloc(n*sizeof(int));
  baa->ibend = (int *)cmalloc(n*sizeof(int));
  baa->jbend = (int *)cmalloc(n*sizeof(int));
  baa->kbend = (int *)cmalloc(n*sizeof(int));
  baa->iselect = (int *)cmalloc(n*sizeof(int));
  baa->ifix = bcg->ifix;
  baa->ibendtype = (int *)cmalloc(n*sizeof(int));
  baa->eqbend = bcg->eqbend;
  baa->fkbend = bcg->fkbend;

  /* so3 selection */
  n2 = baa->nbends2 = bcg->nbends2;
  baa->nsel_so3cgr = 0; /* so3 selection in CG region */
  baa->nsel_so3hr = 0; /* so3 selection in healing region */
  baa->idxbend2 = (int *)cmalloc(n2*sizeof(int));
  baa->ibend2 = (int *)cmalloc(n2*sizeof(int));
  baa->jbend2 = (int *)cmalloc(n2*sizeof(int));
  baa->kbend2 = (int *)cmalloc(n2*sizeof(int));
  baa->isel_so3cgr = (int *)cmalloc(n2*sizeof(int));
  
  /* Order the interactions: first intraCG per CG unit, then interCG */
  
  baa->icgunit = (int *)cmalloc((ncg+1)*sizeof(int));
  /* first filter out the intraCG bends per CG unit */
  for(i=bcg->ncgbends;i<bcg->ncgbends+n;++i){
    bcg->ibend[i] -= ncg;
    bcg->jbend[i] -= ncg;		
    bcg->kbend[i] -= ncg;		
  }
  j = 0;
  for(k=0;k<ncg;k++){
    baa->icgunit[k] = j;
    for(i=bcg->ncgbends;i<bcg->ncgbends+n;++i){
      if((icg[bcg->ibend[i]]==k) && (icg[bcg->jbend[i]]==k) && (icg[bcg->kbend[i]]==k)){
	baa->ibendtype[j] = bcg->ibendtype[i];
	baa->idxbend[j] = bcg->idxbend[i];
	baa->ibend[j] = bcg->ibend[i];
	baa->jbend[j] = bcg->jbend[i];
	baa->kbend[j] = bcg->kbend[i];
	j++;
      }
    }
  }
  baa->icgunit[k] = j;
  /* then add the interCG bends */
  for(i=bcg->ncgbends;i<bcg->ncgbends+n;++i){
    if( (icg[bcg->ibend[i]] != icg[bcg->jbend[i]]) || (icg[bcg->ibend[i]] != icg[bcg->kbend[i]]) ){
      baa->ibendtype[j] = bcg->ibendtype[i];
      baa->idxbend[j] = bcg->idxbend[i];
      baa->ibend[j] = bcg->ibend[i];
      baa->jbend[j] = bcg->jbend[i];
      baa->kbend[j++] = bcg->kbend[i];
    }
  }
  
  for(i=0;i<n2;i++){
    /*   baa->ibendtype[i] = bcg->ibendtype[j]; */ /* what is this thing? */
    baa->idxbend2[i] = bcg->idxbend2[i];
    baa->ibend2[i] = bcg->ibend2[i]-ncg;
    baa->jbend2[i] = bcg->jbend2[i]-ncg;
    baa->kbend2[i] = bcg->kbend2[i]-ncg;
  }
  
  n=bcg->ncgbends;
  bcg->nbends = bcg->nselect = n;
  if(n==0) n=1;
  bcg->ibendtype = (int *)crealloc(bcg->ibendtype,n*sizeof(int));
  bcg->idxbend = (int *)crealloc(bcg->idxbend,n*sizeof(int));
  bcg->ibend = (int *)crealloc(bcg->ibend,n*sizeof(int));
  bcg->jbend = (int *)crealloc(bcg->jbend,n*sizeof(int));
  bcg->kbend = (int *)crealloc(bcg->kbend,n*sizeof(int)); 
  if((bcg->iselect!=NULL) && (bcg->nbends!=0))  
    bcg->iselect = (int *)crealloc(bcg->iselect,n*sizeof(int));
  
  /* free SO3 stuff in CG struct */
  bcg->nsel_so3cgr = bcg->nsel_so3hr = 0;
  if(bcg->nbends2>0){
    free(bcg->isel_so3cgr);
    bcg->isel_so3cgr = bcg->isel_so3hr = NULL;
    free(bcg->idxbend2);
    free(bcg->ibend2);
    free(bcg->jbend2);
    free(bcg->kbend2);
    bcg->nbends2 = 0;
  }
  
  return;
}

/*------------------------------------------------------------------ */
/* split the tors struct into two: one AA part and one CG part     */
void split_tors_in_two(int ncg,int *icg,TORS *bcg, TORS *baa){
  int i,j,n,n2,k;
	
  n = bcg->ntorss - bcg->ncgtorss;
  baa->itypes = bcg->itypes;
  baa->ntorss = baa->nselect = n;
  baa->idxtors = (int *)cmalloc(n*sizeof(int));
  baa->itors = (int *)cmalloc(n*sizeof(int));
  baa->jtors = (int *)cmalloc(n*sizeof(int));
  baa->ktors = (int *)cmalloc(n*sizeof(int));
  baa->ltors = (int *)cmalloc(n*sizeof(int));
  baa->iselect = (int *)cmalloc(n*sizeof(int));
  baa->ifix = bcg->ifix;
  baa->eqtors = bcg->eqtors;
  baa->fktors = bcg->fktors;
  baa->vphi = bcg->vphi;
  baa->ind_pot = bcg->ind_pot;
  baa->ind_pot_num = bcg->ind_pot_num;
  
  /* so3 selection */
  n2 = baa->ntorss2 = bcg->ntorss2;
  baa->nsel_so3cgr = 0; /* so3 selection in CG region */
  baa->nsel_so3hr = 0; /* so3 selection in healing region */
  baa->idxtors2 = (int *)cmalloc(n2*sizeof(int));
  baa->itors2 = (int *)cmalloc(n2*sizeof(int));
  baa->jtors2 = (int *)cmalloc(n2*sizeof(int));
  baa->ktors2 = (int *)cmalloc(n2*sizeof(int));
  baa->ltors2 = (int *)cmalloc(n2*sizeof(int));
  baa->isel_so3cgr = (int *)cmalloc(n2*sizeof(int));
  
  /* Order the interactions: first intraCG per CG unit, then interCG */
  
  baa->icgunit = (int *)cmalloc((ncg+1)*sizeof(int));
  /* first filter out the intraCG torions per CG unit */
  for(i=bcg->ncgtorss;i<bcg->ncgtorss+n;++i){
    bcg->itors[i] -= ncg;
    bcg->jtors[i] -= ncg;		
    bcg->ktors[i] -= ncg;
    bcg->ltors[i] -= ncg;		
  }
  j = 0;
  for(k=0;k<ncg;k++){
    baa->icgunit[k] = j;
    for(i=bcg->ncgtorss;i<bcg->ncgtorss+n;++i){
      if((icg[bcg->itors[i]]==k) && (icg[bcg->jtors[i]]==k) && (icg[bcg->ktors[i]]==k) && (icg[bcg->ltors[i]]==k)){
	baa->idxtors[j] = bcg->idxtors[i];
	baa->itors[j] = bcg->itors[i];
	baa->jtors[j] = bcg->jtors[i];
	baa->ktors[j] = bcg->ktors[i];
	baa->ltors[j] = bcg->ltors[i];
	j++;
      }
    }
  }
  baa->icgunit[k] = j;
  /* then add the interCG torsions */
  for(i=bcg->ncgtorss;i<bcg->ncgtorss+n;++i){
    if( (icg[bcg->itors[i]] != icg[bcg->jtors[i]]) || (icg[bcg->itors[i]] != icg[bcg->ktors[i]]) || (icg[bcg->itors[i]] != icg[bcg->ltors[i]]) ){
      baa->idxtors[j] = bcg->idxtors[i];
      baa->itors[j] = bcg->itors[i];
      baa->jtors[j] = bcg->jtors[i];
      baa->ktors[j] = bcg->ktors[i];
      baa->ltors[j] = bcg->ltors[i];
      j++;
    }
  }
  
  for(i=0;i<n2;i++){
    baa->idxtors2[i] = bcg->idxtors2[i];
    baa->itors2[i] = bcg->itors2[i]-ncg;
    baa->jtors2[i] = bcg->jtors2[i]-ncg;
    baa->ktors2[i] = bcg->ktors2[i]-ncg;
    baa->ltors2[i] = bcg->ltors2[i]-ncg;
  }
	
  n=bcg->ncgtorss;
  bcg->ntorss = bcg->nselect = n;
  if(n==0) n=1;
  bcg->idxtors = (int *)crealloc(bcg->idxtors,n*sizeof(int));
  bcg->itors = (int *)crealloc(bcg->itors,n*sizeof(int));
  bcg->jtors = (int *)crealloc(bcg->jtors,n*sizeof(int));
  bcg->ktors = (int *)crealloc(bcg->ktors,n*sizeof(int));
  bcg->ltors = (int *)crealloc(bcg->ltors,n*sizeof(int));
  if((bcg->iselect != NULL) && (bcg->ntorss!=0))
    bcg->iselect = (int *)crealloc(bcg->iselect,n*sizeof(int));
  
  /* free SO3 stuff in CG struct */
  bcg->nsel_so3cgr = bcg->nsel_so3hr = 0;
  if(bcg->ntorss2>0){
    free(bcg->isel_so3cgr);
    bcg->isel_so3cgr = bcg->isel_so3hr = NULL;
    free(bcg->idxtors2);
    free(bcg->itors2);
    free(bcg->jtors2);
    free(bcg->ktors2);
    free(bcg->ltors2);
    bcg->ntorss2 = 0;
  }
  
  return; 
}

/*------------------------------------------------------------------ */
/* split the onfo struct into two: one AA part and one CG part     */
void split_onfo_in_two(int ncg,int *icg,ONFO *bcg, ONFO *baa){
  int i,j,n,n2,k;

  n = bcg->n14s - bcg->ncg14s;
  baa->itypes = bcg->itypes;
  baa->n14s = baa->nselect = n;
  baa->i14dx = (int *)cmalloc(n*sizeof(int));
  baa->i14 = (int *)cmalloc(n*sizeof(int));
  baa->j14 = (int *)cmalloc(n*sizeof(int));
  baa->iselect = (int *)cmalloc(n*sizeof(int));
  baa->n14table = bcg->n14table ;
  baa->v14tab = bcg->v14tab;
  baa->dv14tab = bcg->dv14tab;
  baa->d2v14tab = bcg->d2v14tab;
  baa->dx214tab = bcg->dx214tab;
  baa->r14min2 = bcg->r14min2;
  baa->r14max2 = bcg->r14max2;
  baa->scale_onefour = bcg->scale_onefour;
  baa->scale_onefour_e = bcg->scale_onefour_e;
	
  /* so3 selection */
  n2 = baa->n14s2 = bcg->n14s2;
  baa->nsel_so3cgr = 0; /* so3 selection in CG region */
  baa->nsel_so3hr = 0; /* so3 selection in healing region */
  baa->i14dx2 = (int *)cmalloc(n2*sizeof(int));
  baa->i142 = (int *)cmalloc(n2*sizeof(int));
  baa->j142 = (int *)cmalloc(n2*sizeof(int));
  baa->isel_so3cgr = (int *)cmalloc(n2*sizeof(int));
  /* Order the interactions: first intraCG per CG unit, then interCG */
  
  baa->icgunit = (int *)cmalloc((ncg+1)*sizeof(int));
  /* first filter out the intraCG onfos per CG unit */
  for(i=bcg->ncg14s;i<bcg->ncg14s+n;++i){
    bcg->i14[i] -= ncg;
    bcg->j14[i] -= ncg;		
  }
  j = 0;
  for(k=0;k<ncg;k++){
    baa->icgunit[k] = j;
    for(i=bcg->ncg14s;i<bcg->ncg14s+n;++i){
      if((icg[bcg->i14[i]]==k) && (icg[bcg->j14[i]]==k)){
	baa->i14dx[j] = bcg->i14dx[i];
	baa->i14[j] = bcg->i14[i];
	baa->j14[j] = bcg->j14[i];
	j++;
      }
    }
  }
  baa->icgunit[k] = j;
  /* then add the interCG onfos */
  for(i=bcg->ncg14s;i<bcg->ncg14s+n;++i){
    if( icg[bcg->i14[i]] != icg[bcg->j14[i]] ){
      baa->i14dx[j] = bcg->i14dx[i];
      baa->i14[j] = bcg->i14[i];
      baa->j14[j++] = bcg->j14[i];
    }
  }
  
  for(i=0;i<n2;i++){
    baa->i14dx2[i] = bcg->i14dx2[i];
    baa->i142[i] = bcg->i142[i]-ncg;
    baa->j142[i] = bcg->j142[i]-ncg;
  }
	
  n=bcg->ncg14s;
  bcg->n14s = bcg->nselect = n;
  if(n==0) n=1;
  bcg->i14dx = (int *)crealloc(bcg->i14dx,n*sizeof(int));
  bcg->i14 = (int *)crealloc(bcg->i14,n*sizeof(int));
  bcg->j14 = (int *)crealloc(bcg->j14,n*sizeof(int));
  if((bcg->iselect!=NULL) && (bcg->n14s!=0))  
    bcg->iselect = (int *)crealloc(bcg->iselect,n*sizeof(int));
  
  /* free SO3 stuff in CG struct */
  bcg->nsel_so3cgr = bcg->nsel_so3hr = 0;
  if(bcg->n14s2>0){
    free(bcg->isel_so3cgr);
    bcg->isel_so3cgr = bcg->isel_so3hr = NULL;
    free(bcg->i14dx2);
    free(bcg->i142);
    free(bcg->j142);
    bcg->n14s2 = 0;
  }
  
  return; 
}

/*
  typedef struct bondxs {
  int itypes,nbondxs,*idxbox,*ibondx,*jbondx,*kbondx,*ifix;
  int *icgselect,*iaaselect,naaselect,ncgselect,ncgbondxs;
  double *eq1bondx,*eq2bondx,*fkbondx;
  } BONDXS;
*/

/*------------------------------------------------------------------ */
/* split the bondxs struct into two: one AA part and one CG part     */
void split_bondxs_in_two(int ncg,int *icg,BONDXS *bcg, BONDXS *baa){
  int i,j,n,n2,k;
  
  n = bcg->nbondxs - bcg->ncgbondxs;
  baa->itypes = bcg->itypes;
  baa->nbondxs = baa->nselect = n;
  baa->idxbox = (int *)cmalloc(n*sizeof(int));
  baa->ibondx = (int *)cmalloc(n*sizeof(int));
  baa->jbondx = (int *)cmalloc(n*sizeof(int));
  baa->kbondx = (int *)cmalloc(n*sizeof(int));
  baa->iselect = (int *)cmalloc(n*sizeof(int));
  baa->ifix = bcg->ifix;
  baa->eq1bondx = bcg->eq1bondx;
  baa->eq2bondx = bcg->eq2bondx;
  baa->fkbondx = bcg->fkbondx;
	
  /* so3 selection */
  n2 = baa->nbondxs2 = bcg->nbondxs2;
  baa->nsel_so3cgr = 0; /* so3 selection in CG region */
  baa->nsel_so3hr = 0; /* so3 selection in healing region */
  baa->idxbox2 = (int *)cmalloc(n2*sizeof(int));
  baa->ibondx2 = (int *)cmalloc(n2*sizeof(int));
  baa->jbondx2 = (int *)cmalloc(n2*sizeof(int));
  baa->kbondx2 = (int *)cmalloc(n2*sizeof(int));
  baa->isel_so3cgr = (int *)cmalloc(n2*sizeof(int));
  /* Order the interactions: first intraCG per CG unit, then interCG */
  
  baa->icgunit = (int *)cmalloc((ncg+1)*sizeof(int));
  /* first filter out the intraCG bonds per CG unit */
  for(i=bcg->ncgbondxs;i<bcg->ncgbondxs+n;++i){
    bcg->ibondx[i] -= ncg;
    bcg->jbondx[i] -= ncg;		
    bcg->kbondx[i] -= ncg;
  }
  j = 0;
  for(k=0;k<ncg;k++){
    baa->icgunit[k] = j;
    for(i=bcg->ncgbondxs;i<bcg->ncgbondxs+n;++i){
      if((icg[bcg->ibondx[i]]==k) && (icg[bcg->jbondx[i]]==k) && (icg[bcg->kbondx[i]]==k)){
	baa->idxbox[j] = bcg->idxbox[i];
	baa->ibondx[j] = bcg->ibondx[i];
	baa->jbondx[j] = bcg->jbondx[i];
	baa->kbondx[j] = bcg->kbondx[i];
	j++;
      }
    }
  }
  baa->icgunit[k] = j;
  /* then add the interCG bondxs */
  for(i=bcg->ncgbondxs;i<bcg->ncgbondxs+n;++i){
    if( (icg[bcg->ibondx[i]] != icg[bcg->jbondx[i]]) || (icg[bcg->ibondx[i]] != icg[bcg->kbondx[i]]) ){
      baa->idxbox[j] = bcg->idxbox[i];
      baa->ibondx[j] = bcg->ibondx[i];
      baa->jbondx[j] = bcg->jbondx[i];
      baa->kbondx[j++] = bcg->kbondx[i];
    }
  }
  
  for(i=0;i<n2;i++){
    baa->idxbox2[i] = bcg->idxbox2[i];
    baa->ibondx2[i] = bcg->ibondx2[i]-ncg;
    baa->jbondx2[i] = bcg->jbondx2[i]-ncg;
    baa->kbondx2[i] = bcg->kbondx2[i]-ncg;
  }
	
  n=bcg->ncgbondxs;
  bcg->nbondxs = bcg->nselect = n;
  if(n==0) n=1;
  bcg->idxbox = (int *)crealloc(bcg->idxbox,n*sizeof(int));
  bcg->ibondx = (int *)crealloc(bcg->ibondx,n*sizeof(int));
  bcg->jbondx = (int *)crealloc(bcg->jbondx,n*sizeof(int));
  bcg->kbondx = (int *)crealloc(bcg->kbondx,n*sizeof(int));
  if((bcg->iselect!=NULL) && (bcg->nbondxs!=0))
    bcg->iselect = (int *)crealloc(bcg->iselect,n*sizeof(int));
  
  /* free SO3 stuff in CG struct */
  bcg->nsel_so3cgr = bcg->nsel_so3hr = 0;
  if(bcg->nbondxs2>0){
    free(bcg->isel_so3cgr);
    bcg->isel_so3cgr = bcg->isel_so3hr = NULL;
    free(bcg->idxbox2);
    free(bcg->ibondx2);
    free(bcg->jbondx2);
    free(bcg->kbondx2);
    bcg->nbondxs2 = 0;
  }
  
  return; 
}



/*------------------------------------------------------------------ */
/* split the coords struct into two: one AA part and one CG part     */
void split_coords_in_two(int ncg, int naa, int *icgunit, COORDS *coords,
			 int *iorder, int nres_maxlevel,int nres_type,int ntherm_tot){
  int i,j,k,l,n,natoms,nrespa;
  double *pcg,*paa;
  
  /* make new COORDS instance, make link and copy some stuff          */
  coords->next = (COORDS *)cmalloc(sizeof(COORDS));
  coords->lcg = 1;
  coords->next->lcg = 0;
  coords->next->next = coords;
  coords->natoms = coords->nselect = ncg;
  coords->ninactive = coords->next->ninactive = 0;
  coords->next->natoms = coords->next->nselect = naa;
  coords->iselect = (int *)crealloc(coords->iselect,(ncg+1)*sizeof(int));
  coords->inactive = coords->iselect+ncg;  /* second part of iselect holds inactive atoms */
  coords->next->iselect = (int *)cmalloc((naa+1)*sizeof(int));
  coords->next->inactive = coords->next->iselect+naa;
  coords->next->zmin_extern = coords->zmin_extern;    /* from read_sim_inputs */
  coords->next->zmax_extern = coords->zmax_extern;
  coords->next->ifreeze = coords->ifreeze;
  
	
  /* allocate memory for positions and velocities variables and copy */  
  natoms=ncg+naa;
  free(coords->fxs);
  paa = (double *)cmalloc(9*naa*sizeof(double));
  pcg = (double *)cmalloc(9*ncg*sizeof(double));
  coords->next->px = paa;
  coords->next->py = paa + naa;
  coords->next->pz = paa + 2*naa;
  coords->next->vx = paa + 3*naa;
  coords->next->vy = paa + 4*naa;
  coords->next->vz = paa + 5*naa;
  coords->next->amass = paa + 6*naa;
  coords->next->qch   = paa + 7*naa;
  coords->next->alpha = paa + 8*naa;
  k=l=0;
  for(j=0;j<9*natoms;j+=natoms){
    for(i=0+j;i<j+ncg;i++) {
      pcg[k++]=coords->px[i];
    }
    for(i=j+ncg;i<j+natoms;i++) {
      paa[l++]=coords->px[i];
    }
  }
  free(coords->px);
  coords->px = pcg;
  coords->py = pcg + ncg;
  coords->pz = pcg + 2*ncg;
  coords->vx = pcg + 3*ncg;
  coords->vy = pcg + 4*ncg;
  coords->vz = pcg + 5*ncg;
  coords->amass = pcg + 6*ncg;
  coords->qch   = pcg + 7*ncg;
  coords->alpha = pcg + 8*ncg;
  
  /* allocate forces */
  n=ncg;
  j=0;
  for(k=0;k<2;k++){
    coords->fx[0] = (double *)cmalloc(18*n*sizeof(double));
    nrespa=0;
    for(i=0;i<nres_maxlevel;i++) if(iorder[i]==2+j) break;
    if((i<nres_maxlevel)||(nres_type!=1)){
      coords->fxl = coords->fx[0]+ (nrespa++)*n;
      coords->fyl = coords->fx[0]+ (nrespa++)*n;
      coords->fzl = coords->fx[0]+ (nrespa++)*n;
      coords->fxs = coords->fx[0]+ (nrespa++)*n;
      coords->fys = coords->fx[0]+ (nrespa++)*n;
      coords->fzs = coords->fx[0]+ (nrespa++)*n;
    }
    for(i=0;i<nres_maxlevel;i++) if(iorder[i]==4+j) break;
    if((i<nres_maxlevel)||(nres_type!=1)){
      coords->fxr = coords->fx[0]+ (nrespa++)*n;
      coords->fyr = coords->fx[0]+ (nrespa++)*n;
      coords->fzr = coords->fx[0]+ (nrespa++)*n;
    }
    for(i=0;i<nres_maxlevel;i++) if(iorder[i]==2+j) break;
    if((i<nres_maxlevel)||(nres_type!=1)){
      coords->fxa = coords->fx[0]+ (nrespa++)*n;
      coords->fya = coords->fx[0]+ (nrespa++)*n;
      coords->fza = coords->fx[0]+ (nrespa++)*n;
    }
    coords->fxt = coords->fx[0]+ (nrespa++)*n;
    coords->fyt = coords->fx[0]+ (nrespa++)*n;
    coords->fzt = coords->fx[0]+ (nrespa++)*n;
    coords->fxb = coords->fx[0]+ (nrespa++)*n;
    coords->fyb = coords->fx[0]+ (nrespa++)*n;
    coords->fzb = coords->fx[0]+ (nrespa++)*n;
  /* following line is a bad idea; unless all fxt, fxa, fxr, fxs pointers are reset... */
  /*    coords->fx[0] = (double *)realloc(coords->fx[0],nrespa*n*sizeof(double)); */
    paa=coords->fx[0];
    for(i=0;i<nrespa*n;i++) paa[i]=0.;
    n=naa;
    coords=coords->next;
    j++;
  }
  
  for(i=0;i<nres_maxlevel+1;i++){
    switch (iorder[i]){
    case 0:
      coords->fx[i] = coords->next->fx[i] = coords->fxl;
      break;
    case 1:
      coords->fx[i] = coords->next->fx[i] = coords->next->fxl;
      break;
    case 2:
      coords->fx[i] = coords->next->fx[i] = coords->fxs;
      break;
    case 3:
      coords->fx[i] = coords->next->fx[i] = coords->next->fxs;
      break;
    case 4:
      coords->fx[i] = coords->next->fx[i] = coords->fxr;
      break;
    case 5:
      coords->fx[i] = coords->next->fx[i] = coords->next->fxr;
      break;
    case 6:
      coords->fx[i] = coords->next->fx[i] = coords->fxa;
      break;
    case 7:
      coords->fx[i] = coords->next->fx[i] = coords->next->fxa;
      break;
    default:
      break;
    }
  }
  coords->next->iperd = coords->iperd;
  
  split_bonds_in_two(ncg,icgunit,&coords->bonds,&coords->next->bonds);
  split_bends_in_two(ncg,icgunit,&coords->bends,&coords->next->bends);
  split_tors_in_two(ncg,icgunit,&coords->tors,&coords->next->tors);
  split_onfo_in_two(ncg,icgunit,&coords->onfo,&coords->next->onfo);
  split_bondxs_in_two(ncg,icgunit,&coords->bondxs,&coords->next->bondxs);
  split_coul_in_two(ncg,naa,&coords->coul,&coords->next->coul);
  
  coords->next->colvar.ncolvar = 0;
  

/* let's try to only barostat the CG system first    */
/*  coords->next->mbar   (initialized in set_baro()  */
/*  coords->next->mvol                               */
/*  coords->bar   (initialized in allocate_extend()  */
/*  coords->vbar                                     */

  i = coords->ntherm;
  coords->next->ntherm = ntherm_tot - i;
  coords->next->feta = &(coords->feta[i]); 
  coords->next->gkt = &(coords->gkt[i]);
  coords->next->mnh = &(coords->mnh[i]); 
  coords->next->eta = &(coords->eta[i]) ;
  coords->next->veta = &(coords->veta[i]);
  coords->next->p2mt = &(coords->p2mt[i]);
  coords->next->ithm = &(coords->ithm[coords->natoms]); 
  for(j=0;j<coords->next->natoms;j++) coords->next->ithm[j] -= coords->ntherm;
  coords->next->itselect = &(coords->itselect[i]);
  coords->ntselect = coords->next->ntselect = 0;
  
  return;
}

/*------------------------------------------------------------------ */

  /* coords->iperd is used to select ewald calculation                      */
  /* for now: set iperd=0 for AA part unless the whole system is treated AA */
void select_ewald(COORDS *coords){
  
  coords->next->iperd = coords->iperd;
  if(coords->next->iperd==3){
    if((coords->next->natoms==coords->next->nselect) &&  /* whole system is AA */
       (coords->nselect == 0 ) ){          /* and whole system is not HR */
      coords->next->iperd=3;
      coords->iperd=0;
    }else{
      coords->next->iperd=0;
      coords->iperd=3;
    }  
    return;
  }
}
/*------------------------------------------------------------------ */

void split_energies_in_two(ENERGIES *energies){

 energies->lcg = 1;
 energies->next = (ENERGIES *)cmalloc(sizeof(ENERGIES));
 energies->next->next = energies;
 energies->next->lcg = 0;

 /* set some stugg to zero */
 energies->zkecg  = 0.0;
 energies->vcgtra = 0.0;
 energies->vcgter = 0.0;

 energies = energies->next;
 energies->zkecg  = 0.0;
 energies->vcgtra = 0.0;
 energies->vcgter = 0.0;
 energies->zken = 0.0;
 energies->zkev = 0.0;
 energies->potn = 0.0;
 energies->potv = 0.0;
 energies = energies->next;
}
