/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

typedef char WORD[MAXWORDLEN];
typedef char LINE[MAXLINELEN];
typedef double VPHI[MAXPOWER_TORS];

/* ----------------system.h defs ------------------*/

typedef struct KEYS {
  char *name,*arg;
  int itype;
  struct KEYS *next;
} KEYS;

typedef struct META_KEY {
  char *meta;
  int nkeys;
  KEYS *keys;
  struct META_KEY *next;
} META_KEY;

/* ----------------system.h defs ------------------*/

typedef struct keys {
  WORD keyword,keyarg;
  struct keys *next; 
} KEY;

typedef struct species {
  WORD filename,thermopt,name,group;
  int nmol,molidx,napm,nbond,nbend,ntors,n14,npath,nbondx,nbendquart;
  struct species *next; 
} SPECIES;

typedef struct coul {
  int ncharge,*icharge,*iselect,nselect,*iselect_ecorr,nselect_ecorr;
  int *nexclude,**exclude;
  int *ka,*kb,*kc,*kup,ktot;
  int necorr,*iecorr,*jecorr;
  double kappa,kappa2,sqrt_pi,ewald_self,ewald_bkg,falp2;
  double vk_ewald,vp_ewald,pot_ecorr,prs_ecorr;
  double prs_ecorr_tensor[9];
  double vp_ewald_tensor[9];
  double *cossc,*sinsc,*helr,*heli;
} COUL;

typedef struct bonds {
  int itypes,nbonds,ncgbonds,*itypbond,*idxbond,*ibond,*jbond,*ifix;
  int nbonds2,*idxbond2,*ibond2,*jbond2;
  int *icgunit; /* length ncg+1; hold number of 1st intraCG bond for each CG unit */
  int *iselect,nselect;
  int *isel_so3hr,nsel_so3hr, *isel_so3cgr,nsel_so3cgr;
  double *eqbond,*fkbond,*bmorse;
} BONDS;

typedef struct bends {
  int itypes,nbends,*idxbend,*ibend,*jbend,*kbend,*ifix,*ibendtype;
  int *iselect,nselect,ncgbends;
  int *isel_so3hr,nsel_so3hr, *isel_so3cgr,nsel_so3cgr;
  int nbends2,*idxbend2,*ibend2,*jbend2,*kbend2;
  int *icgunit; /* length ncg+1; hold number of 1st intraCG bond for each CG unit */
  double *eqbend,*fkbend;
} BENDS;

typedef struct tors {
  int itypes,ntorss,*idxtors,*itors,*jtors,*ktors,*ltors,*ifix;
  int *iselect,*ind_pot,*ind_pot_num,nselect,ncgtorss;
  int *isel_so3hr,nsel_so3hr, *isel_so3cgr,nsel_so3cgr;
  int ntorss2,*idxtors2,*itors2,*jtors2,*ktors2,*ltors2;
  int *icgunit; /* length ncg+1; hold number of 1st intraCG bond for each CG unit */
  double *eqtors,*fktors;
  VPHI *vphi;
} TORS;

typedef struct onfo {
  int itypes,n14s,*i14,*j14,*i14dx,n14table;
  int *iselect,nselect,ncg14s;
  int *isel_so3hr,nsel_so3hr, *isel_so3cgr,nsel_so3cgr;
  int n14s2,*i142,*j142,*i14dx2;
  int *icgunit; /* length ncg+1; hold number of 1st intraCG bond for each CG unit */
  double **v14tab,**dv14tab,**d2v14tab;
  double *r14min2,*r14max2,*dx214tab;
  double scale_onefour,scale_onefour_e;
} ONFO;

typedef struct bondxs {
  int itypes,nbondxs,*idxbox,*ibondx,*jbondx,*kbondx,*ifix;
  int *iselect,nselect,ncgbondxs;
  int *isel_so3hr,nsel_so3hr, *isel_so3cgr,nsel_so3cgr;
  int nbondxs2,*idxbox2,*ibondx2,*jbondx2,*kbondx2;
  int *icgunit; /* length ncg+1; hold number of 1st intraCG bond for each CG unit */
  double *eq1bondx,*eq2bondx,*fkbondx;
} BONDXS;

/* added June 2000, JCS */
typedef struct bendquart {

  int itypes,nbendquarts,*idxbendquart,*ibendquart,*jbendquart,*kbendquart;
  int *ifix;
  double *eqbendquart,*fkbendquart;
} BENDQUART;
/* end of added June 2000, JCS */

typedef struct subspace {
  int nstate,num_subst,stemp_update;
  int *num_vecsub,*num_real,*num_imag;
  int *num_vecsub_max,*num_real_max,*num_imag_max;
  int igetvec,nstate_max;
  double freq_min,freq_max;
  WORD vecfile,vecconf;
  double *q0,*q1,*q2;
  double **d2v,*d2vn;
} SUBSPACE;

typedef struct simparms {
  int icalc_type,istart,nma_type,fstep,istep,nstep;
  int natoms,naa,ncg,nspec,ntypes,ncgtypes,iensemble;
  int ninter,ninnra,ninrra,nbar,ivol,iperd,imin_type,nocell_all;
  int nres_nsteps[MAXLEVEL_RESPA],nres_iorder[MAXLEVEL_RESPA];
  int nres_maxlevel,nres_maxaalevel,nres_maxcglevel,nres_type;
  int nres_aa_nsr,nres_aa_isr,nres_cg_nsr,nres_cg_isr;
  int iunits,ndof,nlen,npoints,iextern,nhc;
  int ntherm,nchain,max_exclude,nfreeze;
  int rank,size,mem_bytes,nyoshida,ipolar,itype_hr,iso3dyn,iprintatoms;
  int *itype,*iatom,*imolecule,*ispecies,*igroup;
  int *icgunit,*napcg,**icglist,*iaacg,*iaacg_b4,iupdate,*iaaregion,naaregion,iHybridSwitch;
  double rmin_hr,rmax_hr,rskin_hr,*faacg,*faacgprev,*dfaacg,*dfaacgprev,fnatoms[2];
  double switch_fmin, switch_fmax,switch_df;
  double *dr,*ddr,*ddr_lrf,p2mbuf,*vbuf,*vds_lrf,*vds_lrfprev,*vcgintern;
  double *ddr_intra,*vds_intra,*vds_intraprev,*ddr_tors,*vds_tors,*vds_torsprev;
  double temp,dt,pext,vlrc[2],wlrc[2],scaleeps,scalecharge,scaletemp,scaleso3vel;
  double min_tol,wallclock,wyosh[3];
  double nres_dt[MAXLEVEL_RESPA];
  WORD *atom,*mole,*group;
} SIMPARMS;

typedef struct sfilename {
  FILE *fconf,*fvel,*fcgconf,*fcgvel,*fham,*feng,*fext,*fforce,*fcolv,*fhybrid;
  WORD restart,initfile,configs,velfile,instham,insteng,setfile,forcefile;
  WORD nmfile,specfile,molvec,sysvec,extfile,instext,partrat,colvarcntrfile;
  WORD command,infile,outfile,errfile,colvfile,hillfile,cgconfigs,cgvelfile;
  WORD hybridinputfile;
} FILENAMES;

typedef struct senergies {
  int estep;
  double avham,avpotra,avpoter,avzke,avpotn,avzken,avpotv,avzkev;
  double avtemp,avprs,avvol,acpu,bcpu;
  double avprs_tensor[9],prs_tensor[9];
  double prsi,tiout,ham;
  double tiout_hr,tiout_aa,tiout_cg,zkecg,vcgtra,vcgter;
  double avtemp_hr,avtemp_aa,avtemp_cg,avzkecg,avcgtra,avcgter;
  double poth,potra,poter,zke,zken,potn,zkev,potv;
  double pot_inter,pot_bond,pot_bend,pot_bendquart,pot_tors,pot_extern,pot_extern_e;
  double pot_onfo,pot_onfo_e,pot_elec,pot_recip;
  double so3_bond,so3_bend,so3_onfo,so3_onfo_e,so3_xbond,so3_tors,so3_ekin,so3_potra,so3_poter,so3_ham;
  double avso3_bond,avso3_bend,avso3_onfo,avso3_onfo_e,avso3_xbond,avso3_tors,avso3_ekin,avso3_potra,avso3_poter,avso3_ham;
  double so3_temp, avso3_temp,so3_vzero_ngbr,avso3_vzero_ngbr;
  double avpot_inter,avpot_bond,avpot_bend,avpot_bendquart,avpot_tors;
  double avpot_onfo,avpot_onfo_e,avpot_elec,avpot_recip;
  double avpot_extern,avpot_extern_e;
  double econv,econ2;
  double econi,deltae2;
  double vinter,vhinter,velec,vewald,vbond,vbend,vbendquart,vtors,vonfo,vonfo_e;
  double vextern,vextern_e,vbondx;
  double winter,welec,wewald,wbond,wbend,wbendquart,wtors,wonfo,wonfo_e,wbondx;
  double wintertensor[9],welectensor[9],wbondtensor[9],wbondxtensor[9];
  double wbendtensor[9],wbendquarttensor[9],wonfotensor[9],wonfo_etensor[9];
  double wtorstensor[9],wewaldtensor[9],WItensor[9],Wtensor[9];
  double wextern,wextern_e;
  double US,U,UI,W,WI;
  double colvar;
  struct senergies *next;
  int lcg;
} ENERGIES;

typedef struct satom_topol {
  int atm_idx,icgflag,cg_ind;
  WORD type,group,cg_typ;
  double mass,charge,alpha;
} ATOM_TOPOL;

typedef struct swrite_step {
  int nscrn,ndump,nvel,nconf,ncgvel,ncgconf,ninst,nupdss,nforce,ncolv;
  int resamvel,psysvec,rescalevel,peigval;
} WRITE_STEP;

typedef struct colvar {
  int ncolvar,lrestart,nhills,ltunehills;
  int maxstephill,minstephill,lasthill;/* max and min of step before we place a hill */
  int *lconst_vel,*lhills, *type;
  int *ndim,**natom,***atom; /* ndim[ncolvar] # of centers in your constraint */
  /* natom[ncolvar,ndim] # of atoms in each center */
  /* atom[ncolvar,ndim,natom] array of atom indexes */
  double ecolvar;
  double *amass,*fk,*temp,*thrmmass;
  double *pcolvar,*vcolvar,*fcolvar,*pistcolvar;
  double *pcolvarlast,*pcolvarmin,*pcolvarmax,*vcolvarlast,*pcolvarmean,*pcolvarstd;
  double *f_hills,*f_spring;
  double *hillwidth, hilldepth;
  double **t_pcolvar;  /* t_pcolvar[ncolvar,nhills] */
  double **t_hillwidth, *t_hilldepth;
  /* t_hillwidth[ncolvar,nhills] t_hilldepth[nhills] */
  double *min_val, *max_val;   /* [ncolvar] coll.var. boundary wall parameters */
  double *mindwidth; /* min distance relative to hillwidth before we place a hill */
  int *istep;
  double **switch_r,*switch_rc;
  int *nswitch_r,*switch_n,*switch_m;
} COLVAR;


typedef struct scoords {
  int nbreak,*ibreak; /* nlength scratch */
  int *ithm,*ifreeze,ntherm,ntselect,*itselect;
  int *indx,*jndx,*ityp,*jtyp,*intr;
  double *px,*py,*pz,*vx,*vy,*vz,*amass,*qch,*alpha;
  double *ux,*uy,*uz,cmx,cmy,cmz;
  double **eta,**veta,**feta,**mnh,**gkt,*p2mt;
  double *bar,*vbar,*mbar;
  double *fxs,*fys,*fzs,*fxl,*fyl,*fzl,*fxr,*fyr,*fzr;
  double *fxa,*fya,*fza,*fxt,*fyt,*fzt,*fxb,*fyb,*fzb;
  double *fx[MAXLEVEL_RESPA];
  double zmin_extern,zmax_extern,dxtab_extern;
  double **dvtab_extern,**vtab_extern,*dvtab_extern_e,*vtab_extern_e;
  double pvol,vvol,fvol,mvol;
  double vvol9[9],fvol9[9];
  double hmat[9],hmati[9];
  double *scr_buf,*scr_rec;
  double *xold,*yold,*zold,cell_old;
  double *xdis,*ydis,*zdis;  /* nlength scratch */
  int lcg, natoms, nselect, *iselect, ninactive, *inactive, nschizo, *ischizo, iperd;
  struct scoords *next;
  COUL coul;
  BONDS bonds;
  BENDS bends;
  TORS tors;
  ONFO onfo;
  BONDXS bondxs;
  BENDQUART bendquarts;
  COLVAR colvar;
} COORDS;

typedef struct sngbr {
  int npso,npto,npao,npro,noffs,nofft,noffa,noffr;
  int istart,ncell_now,ncell_nowe;
  int update,ilist,onelist;
  int nprs,nprt,nprse,nprte,npra,nprr;
  int nprst,nprtt,nprset,nprtet;
  int *inps,*jnps,*inpt,*jnpt,*offs,*offl,*offt;
  int *inpr,*jnpr,*inpa,*jnpa,*offa,*offr;
  int *inpse,*jnpse,*inpte,*jnpte,*offse,*offte;

  int ncells,*lnklist,*hlist;
  int npaircell,*icell,*jcell,npaircelle,*icelle,*jcelle;
  int lcg;
  struct sngbr *next;
} NGBR;

typedef struct sinter {
  int ntable,ntype, ncgtype;
  int *itype,**map,*nexclude,**exclude;
  double **vtab,**dvtab,**d2vtab,**switchs;
  double *rminl_cut2,*rmaxl_cut2,*dx2tab;
  double *rmins_cut2,*rmaxs_cut2;
  double *vlrc, *wlrc;
  double *rmaxs_skin2,*rmaxl_skin2,*rminl_skin2;
  double *dxn,*dyn,*dzn;
  double skin_ter,rheal;
  /* electrostatic */
  double *coultab,*dcoultab,*switchse,dx2coul,*coultab_aa,*dcoultab_aa;
  double rmins_coul2,rmaxs_coul2,rminl_coul2,rmaxl_coul2;
  double rmaxs_skin2_e,rmaxl_skin2_e,rminl_skin2_e;
  double *rmax_rblob, *rmax_so3;
#ifdef FIXHYBRIDSWAPS
	int nswptot,nswpmax,*iatswp,*jatswp,*iswp,nswp;
	double *swppot;
#endif	
  int lcg;
  struct sinter *next;
} INTER;

typedef struct sstartup {
  WORD inter_file,bond_file,bend_file,tors_file,onfo_file;
  SPECIES spec_root;
  int i,nspec,kmax,ishift,ivlrc;
  double tau_nhc,tau_vol,alp_ewald;
  double scale_onfo,scale_onfo_e;
  double rcute_max,rcute_min,rcute_resp;
  LINE nvecstore,nvecreal,nvecimag;
} STARTUP;

typedef struct nmodes {
  int nconf,nconf_off,flo,ier,imodes;
  int nspec,*napm,*nmol;
  double maxfreq,minfreq,acpu,df;
  double **fcmat,*dn,*en,*en2,*freq,*fric,*fpos,*ftot;
  double **molmat,**decmod,**permod,*fricmod,*fricmodq,*fricmodcq;
  double **decmodd, **decmodq, **decmoddq,**decmodrr,**decmod_isorr;
  double *aniso_raman,*iso_raman,*raman,*redram,*oke,*iso;
  double *suspx,*suspy,*suspz;
  double *avg_freq,*avg_strn;
  double *part,*fpart;
  double *A_xx1,*A_zz1,*A_xz1,*A_zx1;
  double *A_xx2,*A_zz2,*A_xz2,*A_zx2;
  double *mux1,*muy1,*muz1,*mux2,*muy2,*muz2;
} NMODES;

typedef struct so3 {
  double vzero_ngbr;  /* buffer for energy differences due to neighborlist updates */
  double tau;
  double (*omega)[3], (*tor)[3], (*tor_s)[3], (*mit)[3][3], (*accel)[3];
} SO3; 


#define MAXCLOCKS 32
#define MAXLENNAME 64

typedef struct mytiming {
  int nclocks, iclock[MAXCLOCKS], ncalls[MAXCLOCKS], iset_clock[MAXCLOCKS];
  char clocknames[MAXCLOCKS][MAXLENNAME];
  double timeon[MAXCLOCKS], timing_last[MAXCLOCKS], timing_sum[MAXCLOCKS];
} MYTIMING;

