/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

  Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* subroutines to read in hybrid AA/CG dynamics parameters */


#include "md.h"
#define NKEYS 7
#define REGIONKEY "hybridregions"
#define DEBUG
#define PRINTSTUFF printf("%4d%5d%5d |%5d%5d%8s |%5d%8s   %5d  %5d%12s \n",idim,i,iat,simparms->itype[iat],simparms->iatom[iat],simparms->atom[simparms->itype[iat]],simparms->imolecule[iat],simparms->mole[simparms->ispecies[iat]],simparms->ispecies[iat],simparms->igroup[iat],simparms->group[simparms->igroup[iat]]);
			   
void parse_aasphere(char *line,SIMPARMS *simparms,double ngbrlistskin){
  char *pt;
  
  simparms->itype_hr = 0;
  pt = strtok(line,",");
  sscanf(pt,"%lf",&simparms->rmin_hr);
  pt = strtok(NULL,",");
  sscanf(pt,"%lf",&simparms->rmax_hr);
  pt = strtok(NULL,",");
  if(pt==NULL){
    printf("No healing region skin found; set to half the ngbrlist skin instead\n");
    simparms->rskin_hr = 0.5*ngbrlistskin;
  }else{
    sscanf(pt,"%lf",&simparms->rskin_hr);
  }
  
  if(simparms->rmax_hr <= simparms->rmin_hr){
    simparms->rmax_hr = simparms->rmin_hr;
    md_warning("The healing region radius is smaller than the all atom region radius;\n no healing region used!");
  }
  
  if(simparms->rskin_hr <= 0.){
    md_warning("Setting the HR skin to zero is usually a bad idea!! (good choice would be half the ngbrlist skin");
  }
}

void parse_aaslice(char *line,SIMPARMS *simparms,double ngbrlistskin){
  char *pt;
  int j;
  char dir;
  WORD line2;

  pt = strtok(line,",");
  sscanf(pt,"%c",&dir);
  if((dir!='x')&&(dir!='y')&&(dir!='z')){
    sprintf(line2,"Expected x,y or z in AAslice key, but found %c",dir);
    md_error(line2);
  }
  j= dir - 'x';
  simparms->itype_hr = 1+j;
  pt = strtok(NULL,",");
  sscanf(pt,"%lf",&simparms->rmin_hr);
  pt = strtok(NULL,",");
  sscanf(pt,"%lf",&simparms->rmax_hr);
  if(simparms->rmax_hr <= simparms->rmin_hr){
    simparms->rmax_hr = simparms->rmin_hr;
    sprintf(line2,"The healing region radius is smaller than the all atom region radius;\n no healing region used in %c-direction!",dir);
    md_warning(line2);
  }
  pt = strtok(NULL,",");
  if(pt==NULL){
    printf("No healing region skin found; set to half the ngbrlist skin instead\n");
    simparms->rskin_hr = 0.5*ngbrlistskin;
  }else{
    sscanf(pt,"%lf",&simparms->rskin_hr);
  }
  
  if(simparms->rskin_hr <= 0.){
    md_warning("Setting the HR skin to zero is usually a bad idea!! (good choice would be half the ngbrlist skin");
  }
  
}

void parse_aacylinder(char *line,SIMPARMS *simparms,double ngbrlistskin){
  char *pt;
  int j;
  char dir;
  WORD line2;

  pt = strtok(line,",");
  sscanf(pt,"%c",&dir);
  if((dir!='x')&&(dir!='y')&&(dir!='z')){
    sprintf(line2,"Expected x,y or z in AAslice key, but found %c",dir);
    md_error(line2);
  }
  j= dir - 'x';
  simparms->itype_hr = 5+j;
  pt = strtok(NULL,",");
  sscanf(pt,"%lf",&simparms->rmin_hr);
  pt = strtok(NULL,",");
  sscanf(pt,"%lf",&simparms->rmax_hr);
  if(simparms->rmax_hr <= simparms->rmin_hr){
    simparms->rmax_hr = simparms->rmin_hr;
    sprintf(line2,"The healing region radius is smaller than the all atom region radius;\n no healing region used in %c-direction!",dir);
    md_warning(line2);
  }
  pt = strtok(NULL,",");
  if(pt==NULL){
    printf("No healing region skin found; set to half the ngbrlist skin instead\n");
    simparms->rskin_hr = 0.5*ngbrlistskin;
  }else{
    sscanf(pt,"%lf",&simparms->rskin_hr);
  }
  
  if(simparms->rskin_hr <= 0.){
    md_warning("Setting the HR skin to zero is usually a bad idea!! (good choice would be half the ngbrlist skin");
  }
  
}


int parse_region(SIMPARMS *simparms,char *region,int iflg){
  char *pt;
  int i, iatfirst, iatlast, n, idim, ndim, iat, imol, ion, ioff, natoms;
  
  natoms=0;
  if(iflg==0){
    ion=0; ioff=2;
  }else if(iflg==1){
    ion=2; ioff=0; 
  }else{  /* hybrid switch */
//    simparms->itype_hr = 4;
    simparms->iHybridSwitch = 1;
    pt = strtok(region,",");
    sscanf(pt,"%lf",&simparms->switch_fmin);
    pt = strtok(NULL,",");
    sscanf(pt,"%lf",&simparms->switch_fmax);
    i = simparms->nstep;
    if(simparms->istart==3) i -= simparms->istep;
    simparms->switch_df = (simparms->switch_fmax-simparms->switch_fmin)/(double)(i-1.);
#ifdef DEBUG
    printf("AR+HR selection is to switch hybrid value from %g to %g in %d steps (ds=%g)\n",
	   simparms->switch_fmin,simparms->switch_fmax,i,simparms->switch_df);	
#endif
    if(simparms->istart==3) simparms->switch_fmin -= simparms->switch_df*(simparms->istep-1);
    return simparms->naaregion;
  }
  
  for(iat=0;iat<simparms->ncg;iat++)
    simparms->iaacg[iat] = ioff;
  
  pt = strtok(region,",");
  sscanf(pt,"%d",&ndim);
#ifdef DEBUG
  if(iflg){
    printf("\nSelected CG units in CG region:\n");
  }else{
    printf("\nSelected CG units in AA region:\n");
  }
  printf(" idim iat num  | type atomno  name | molno molname  specno groupno groupname\n");
#endif
  for(idim=0;idim<ndim;++idim){
    pt = strtok(NULL,",");
    if(strcasecmp(pt,"all")==0){
      for(iat=0;iat<simparms->ncg;iat++){
	simparms->iaacg[iat] = ion;
      }
      natoms = simparms->ncg;
#ifdef DEBUG
      printf(" ** ALL CG ATOMS SELECTED (list surpressed) ** \n");
#endif
    }else if(strcasecmp(pt,"atomlist")==0){
      pt = strtok(NULL,",");
      sscanf(pt,"%d",&n);
      for(i=0;i<n;++i){
	pt = strtok(NULL,",");
	sscanf(pt,"%d",&iat);
	simparms->iaacg[iat] = ion;
#ifdef DEBUG
	PRINTSTUFF
#endif
	  }
      natoms += n;
    }else if(strcasecmp(pt,"atomrange")==0){
      pt = strtok(NULL,",");
      sscanf(pt,"%d",&iatfirst);
      pt = strtok(NULL,",");
      sscanf(pt,"%d",&iatlast);
      for(iat=iatfirst;iat<iatlast+1;++iat){
	simparms->iaacg[iat] = ion;
	natoms++;
#ifdef DEBUG
	PRINTSTUFF
#endif
	  }
	}else if(strcasecmp(pt,"mollist")==0){
      pt = strtok(NULL,",");
      sscanf(pt,"%d",&n);
      for(i=0;i<n;++i){
	pt = strtok(NULL,",");
	sscanf(pt,"%d",&imol);
	for(iat=0;iat<simparms->ncg;iat++){
	  if(imol==simparms->imolecule[iat]){
	    simparms->iaacg[iat] = ion;
	    natoms++;
#ifdef DEBUG
	    PRINTSTUFF
#endif
	      }
	}
      }
    }else if(strcasecmp(pt,"none")==0){
#ifdef DEBUG
      printf(" *** NO CG ATOMS SELECTED *** \n");	
#endif
    } else {
      md_error("Please call Steve to purchase a new token, they are sold at 60th and Market at 1am, ask for Pete.");
    }
  }
  
  if(iflg==1) natoms = simparms->ncg - natoms;
  return natoms;
}


void read_hybridfile(SIMPARMS *simparms,char *hybridfile,COORDS *coords,INTER *inter)
{
  int i,j,ich,nline,nch,num_keys,icase;
  
  char line[MAXLINELEN];
  int  num_dict=NKEYS,num_found[NKEYS],ifound;
  WORD metakey,dict[NKEYS],region,region2,aashape;
  KEY key_root,*key_now;
  FILE *fp;
  
  strcpy(dict[0],"AAregion");     strcpy(dict[1],"CGregion");
  strcpy(dict[2],"AAsphere");     strcpy(dict[3],"AAslice");
  strcpy(dict[4],"update");	  strcpy(dict[5],"HybridSwitch");
  strcpy(dict[6],"AAcylinder");
  
  /* iaacg[natoms] is allocated in read_parmfile() and defaulted to AA region=1. */
  
  /* open hybridinputfile */
  if ((fp=fopen(hybridfile,"r"))==NULL){
    sprintf(line,"can't open hybrid file \"%s\"",hybridfile);
    md_error(line);
  } else {
    if(simparms->rank==0){
      sprintf(line,"Reading in hybrid file \"%s\"",hybridfile);
      md_stdout(line);
    }
  }
  /* loop until end of file */
  nline = nch = ifound = 0;
  while( (ich = fgetc(fp)) != EOF) {
    line[nch++] = (char)ich;
    if(ich == META_KEY_CHAR){
      
      get_meta_key(fp,hybridfile,line,&nch,&nline,metakey,&key_root,&num_keys);
      
#ifdef DEBUG
      md_stdout(metakey);
      for(i=0,key_now=&key_root;i<num_keys;i++,key_now = key_now->next){
	sprintf(line,"%d %s %s",i,key_now->keyword,key_now->keyarg);
	md_stdout(line);
      }
#endif
      if(!strcasecmp(metakey,REGIONKEY)){
	
	for(i=0;i<num_dict;i++) num_found[i] = 0;
	
	for(i=0,key_now=&key_root;i<num_keys;i++,key_now = key_now->next){
	  icase = get_dict_num(num_dict,dict,key_now->keyword);
	  if(icase == -1){
	    print_dict(key_now->keyword,num_dict,dict);
	  }
	  num_found[icase]++;
	  
	  switch (icase) {
	  case 0:strcpy(region,key_now->keyarg);  /* all atom region */
	    ifound++;
	    break;
	  case 1:strcpy(region,key_now->keyarg);  /* coarse grain region */
	    ifound++;
	    break;
	  case 2:strcpy(aashape,key_now->keyarg);  /* AA sphere  */
	    break;
	  case 3:strcpy(aashape,key_now->keyarg);  /* AA slice */
	    break;
	  case 4: sscanf(key_now->keyarg,"%d",&simparms->iupdate);  /* update regions */
	    break;
	  case 5:strcpy(region2,key_now->keyarg);  /* hybrid switch */
	    ifound++;
	    break;
	  case 6:strcpy(aashape,key_now->keyarg);  /* AA cylinder */
	    break;
	  default:
	    print_dict(key_now->keyword,num_dict,dict); break;
	  }
	}
	
	/*  Make sure nothing specified more then once */ 
	for(i=0;i<NKEYS;i++){
	  if(num_found[i] > 1) {
	    sprintf(line,"You specified %s %d times",dict[i],num_found[i]);
	    md_error(line);
	  }
	}
	
	/* other checks  */
	if(num_found[0]+num_found[1] != 1){
	  md_error("select one keyword of either AAregion, CGregion to specify region...");
	}
	if((num_found[2]==1)&&(num_found[3]==1)){
	  md_error("select either AAslice or AAsphere to specify region, not both...");
	}
//	if(((num_found[2]==1)||(num_found[3]==1))&&(num_found[5]==1)){
//	  md_error("Dynamic hybrid regions (AAslice or AAsphere) are not suported with HybridSwitch");
//	}
	
	
	/* copy inputs to actually variables */
	if(num_found[0]) simparms->naaregion = parse_region(simparms,region,0);  /* parse AAregion */
	if(num_found[1]) simparms->naaregion = parse_region(simparms,region,1);  /* parse CGregion */
	if(num_found[5]) simparms->naaregion = parse_region(simparms,region2,2);  /* parse hybrid switch */
	if(num_found[2]) parse_aasphere(aashape,simparms,inter->skin_ter);         /* parse AA sphere */
	if(num_found[3]) parse_aaslice(aashape,simparms,inter->skin_ter);          /* parse AA slice  */
	if(num_found[6]) parse_aacylinder(aashape,simparms,inter->skin_ter);          /* parse AA cylinder  */
	
	simparms->iaaregion = (int *)cmalloc(simparms->naaregion*sizeof(int));
	for(i=0,j=0;i<simparms->ncg;++i)
	  if(simparms->iaacg[i]==0) simparms->iaaregion[j++] = i;
	if(j!=simparms->naaregion) md_error("Failed to count number of CG atoms in AA region");
	

	/* some more checks  */
	if((coords->coul.ncharge) && (simparms->iperd) && 
	   ((simparms->itype_hr==1)||(simparms->itype_hr==2)||(simparms->itype_hr==3)) )
	  md_error("The system has charges and you specified a 2D periodic healing region...Please stop doing that...");
	if((coords->coul.ncharge) && (0.5*inter->rmaxl_coul2 < simparms->rmax_hr*simparms->rmax_hr)) 
	  md_error("The system has charges and the cut-off is smaller than the healing region. Increase rcute_max.");
	
	
      }
      free_key(num_keys,&key_root);
    }
    if(ich == NEWLINE){nch = 0; nline++;}
  }
  fclose(fp);
  
  if(ifound == 0){
    sprintf(line,"Failed to find any AA/HR/CG region definitions in file %s",hybridfile);
    md_error(line);
  }
  
}
 
/*--------------------------------------------------------------*/

