/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore
   
Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include "md.h"

/*------------------------------------------------------*/
void UpdateVelocities(COORDS *coords, double dt, int *icgunit, double *fxcg, double *fxaa)
{
  int ncg=coords->natoms;
  int naa=coords->next->natoms;
  
  /* update CG velocities in CG region */
  multiply_selection(coords->vx,dt,fxcg,coords->iselect,coords->nselect-coords->nschizo,ncg);
  
  /* update AA velocities in AA+HR regions */
  coords = coords->next;
  distribute_cgaccel(coords->vx,dt,ncg,naa,icgunit,coords->iselect,coords->nselect,fxcg);
  multiply_selection(coords->vx,dt,fxaa,coords->iselect,coords->nselect,naa);
  coords = coords->next;
}
/*------------------------------------------------------*/
void UpdatePositions(COORDS *coords, double dt, int **icglist,int *napcg)
{
  /* update CG positions in CG region */
  multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
  
  /* update AA posititions in AA+HR regions */
  coords = coords->next;
  multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect,coords->natoms);
  coords = coords->next;
  
  /* recompute c.o.m CG positions in AA+HR regions */
  get_cgpos(coords,icglist,napcg,coords->ninactive+coords->nschizo,coords->ischizo);
}

/*------------------------------------------------------*/

void step_init(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies,
               INTER *inter,NGBR *ngbr,MYTIMING *timing)
{
  /*  solvent - solvent potential cutoff  */
  if(simparms->rank==0){
    md_stdout("Calculating initial forces");
  }
  ngbr->update = -1;
	
  get_ngbr(simparms,coords,inter,ngbr);
  
#ifdef DEBUG
  check_distance(simparms,coords,inter);
#endif
	
  force(simparms,coords,inter,ngbr,timing,0);
  force(simparms,coords,inter,ngbr,timing,2);
  force(simparms,coords,inter,ngbr,timing,3);
	
  if(simparms->rank==0){
    md_stdout("Calculating initial potential energies");
  }
	
  getvireal(simparms,coords,energies,inter,ngbr,(SO3 *)0,1);
  geteng(simparms,coords,energies);
  gettngbr(ngbr);
  if(simparms->ivol) forceV(simparms,coords,energies);
}
/*------------------------------------------------------*/
void step_hybrid_init(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies,
		      INTER *inter,NGBR *ngbr,SO3 *so3,MYTIMING *timing)
{
  int i;
  /*  solvent - solvent potential cutoff  */
  if(simparms->rank==0){
    md_stdout("Calculating initial forces");
  }
	
  if(ngbr->onelist == 0){
    /* make first CG neighborlist and use that for AA neighborlist */
    ngbr->update = ngbr->next->update = -1;
    get_ngbr(simparms,coords->next,inter->next,ngbr->next);
    get_ngbr(simparms,coords,inter,ngbr);
    getsvirterso3(&so3->vzero_ngbr);
    printf(" TEST evoor= %g ",so3->vzero_ngbr);
    so3->vzero_ngbr -= getvnonbond_so3(simparms,coords,inter,so3,ngbr->nprr,ngbr->inpr,ngbr->jnpr,0);
    printf("  evoor-ena = %g\n",so3->vzero_ngbr);
	}else if(ngbr->onelist == 1){
    /* make AA or CG neighborlist */
    ngbr->update = -1;
    get_ngbr(simparms,coords,inter,ngbr);
  }else if(ngbr->onelist == 2){
    /* CG neighborlist has already been setup; do nothing */
  }
  
#ifdef DEBUG
  check_distance(simparms,coords,inter);
#endif
	
  if( (coords->fxl!=NULL) && (coords->fxs!=NULL) ){
    force_hybrid(simparms,coords,inter,ngbr,so3,timing,0);   /* VdW + elec long range and short range (case 0) */
  }
  if(coords->fxr!=NULL) force_hybrid(simparms,coords,inter,ngbr,so3,timing,2);   /* torsions + one-fours */
  if(coords->fxa!=NULL) force_hybrid(simparms,coords,inter,ngbr,so3,timing,3);   /* bonds + bends */
	
  if(simparms->rank==0){
    md_stdout("Calculating initial potential energies");
  }
	
  for(i=0;i<simparms->natoms;i++) simparms->vbuf[i]=0;
  getvireal(simparms,coords,energies,inter,ngbr,so3,1);
  geteng(simparms,coords,energies);
  gettngbr(ngbr);
  if(simparms->ivol) forceV(simparms,coords,energies);
}
/*------------------------------------------------------*/

void step(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies,
          INTER *inter,NGBR *ngbr,MYTIMING *timing)
{
  int i,intr,intra,inrra,n3,ninter,ninter1,natoms;
  double dtl2,dts2,dta2,dtr2,dter,dter1;
  double *pos,*vel,*acel,*acel2;
	
  int ncolvar = coords->colvar.ncolvar;
  natoms = simparms->natoms;
  ninter = simparms->ninter;
  ninter1= ninter-1;
  dtl2 = .5*simparms->dt;
  dts2 = dtl2/(double)simparms->ninter;
  dtr2 = dts2/(double)simparms->ninrra;
  dta2 = dtr2/(double)simparms->ninnra;
  dter = (double)ninter*dts2;
  dter1= (double)(1-ninter)*dts2;
	
  n3 = 3*natoms;
  pos = coords->px;
  vel = coords->vx;
	
  /* get half step */
  if(simparms->ivol) {
#ifdef ISO
    coords->vvol += dtl2*coords->fvol;
#endif
    for(i=0;i<9;i++) coords->vvol9[i] += dtl2*coords->fvol9[i];
  } else {
    coords->vvol = 0.;
    for(i=0;i<9;i++) coords->vvol9[i] = 0.;
  }
  	
  for(intr=0;intr<ninter;intr++){
    /* long and short range intermolecular forces */
    if(intr==0){
      acel  = coords->fxl;  acel2 = coords->fxs;
      for(i=0;i<n3;i++) vel[i] += dter*acel[i]+dter1*acel2[i];
    } else {
      acel= coords->fxs;
      for(i=0;i<n3;i++) vel[i] += dts2*acel[i];
    }
		
    /* intRa force (torsion one-four)*/
    for(inrra=0;inrra<simparms->ninrra;inrra++){
      acel= coords->fxr;
      for(i=0;i<n3;i++) vel[i] += dtr2*acel[i];
			
      /* intra force (bonds bends xbonds) */
      for(intra=0;intra<simparms->ninnra;intra++){
  myclockon(timing,"RESPA INTRA",9);
        acel = coords->fxa;
				
        for(i=0;i<n3;i++) vel[i] += dta2*acel[i];
        for(i=0;i<ncolvar;++i) coords->colvar.vcolvar[i] += dta2*coords->colvar.fcolvar[i];
        for(i=0;i<n3;i++) pos[i] += dta2*vel[i];
				
        for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dta2*coords->colvar.vcolvar[i];
        /* Volume coupled equations */
        if(simparms->ivol) intbar(simparms,coords,dta2,0);
        if(simparms->ntherm) intnhc(simparms,coords,dta2,-1);
        if(simparms->ivol) intbar(simparms,coords,dta2,1);
				
        for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dta2*coords->colvar.vcolvar[i];
        for(i=0;i<n3;i++) pos[i] += dta2*vel[i];
				
        force(simparms,coords,inter,ngbr,timing,3);
        acel = coords->fxa;
        for(i=0;i<n3;i++) vel[i] += dta2*acel[i];
        for(i=0;i<ncolvar;++i) coords->colvar.vcolvar[i] += dta2*coords->colvar.fcolvar[i];
    myclockoff(timing,"RESPA INTRA",9);
      }
      force(simparms,coords,inter,ngbr,timing,2);
      acel= coords->fxr;
      for(i=0;i<n3;i++) vel[i] += dtr2*acel[i];
    }
    nrefsh(simparms,coords,inter,ngbr,0,timing);
		
    if(intr==ninter1){
      force(simparms,coords,inter,ngbr,timing,0);
      acel  = coords->fxl;  acel2 = coords->fxs;
      for(i=0;i<n3;i++) vel[i] += dter*acel[i]+dter1*acel2[i];
    } else {
      force(simparms,coords,inter,ngbr,timing,1);
      acel= coords->fxs;
      for(i=0;i<n3;i++) vel[i] += dts2*acel[i];
    }
  }
  
  /* long range force  */
  getvireal(simparms,coords,energies,inter,ngbr,(SO3 *)0,1);
  if(simparms->ivol){
    forceV(simparms,coords,energies);
#ifdef ISO
    coords->vvol += dtl2*coords->fvol;
#endif
    for(i=0;i<9;i++) coords->vvol9[i] += dtl2*coords->fvol9[i];
  }
  gettngbr(ngbr);
}

/*-----------------------------------------------------------*/

void step_hybrid_cg_aa_paired(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies,
															INTER *inter,NGBR *ngbr,SO3 *so3,MYTIMING *timing)
{
  int i,intr,intra,inrra,ninter,ninter1,*isel;
  double dtl2,dts2,dta2,dtr2,dter,dter1;
	
  int ncolvar = coords->colvar.ncolvar;
  ninter = simparms->ninter;
  ninter1= ninter-1;
  dtl2 = .5*simparms->dt;
  dts2 = dtl2/(double)simparms->ninter;
  dtr2 = dts2/(double)simparms->ninrra;
  dta2 = dtr2/(double)simparms->ninnra;
  dter = (double)ninter*dts2;
  dter1= (double)(1-ninter)*dts2;
	
  /* get half step */
  if(simparms->ivol) {
#ifdef ISO
    coords->vvol += dtl2*coords->fvol;
#endif
    for(i=0;i<9;i++) coords->vvol9[i] += dtl2*coords->fvol9[i];
  } else {
    coords->vvol = 0.;
    for(i=0;i<9;i++) coords->vvol9[i] = 0.;
  }
  
  for(intr=0;intr<ninter;intr++){
    /* long and short range intermolecular forces */
    if(intr==0){
      UpdateVelocities(coords,dter,simparms->icgunit,coords->fxl,coords->next->fxl);			
      UpdateVelocities(coords,dter1,simparms->icgunit,coords->fxs,coords->next->fxs);
//      if(simparms->iso3dyn)
//        UpdateOmega(simparms,so3,coords->nselect-coords->nschizo,coords->iselect,dter);
    } else {
      UpdateVelocities(coords,dts2,simparms->icgunit,coords->fxs,coords->next->fxs);			
//      if(simparms->iso3dyn)
//        UpdateOmega(simparms,so3,coords->nselect-coords->nschizo,coords->iselect,dts2);
    }
    
    /* intRa force (torsion one-four)*/
    for(inrra=0;inrra<simparms->ninrra;inrra++){
      UpdateVelocities(coords,dtr2,simparms->icgunit,coords->fxr,coords->next->fxr);
      
      /* intra force (bonds bends xbonds) */
      for(intra=0;intra<simparms->ninnra;intra++){
  myclockon(timing,"RESPA INTRA",9);
	UpdateVelocities(coords,dta2,simparms->icgunit,coords->fxa,coords->next->fxa);
	if(simparms->iso3dyn)
	  UpdateOmega(simparms,so3,coords->nselect-coords->nschizo,coords->iselect,dta2);
	
	for(i=0;i<ncolvar;++i) coords->colvar.vcolvar[i] += dta2*coords->colvar.fcolvar[i];
	
	UpdatePositions(coords,dta2,simparms->icglist,simparms->napcg);
	for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dta2*coords->colvar.vcolvar[i];
	
	/* Volume coupled equations */
  myclockon(timing,"EXTENDED VARIABLES",22);
	if(simparms->ivol) intbar(simparms,coords,dta2,0);
	if(simparms->ntherm){
	  intnhc(simparms,coords,dta2,-1);
	  intnhc(simparms,coords->next,dta2,-1-coords->ntherm);
	}
	if(simparms->ivol) intbar(simparms,coords,dta2,1);
  myclockoff(timing,"EXTENDED VARIABLES",22);
	
	for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dta2*coords->colvar.vcolvar[i];
	
	UpdatePositions(coords,dta2,simparms->icglist,simparms->napcg);
	if(simparms->iso3dyn){
      myclockon(timing,"SO3 ROTATION",12);
	  UpdateSO3Coords(simparms,coords->next,so3,2.*dta2);				
      myclockoff(timing,"SO3 ROTATION",12);
  }
  
  myclockon(timing,"SET SWITCHFUNCTION",11);
	set_switchfunction(simparms,coords);
  myclockoff(timing,"SET SWITCHFUNCTION",11);
	isel = coords->ischizo;
	for(i=0;i<coords->ninactive+coords->nschizo;i++) simparms->vcgintern[isel[i]]=0.;
		
  myclockon(timing,"ATOMISTIC FORCE BONDS BENDS",14);
	force_hybrid(simparms,coords->next,inter->next,ngbr->next,so3,timing,3);
  myclockoff(timing,"ATOMISTIC FORCE BONDS BENDS",14);
  myclockon(timing,"CG FORCE BONDS BENDS",18);
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,3);
  myclockoff(timing,"CG FORCE BONDS BENDS",18);
  
	if(simparms->iso3dyn){
    myclockon(timing,"SO3 ACCELERATION",13);
	  getAccel(coords->nselect-coords->nschizo,coords->iselect,so3,2.*dta2,0);
    myclockoff(timing,"SO3 ACCELERATION",13);
	  UpdateOmega(simparms,so3,coords->nselect-coords->nschizo,coords->iselect,dta2);
	}
	UpdateVelocities(coords,dta2,simparms->icgunit,coords->fxa,coords->next->fxa);			
	for(i=0;i<ncolvar;++i) coords->colvar.vcolvar[i] += dta2*coords->colvar.fcolvar[i];
    myclockoff(timing,"RESPA INTRA",9);
      }
			
      myclockon(timing,"CG FORCE TORSION ONE-FOUR",19);
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,2);
      myclockoff(timing,"CG FORCE TORSION ONE-FOUR",19);
      myclockon(timing,"ATOMISTIC FORCE TORSION ONE-FOUR",15);
      force_hybrid(simparms,coords->next,inter->next,ngbr->next,so3,timing,2);
      myclockoff(timing,"ATOMISTIC FORCE TORSION ONE-FOUR",15);
      UpdateVelocities(coords,dtr2,simparms->icgunit,coords->fxr,coords->next->fxr);
    }
    nrefsh(simparms,coords,inter,ngbr,so3,timing);
    
    if(intr==ninter1){
      myclockon(timing,"ATOMISTIC FORCE LONG RANGE",17);
      force_hybrid(simparms,coords->next,inter->next,ngbr->next,so3,timing,0);
      myclockoff(timing,"ATOMISTIC FORCE LONG RANGE",17);
      myclockon(timing,"CG FORCE LONG RANGE",21);
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,0);
      myclockoff(timing,"CG FORCE LONG RANGE",21);
      UpdateVelocities(coords,dter,simparms->icgunit,coords->fxl,coords->next->fxl);
      UpdateVelocities(coords,dter1,simparms->icgunit,coords->fxs,coords->next->fxs);
//      if(simparms->iso3dyn){
//        getAccel(coords->nselect-coords->nschizo,coords->iselect,so3,2.*dter,1);
//        UpdateOmega(simparms,so3,coords->nselect-coords->nschizo,coords->iselect,dter);
//      }
    } else {
      myclockon(timing,"ATOMISTIC FORCE SHORT RANGE",16);
      force_hybrid(simparms,coords->next,inter->next,ngbr->next,so3,timing,1);
      myclockoff(timing,"ATOMISTIC FORCE SHORT RANGE",16);
      myclockon(timing,"CG FORCE SHORT RANGE",20);
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,1);
      myclockoff(timing,"CG FORCE SHORT RANGE",20);
      UpdateVelocities(coords,dts2,simparms->icgunit,coords->fxs,coords->next->fxs);
//      if(simparms->iso3dyn){
//        getAccel(coords->nselect-coords->nschizo,coords->iselect,so3,2.*dts2,0);
//        UpdateOmega(simparms,so3,coords->nselect-coords->nschizo,coords->iselect,dts2);
//      }
    }
  }
  
  /* long range force  */
  myclockon(timing,"GET VIREAL",10);
  getvireal(simparms,coords,energies,inter,ngbr,so3,1);
  getvireal(simparms,coords->next,energies->next,inter->next,ngbr->next,so3,1);

  if(simparms->ivol){
    forceV(simparms,coords,energies);
#ifdef ISO
    coords->vvol += dtl2*coords->fvol;
#endif
    for(i=0;i<9;i++) coords->vvol9[i] += dtl2*coords->fvol9[i];
  }
  gettngbr(ngbr);
  gettngbr(ngbr->next);
  myclockoff(timing,"GET VIREAL",10);
}


/*-----------------------------------------------------------*/

void step_hybrid_cg_aa_paired_ORIGBKUP(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies,
			      INTER *inter,NGBR *ngbr,SO3 *so3,MYTIMING *timing)
{
  int i,l,intr,intra,inrra,ninter,ninter1,*isel;
  double dtl2,dts2,dta2,dtr2,dter,dter1;
	
  int ncolvar = coords->colvar.ncolvar;
  ninter = simparms->ninter;
  ninter1= ninter-1;
  dtl2 = .5*simparms->dt;
  dts2 = dtl2/(double)simparms->ninter;
  dtr2 = dts2/(double)simparms->ninrra;
  dta2 = dtr2/(double)simparms->ninnra;
  dter = (double)ninter*dts2;
  dter1= (double)(1-ninter)*dts2;
  l=16; /*debug atom*/
	
  /* get half step */
  if(simparms->ivol) {
#ifdef ISO
    coords->vvol += dtl2*coords->fvol;
#endif
    for(i=0;i<9;i++) coords->vvol9[i] += dtl2*coords->fvol9[i];
  } else {
    coords->vvol = 0.;
    for(i=0;i<9;i++) coords->vvol9[i] = 0.;
  }
  
  for(intr=0;intr<ninter;intr++){
    /* long and short range intermolecular forces */
    if(intr==0){
#ifdef DEBUG2
      printf(" %d %d %lf  %lf %lf %lf | %lf %lf %lf\n",0,intr,dter,coords->px[l],coords->vx[l],coords->fxl[l],coords->next->px[l],coords->next->vx[l],coords->next->fxl[l]);
      printf("+%d %d %lf  %lf %lf %lf | %lf %lf %lf\n",0,intr,dter1,coords->px[l],coords->vx[l],coords->fxs[l],coords->next->px[l],coords->next->vx[l],coords->next->fxs[l]);
#endif
      multiply_selection(coords->vx,dter,coords->fxl,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dter,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,coords->fxl);
      multiply_selection(coords->vx,dter1,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dter1,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,coords->fxs);
      coords = coords->next;
      multiply_selection(coords->vx,dter,coords->fxl,coords->iselect,coords->nselect,coords->natoms);
      multiply_selection(coords->vx,dter1,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next;
    } else {
#ifdef DEBUG2
      printf(" %d %d %lf  %lf %lf %lf | %lf %lf %lf\n",1,intr,dter,coords->px[l],coords->vx[l],coords->fxs[l],coords->next->px[l],coords->next->vx[l],coords->next->fxs[l]);
#endif
      multiply_selection(coords->vx,dts2,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dts2,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,coords->fxs);
      coords = coords->next;
      multiply_selection(coords->vx,dts2,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next;
    }
		
    /* intRa force (torsion one-four)*/
    for(inrra=0;inrra<simparms->ninrra;inrra++){
#ifdef DEBUG2
      printf(" %d %d %lf  %lf %lf %lf | %lf %lf %lf\n",2,inrra,dtr2,coords->px[l],coords->vx[l],coords->fxr[l],coords->next->px[l],coords->next->vx[l],coords->next->fxr[l]);
#endif
      multiply_selection(coords->vx,dtr2,coords->fxr,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dtr2,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,coords->fxr);
      coords = coords->next;
      multiply_selection(coords->vx,dtr2,coords->fxr,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next;
			
      /* intra force (bonds bends xbonds) */
      for(intra=0;intra<simparms->ninnra;intra++){
#ifdef DEBUG2
	printf(" %d %d %lf  %lf %lf %lf | %lf %lf %lf\n",3,intra,dta2,coords->px[l],coords->vx[l],coords->fxa[l],coords->next->px[l],coords->next->vx[l],coords->next->fxa[l]);
#endif
	multiply_selection(coords->vx,dta2,coords->fxa,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dta2,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxa);
	coords = coords->next;
	multiply_selection(coords->vx,dta2,coords->fxa,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next;
				
	for(i=0;i<ncolvar;++i) coords->colvar.vcolvar[i] += dta2*coords->colvar.fcolvar[i];
	
	multiply_selection(coords->px,dta2,coords->vx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	coords = coords->next;
	multiply_selection(coords->px,dta2,coords->vx,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next;
	get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive+coords->nschizo,coords->ischizo);
#ifdef DEBUG2
	printf("P%d %d %lf  %lf %lf %lf | %lf %lf %lf\n",3,intra,dta2,coords->px[l],coords->vx[l],coords->fxa[l],coords->next->px[l],coords->next->vx[l],coords->next->fxa[l]);
#endif
	for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dta2*coords->colvar.vcolvar[i];
				
	/* Volume coupled equations */
	if(simparms->ivol) intbar(simparms,coords,dta2,0);
	if(simparms->ntherm){
	  intnhc(simparms,coords,dta2,-1);
	  intnhc(simparms,coords->next,dta2,-1-coords->ntherm);
	}
	if(simparms->ivol) intbar(simparms,coords,dta2,1);
				
	for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dta2*coords->colvar.vcolvar[i];
				
	coords = coords->next;
	multiply_selection(coords->px,dta2,coords->vx,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next;
	multiply_selection(coords->px,dta2,coords->vx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive+coords->nschizo,coords->ischizo);
				
	set_switchfunction(simparms,coords);
	isel = coords->ischizo;
	for(i=0;i<coords->ninactive+coords->nschizo;i++) simparms->vcgintern[isel[i]]=0.;
				
#ifdef DEBUG2
	printf(">%d %d %lf  %lf %lf %lf | %lf %lf %lf\n",3,intra,dta2,coords->px[l],coords->vx[l],coords->fxa[l],coords->next->px[l],coords->next->vx[l],coords->next->fxa[l]);
#endif
	coords = coords->next; inter = inter->next; ngbr = ngbr->next;
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,3);
	multiply_selection(coords->vx,dta2,coords->fxa,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next; inter = inter->next; ngbr = ngbr->next;
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,3);  /*moved */
	multiply_selection(coords->vx,dta2,coords->fxa,coords->iselect,coords->nselect-coords->nschizo,coords->natoms); /* moved */
	distribute_cgaccel(coords->next->vx,dta2,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxa);
	for(i=0;i<ncolvar;++i) coords->colvar.vcolvar[i] += dta2*coords->colvar.fcolvar[i];
      }
			
#ifdef DEBUG2
      printf(">%d %d %lf  %lf %lf %lf | %lf %lf %lf\n",2,inrra,dtr2,coords->px[l],coords->vx[l],coords->fxr[l],coords->next->px[l],coords->next->vx[l],coords->next->fxr[l]);
#endif
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,2);
      multiply_selection(coords->vx,dtr2,coords->fxr,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,2); /* moved */
      multiply_selection(coords->vx,dtr2,coords->fxr,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dtr2,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,coords->fxr);
			
    }
    nrefsh(simparms,coords,inter,ngbr,so3,timing);
    
    if(intr==ninter1){
#ifdef DEBUG2
			
      printf(">%d %d %lf  %lf %lf %lf | %lf %lf %lf\n",0,intr,dter,coords->px[l],coords->vx[l],coords->fxl[l],coords->next->px[l],coords->next->vx[l],coords->next->fxl[l]);
      printf("+%d %d %lf  %lf %lf %lf | %lf %lf %lf\n",0,intr,dter1,coords->px[l],coords->vx[l],coords->fxs[l],coords->next->px[l],coords->next->vx[l],coords->next->fxs[l]);
#endif
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,0);
      multiply_selection(coords->vx,dter,coords->fxl,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dter,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,coords->fxl);
      multiply_selection(coords->vx,dter1,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dter1,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,coords->fxs);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next; 
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,0);
      multiply_selection(coords->vx,dter,coords->fxl,coords->iselect,coords->nselect,coords->natoms);
      multiply_selection(coords->vx,dter1,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next; 
    } else {
#ifdef DEBUG2
      printf(">%d %d %lf  %lf %lf %lf | %lf %lf %lf\n",1,intr,dts2,coords->px[l],coords->vx[l],coords->fxs[l],coords->next->px[l],coords->next->vx[l],coords->next->fxs[l]);
#endif
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,1);
      multiply_selection(coords->vx,dts2,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dts2,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,coords->fxs);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,1);
      multiply_selection(coords->vx,dts2,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
    }
  }
	
  /* long range force  */
  getvireal(simparms,coords,energies,inter,ngbr,so3,1);
  getvireal(simparms,coords->next,energies->next,inter->next,ngbr->next,so3,1);
  if(simparms->ivol){
    forceV(simparms,coords,energies);
#ifdef ISO
    coords->vvol += dtl2*coords->fvol;
#endif
    for(i=0;i<9;i++) coords->vvol9[i] += dtl2*coords->fvol9[i];
  }
  gettngbr(ngbr);
  gettngbr(ngbr->next);
}
 
/*-----------------------------------------------------------*/
/*  AA and CG long+short range interaction are same RESPA level */
/*-----------------------------------------------------------*/
/* perform RESPA multi-timestep                              */
/* nres_iorder[ilevel] keeps the quantity to be updated:     */
/*  nres_iorder[ilevel]  quantity                            */
/*      2       CG long range intermolecular interactions    */
/*              AA long range intermolecular interactions    */
/*              CG short range intermolecular interactions   */
/*              AA short range intermolecular interactions   */
/*															 */
/*      4       CG torsions + one-fours                      */
/*      5       AA torsions + one-fours                      */
/*      6       CG bonds, bends, crossbonds                  */
/*      7       AA bonds, bends, crossbonds                  */
/*      8       Nose Hoover variables                        */
/*      9       Barostat variables                           */
/*     10       Collective variables                         */
void step_hybrid_fixed_inter(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies,
			     INTER *inter,NGBR *ngbr,int ilevel,SO3 *so3,MYTIMING *timing)
{
  int i;
  int istep, nsteps, iorder, iforce, *isel;
  double dt,dt1,dt2, *fx;
  int ncolvar = coords->colvar.ncolvar; 
   
  ilevel++;
	
  iorder = simparms->nres_iorder[ilevel];
  dt = simparms->nres_dt[ilevel];
  nsteps = simparms->nres_nsteps[ilevel];
  dt1 = dt*nsteps;
  dt2 = dt*(1-nsteps);
  fx = coords->fx[ilevel];
	
  if(ilevel==0){
    if(simparms->ivol) {
      for(i=0;i<9;i++) coords->vvol9[i] += dt*coords->fvol9[i];
    } else {
      coords->vvol = 0.;
      for(i=0;i<9;i++) coords->vvol9[i] = 0.;
    }
  }
	
	
  for(istep=0;istep<nsteps;istep++){

#ifdef DEBUG2
    printf(" %d %d %d (%lf %lf) %lf %lf (%lf %lf) | %lf %lf (%lf)\n", ilevel,iorder,istep,dt,dt1,coords->px[l],coords->vx[l],fx[l],coords->fxl[l],coords->next->px[l],coords->next->vx[l],coords->next->fxl[l]);
    printf(" %d %d %d (%lf %lf) %lf %lf (%lf %lf) | %lf %lf (%lf)\n", ilevel,iorder,istep,dt,dt2,coords->px[l],coords->vx[l],fx[l],coords->fxs[l],coords->next->px[l],coords->next->vx[l],coords->next->fxs[l]);
#endif
		
    /* update first halve of velocities */
		
    switch(iorder){
    case 0:  /* CG long range inter     */
      md_error("No CG seperate long range");
      break;
    case 2:  /* CG short range inter    */
      if(istep==0){
	multiply_selection(coords->vx,dt1,coords->fxl,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt1,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxl);
	multiply_selection(coords->vx,dt2,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt2,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxs);
	coords = coords->next;
	multiply_selection(coords->vx,dt1,coords->fxl,coords->iselect,coords->nselect,coords->natoms);
	multiply_selection(coords->vx,dt2,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next;
      }else{
	multiply_selection(coords->vx,dt,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxs);
	coords = coords->next;
	multiply_selection(coords->vx,dt,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next;
      }
      break;
    case 4:  /* CG torsions + one-fours */
    case 6:  /* CG bonds + bends        */
      multiply_selection(coords->vx,dt,fx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      break;
    case 1:  /* AA long range inter     */
      md_error("No seperate AA long range");
      break;
    case 3:  /* AA short range inter    */
      md_error("No seperate AA short range");
    case 5:  /* AA torsions + one-fours */
    case 7:  /* AA bonds + bends        */
      coords = coords->next;
      multiply_selection(coords->vx,dt,fx,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next;
      break;
    case 8:   /* Nose Hoover                        */
      break;
    case 9:  /* Barostat                           */
      break;
    case 10:  /* Metadynamics collective variables  */
      for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dt*coords->colvar.vcolvar[i];
      break;
    }
		
    /* update first halve of positions or step one RESPA level up */
		
    if(ilevel == simparms->nres_maxcglevel){
      multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive+coords->nschizo,coords->ischizo);
    }else if(ilevel == simparms->nres_maxaalevel){
      coords = coords->next;
      multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next;
    }
				
    if(ilevel == simparms->nres_maxlevel){
      /* Volume coupled equations */
      if(simparms->ivol) intbar(simparms,coords,dt,0);
      if(simparms->ntherm){
	intnhc(simparms,coords,dt,-1);	
	intnhc(simparms,coords->next,dt,-1-coords->ntherm);	
      }
      if(simparms->ivol) intbar(simparms,coords,dt,1);
    }else{
      step_hybrid_fixed_inter(simparms,coords,energies,inter,ngbr,ilevel,so3,timing);
    }
		
    /* update second halve of positions */	
    if(ilevel == simparms->nres_maxcglevel){
      multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive+coords->nschizo,coords->ischizo);
      set_switchfunction(simparms,coords);
    }else if(ilevel == simparms->nres_maxaalevel){
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      isel = coords->ischizo;
      for(i=0;i<coords->ninactive+coords->nschizo;i++) simparms->vcgintern[isel[i]]=0.;
    }
		
#ifdef DEBUG2
    printf(" %d %d %d (%lf %lf) %lf %lf (%lf %lf) | %lf %lf (%lf)\n", ilevel,iorder,istep,dt,dt1,coords->px[l],coords->vx[l],fx[l],coords->fxl[l],coords->next->px[l],coords->next->vx[l],coords->next->fxl[l]);
    printf(" %d %d %d (%lf %lf) %lf %lf (%lf %lf) | %lf %lf (%lf)\n", ilevel,iorder,istep,dt,dt2,coords->px[l],coords->vx[l],fx[l],coords->fxs[l],coords->next->px[l],coords->next->vx[l],coords->next->fxs[l]);
#endif
    /* compute forces and update second halve of velocities */
    iforce = iorder/2;
    switch(iorder){
    case 2:  /* CG short range intermolecular */
      nrefsh(simparms,coords,inter,ngbr,so3,timing);
      if(istep==nsteps-1){
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,0);
	multiply_selection(coords->vx,dt1,coords->fxl,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt1,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxl);
	multiply_selection(coords->vx,dt2,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt2,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxs);
	coords = coords->next; inter = inter->next; ngbr = ngbr->next; 
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,0);
	multiply_selection(coords->vx,dt1,coords->fxl,coords->iselect,coords->nselect,coords->natoms);
	multiply_selection(coords->vx,dt2,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      }else{
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,1);
	multiply_selection(coords->vx,dt,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxs);
	coords = coords->next; inter = inter->next; ngbr = ngbr->next;
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,1);
	multiply_selection(coords->vx,dt,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next; inter = inter->next; ngbr = ngbr->next;		  
      }
      break;
    case 0:  /* CG long range intermolecular */
      md_error("No CG seperate long range");
    case 4:  /* CG torsions, 1-4           */
    case 6:  /* CG bonds, bends */
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,iforce);
      multiply_selection(coords->vx,dt,fx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      break;	
    case 3:  /* AA short range intermolecular */
      md_error("No AA seperate short range");
      break;	
    case 1:  /* AA long range intermolecular */
      md_error("No AA seperate long range");
      break;	
    case 5:  /* AA torsions, 1-4           */
    case 7:  /* AA bonds, bends, colvar */
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,iforce);
      multiply_selection(coords->vx,dt,fx,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      break;	
    case 8:   /* Nose Hoover                        */
      break;
    case 9:   /* Barostat                           */
      break;
    case 10:  /* Metadynamics collective variables  */
      for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dt*coords->colvar.vcolvar[i];
      for(i=0;i<ncolvar;++i) coords->colvar.vcolvar[i] += dt*coords->colvar.fcolvar[i];
      break;
    }
  }
	
  if(ilevel==0){  
    getvireal(simparms,coords,energies,inter,ngbr,so3,1);
    getvireal(simparms,coords->next,energies->next,inter->next,ngbr->next,so3,1);
    if(simparms->ivol){
      forceV(simparms,coords,energies);
      for(i=0;i<9;i++) coords->vvol9[i] += dt*coords->fvol9[i];
    }
    gettngbr(ngbr);
    gettngbr(ngbr->next);
  }
}
/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/
/* perform RESPA multi-timestep                              */
/* nres_iorder[ilevel] keeps the quantity to be updated:     */
/*  nres_iorder[ilevel]  quantity                            */
/*      0       CG long range intermolecular interactions    */
/*      1       AA long range intermolecular interactions    */
/*      2       CG short range intermolecular interactions   */
/*      3       AA short range intermolecular interactions   */
/*															 */
/*      4       CG torsions + one-fours                      */
/*      5       AA torsions + one-fours                      */
/*      6       CG bonds, bends, crossbonds                  */
/*      7       AA bonds, bends, crossbonds                  */
/*      8       Nose Hoover variables                        */
/*      9       Barostat variables                           */
/*     10       Collective variables                         */
void step_hybrid(SIMPARMS *simparms,COORDS *coords,ENERGIES *energies,
		 INTER *inter,NGBR *ngbr,int ilevel,SO3 *so3,MYTIMING *timing)
{
  int i;
  int istep, nsteps, iorder, iforce, *isel;
  double dt,dt1,dt2, *fx;
  int ncolvar = coords->colvar.ncolvar;
  
  ilevel++;
	
  iorder = simparms->nres_iorder[ilevel];
  dt = simparms->nres_dt[ilevel];
  nsteps = simparms->nres_nsteps[ilevel];
  dt1 = dt*nsteps;
  dt2 = dt*(1-nsteps);
  fx = coords->fx[ilevel];
	
  if(ilevel==0){
    if(simparms->ivol) {
      for(i=0;i<9;i++) coords->vvol9[i] += dt*coords->fvol9[i];
    } else {
      coords->vvol = 0.;
      for(i=0;i<9;i++) coords->vvol9[i] = 0.;
    }
  }
	
	
  for(istep=0;istep<nsteps;istep++){
#ifdef DEBUG2
    printf(" %d %d %d (%lf %lf) %lf %lf (%lf %lf) | %lf %lf (%lf)\n", ilevel,iorder,istep,dt,dt1,coords->px[l],coords->vx[l],fx[l],coords->fxl[l],coords->next->px[l],coords->next->vx[l],coords->next->fxl[l]);
    printf(" %d %d %d (%lf %lf) %lf %lf (%lf %lf) | %lf %lf (%lf)\n", ilevel,iorder,istep,dt,dt2,coords->px[l],coords->vx[l],fx[l],coords->fxs[l],coords->next->px[l],coords->next->vx[l],coords->next->fxs[l]);
#endif
		
    /* update first halve of velocities */
		
    switch(iorder){
    case 0:  /* CG long range inter     */
      md_error("No CG seperate long range");
      break;
    case 2:  /* CG short range inter    */
      if(istep==0){
	multiply_selection(coords->vx,dt1,coords->fxl,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt1,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxl);
	multiply_selection(coords->vx,dt2,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt2,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxs);
      }else{
	multiply_selection(coords->vx,dt,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxs);
      }
      break;
    case 4:  /* CG torsions + one-fours */
    case 6:  /* CG bonds + bends        */
      multiply_selection(coords->vx,dt,fx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dt,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,fx);
      break;
    case 1:  /* AA long range inter     */
      md_error("No seperate AA long range");
      break;
    case 3:  /* AA short range inter    */
      if(istep==0){
	coords = coords->next;
	multiply_selection(coords->vx,dt1,coords->fxl,coords->iselect,coords->nselect,coords->natoms);
	multiply_selection(coords->vx,dt2,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next;
      }else{
	coords = coords->next;
	multiply_selection(coords->vx,dt,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next;
      }
      break;
    case 5:  /* AA torsions + one-fours */
    case 7:  /* AA bonds + bends        */
      coords = coords->next;
      multiply_selection(coords->vx,dt,fx,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next;
      break;
    case 8:   /* Nose Hoover                        */
      break;
    case 9:  /* Barostat                           */
      break;
    case 10:  /* Metadynamics collective variables  */
      for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dt*coords->colvar.vcolvar[i];
      break;
    }
		
    /* update first halve of positions or step one RESPA level up */
		
    if(ilevel == simparms->nres_maxcglevel){
      multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive+coords->nschizo,coords->ischizo);
    }else if(ilevel == simparms->nres_maxaalevel){
      coords = coords->next;
      multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next;
    }
		
    if(ilevel == simparms->nres_maxlevel){
      /* Volume coupled equations */
      if(simparms->ivol) intbar(simparms,coords,dt,0);
      if(simparms->ntherm){
	intnhc(simparms,coords,dt,-1);	
	intnhc(simparms,coords->next,dt,-1-coords->ntherm);	
      }
      if(simparms->ivol) intbar(simparms,coords,dt,1);
    }else{
      step_hybrid(simparms,coords,energies,inter,ngbr,ilevel,so3,timing);
    }
		
    /* update second halve of positions */	
    if(ilevel == simparms->nres_maxcglevel){
      multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive+coords->nschizo,coords->ischizo);
      set_switchfunction(simparms,coords);
    }else if(ilevel == simparms->nres_maxaalevel){
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      multiply_selection(coords->px,dt,coords->vx,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      isel = coords->ischizo;
      for(i=0;i<coords->ninactive+coords->nschizo;i++) simparms->vcgintern[isel[i]]=0.;
    }
		
#ifdef DEBUG2
    printf(" %d %d %d (%lf %lf) %lf %lf (%lf %lf) | %lf %lf (%lf)\n", ilevel,iorder,istep,dt,dt1,coords->px[l],coords->vx[l],fx[l],coords->fxl[l],coords->next->px[l],coords->next->vx[l],coords->next->fxl[l]);
    printf(" %d %d %d (%lf %lf) %lf %lf (%lf %lf) | %lf %lf (%lf)\n", ilevel,iorder,istep,dt,dt2,coords->px[l],coords->vx[l],fx[l],coords->fxs[l],coords->next->px[l],coords->next->vx[l],coords->next->fxs[l]);
#endif
    /* compute forces and update second halve of velocities */
    iforce = iorder/2;
    switch(iorder){
    case 2:  /* CG short range intermolecular */
      nrefsh(simparms,coords,inter,ngbr,so3,timing);
      if(istep==nsteps-1){
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,0);
	multiply_selection(coords->vx,dt1,coords->fxl,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt1,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxl); 
	multiply_selection(coords->vx,dt2,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt2,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxs);
      }else{
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,1);
	multiply_selection(coords->vx,dt,coords->fxs,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
	distribute_cgaccel(coords->next->vx,dt,simparms->ncg,simparms->naa,simparms->icgunit,
			   coords->next->iselect,coords->next->nselect,coords->fxs);
      }
      break;
    case 0:  /* CG long range intermolecular */
      md_error("No CG seperate long range");
    case 4:  /* CG torsions, 1-4           */
    case 6:  /* CG bonds, bends */
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,iforce);
      multiply_selection(coords->vx,dt,fx,coords->iselect,coords->nselect-coords->nschizo,coords->natoms);
      distribute_cgaccel(coords->next->vx,dt,simparms->ncg,simparms->naa,simparms->icgunit,
			 coords->next->iselect,coords->next->nselect,fx);
      break;	
    case 3:  /* AA short range intermolecular */
      nrefsh(simparms,coords,inter,ngbr,so3,timing); 
      if(istep==nsteps-1){
	coords = coords->next; inter = inter->next; ngbr = ngbr->next; 
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,0);
	multiply_selection(coords->vx,dt1,coords->fxl,coords->iselect,coords->nselect,coords->natoms);
	multiply_selection(coords->vx,dt2,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      }else{
	coords = coords->next; inter = inter->next; ngbr = ngbr->next;
	force_hybrid(simparms,coords,inter,ngbr,so3,timing,1);
	multiply_selection(coords->vx,dt,coords->fxs,coords->iselect,coords->nselect,coords->natoms);
	coords = coords->next; inter = inter->next; ngbr = ngbr->next;		  
      }
      break;
    case 1:  /* AA long range intermolecular */
      md_error("No AA seperate long range");
    case 5:  /* AA torsions, 1-4           */
    case 7:  /* AA bonds, bends, colvar */
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      force_hybrid(simparms,coords,inter,ngbr,so3,timing,iforce);
      multiply_selection(coords->vx,dt,fx,coords->iselect,coords->nselect,coords->natoms);
      coords = coords->next; inter = inter->next; ngbr = ngbr->next;
      break;	
    case 8:   /* Nose Hoover                        */
      break;
    case 9:   /* Barostat                           */
      break;
    case 10:  /* Metadynamics collective variables  */
      for(i=0;i<ncolvar;++i) coords->colvar.pcolvar[i] += dt*coords->colvar.vcolvar[i];
      for(i=0;i<ncolvar;++i) coords->colvar.vcolvar[i] += dt*coords->colvar.fcolvar[i];
      break;
    }
  }
	
  if(ilevel==0){  
    getvireal(simparms,coords,energies,inter,ngbr,so3,1);
    getvireal(simparms,coords->next,energies->next,inter->next,ngbr->next,so3,1);
    if(simparms->ivol){
      forceV(simparms,coords,energies);
      for(i=0;i<9;i++) coords->vvol9[i] += dt*coords->fvol9[i];
    }
    gettngbr(ngbr);
    gettngbr(ngbr->next);
  }
}
/*-----------------------------------------------------------*/
