	/* 
Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore
 
Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* controling subroutine for molecular dynamcis 
   using nose thermostats (NVT) and extened system variables
   for the volume
*/

#include "md.h"

/* #define USER_FUNC */

/* might get different numbers from the eigensolver */
/* ADDED BY BERND 16 Aug 2004: rotation messes with torsion_switches */
/* BTW, seems that the rotation also transforms structure into mirror image ....*/
/* ...which is cheaply fixed by calling rotate twice :-) */
#define NO_ROTATE 

#define __INIT_OMEGA_TO_ZERO /* if not defined: use atomistic velocities to init so3->omega */

/*------------------------------------------------------------------*/
void hybridcntr(char *command,FILENAMES *filenames,SIMPARMS *simparms,
		WRITE_STEP *write_step,COORDS *coords,INTER *inter,
		NGBR *ngbr,SO3 *so3, MYTIMING *timing)
{
  int exitall=0,exitsum=0;
  double acttime,oldtime,difftime,starttime;
  ENERGIES *energies;
	
  energies = (ENERGIES *)cmalloc(sizeof(ENERGIES));
  allocate_hybrid_lists(simparms);
  split_energies_in_two(energies);
  split_ngbr_in_two(simparms->ncg,simparms->naa,ngbr);
  split_coords_in_two(simparms->ncg,simparms->naa,simparms->icgunit,coords,simparms->nres_iorder,
		      simparms->nres_maxlevel,simparms->nres_type,simparms->ntherm);
  split_inter_in_two(simparms->ncg,simparms->naa,simparms->ntypes,simparms->ncgtypes,inter,simparms->nlen);
	
  /* allocate SO3 dynamics */
  so3->tau = simparms->dt/(double)(simparms->ninter*simparms->ninrra*simparms->ninnra);
  so3->omega = (double (*)[3])malloc(simparms->ncg*sizeof(double [3]));
  so3->accel = (double (*)[3])malloc(simparms->ncg*sizeof(double [3]));
  so3->tor   = (double (*)[3])malloc(simparms->ncg*sizeof(double [3]));
  so3->tor_s = (double (*)[3])malloc(simparms->ncg*sizeof(double [3]));
  so3->mit   = (double (*)[3][3])malloc(simparms->ncg*sizeof(double [3][3]));
/*  so3->mitinv = (double (*)[3][3])malloc(simparms->ncg*sizeof(double [3][3]));	*/
  
  if(simparms->rank==0){
    readpos(command,filenames->initfile,simparms,
	    &energies->estep,energies,coords,so3);
    read_hybridfile(simparms,filenames->hybridinputfile,coords,inter);
    select_hybridregions(simparms,coords,inter,ngbr,so3,1,1);
    
    if(write_step->ncolv){
      colvar_init(coords->colvar.ncolvar,&(coords->colvar));    
      read_colvarfile(simparms,filenames->colvarcntrfile,coords);
    }
    if(simparms->istart == 1){ /* new start */
      samvel(simparms,coords->next,1);
      samveta(simparms,coords);
      get_cgvel(coords,simparms->icglist,simparms->napcg,coords->natoms,coords->iselect);
      if(simparms->ivol) samvbar(simparms,coords->next);
    }
    if(write_step->ncolv && simparms->istart == 3){ /* read in previous hill run */      
      read_hills(filenames->hillfile,simparms->istep,simparms->dt,&coords->colvar);
    }
    
    if(simparms->iperd==0 && simparms->istart!=3){
      get_relat_aapos(coords,simparms->icglist,simparms->napcg,coords->iselect,
		      coords->nselect-coords->nschizo);
      get_relat_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,
		      coords->nselect-coords->nschizo);
#ifndef NO_ROTATE 
      rotate(simparms->naa,coords->px,coords->py,coords->pz,
	     coords->amass);
#endif
      /* zero ang. + trans. momenta for whole CG system */
      zero_angm(simparms->ncg,coords->px,coords->py,coords->pz,
		coords->vx,coords->vy,coords->vz,coords->amass);
      
      /* rescale velocities */
      if(simparms->istart==1) samvel(simparms,coords,0);
      get_abs_aapos(coords,simparms->icglist,simparms->napcg,coords->iselect, 
		    coords->nselect-coords->nschizo);
      get_abs_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,
		    coords->nselect-coords->nschizo);
    }
    get_relat_aapos(coords,simparms->icglist,simparms->napcg,coords->iselect,
		    coords->nselect-coords->nschizo);
    get_relat_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,
		    coords->nselect-coords->nschizo);
    
    zero_averages(simparms,energies);
    zero_averages(simparms,energies->next);
    
  }
  /* broadcast positions and velocities, masses and charge 
     use the fact that they are contiguous in memory */
#ifdef PARA
  bcast_system(simparms,coords);
  bcast_colvar(simparms,coords);
#endif
  
  gethinv9(coords->hmat,coords->hmati);
  gethinv9(coords->next->hmat,coords->next->hmati);
  
  if(!simparms->ivol){
    if(simparms->nbar > 1){
      md_error("you can't barostat a constant volume!\n \tChange either ivol or nbar in the control file.\n");
      exit(1);
    }
  }
  
  /* setup integrator and get energies */
  energies->acpu = 0.;
  energies->acpu = cputime();
  
  get_cgpot_intern(simparms,coords->next,inter->next);
  
  step_hybrid_init(simparms,coords->next,energies->next,inter->next,ngbr->next,so3,timing);
  step_hybrid_init(simparms,coords,energies,inter,ngbr,so3,timing);
  
  if( simparms->iso3dyn ){
//    getOmegaInit(simparms,coords->next,so3);  /* moved to readpos() */
    getAlphaInit(so3,coords->nselect-coords->nschizo,coords->iselect);
  }

  if(simparms->iunits==1){
    getengn_cheese(simparms,coords,&energies->zken,&energies->potn);
    getengv_cheese(simparms,coords,&energies->zkev,&energies->potv,
		   &energies->zken);
  } else {
    getengn(simparms,coords,&energies->zken,&energies->potn);
    getengv(simparms,coords,&energies->zkev,&energies->potv);
  }
  
  /* output stuff the user might want to know */
  if(simparms->rank==0){
    output_simparm(simparms,write_step,simparms->istep);
    
    output_initval(simparms,coords,ngbr,energies);
    
    if(check_vals(9,energies->potra,energies->poter,energies->zke,
		  energies->potn,energies->zken,energies->potv,energies->zkev,
		  energies->tiout,energies->prsi)){
      free_all_pointers(simparms,coords,inter,ngbr);
      exit(1);
    }
    
    set_econi(energies);        /* sum up energies->ham*/
    set_econi(energies->next);
  }
  
#ifdef USER_FUNC
  md_stdout("User Function defined");
  user_initial(simparms->istep,simparms->natoms,simparms->dt,
	       coords->px,coords->py,coords->pz,
	       coords->vx,coords->vy,coords->vz,
	       coords->amass,coords->qch,coords->hmat);
#endif
  
  /* initial energies */
  energies->acpu = cputime();
  simparms->fstep=simparms->istep;
  if(simparms->istart == 3) simparms->istep++;
  if(simparms->rank==0){
    writescreen(simparms,ngbr,energies,coords->hmat,coords->natoms);
    writescreen(simparms,ngbr->next,energies->next,coords->hmat,coords->next->natoms);
    /* **************  Commence MD Steps ******************** */
    md_stdout("\n ***** Begining Molecular Dynamics Steps *****");
  }
  energies->bcpu = cputime();
  energies->acpu += energies->bcpu; 
  oldtime=starttime=realtime();
  myclockoff(timing,"INITIALIZATION",1);
  
  for(;simparms->istep<=simparms->nstep;simparms->istep++){
    
    myclockon(timing,"MD STEP",2);
    
    if(simparms->nres_type==1){
      step_hybrid(simparms,coords,energies,inter,ngbr,-1,so3,timing); 
    }else if(simparms->nres_type==0){
      step_hybrid_cg_aa_paired(simparms,coords,energies,inter,ngbr,so3,timing);
    }else if(simparms->nres_type==2){
      step_hybrid_fixed_inter(simparms,coords,energies,inter,ngbr,-1,so3,timing);
    }else{
      md_error("Non-existing RESPA type option. Game over.");
    }

    myclockoff(timing,"MD STEP",2);
    myclockon(timing,"WRITE DATA",8);
    
    geteng(simparms,coords,energies);
    geteng(simparms,coords->next,energies->next);
    gettngbr(ngbr);
    gettngbr(ngbr->next);
    
    if(simparms->iunits==1){
      getengn_cheese(simparms,coords,&energies->zken,&energies->potn);
      getengv_cheese(simparms,coords,&energies->zkev,&energies->potv,
		     &energies->zken);
    } else {
      getengn(simparms,coords,&energies->zken,&energies->potn);
      getengv(simparms,coords,&energies->zkev,&energies->potv);
    }
    
    if(simparms->rank==0){
      accumulate(coords->hmat,energies);
      accumulate(coords->hmat,energies->next);
			      
      /* dump restart info and output averages every ndump passes */
      if(write_step->nscrn && !(simparms->istep%(write_step->nscrn))){
	energies->bcpu = cputime();
	energies->acpu += energies->bcpu; 
	writescreen(simparms,ngbr,energies,coords->hmat,coords->natoms);
	writescreen(simparms,ngbr->next,energies->next,coords->hmat,coords->next->natoms);
	print_timing(timing);
      }
      if(write_step->ndump && !(simparms->istep%(write_step->ndump))){
	get_abs_aapos(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
	get_abs_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
	/* check if the next 2 lines are still required (I think I am doing this at other places every step) */
	get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive,coords->inactive);
	get_cgvel(coords,simparms->icglist,simparms->napcg,coords->ninactive,coords->inactive);
	writestart(filenames,simparms->istep,simparms,energies,coords,so3);
	get_relat_aapos(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
	get_relat_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
      }
      if(write_step->rescalevel&&!(simparms->istep%(write_step->rescalevel))){
	get_abs_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
//	samvel(simparms,coords->next,0);
  simparms->temp *= simparms->scaletemp;
  scalevel_region(simparms,coords,0);
  scalevel_region(simparms,coords,1);
  scalevel_region(simparms,coords,2);
	get_cgvel(coords,simparms->icglist,simparms->napcg,coords->natoms,coords->iselect);
	get_relat_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,simparms->ncg-coords->nschizo);
	if(simparms->iperd==0){
	  zero_angm(simparms->ncg,coords->px,coords->py,coords->pz,
		    coords->vx,coords->vy,coords->vz,coords->amass);
	}
      }
      if(write_step->resamvel && !(simparms->istep%(write_step->resamvel))){
	get_abs_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
	samvel(simparms,coords->next,1);
	get_cgvel(coords,simparms->icglist,simparms->napcg,coords->natoms,coords->iselect);
	get_relat_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,simparms->ncg-coords->nschizo);
	if(simparms->iperd==0){
	  zero_angm(simparms->ncg,coords->px,coords->py,coords->pz,
		    coords->vx,coords->vy,coords->vz,coords->amass);
	}
	samveta(simparms,coords);
	if(simparms->ivol) samvbar(simparms,coords); 
      }
      /* -------  record instantaneous properties ------- */
      
      if(write_step->nconf && simparms->istep%(write_step->nconf) == 0){
	get_abs_aapos(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
	save_conf_hybrid(filenames->fconf,simparms,coords->next);
	get_relat_aapos(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
      }
      if(write_step->nvel && simparms->istep%(write_step->nvel) == 0){
       get_abs_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
       save_vel_hybrid(filenames->fvel,simparms,coords->next);
       get_relat_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
      }
      
      if(write_step->ncgconf && simparms->istep%(write_step->ncgconf) == 0){
	get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive,coords->inactive);
	save_conf_hybrid(filenames->fcgconf,simparms,coords);
      }
      if(write_step->ncgvel && simparms->istep%(write_step->ncgvel) == 0){
	get_cgvel(coords,simparms->icglist,simparms->napcg,coords->ninactive,coords->inactive);
	save_vel_hybrid(filenames->fcgvel,simparms,coords);
      }
      
      if(write_step->nforce && simparms->istep%(write_step->nforce) == 0){
	save_force(filenames->fforce,simparms->natoms,coords->next);
      }
      
      if(write_step->ncolv && simparms->istep%(write_step->ncolv) == 0){
	save_colv(filenames->fcolv,simparms->istep,simparms->dt,&coords->colvar);
	save_hills(filenames->hillfile,simparms->istep,simparms->dt,&coords->colvar);
      }
      
      
      if(write_step->ninst && simparms->istep%(write_step->ninst) == 0){
	save_inst_hybrid(filenames->fham,filenames->feng,filenames->fext,
			 simparms->istep,simparms->dt,energies,coords->hmat);
      }
      
#ifdef USER_FUNC
      user_function(simparms->istep,simparms->natoms,simparms->dt,
		    coords->px,coords->py,coords->pz,
		    coords->vx,coords->vy,coords->vz,
		    coords->amass,coords->qch,coords->hmat);
#endif 
    }
    /* calculate how much real time has passed and see if we need to exit */
    acttime=realtime();
    difftime=acttime-oldtime; if (difftime<0) difftime= -difftime;
    oldtime=acttime;
#ifdef DEBUG
    printf("here comes wallclock and acttime %g,%g,%g %g %g \n",
	   simparms->wallclock,difftime,acttime-starttime,
	   starttime,acttime);
#endif
    if ((acttime-starttime+difftime) > simparms->wallclock) exitall=1;
#ifdef PARA
    if(exitall) fprintf(stdout,"Rank %d has exceeded the time limit %g\n",
			simparms->rank,simparms->wallclock);
    MPI_Allreduce(&exitall,&exitsum,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
    bcast_system(simparms,coords);
#else
    exitsum=exitall;
#endif
    myclockoff(timing,"WRITE DATA",8);
    if (exitsum) break;
  } /* for istep loop */
  
  if(simparms->rank==0){
    if (exitsum) md_warning("Kicked out by timelimit!");
    md_stdout("\n *** Finnished Molecular Dynamics Steps ***\n");
    if (!exitall) simparms->istep -= 1;
    get_abs_aapos(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
    get_abs_aavel(coords,simparms->icglist,simparms->napcg,coords->iselect,coords->nselect-coords->nschizo);
    get_cgpos(coords,simparms->icglist,simparms->napcg,coords->ninactive,coords->inactive);
    get_cgvel(coords,simparms->icglist,simparms->napcg,coords->ninactive,coords->inactive);
    finish(filenames,simparms,energies,coords,so3);
  }
}

