/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include "md.h"
/* #define ENERGIES  */

/*---------------------------------------------------------*/
void output_initval(SIMPARMS *simparms,COORDS *coords,NGBR *ngbr,
		    ENERGIES *energies)
{
  double total,ratoms,tmass,volume,nsel;
  
/*  ratoms = 1./(double)(simparms->natoms);  */
  nsel = coords->nselect;
  ratoms = 1./(double)(nsel);
  total = (energies->poth+energies->potra+energies->zke+
	   energies->potn+energies->zken+energies->potv+energies->zkev);
/*  tmass = dsum1(simparms->natoms,coords->amass);  */
  tmass = dsum2(nsel,coords->iselect,coords->amass);
  volume = get_deth(coords->hmat);

  if (simparms->iunits==1){
    fprintf(stdout,"\n  *** initial report ***\n");
    fprintf(stdout,"number of pairs  = s=%d t=%d se=%d te=%d\n",
	    ngbr->nprs,ngbr->nprt,ngbr->nprse,ngbr->nprte);
    fprintf(stdout,"\n");
    fprintf(stdout,"econv            =     %15.6f\n",(total)/ECONV); 
    fprintf(stdout,"temperature      =     %15.6f\n",
	    energies->zke*ratoms*3./2.);
#ifdef ENERGIES
    fprintf(stdout,"part kin         =     %15.6f\n",energies->zke/(ECONV));
    fprintf(stdout,"part vter        =     %15.6f\n",energies->poth/(ECONV));
    fprintf(stdout,"part vtra        =     %15.6f\n",energies->potra/(ECONV));
    fprintf(stdout,"pressure         =     %15.6f\n",energies->prsi*PCONV);
    fprintf(stdout,"nhc ke           =     %15.6f\n",energies->zken*2.0/
	    (double)(simparms->nchain*(simparms->ntherm+1)));
    fprintf(stdout,"vol ke           =     %15.6f\n",energies->zkev*2.0);
#endif
    fprintf(stdout,"volume           =     %15.6f\n",
	    volume*LCONV*LCONV*LCONV);
    fprintf(stdout,"\n");
  }else{
    fprintf(stdout,"\n  *** Initial Report ***\n");
    fprintf(stdout,"npairs = s=%d t=%d se=%d te=%d\n",
	    ngbr->nprs,ngbr->nprt,ngbr->nprse,ngbr->nprte);
#ifdef ENERGIES
    fprintf(stdout,"  Ura/N    = %g K\n",energies->potra*ratoms);
    fprintf(stdout,"  Uer/N    = %g K\n",energies->poter*ratoms);
    fprintf(stdout,"  K/N      = %g K\n",energies->zke*ratoms);
    if(simparms->ntherm){
      fprintf(stdout,"  U_eta/nc = %g K\n",energies->potn/
	      (double)(simparms->nchain*simparms->ntherm));
      fprintf(stdout,"  K_eta/nc = %g K\n",energies->zken/
	      (double)(simparms->nchain*simparms->ntherm));
    }
    if(simparms->ivol){
      fprintf(stdout,"  U_bar    = %g K\n",energies->potv/
	      ((double)simparms->nbar+1.));
      fprintf(stdout,"  K_bar    = %g K\n",energies->zkev/
	      ((double)simparms->nbar+1.));
    }
#endif
    fprintf(stdout,"  density  = %g g/cc\n",DCONV*tmass/volume);
    fprintf(stdout,"  Volume   = %g A^3 (cbrt = %g)\n",volume,cbrt(volume));
    fprintf(stdout,"  Temp     = %g K\n",energies->tiout);
    fprintf(stdout,"  Pressure = %g K/A^3 = %g ATM\n",energies->prsi,
	    energies->prsi*PCONV);
    fprintf(stdout,"\n");
  }
}
/*---------------------------------------------------------*/
