/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#define PRINTMASS
#include "md.h"
/*---------------------------------------------------------*/
/* configuration */
void save_conf(FILE *fconf,int natoms,COORDS *coords)
{
  int ii;
	
  if(fconf==NULL){
    md_warning("No file name for configuration file, specify in input file\n");
    return;
  }
	
  for(ii=0;ii<natoms;ii++){
    fprintf(fconf,"%g %g %g\n",coords->px[ii],coords->py[ii],coords->pz[ii]);
  }
  fprintf(fconf,"%g %g %g %g %g %g %g %g %g\n",
          coords->hmat[0],coords->hmat[1],coords->hmat[2],
          coords->hmat[3],coords->hmat[4],coords->hmat[5],
          coords->hmat[6],coords->hmat[7],coords->hmat[8]);
	
  fflush(fconf);
}
/*---------------------------------------------------------*/
/* configuration TO BE REMOVED */
void save_conf_old(FILE *fconf,int natoms,COORDS *coords,int *icg,double *faacg,int icalc_type)
{
  int ii;
	
  if(fconf==NULL){
    md_warning("No file name for configuration file, specify in input file\n");
    return;
  }
	
  /* if hybrid MD simulation print 4th column with faacg to set transparancy in xmov */
  if(icalc_type == 5){
    if(coords->lcg == 1){  /* CG confp */
      for(ii=0;ii<natoms;ii++){
        fprintf(fconf,"%g %g %g %g\n",coords->px[ii],coords->py[ii],coords->pz[ii],
								faacg[ii]);
      }
		}else{                 /* AA confp */
      for(ii=0;ii<natoms;ii++){
        fprintf(fconf,"%g %g %g %g\n",coords->px[ii],coords->py[ii],coords->pz[ii],
								1.0-faacg[icg[ii]]);
      }
		}
  }else{
    for(ii=0;ii<natoms;ii++){
      fprintf(fconf,"%g %g %g\n",coords->px[ii],coords->py[ii],coords->pz[ii]);
    }
  }    
fprintf(fconf,"%g %g %g %g %g %g %g %g %g\n",
				coords->hmat[0],coords->hmat[1],coords->hmat[2],
				coords->hmat[3],coords->hmat[4],coords->hmat[5],
				coords->hmat[6],coords->hmat[7],coords->hmat[8]);

fflush(fconf);
}
/*--------------------------------------------------------*/
/* hybrid configuration selection */
/* print first line with numbers of molecules and then the selected molecule positions*/
/*	iprintatoms = 0  -> print positions+velocities of all CG+AA atoms                  */
/*	iprintatoms = 1  -> (default) print all CG atoms and active (AA+HR) atomistic pos+vel */
/*	iprintatoms = 2  -> print active (CG+HR) CG atoms and active (AA+HR) atomistic atoms */
void save_conf_hybrid(FILE *fconf,SIMPARMS *simparms,COORDS *coords)
{
  int i,j,*iselect,*nselect,ifound,ispecies,imolecule,istart,isp0;
	int *ispec=simparms->ispecies+simparms->ncg*(1-coords->lcg);
	int *imol=simparms->imolecule+simparms->ncg*(1-coords->lcg);
	double *faacg = simparms->faacg+simparms->ncg*(1-coords->lcg);

  if(fconf==NULL){
    md_warning("No file name for configuration file, specify in input file\n");
    return;
  }

	if(simparms->iprintatoms-coords->lcg < 1){
		/* print all positions and at 4th column the faacg to set transparancy in xmov */
    if(coords->lcg){
      for(i=0;i<coords->natoms;i++){
        fprintf(fconf,"%g %g %g %g\n",coords->px[i],coords->py[i],coords->pz[i],
                faacg[i]);
      }
    }else{
      for(i=0;i<coords->natoms;i++){
        fprintf(fconf,"%g %g %g %g\n",coords->px[i],coords->py[i],coords->pz[i],
                faacg[i]);
      }
		}
	}else{
		/* count number of active molecules and print only those */
		nselect = (int *)malloc(simparms->nspec*sizeof(int));
		iselect = (int *)malloc(coords->natoms*sizeof(int));
		for(i=0;i<simparms->nspec;i++) nselect[i]=0;
		for(i=0;i<coords->natoms;i++) iselect[i]=0;
		ifound=0;
		ispecies=isp0=ispec[0];; imolecule=0;
		istart=0;
		for(i=0;i<coords->natoms;i++){
			if(ispecies!=ispec[i]){
				ispecies++;
				imolecule=0;
				if(ifound){
					for(j=istart;j<i;j++) iselect[j]=1;
					nselect[ispecies-1] += i-istart;					
				}
				ifound=0;
				istart=i;
			}
			if(imolecule!=imol[i]){
				if(ifound){
					for(j=istart;j<i;j++) iselect[j]=1;
					nselect[ispecies] += i-istart;					
				} 
				ifound=0;
				imolecule++;
				istart=i;
			}
			if(faacg[i]>0) ifound=1;
		}
		if(ifound){
			for(j=istart;j<i;j++) iselect[j]=1;
			nselect[ispecies] += i-istart;					
		}
		fprintf(fconf,"#");
		for(i=0;i<simparms->nspec/2;i++) fprintf(fconf," %d ",nselect[isp0+i]);
		fprintf(fconf,"\n");
		for(i=0;i<coords->natoms;i++){
			if(iselect[i]) fprintf(fconf,"%d %g %g %g %g\n",i,
														 coords->px[i],coords->py[i],coords->pz[i],faacg[i]);
		}
		if(simparms->iprintatoms==0){
			free(iselect);
			free(nselect);
		}		
	}
	
  fprintf(fconf,"%g %g %g %g %g %g %g %g %g\n",
	  coords->hmat[0],coords->hmat[1],coords->hmat[2],
	  coords->hmat[3],coords->hmat[4],coords->hmat[5],
	  coords->hmat[6],coords->hmat[7],coords->hmat[8]);

  fflush(fconf);

}
/*--------------------------------------------------------*/
void save_vel(FILE *fvel,int natoms,COORDS *coords)
{
  int ii;

  if(fvel==NULL){
    md_warning("No file name for velocity file, specify in input file\n");
    return;
  }

  for(ii=0;ii<natoms;ii++){
#ifdef PRINTMASS  
    fprintf(fvel,"%g %g %g %g\n",coords->vx[ii],coords->vy[ii],coords->vz[ii],coords->amass[ii]);
#else
    fprintf(fvel,"%g %g %g\n",coords->vx[ii],coords->vy[ii],coords->vz[ii]);
#endif
  }
  fflush(fvel);
}
/*--------------------------------------------------------*/
/* hybrid configuration selection */
/* print first line with numbers of molecules and then the selected molecule positions */
/*	iprintatoms = 0  -> print positions+velocities of all CG+AA atoms                  */
/*	iprintatoms = 1  -> (default) print all CG atoms and active (AA+HR) atomistic pos+vel */
/*	iprintatoms = 2  -> print active (CG+HR) CG atoms and active (AA+HR) atomistic atoms */
void save_vel_hybrid(FILE *fvel,SIMPARMS *simparms,COORDS *coords)
{
  int i,j,*iselect,*nselect,ifound,ispecies,imolecule,istart,isp0;
	int *ispec=simparms->ispecies+simparms->ncg*(1-coords->lcg);
	int *imol=simparms->imolecule+simparms->ncg*(1-coords->lcg);
	double *faacg = simparms->faacg+simparms->ncg*(1-coords->lcg);
	
  if(fvel==NULL){
    md_warning("No file name for velocity file, specify in input file\n");
    return;
  }
	
	if(simparms->iprintatoms-coords->lcg < 1){
		/* print all velocities */
		for(i=0;i<coords->natoms;i++){
#ifdef PRINTMASS  
			fprintf(fvel,"%g %g %g %g\n",coords->vx[i],coords->vy[i],coords->vz[i],coords->amass[i]);
#else
			fprintf(fvel,"%g %g %g\n",coords->vx[i],coords->vy[i],coords->vz[i]);
#endif
		}
	}else{
		/* count number of active molecules and print only those */
		nselect = (int *)malloc(simparms->nspec*sizeof(int));
		iselect = (int *)malloc(coords->natoms*sizeof(int));
		for(i=0;i<simparms->nspec;i++) nselect[i]=0;
		for(i=0;i<coords->natoms;i++) iselect[i]=0;
		ifound=0;
		ispecies=isp0=ispec[0];; imolecule=0;
		istart=0;
		for(i=0;i<coords->natoms;i++){
			if(ispecies!=ispec[i]){
				ispecies++;
				imolecule=0;
				if(ifound){
					for(j=istart;j<i;j++) iselect[j]=1;
					nselect[ispecies-1] += i-istart;					
				}
				ifound=0;
				istart=i;
			}
			if(imolecule!=imol[i]){
				if(ifound){
					for(j=istart;j<i;j++) iselect[j]=1;
					nselect[ispecies] += i-istart;					
				} 
				ifound=0;
				imolecule++;
				istart=i;
			}
			if(faacg[i]>0) ifound=1;
		}
		if(ifound){
			for(j=istart;j<i;j++) iselect[j]=1;
			nselect[ispecies] += i-istart;					
		}
		fprintf(fvel,"#");
		for(i=0;i<simparms->nspec/2;i++) fprintf(fvel," %d ",nselect[isp0+i]);
		fprintf(fvel,"\n");
		for(i=0;i<coords->natoms;i++){
			if(iselect[i]){
#ifdef PRINTMASS  
				fprintf(fvel,"%d %g %g %g %g\n",i,coords->vx[i],coords->vy[i],coords->vz[i],coords->amass[i]);
#else
				fprintf(fvel,"%d %g %g %g\n",i,coords->vx[i],coords->vy[i],coords->vz[i]);
#endif
			}
		}
		if(simparms->iprintatoms==0){
			free(iselect);
			free(nselect);
		}
	}
	
	
  fflush(fvel);
	
}
/*--------------------------------------------------------*/
void save_force(FILE *fforce,int natoms,COORDS *coords)
{
  int ii;
  double am;
  double *fxl,*fyl,*fzl;
  double *fxr,*fyr,*fzr;
  double *fxa,*fya,*fza;
  
  if(fforce==NULL){
    md_warning("No file name for force file, specify in input file\n");
    return;
  }

/*
  fx = coords->fxt;  fy = coords->fyt;  fz = coords->fzt;
  for(ii=0;ii<natoms;ii++) fx[ii] = fy[ii] = fz[ii] = 0.;

  for(ii=0;ii<natoms;ii++){
    fx[ii] += coords->fxl[ii]+coords->fxa[ii]+coords->fxr[ii];
    fy[ii] += coords->fyl[ii]+coords->fya[ii]+coords->fyr[ii];
    fz[ii] += coords->fzl[ii]+coords->fza[ii]+coords->fzr[ii];
  }

  for(ii=0;ii<natoms;ii++){
    am = coords->amass[ii];
    fprintf(fforce,"%g %g %g\n",fx[ii]*am,fy[ii]*am,fz[ii]*am);
  }
*/

  fxl = coords->fxl;  fyl = coords->fyl;  fzl = coords->fzl;
  fxr = coords->fxr;  fyr = coords->fyr;  fzr = coords->fzr;
  fxa = coords->fxa;  fya = coords->fya;  fza = coords->fza;

  for(ii=0;ii<natoms;ii++){
    am = coords->amass[ii];
    fprintf(fforce,"%g %g %g\n",(fxl[ii]+fxa[ii]+fxr[ii])*am,(fyl[ii]+fxa[ii]+fyr[ii])*am,(fzl[ii]+fza[ii]+fzr[ii])*am);
  }

  fflush(fforce);
}
/*--------------------------------------------------------*/
void save_eigvec(char *filename,int m,int n,double *dn,double **d2v)
{
  char line[MAXLINELEN];
  int i,j;
  FILE *fp;
  
  if((fp=fopen(filename,"w"))==NULL){
    sprintf(line,"can't open \"%s\" for saving eigenvalues",filename);
    md_error(line);
  }
  /* write header */
  fprintf(fp,"%d %d\n",m,n);

  for(i=0;i<m;i++){
    fprintf(fp,"%g\n",dn[i]);
  }
  for(i=0;i<m;i++){
    fprintf(fp,"\n");
    for(j=0;j<n;j++){
      fprintf(fp,"%g\n",d2v[j][i]);
    }
  }
  fclose(fp);
}
/*---------------------------------------------------------*/
#ifndef BUFSIZ
#define BUFSIZ 1024
#endif

void keep(char *name)
{
  char save[MAXLINELEN];
  int n,buf[BUFSIZ];
  FILE *fpi,*fpo;

  sprintf(save,"%s.O",name);

  fprintf(stdout,"Saving file initial files as %s\n",save);
  if((fpi= fopen(name,"r"))==NULL){
    fprintf(stderr,"ERROR: can't open file %s for reading\n",name);
    return;
  }
  if((fpo= fopen(save,"w"))==NULL){
    fprintf(stderr,"ERROR: can't open file %s for writing\n",save);
    return;
  }
  while( (n = fread(buf,sizeof(int),BUFSIZ,fpi)) > 0 ){
    fwrite(buf,sizeof(int),n,fpo);
  }
  fwrite(buf,sizeof(int),n,fpo);

  fclose(fpi);
  fclose(fpo);
}
/*-----------------------------------------------------------*/
