/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include "md.h"

/*---------------------------------------------------------*/
void output_simparm(SIMPARMS *simparms,WRITE_STEP *write_step,int istep)
{
  int i;
  
  /* output stuff that the user might want to know */
  fprintf(stdout,"\n");
  fprintf(stdout,"type of calculation   = %d\n",simparms->istart);
  fprintf(stdout,"units                 = %d\n",simparms->iunits);
  fprintf(stdout,"periodicity           = %d\n",simparms->iperd);
  fprintf(stdout,"# of atoms            = %d\n",simparms->natoms);
  fprintf(stdout,"# of frozen atoms     = %d\n",simparms->nfreeze);
  fprintf(stdout,"MD time step          = %g ps\n",simparms->dt);
  fprintf(stdout,"RESPA type %d:\n",simparms->nres_type);
  if(simparms->nres_type==0){
    fprintf(stdout,"# of short range inter time steps = %d (dt = %g)\n",               
	  simparms->ninter,simparms->dt/simparms->ninter);                      
    fprintf(stdout,"# of Torsion steps                = %d (dt = %g)\n",simparms->ninrra,
  	  simparms->dt/(simparms->ninter*simparms->ninrra));                     
    fprintf(stdout,"# of Bond/Bend steps              = %d (dt = %g)\n",simparms->ninnra,
	  simparms->dt/(simparms->ninter*simparms->ninrra*simparms->ninnra));
  }else if(simparms->nres_type==1){
    for(i=0;i<simparms->nres_maxlevel+1;i++){
      switch (simparms->nres_iorder[i]){
	    case 2:
	      fprintf(stdout,"# of CG Long range inter time steps = %d (dt = %g)\n",
			1, 2*simparms->nres_dt[i]*simparms->nres_nsteps[i]);
	      fprintf(stdout,"# of Short inter time steps         = ");
		  break;
	    case 3:
	      fprintf(stdout,"# of AA Long range inter time steps = %d (dt = %g)\n",
			1, 2*simparms->nres_dt[i]*simparms->nres_nsteps[i]);
	      fprintf(stdout,"# of AA Short range inter time steps= ");
		  break;
	    case 4:
	      fprintf(stdout,"# of CG Torsion time steps          = ");
		  break;
	    case 5:
	      fprintf(stdout,"# of AA Torsion time steps          = ");
		  break;
	    case 6:
	      fprintf(stdout,"# of CG Bond/Bend time steps        = ");
		  break;
	    case 7:
	      fprintf(stdout,"# of AA Bond/Bend time steps        = ");
		  break;
	  }
	  fprintf(stdout,"%d (dt = %g)\n",simparms->nres_nsteps[i],
	    2*simparms->nres_dt[i]); 
	}
  }else if(simparms->nres_type==2){
    for(i=0;i<simparms->nres_maxlevel+1;i++){
      switch (simparms->nres_iorder[i]){
	    case 2:
	      fprintf(stdout,"# of CG+AA Long range inter time steps  = %d (dt = %g)\n",
			1, 2*simparms->nres_dt[i]*simparms->nres_nsteps[i]);
	      fprintf(stdout,"# of CG+AA Short range inter time steps = ");
		  break;
	    case 4:
	      fprintf(stdout,"# of CG Torsion time steps              = ");
		  break;
	    case 5:
	      fprintf(stdout,"# of AA Torsion time steps              = ");
		  break;
	    case 6:
	      fprintf(stdout,"# of CG Bond/Bend time steps            = ");
		  break;
	    case 7:
	      fprintf(stdout,"# of AA Bond/Bend time steps            = ");
		  break;
	  }
	  fprintf(stdout,"%d (dt = %g)\n",simparms->nres_nsteps[i],
	    2*simparms->nres_dt[i]); 
	}
  }else{
    for(i=0;i<simparms->nres_maxlevel+1;i++){
      switch (simparms->nres_iorder[i]){
	    case 0:
	      fprintf(stdout,"# of Long range inter time steps    = ");
	  	  break;
	    case 1:
	      fprintf(stdout,"# of AA Long range inter time steps = ");
		  break;
	    case 2:
	      fprintf(stdout,"# of Short inter time steps         = ");
		  break;
	    case 3:
	      fprintf(stdout,"# of AA Short inter time steps      = ");
		  break;
	    case 4:
	      fprintf(stdout,"# of Torsion time steps             = ");
		  break;
	    case 5:
	      fprintf(stdout,"# of AA Torsion time steps          = ");
		  break;
	    case 6:
	      fprintf(stdout,"# of Bond/Bend time steps           = ");
		  break;
	    case 7:
	      fprintf(stdout,"# of AA Bond/Bend time steps        = ");
		  break;
	  }
	  fprintf(stdout,"%d (dt = %g)\n",simparms->nres_nsteps[i],
	    2*simparms->nres_dt[i]); 
	}
  }
  fprintf(stdout,"# of thermostats      = %d\n",simparms->ntherm);
  fprintf(stdout,"# of nh chains links  = %d\n",simparms->nchain);
  fprintf(stdout,"# of barostats        = %d\n",simparms->nbar);
  if(simparms->ntherm)
    fprintf(stdout,"Temperature           = %g K\n",simparms->temp);
  else
    fprintf(stdout,"Temperature is not thermostated\n");

  if(simparms->ivol)
    fprintf(stdout,"Pressure              = %g K/A^3\n",simparms->pext);
  else
    fprintf(stdout,"Volume will not fluctuate.\n");

  fprintf(stdout,"\n");
  fprintf(stdout,"Averages reported every       %d ts\n",write_step->nscrn);
  fprintf(stdout,"Restart info dumped every     %d ts\n",write_step->ndump);
  fprintf(stdout,"A config is save every        %d ts\n",write_step->nconf);
  fprintf(stdout,"A velocity file is save every %d ts\n",write_step->nvel);
  fprintf(stdout,"A force file is save every    %d ts\n",write_step->nforce);
  fprintf(stdout,"Instant quantaties recorded   %d ts\n",write_step->ninst);
  fprintf(stdout,"\n");
  fprintf(stdout,"Calcutaion time step %d to %d.\n",istep,simparms->nstep);
  fprintf(stdout,"total time of simulation %g ps.\n",
	  simparms->nstep*simparms->dt);
  fprintf(stdout,"\n");
}
/*---------------------------------------------------------*/ 
