/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* output things that the user might want to know */

#include "md.h"

/*-------------------------------------------------------------------*/
void output_subparm(SIMPARMS *simparms,WRITE_STEP *write_step,
		    SUBSPACE *subspace,int istep)
{
  int i,nr,ni,nt;

  /* output stuff that the user might want to know */
  fprintf(stdout,"\n");
  fprintf(stdout,"type of calculation   = %d\n",simparms->istart);
  fprintf(stdout,"units                 = %d\n",simparms->iunits);
  fprintf(stdout,"periodicity           = %d\n",simparms->iperd);
  fprintf(stdout,"# of atoms            = %d\n",simparms->natoms);
  fprintf(stdout,"# of states           = %d\n",subspace->nstate);
  fprintf(stdout,"# of inter time steps = %d\n",simparms->ninter);
  fprintf(stdout,"# of intra time steps = %d\n",simparms->ninnra);
  fprintf(stdout,"# of thermostats      = %d\n",simparms->ntherm);
  fprintf(stdout,"# of nh chains links  = %d\n",simparms->nchain);
  fprintf(stdout,"# of barostats        = %d\n",simparms->nbar);
  fprintf(stdout,"MD time step          = %g ps\n",simparms->dt);
  if(simparms->ntherm)
    fprintf(stdout,"Temperature           = %g K\n",simparms->temp);
  else
    fprintf(stdout,"Temperature is not thermostated\n");

  if(simparms->ivol)
    fprintf(stdout,"Pressure              = %g K/A^3\n",simparms->pext);
  else
    fprintf(stdout,"Volume will not fluctuate.\n");

  fprintf(stdout,"\n");
  fprintf(stdout,"Averages reported every       %d ts\n",write_step->nscrn);
  fprintf(stdout,"Restart info dumped every     %d ts\n",write_step->ndump);
  fprintf(stdout,"A config is save every        %d ts\n",write_step->nconf);
  fprintf(stdout,"A velocity file is save every %d ts\n",write_step->nvel);
  fprintf(stdout,"Instant quantaties recorded   %d ts\n",write_step->ninst);
  fprintf(stdout,"\n");
  fprintf(stdout,"Calcutaion time step %d to %d.\n",istep,simparms->nstep);
  fprintf(stdout,"total time of simulation %g ps.\n",
	  simparms->nstep*simparms->dt);
  fprintf(stdout,"\n");


  nr = ni = nt = 0;
  fprintf(stdout,"----------- %d subvectors --------------\n",subspace->num_subst);
  fprintf(stdout,"     #  Vectors  Real  Imaginary\n");
  for(i=0;i<subspace->num_subst;i++){
    fprintf(stdout,"%6d %6d %6d %6d\n",i+1,subspace->num_vecsub_max[i],
	   subspace->num_real_max[i],subspace->num_imag_max[i]);
    nr += subspace->num_real_max[i];
    ni += subspace->num_imag_max[i];
    nt += subspace->num_vecsub_max[i];
  }
  fprintf(stdout,"----------------------------------------------\n");
  fprintf(stdout,"total  %6d %6d %6d\n",nt,nr,ni);
}
/*---------------------------------------------------------*/ 
