/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "md.h"

#define REPORT_ENERGIES
#define PRESSURE
/*--------------------------------------------------------*/
void writescreen(SIMPARMS *simparms,NGBR *ngbr,
		 ENERGIES *energies,double *hmat, int natoms)
{     
  int lhybrid;
  double update,vol,*apt,*pt;
  double stddev,unit_conv,step,ratoms,ratoms_tot;
  WORD sunit;

  lhybrid = (simparms->icalc_type==5);
  vol = get_deth(hmat);
  if(simparms->istep==0) { 
    step = 1.;
  } else {
    step = (double)simparms->istep;
  }
/*  ratoms = 1./(double)simparms->natoms; */
  ratoms_tot = 1./(double)natoms;   /* temporary solution, fix later */
  ratoms = 1./simparms->fnatoms[1-energies->lcg];
  if(simparms->fnatoms[1-energies->lcg]==0) ratoms = ratoms_tot;
  
  if(ngbr->update==0) {
    update=1.;
  } else {
    update = (double)ngbr->update;
  }
  
  if(lhybrid){
    if(energies->lcg){
      printf("========================================================================\n");
      printf("=====    Coarse grain summary                                      =====\n");
    }else{
      printf("=====    All atom summary                                          =====\n");
    }
  }
  
  fprintf(stdout,"time steps= %d, time= %g ps, <cpu>= %g, <cpu/ps>= %g, cpu= %g\n",
	  simparms->istep,step*simparms->dt,
	  energies->acpu/
	  ((step-simparms->fstep)==0?1.:(step-simparms->fstep)),
	  energies->acpu/
	  (((step-simparms->fstep)==0?1.:(step-simparms->fstep))
	   *simparms->dt),energies->bcpu);
  fprintf(stdout,"nprs=%d, nprt=%d, nprse=%d, nprte=%d, <ts/update>=%g\n",
	  ngbr->nprst,ngbr->nprtt,ngbr->nprset,ngbr->nprtet,
	  (step-simparms->fstep)/update);
  
  switch(simparms->iunits) {
    /*-----------------------------------------------------*/
  default : case 0: /* Kelvin Angstroms picoseconds */
    unit_conv = 1.;  
    strcpy(sunit,"K"); 
    break;
  case 1: /* AU */
    unit_conv = 1./ECONV;
    strcpy(sunit,"AU"); 
    break;
  case 2:   /* Kcal Units */
    unit_conv = 1./KCAL;
    strcpy(sunit,"KCAL/mol"); 
    break;
  }
  
  if(energies->potra){
    if(isnan(unit_conv*energies->avpotra/step) != 0 ||
       isnan(unit_conv*energies->potra) != 0){
      fprintf(stdout,"###NAN ERROR###<Ura>    \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpotra/step,
	      unit_conv*energies->potra,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<Ura>    \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpotra/step,
	      unit_conv*energies->potra,sunit);
    }
  }
  if(energies->poter){
    if(isnan(unit_conv*energies->avpoter/step) != 0 ||
       isnan(unit_conv*energies->poter) !=0){
      fprintf(stdout,"###NAN ERROR###<Uer>    \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpoter/step,
	      unit_conv*energies->poter,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<Uer>    \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpoter/step,
	      unit_conv*energies->poter,sunit);
    }
  }
    if(energies->zke){  /* divided by all atoms; not correct for AA part...*/
      if(isnan(unit_conv*energies->avzke/step) != 0 ||
	 isnan(unit_conv*energies->zke)!=0){
	fprintf(stdout,"###NAN ERROR###<Ke >    \t= %15.6f\t~ %15.6f %s\n",
		unit_conv*energies->avzke/step,
		unit_conv*energies->zke,sunit);
	exit(99);
      }else{
	fprintf(stdout,"<Ke >    \t= %15.6f\t~ %15.6f %s\n",
		unit_conv*energies->avzke/step,
		unit_conv*energies->zke,sunit);
      }
    }
#ifdef REPORT_ENERGIES
  if(energies->pot_extern){
    if(isnan(unit_conv*energies->avpot_extern/step) != 0 ||
       isnan(unit_conv*energies->pot_extern) != 0){
      fprintf(stdout,"###NAN ERROR###<extern>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_extern/step,
	      unit_conv*energies->pot_extern,sunit);
      exit(99);
    }else{
    fprintf(stdout,"<extern>  \t= %15.6f\t~ %15.6f %s\n",
	    unit_conv*energies->avpot_extern/step,
	    unit_conv*energies->pot_extern,sunit);
    }
  }
  if(energies->pot_extern_e){
    if(isnan(unit_conv*energies->avpot_extern_e/step) != 0 ||
       isnan(unit_conv*energies->pot_extern_e) != 0){
      fprintf(stdout,"###NAN ERROR###<extern_e>\t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_extern_e/step,
	      unit_conv*energies->pot_extern_e,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<extern_e>\t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_extern_e/step,
	      unit_conv*energies->pot_extern_e,sunit);
    }
  }
  if(energies->pot_inter){
    if(isnan(unit_conv*energies->avpot_inter/step) != 0 ||
       isnan(unit_conv*energies->pot_inter) !=0){
      fprintf(stdout,"###NAN ERROR <inter>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_inter/step,
	      unit_conv*energies->pot_inter,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<inter>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_inter/step,
	      unit_conv*energies->pot_inter,sunit);
    }
  }
  if(energies->pot_bond){
    if(isnan(unit_conv*energies->avpot_bond/step) != 0 ||
       isnan(unit_conv*energies->pot_bond) !=0){
      fprintf(stdout,"###NAN ERROR <bond>   \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_bond/step,
	      unit_conv*energies->pot_bond,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<bond>   \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_bond/step,
	      unit_conv*energies->pot_bond,sunit);
    }
  }
  if(energies->pot_bend){
    if(isnan(unit_conv*energies->avpot_bend/step) != 0 ||
       isnan(unit_conv*energies->pot_bend) !=0){
      fprintf(stdout,"###NAN ERROR <bend>   \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_bend/step,
	      unit_conv*energies->pot_bend,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<bend>   \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_bend/step,
	      unit_conv*energies->pot_bend,sunit);
    }
  }
  if(energies->pot_tors){
    if(isnan(unit_conv*energies->avpot_tors/step) != 0 ||
       isnan(unit_conv*energies->pot_tors) !=0){
      fprintf(stdout,"###NAN ERROR <tors>   \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_tors/step,
	      unit_conv*energies->pot_tors,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<tors>   \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_tors/step,
	      unit_conv*energies->pot_tors,sunit);
    }
  }
  if(energies->pot_onfo){
    if(isnan(unit_conv*energies->avpot_onfo/step) != 0 ||
       isnan(unit_conv*energies->pot_onfo) !=0){
      fprintf(stdout,"###NAN ERROR <onefour>\t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_onfo/step,
	      unit_conv*energies->pot_onfo,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<onefour>\t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_onfo/step,
	      unit_conv*energies->pot_onfo,sunit);
    }
  }
  if(energies->pot_onfo_e){
    if(isnan(unit_conv*energies->avpot_onfo_e/step) != 0 ||
       isnan(unit_conv*energies->pot_onfo_e) !=0){
      fprintf(stdout,"###NAN ERROR <onfo_e> \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_onfo_e/step,
	      unit_conv*energies->pot_onfo_e,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<onfo_e> \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_onfo_e/step,
	      unit_conv*energies->pot_onfo_e,sunit);
    }
  }
  if(energies->pot_elec){
    if(isnan(unit_conv*energies->avpot_elec/step) != 0 ||
       isnan(unit_conv*energies->pot_elec) !=0){
      fprintf(stdout,"###NAN ERROR <elec>   \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_elec/step,
	      unit_conv*energies->pot_elec,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<elec>   \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_elec/step,
	      unit_conv*energies->pot_elec,sunit);
    }
  }
  if(energies->pot_recip){
    if(isnan(unit_conv*energies->avpot_recip/step) != 0 ||
       isnan(unit_conv*energies->pot_recip) !=0){
      fprintf(stdout,"###NAN ERROR <recip>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_recip/step,
	      unit_conv*energies->pot_recip,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<recip>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avpot_recip/step,
	      unit_conv*energies->pot_recip,sunit);
    }
  }
  if(energies->zkecg){
    if(isnan(unit_conv*energies->avzkecg/step) != 0 ||
       isnan(unit_conv*energies->zkecg) !=0){
      fprintf(stdout,"###NAN ERROR###<Kcg>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avzkecg/step,
	      unit_conv*energies->zkecg,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<Kcg>    \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avzkecg/step,
	      unit_conv*energies->zkecg,sunit);
    }
  }
  if(energies->vcgtra){
    if(isnan(unit_conv*energies->avcgtra/step) != 0 ||
       isnan(unit_conv*energies->vcgtra) !=0){
      fprintf(stdout,"###NAN ERROR###<Vcgtra>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avcgtra/step,
	      unit_conv*energies->vcgtra,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<Vcgtra>  \t= %15.6f\t~ %15.6f %s\n",  /* perhaps Ucg per no. CG units ?? */
	      unit_conv*energies->avcgtra/step,
	      unit_conv*energies->vcgtra,sunit);
    }
  }
  if(energies->vcgter){
    if(isnan(unit_conv*energies->avcgter/step) != 0 ||
       isnan(unit_conv*energies->vcgter) !=0){
      fprintf(stdout,"###NAN ERROR###<Vcgter>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avcgter/step,
	      unit_conv*energies->vcgter,sunit);
      exit(99);
    }else{
      fprintf(stdout,"<Vcgter>  \t= %15.6f\t~ %15.6f %s\n",
	      unit_conv*energies->avcgter/step,
	      unit_conv*energies->vcgter,sunit);
    }
  }  
#endif
  if(simparms->ntherm){
    fprintf(stdout,"<Ueta/NC>\t= %15.6f\t~ %15.6f %s\n",
	    unit_conv*energies->avpotn/
	    (step*(double)simparms->nchain*(double)simparms->ntherm),
	    unit_conv*energies->potn/
	    ((double)simparms->nchain*(double)simparms->ntherm),sunit);
    fprintf(stdout,"<Keta/NC>\t= %15.6f\t~ %15.6f %s\n",
	    unit_conv*energies->avzken*2./
	    (step*(double)(simparms->nchain*(simparms->ntherm+1))),
	    unit_conv*energies->zken*2./
	    ((double)simparms->nchain*((double)simparms->ntherm+1.)),sunit);
  }
  if(simparms->ivol){
    fprintf(stdout,"<Ubar/NB>\t= %15.6f\t~ %15.6f %s\n",
	    unit_conv*energies->avpotv/(step*(double)(simparms->nbar+1)),
	    unit_conv*energies->potv/(double)(simparms->nbar+1),sunit);
    fprintf(stdout,"<Kbar/NB>\t= %15.6f\t~ %15.6f %s\n",
	    unit_conv*energies->avzkev/(step*(double)(simparms->nbar+1)),
	    unit_conv*energies->zkev/(double)(simparms->nbar+1),sunit);
    fprintf(stdout,"<vol>    \t= %15.6f\t~ %15.6f A^3 ",
	    energies->avvol/step,vol);
    fprintf(stdout,"\t cbrt(vol)=%15.6f A\n",cbrt(vol));
  }
  fprintf(stdout,"<Ham>    \t= %15.6f\t~ %15.6f %s\n",
	  unit_conv*energies->avham/step,
	  unit_conv*energies->ham,sunit);
  fprintf(stdout,"<temp>   \t= %15.6f\t~ %15.6f K\n",
	  energies->avtemp/step,energies->tiout);
  if(lhybrid){
    fprintf(stdout,"<temp aa> \t= %15.6f\t~ %15.6f K\n",
		energies->avtemp_aa/step,energies->tiout_aa);
    fprintf(stdout,"<temp hr> \t= %15.6f\t~ %15.6f K\n",
		energies->avtemp_hr/step,energies->tiout_hr);
    fprintf(stdout,"<temp cg> \t= %15.6f\t~ %15.6f K\n",
		energies->avtemp_cg/step,energies->tiout_cg);  
  }
  fprintf(stdout,"<pres>   \t= %15.6f\t~ %15.6f %s/A^3\n",
	  unit_conv*energies->avprs/step,
	  unit_conv*energies->prsi,sunit);
  fprintf(stdout,"<pres>   \t= %15.6f\t~ %15.6f atm\n",
	  energies->avprs*PCONV/(step),energies->prsi*PCONV);

#ifdef PRESSURE
  pt = energies->prs_tensor;
  apt = energies->avprs_tensor;
  
  fprintf(stdout,"Pressure Tensor\n");
  fprintf(stdout,"%10f %10f %10f ~ %10f %10f %10f atm\n",
     apt[0]*PCONV/(step), apt[1]*PCONV/(step), apt[2]*PCONV/(step),
	  pt[0]*PCONV,	  pt[1]*PCONV,	  pt[2]*PCONV);
  fprintf(stdout,"%10f %10f %10f ~ %10f %10f %10f atm\n",
     apt[3]*PCONV/(step),     apt[4]*PCONV/(step),     apt[5]*PCONV/(step),
	  pt[3]*PCONV,	  pt[4]*PCONV,	  pt[5]*PCONV);
  fprintf(stdout,"%10f %10f %10f ~ %10f %10f %10f atm\n",
     apt[6]*PCONV/(step),     apt[7]*PCONV/(step),     apt[8]*PCONV/(step),
	  pt[6]*PCONV,	  pt[7]*PCONV,	  pt[8]*PCONV);
#endif
  
#ifdef PRESSURE_1
  fprintf(stdout,"<pres xx>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[0]*PCONV/(step),
	  energies->prs_tensor[0]*PCONV);
  fprintf(stdout,"<pres xy>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[1]*PCONV/(step),
	  energies->prs_tensor[1]*PCONV);
  fprintf(stdout,"<pres xz>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[2]*PCONV/(step),
	  energies->prs_tensor[2]*PCONV);
  fprintf(stdout,"<pres yx>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[3]*PCONV/(step),
	  energies->prs_tensor[3]*PCONV);
  fprintf(stdout,"<pres yy>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[4]*PCONV/(step),
     energies->prs_tensor[4]*PCONV);
  fprintf(stdout,"<pres yz>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[5]*PCONV/(step),
	  energies->prs_tensor[5]*PCONV);
  fprintf(stdout,"<pres zx>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[6]*PCONV/(step),
	  energies->prs_tensor[6]*PCONV);
  fprintf(stdout,"<pres zy>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[7]*PCONV/(step),
	  energies->prs_tensor[7]*PCONV);
  fprintf(stdout,"<pres zz>   \t= %15.6f\t~ %15.6f atm\n",
     energies->avprs_tensor[8]*PCONV/(step),
	  energies->prs_tensor[8]*PCONV);
#endif
  
  if(simparms->istep>1 && energies->estep!=0.0){
    stddev = fabs((step*energies->econ2-energies->econv*energies->econv)/
		  ((step-1.)*step));
    fprintf(stdout,"<dE>     \t= %15.6f\t~%15.6f\n",
	    sqrt(stddev)*step/fabs(energies->econv),
	    sqrt(energies->deltae2)/(double)(energies->estep));
  }
#ifdef GCODE
  if(simparms->ntherm){
    fprintf(stdout,"NCHAIN KE   \t= %15.6f\t~ %15.6f K\n",
	    energies->avzken*2./(step*(double)simparms->nchain*
				 ((double)simparms->ntherm+1.)),
	    energies->zken*2./((double)simparms->nchain*
			       ((double)simparms->ntherm+1.)));
    fprintf(stdout,"NCHAIN PE   \t= %15.6f\t~ %15.6f au\n",
	    energies->avpotn/(step*ECONV),energies->potn/ECONV);
  }
  if(simparms->ivol){
    fprintf(stdout,"VOL KE   \t= %15.6f\t~ %15.6f K\n",
	    2.0*energies->avzkev/step,2.0*energies->zkev);
    fprintf(stdout,"VOL PE   \t= %15.6f\t~ %15.6f au\n",
	    energies->avpotv/(step*ECONV),energies->potv/ECONV);
    fprintf(stdout,"VOLUME   \t= %15.6f\t~ %15.6f A^3\n",
	    energies->avvol*LCONV3/step,vol*LCONV3);
  }
#endif

  /* writeout SO3 dynamics energies after "All atom summary" in hybrid MD simulation */
  if(lhybrid && !energies->lcg && simparms->iso3dyn){
    printf("========================================================================\n");
    printf("=====    SO3 dynamics summary                                      =====\n");
    fprintf(stdout,"nprr=%d, (CG-nprr=%d) \n",
            ngbr->nprr,ngbr->next->nprr);
    
    if(energies->so3_potra){
      if(isnan(unit_conv*energies->avso3_potra/step) != 0 ||
         isnan(unit_conv*energies->so3_potra) != 0){
        fprintf(stdout,"###NAN ERROR###<Ura>    \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_potra/step,
                unit_conv*energies->so3_potra,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<Ura>    \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_potra/step,
                unit_conv*energies->so3_potra,sunit);
      }
    }    
    if(energies->so3_poter){
      if(isnan(unit_conv*energies->avso3_poter/step) != 0 ||
         isnan(unit_conv*energies->so3_poter) !=0){
        fprintf(stdout,"###NAN ERROR###<Uer>    \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_poter/step,
                unit_conv*energies->so3_poter,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<Uer>    \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_poter/step,
                unit_conv*energies->so3_poter,sunit);
      }
    }
    if(energies->so3_vzero_ngbr){
      if(isnan(unit_conv*energies->avso3_vzero_ngbr/step) != 0 ||
         isnan(unit_conv*energies->so3_vzero_ngbr) !=0){
        fprintf(stdout,"###NAN ERROR###<Vzero_ngbr>    = %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_vzero_ngbr/step,
                unit_conv*energies->so3_vzero_ngbr,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<Vzero_ngbr>    = %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_vzero_ngbr/step,
                unit_conv*energies->so3_vzero_ngbr,sunit);
      }
    }
    if(energies->so3_ekin){
      if(isnan(unit_conv*energies->avso3_ekin/step) != 0 ||
         isnan(unit_conv*energies->so3_ekin)!=0){
        fprintf(stdout,"###NAN ERROR###<Ke >    \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_ekin/step,
                unit_conv*energies->so3_ekin,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<Ke>    \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_ekin/step,
                unit_conv*energies->so3_ekin,sunit);
      }
    }
    
#ifdef REPORT_ENERGIES
    if(energies->so3_bond){
      if(isnan(unit_conv*energies->avso3_bond/step) != 0 ||
         isnan(unit_conv*energies->so3_bond) !=0){
        fprintf(stdout,"###NAN ERROR <bond>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_bond/step,
                unit_conv*energies->so3_bond,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<bond>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_bond/step,
                unit_conv*energies->so3_bond,sunit);
      }
    }
    if(energies->so3_bend){
      if(isnan(unit_conv*energies->avso3_bend/step) != 0 ||
         isnan(unit_conv*energies->so3_bend) !=0){
        fprintf(stdout,"###NAN ERROR <bend>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_bend/step,
                unit_conv*energies->so3_bend,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<bend>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_bend/step,
                unit_conv*energies->so3_bend,sunit);
      }
    }
    if(energies->so3_tors){
      if(isnan(unit_conv*energies->avso3_tors/step) != 0 ||
         isnan(unit_conv*energies->so3_tors) !=0){
        fprintf(stdout,"###NAN ERROR <tors>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_tors/step,
                unit_conv*energies->so3_tors,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<tors>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_tors/step,
                unit_conv*energies->so3_tors,sunit);
      }
    }
    if(energies->so3_onfo){
      if(isnan(unit_conv*energies->avso3_onfo/step) != 0 ||
         isnan(unit_conv*energies->so3_onfo) !=0){
        fprintf(stdout,"###NAN ERROR <onefour>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_onfo/step,
                unit_conv*energies->so3_onfo,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<onefour>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_onfo/step,
                unit_conv*energies->so3_onfo,sunit);
      }
    }
    if(energies->so3_onfo_e){
      if(isnan(unit_conv*energies->avso3_onfo_e/step) != 0 ||
         isnan(unit_conv*energies->so3_onfo_e) !=0){
        fprintf(stdout,"###NAN ERROR <onefour_e>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_onfo_e/step,
                unit_conv*energies->so3_onfo_e,sunit);
        exit(99);
      }else{
        fprintf(stdout,"<onefour_e>   \t= %15.6f\t~ %15.6f %s\n",
                unit_conv*energies->avso3_onfo_e/step,
                unit_conv*energies->so3_onfo_e,sunit);
      }
    }
    fprintf(stdout,"<temp>   \t= %15.6f\t~ %15.6f K\n",
            energies->avso3_temp/step,energies->so3_temp);
#endif
    fprintf(stdout,"<Ham>    \t= %15.6f\t~ %15.6f %s\n",
            unit_conv*energies->avso3_ham/step,
            unit_conv*energies->so3_ham,sunit);
  }
  fflush(stdout); fflush(stderr);
}
