/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* function prototypes for mathlib directory */

double gethinv9(double hmat[], double hmati[]);
double get_deth(double []);
double ddot1(int n,double *a,double *b);
double dsum1(int n,double *a);
double dsum2(int n,int *iselect,double *a);
double ddot1a(int n,double *a,int astep,double *b,int bstep);
double dsum1a(int n,double *a,int astep);
int locate(int *,int,int);
int insert(int *ia,int *n,int jpoint);
void splint(double [], double [], double [], int , double , double *);
void spline(double [], double [], int , double , double , double []);
void polint(double xa[], double ya[], int n, double x, double *y, double *dy);

double *dvector(int,int);
double **dmatrix(int,int,int,int);
double **drematrix(double **,int,int,int,int,int,int,int,int);
double ***d3tensor(int nrl,int nrh,int ncl,int nch,int ndl,int ndh);

double cputime(void);
double realtime(void);
void init_myclocks(MYTIMING *timing);
void myclockon(MYTIMING *timing,char *name,int index);
void myclockoff(MYTIMING *timing,char *name,int index);
void print_timing(MYTIMING *timing);

void free_dvector(double *,int,int);
void free_dmatrix(double **,int,int,int,int);
void free_d3tensor(double ***,int,int,int,int,int,int);

void srandme(int);
void ggauss(int ,double *);
void *cmalloc(size_t);
void *crealloc(void *,size_t);
void decomp1d(int,int,int,long *,long *);
void decomp_trig(int,int,int,long *,long *);

void rs_me(int n,double *,double **,int);
void ctred2(double **,int,double [],double []);
void ctred2v(double **,int,double [],double []);
void ctqli(double [],double [],int,double **);
void ctqliv(double [],double [],int,double **);
void ceigsrt(double [],double **,int);
void ceigsrtv(double [],double **,int);
void invres(int,double **,double **);
void vsrtasnd(double d[],double **v,int m,int n);

/*-----------------------matopt.c-------------------------------*/
void matmul(int n,double **a,double **b,double **c);
void vecmat(int n,double **a,double *x,double *y);

void matmulv(int n,double a[],double b[],double c[]);
void transmatv(int n,double a[],double b[]);
void sym33(double mat[9]);
void test_anint(void);

/*-------------------------util.c------------------------------*/
int same_mole(int iatom,int jatom,int *itype_species,int *itype_molecules);
FILE *cfopenw(char *name);
void *cmalloc(size_t mem);
void allocate_coords(SIMPARMS *simparms,COORDS *coords);
void allocate_extend(SIMPARMS *simparms,COORDS *coords);
void allocate_exttab(SIMPARMS *simparms,COORDS *coords,int);
void rmass(int natoms,double *fx,double *fy,double *fz,double *amass);
void rmassselect(int *iselect,int nselect,double *fx,double *fy,double *fz,double *amass);
void multiply_selection(double *,double,double *,int *,int,int);
void free_exttab(COORDS *coords);
void free_coords(COORDS *coords);
void free_extend(SIMPARMS *,COORDS *coords);
void free_all_pointers(SIMPARMS *simparms,COORDS *coords,INTER *,NGBR *);

/*-------------------------checkvals.c------------------------------*/
int check_vals(int,...);
/*-------------------------checkvals.c------------------------------*/

/*------------------------md_char.c----------------------------*/
int get_dict_num(int,WORD [],char *);
void set_sim_keyword(int *,WORD **dict);
void get_meta_key(FILE *,char *,char *,int *,int *,char *,KEY *,int *);
void get_keys(FILE *,char *,char *,int *,int *,int *,KEY *);
void free_key(int,KEY *);
void syntax_error(const char *,FILE *,int ,char *,int);
void print_dict(WORD,int,WORD []);
void err_found(WORD,int,char *,int);
void write_key(KEY *);
void get_sim_keys(const char *,const char *,int *,KEY *);
int getnxtint(char *);
/*------------------------md_char.c----------------------------*/

/* -----------------------so3utils.c---------------------------*/
double VecMag(double vec[3]);
void VecSub(double vec1[3], double vec2[3], double vec3[3]);
void VecAdd(double vec1[3], double vec2[3], double vec3[3]);
void MultScalarMat(double c, double mat[3][3]);
void MultMatVec(double mat[3][3], double vec1[3], double vec2[3]);
void CrossProduct(double vec1[3], double vec2[3], double vec3[3]);
void DotProduct(double vec1[3], double vec2[3], double *scalar);
void MatAdd3(double mat1[3][3], double mat2[3][3], double mat3[3][3], double mat4[3][3]);
void MatAdd(double mat1[3][3], double mat2[3][3], double mat3[3][3]);
void MatSub(double mat1[3][3], double mat2[3][3], double mat3[3][3]);
void MatMult(double mat1[3][3], double mat2[3][3], double mat3[3][3]);
void Skew(double vec[3], double mat[3][3]);
void SkewSquare(double vec[3], double mat[3][3]);
void Imat33(double mat[3][3]);
void Rodrigues(double vec[3], double mat[3][3]);
void Invert33(double mat[3][3],double mati[3][3]);
void ludcmp33(double a[3][3],int indx[3],double *d);
void lubksb33(double a[3][3],int indx[3],double b[3]);
/* -----------------------so3utils.c---------------------------*/

