
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

double VecMag(double vec[3]){
  double mag = sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2]); 
  return mag;
}

void VecSub(double vec1[3], double vec2[3], double vec3[3]){
  vec3[0]=vec1[0]-vec2[0];
  vec3[1]=vec1[1]-vec2[1];
  vec3[2]=vec1[2]-vec2[2];
}

void VecAdd(double vec1[3], double vec2[3], double vec3[3]){
  vec3[0]=vec1[0]+vec2[0];
  vec3[1]=vec1[1]+vec2[1];
  vec3[2]=vec1[2]+vec2[2];
}

void MultScalarMat(double c, double mat[3][3]){
  int i,j;
    for(i=0;i<3;++i){
      for(j=0;j<3;++j){
	mat[i][j]*=c;
      }
    }
}

void MultMatVec(double mat[3][3], double vec1[3], double vec2[3]){
  int i,j;  /* this routine does vec2 = mat * vec1  */
  for(i=0;i<3;++i){
    vec2[i]=0.0;
    for(j=0;j<3;++j){
      vec2[i]+=mat[i][j]*vec1[j];
    }
  }
}

void CrossProduct(double vec1[3], double vec2[3], double vec3[3]){
  vec3[0]=vec1[1]*vec2[2]-vec1[2]*vec2[1];
  vec3[1]=vec1[2]*vec2[0]-vec1[0]*vec2[2];
  vec3[2]=vec1[0]*vec2[1]-vec1[1]*vec2[0];
}

void DotProduct(double vec1[3], double vec2[3], double *scalar){
  *scalar=vec1[0]*vec2[0]+vec1[1]*vec2[1]+vec1[2]*vec2[2];
}


void MatAdd3(double mat1[3][3], double mat2[3][3], double mat3[3][3], double mat4[3][3]){
  int i,j;  /* this routine does mat4 = mat1+mat2+mat3   */
    for(i=0;i<3;++i){
      for(j=0;j<3;++j){
	mat4[i][j]=mat1[i][j]+mat2[i][j]+mat3[i][j];
      }
    }
}

void MatAdd(double mat1[3][3], double mat2[3][3], double mat3[3][3]){
  int i,j;  /* this routine does mat3 = mat1+mat2  */
  for(i=0;i<3;++i){
    for(j=0;j<3;++j){
      mat3[i][j]=mat1[i][j]+mat2[i][j];
    }
  }
}

void MatSub(double mat1[3][3], double mat2[3][3], double mat3[3][3]){
  int i,j;  /* this routine does mat3 = mat1-mat2  */
  for(i=0;i<3;++i){
    for(j=0;j<3;++j){
      mat3[i][j]=mat1[i][j]-mat2[i][j];
    }
  }
}

void MatMult(double mat1[3][3], double mat2[3][3], double mat3[3][3]){
  int i,j,k; /* this routine does mat3=mat1*mat2 */
  for(i=0;i<3;++i){
    for(j=0;j<3;++j){
      mat3[i][j]=0.0;
      for(k=0;k<3;++k){
	mat3[i][j]+=mat1[i][k]*mat2[k][j];
      }
    }
  }
}

void Skew(double vec[3], double mat[3][3]){
  mat[0][0]=0.0;  /* creates the skew-symmetric matrix from a vector  */
  mat[0][1]=-vec[2];
  mat[0][2]=vec[1];
  mat[1][0]=vec[2];
  mat[1][1]=0.0;
  mat[1][2]=-vec[0];
  mat[2][0]=-vec[1];
  mat[2][1]=vec[0];
  mat[2][2]=0.0;
}

void SkewSquare(double vec[3], double mat[3][3]){
  mat[0][0]=-vec[2]*vec[2]-vec[1]*vec[1];  /* creates the square of the */  
  mat[0][1]=vec[0]*vec[1]; /*skew-symmetric matrix from a vector  */
  mat[0][2]=vec[0]*vec[2];
  mat[1][0]=vec[1]*vec[0];
  mat[1][1]=-vec[2]*vec[2]-vec[0]*vec[0];
  mat[1][2]=vec[1]*vec[2];
  mat[2][0]=vec[2]*vec[0];
  mat[2][1]=vec[2]*vec[1];
  mat[2][2]=-vec[1]*vec[1]-vec[0]*vec[0];
}

void Imat33(double mat[3][3]){
  int i,j;           /* this routine creates the 3x3 identity matrix  */
  for(i=0;i<3;++i){
    for(j=0;j<3;++j){
      mat[i][j] = 0.0;
    }
    mat[i][i] = 1.0;
  }
}

double det(double mat[3][3])
{
  /* gets det of mat */
  
  return (mat[0][0]*(mat[1][1]*mat[2][2] - mat[1][2]*mat[2][1]) + 
	  mat[0][1]*(mat[1][2]*mat[2][0] - mat[1][0]*mat[2][2]) + 
	  mat[0][2]*(mat[1][0]*mat[2][1] - mat[1][1]*mat[2][0]));
} 
/*---------------------------------------------------------------- */

void Invert33(double mat[3][3],double mati[3][3]){
  
  double rvol = 1./det(mat);
  mati[0][0] = (mat[1][1]*mat[2][2] - mat[1][2]*mat[2][1])*rvol;
  mati[0][1] = (mat[0][2]*mat[2][1] - mat[0][1]*mat[2][2])*rvol;
  mati[0][2] = (mat[0][1]*mat[1][2] - mat[0][2]*mat[1][1])*rvol;
  mati[1][0] = (mat[1][2]*mat[2][0] - mat[1][0]*mat[2][2])*rvol;
  mati[1][1] = (mat[0][0]*mat[2][2] - mat[0][2]*mat[2][0])*rvol;
  mati[1][2] = (mat[0][2]*mat[1][0] - mat[0][0]*mat[1][2])*rvol;
  mati[2][0] = (mat[1][0]*mat[2][1] - mat[1][1]*mat[2][0])*rvol;
  mati[2][1] = (mat[0][1]*mat[2][0] - mat[0][0]*mat[2][1])*rvol;
  mati[2][2] = (mat[0][0]*mat[1][1] - mat[0][1]*mat[1][0])*rvol;
}

void Rodrigues(double vec[3], double mat[3][3]){
  int i;
  double mag, unitvec[3], mat1[3][3], mat2[3][3];
  mag = VecMag(vec);
  Imat33(mat);  /* initializes mat = I */
  if(mag==0.0){
    return;        /* return identity matrix if vec={0,0,0}  */
  }
  for(i=0;i<3;++i){
    unitvec[i]=vec[i]/mag;   /* normalize vec  */
  }
  Skew(unitvec,mat1);
  SkewSquare(unitvec,mat2);
  MultScalarMat((sin(mag)),mat1);  /* the matrix is overwritten */
  MultScalarMat((1.0-cos(mag)),mat2);  /* the matrix is overwritten  */
  MatAdd3(mat,mat1,mat2,mat); /* add the first 3 matrices, answer is 4th matrix */
}

/*------------------------------------------------------*/
void luerr33(char *s){fprintf(stderr,"ERROR: %s\n",s); exit(1);}

#define TINY 1e-20;

/*-------------------------------------------------------*/
/* lower upper decomposition */
void ludcmp33(double a[3][3],int indx[3],double *d)
{
  int i,j,k,imax=0;
  double big,dum,sum,temp,vv[3];

  *d=1.0;
  for (i=0;i<3;i++) {
    big=0.0;
    for (j=0;j<3;j++)
      if ((temp=fabs(a[i][j])) > big) big=temp;
    if (big == 0.0) luerr33("Singular matrix in routine LUDCMP33");
    vv[i]=1.0/big;
  }
  for (j=0;j<3;j++) {
    for (i=0;i<j;i++) {
      sum=a[i][j];
      for (k=0;k<i;k++) sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
    }
    big=0.0;
    for (i=j;i<3;i++) {
      sum=a[i][j];
      for (k=0;k<j;k++) sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
      if ( (dum=vv[i]*fabs(sum)) >= big) {
	big=dum;
	imax=i;
      }
    }
    if (j != imax) {
      for (k=0;k<3;k++) {
	dum=a[imax][k];
	a[imax][k]=a[j][k];
	a[j][k]=dum;
      }
      *d = -(*d);
      vv[imax]=vv[j];
    }
    indx[j]=imax;
    if (a[j][j] == 0.0) a[j][j]=TINY;
    if (j != 3) {
      dum=1.0/(a[j][j]);
      for (i=j+1;i<3;i++) a[i][j] *= dum;
    }
  }
}

/*--------------------------------------------------------*/

void lubksb33(double a[3][3],int indx[3],double b[3])
{
  int i,j,ii=-1,ip;
  double sum;

  for (i=0;i<3;i++) {
    ip=indx[i];
    sum=b[ip];
    b[ip]=b[i];
    if (ii != -1) for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
    else if (sum) ii=i;
    b[i]=sum;
  }

  for (i=3-1;i>=0;i--) {
    sum=b[i];
    for (j=i+1;j<3;j++) sum -= a[i][j]*b[j];
    b[i]=sum/a[i][i];
  }
}


/* int main() */
/* { */
/*   double vec[3] = {3,2,1}; */
/*   double mat[3][3]; */

/*   Rodrigues(vec,mat); */

/*   int i,j; */
/*   for(i=0;i<3;++i){ */
/*     for(j=0;j<3;++j){ */
/*       fprintf(stdout,"%d %d %lg\n",i,j,mat[i][j]); */
/*     } */
/*   } */
/* } */
