/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* subroutine to time processes */

#include <sys/types.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include "md.h"


#define MAXTIME 2147.483648
#define MINTIME (-MAXTIME)

#ifdef PARA
#include <mpi.h>
#define WALL
#endif

#ifndef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC  1000000
#endif

double cputime(void)
{
  static clock_t itime=0;
  static double to=0.,tn=0.,dtime;
#ifdef WALL
  time_t t_time=1;
#endif

#ifdef WALL
#ifdef PARA
  if(!itime){
    itime = 1;
    tn = to = (double)MPI_Wtime();
  } else {
    tn = (double)MPI_Wtime();
  }
  dtime = tn-to;
#else 
  if(!itime){
    itime = 1;
    tn = to = (double)time(&t_time);
  } else {
    tn = (double)time(&t_time);
  }
  dtime = tn-to;
#endif
#else
  itime = clock();
  tn = (double)((double)itime/(double)CLOCKS_PER_SEC);
  
  if(to>tn) dtime = (MAXTIME-to)+(tn-MINTIME);
  else dtime = tn-to;
#endif

  to = tn;
  return dtime;
}
/*-------------------------------------------------*/
double realtime(void)
{
  static double dtime;
  time_t t_time=1;

#ifdef PARA
  dtime = (double)MPI_Wtime();
#else 
  dtime = (double)time(&t_time);
#endif

  return dtime;
}

/*-------------------------------------------------*/
double cputotaltime(void)
{
  static clock_t itime=0;
  static double to=0.,tn=0.,dtime,tottime;

  itime = clock();
  tn = (double)((double)itime/(double)CLOCKS_PER_SEC);
  
  if(to>tn) {
    dtime = (MAXTIME-to)+(tn-MINTIME);
  }else {
    dtime = tn-to;
  }
  to = tn;
  tottime += dtime; 

  return tottime;
}

/*-------------------------------------------------*/
void init_myclocks(MYTIMING *timing){
  int i;
  
  timing->nclocks = 26;
  strcpy(timing->clocknames[0],"TOTAL TIME");
  strcpy(timing->clocknames[1],"INITIALIZATION");
  strcpy(timing->clocknames[2],"MD STEP");
  strcpy(timing->clocknames[3],"FORCE LONG RANGE");
  strcpy(timing->clocknames[4],"FORCE SHORT RANGE");
  strcpy(timing->clocknames[5],"FORCE BONDS BENDS");
  strcpy(timing->clocknames[6],"FORCE TORSION ONE-FOUR");
  strcpy(timing->clocknames[7],"NEIGHBORLIST");
  strcpy(timing->clocknames[8],"WRITE DATA");
  strcpy(timing->clocknames[9],"RESPA INTRA");
  strcpy(timing->clocknames[10],"GET VIREAL");
  strcpy(timing->clocknames[11],"SET SWITCHFUNCTION");
  strcpy(timing->clocknames[12],"SO3 ROTATION");
  strcpy(timing->clocknames[13],"SO3 ACCELERATION");
  strcpy(timing->clocknames[14],"ATOMISTIC FORCE BONDS BENDS");
  strcpy(timing->clocknames[15],"ATOMISTIC FORCE TORSION ONE-FOUR");
  strcpy(timing->clocknames[16],"ATOMISTIC FORCE SHORT RANGE");
  strcpy(timing->clocknames[17],"ATOMISTIC FORCE LONG RANGE");
  strcpy(timing->clocknames[18],"CG FORCE BONDS BENDS");
  strcpy(timing->clocknames[19],"CG FORCE TORSION ONE-FOUR");
  strcpy(timing->clocknames[20],"CG FORCE SHORT RANGE");
  strcpy(timing->clocknames[21],"CG FORCE LONG RANGE");
  strcpy(timing->clocknames[22],"EXTENDED VARIABLES");
  strcpy(timing->clocknames[23],"SELECT HYBRID REGIONS");
  strcpy(timing->clocknames[24],"NEIGBORLIST CG");
  strcpy(timing->clocknames[25],"NEIGBORLIST ATOMISTIC");
  
  for(i=0; i<timing->nclocks; i++){
    timing->iset_clock[i] = 0;
    timing->ncalls[i] = 0;
    timing->timing_sum[i] = 0.0;
  }
  return;
}

/*-------------------------------------------------*/
void myclockon(MYTIMING *timing,char *name,int index){
  double time;
  char line[MAXLINELEN];

  if( strcmp(name,timing->clocknames[index]) != 0 ) {
    sprintf(line,"Error with clock[%d]: called with name %s instead of %s",index,name,timing->clocknames[index]);
    md_error(line);
  }
  
  if( timing->iset_clock[index] != 0 ){
    sprintf(line,"Error with clock[%d] (%s): tried to start clock but was found already running?",index,name);
    md_error(line);
  }

  timing->iset_clock[index]++;
  time = cputotaltime();
  timing->timeon[index] = time;

  return;
}

/*-------------------------------------------------*/
void myclockoff(MYTIMING *timing,char *name,int index){
  double time;
  char line[MAXLINELEN];

  if( strcmp(name,timing->clocknames[index]) != 0 ) {
    sprintf(line,"Error with clock[%d]: called with name %s instead of %s",index,name,timing->clocknames[index]);
    md_error(line);
  }
  
  if( timing->iset_clock[index] != 1 ){
    sprintf(line,"Error with clock[%d] (%s): tried to stop clock but was found already stopped?",index,name);
    md_error(line);
  }

  timing->ncalls[index]++;
  time = cputotaltime();
  time -= timing->timeon[index];
  timing->timing_last[index] = time;
  timing->timing_sum[index] += time;
  timing->iset_clock[index] = 0;

  return;
}
/*-------------------------------------------------*/
void print_timing(MYTIMING *timing){
  int i;
  double totaltime;
  int *ncalls = timing->ncalls;
  double *tlast = timing->timing_last;
  double *tsum = timing->timing_sum;
  
  totaltime = cputotaltime() - timing->timeon[0];
  
  printf("---------------------- TIMINGS -------------------------------------------\n");
  printf(" ROUTINE                                LAST     TOTAL  PERCENTAGE  #CALLS\n");
  for(i=1; i<timing->nclocks; i++){
    printf("  %-32s: %10.6lf %10.6lf %6.2lf %6d\n",timing->clocknames[i],tlast[i],
	   tsum[i],100.0*tsum[i]/totaltime,ncalls[i]);
  }
  printf("  %-32s: %10.6lf %10.6lf %6.2lf %6d\n","TOTAL TIME",totaltime,totaltime,100.0,1);
  printf("--------------------------------------------------------------------------\n");

  return;
}

